Package: correspondenceTables
Type: Package
Title: Creating Correspondence Tables Between Two Statistical
        Classifications
Date: 2026-05-15
Version: 1.0.2
Authors@R: c(person("Vasilis", "Chasiotis", role = "aut", comment="Department of Statistics, Athens University of Economics and Business"), 
      person("Photis", "Stavropoulos", role = "aut", comment="Quantos S.A. Statistics and Information Systems"),
      person("Martin", "Karlberg", role = "aut"),
      person("Mátyás", "Mészáros", email = "matyas.meszaros@ec.europa.eu", role = "cre"),
      person("Martina", "Patone", role = "aut"),
      person("Erkand", "Muraku", role = "aut"), 
      person("Clement", "Thomas", role = "aut"), 
      person("Loic", "Bienvenu", role = "aut"),
      person("Mauro", "Baldacchini", role = "aut"),
      person("Khadija", "Sossey-Lallemand", role = "aut"),
      person("Mark", "van der Loo", role = "aut"))
Description: 
    A candidate correspondence table between two classifications can be created when there are correspondence tables leading from the first classification to the second one via intermediate 'pivot' classifications. 
    The correspondence table between two statistical classifications can be updated when one of the classifications gets updated to a new version.
License: EUPL
Encoding: UTF-8
Imports: data.table, httr, stringr, igraph
Suggests: knitr, rmarkdown, tinytest
VignetteBuilder: knitr
NeedsCompilation: no
URL: https://github.com/eurostat/correspondenceTables
BugReports: https://github.com/eurostat/correspondenceTables/issues
Maintainer: Mátyás Mészáros <matyas.meszaros@ec.europa.eu>
RoxygenNote: 7.3.3
Packaged: 2026-05-15 13:15:57 UTC; rstudio
Author: Vasilis Chasiotis [aut] (Department of Statistics, Athens University of
    Economics and Business),
  Photis Stavropoulos [aut] (Quantos S.A. Statistics and Information
    Systems),
  Martin Karlberg [aut],
  Mátyás Mészáros [cre],
  Martina Patone [aut],
  Erkand Muraku [aut],
  Clement Thomas [aut],
  Loic Bienvenu [aut],
  Mauro Baldacchini [aut],
  Khadija Sossey-Lallemand [aut],
  Mark van der Loo [aut]
Repository: CRAN
Date/Publication: 2026-05-15 15:50:15 UTC
Built: R 4.6.0; ; 2026-05-15 18:28:33 UTC; unix

# Generated by roxygen2: do not edit by hand

export(aggregateCorrespondenceTable)
export(analyseCorrespondenceTable)
export(classificationList)
export(classificationQC)
export(correspondenceTableList)
export(dataStructure)
export(newCorrespondenceTable)
export(prefixList)
export(retrieveClassificationTable)
export(retrieveCorrespondenceTable)
export(updateCorrespondenceTable)
import(httr)
import(igraph)
importFrom(data.table,fread)
importFrom(data.table,fwrite)
importFrom(httr,GET)
importFrom(httr,RETRY)
importFrom(httr,accept)
importFrom(httr,add_headers)
importFrom(httr,content)
importFrom(httr,stop_for_status)
importFrom(httr,timeout)
importFrom(httr,user_agent)
importFrom(igraph,decompose.graph)
importFrom(igraph,graph.data.frame)
importFrom(stats,aggregate)
importFrom(stats,na.omit)
importFrom(stats,setNames)
importFrom(utils,capture.output)
importFrom(utils,head)
importFrom(utils,menu)
importFrom(utils,read.csv)
importFrom(utils,setTxtProgressBar)
importFrom(utils,tail)
importFrom(utils,txtProgressBar)
importFrom(utils,write.csv)

<!-- badges: start -->
[![last commit](https://img.shields.io/github/last-commit/eurostat/correspondenceTables?style=flat)](https://github.com/eurostat/correspondenceTables/commits/)
[![R build
status](https://github.com/eurostat/correspondenceTables/workflows/R-CMD-check/badge.svg)](https://github.com/eurostat/correspondenceTables/actions)
[![dependencies](https://tinyverse.netlify.app/badge/correspondenceTables)](https://CRAN.R-project.org/package=correspondenceTables)
[![CRAN version](https://www.r-pkg.org/badges/version/correspondenceTables)](https://CRAN.R-project.org/package=correspondenceTables)
[![CRAN status](https://badges.cranchecks.info/flavor/r-release-linux-x86_64/correspondenceTables.svg)](https://cran.r-project.org/web/checks/check_results_correspondenceTables.html)
[![license](https://img.shields.io/badge/license-EUPL-success)](https://interoperable-europe.ec.europa.eu/collection/eupl/eupl-text-eupl-12)
[![weekly downloads](https://cranlogs.r-pkg.org/badges/last-week/correspondenceTables)](https://CRAN.R-project.org/package=correspondenceTables)
[![monthly downloads](https://cranlogs.r-pkg.org/badges/correspondenceTables)](https://CRAN.R-project.org/package=correspondenceTables)
[![all downloads](https://cranlogs.r-pkg.org/badges/grand-total/correspondenceTables)](https://CRAN.R-project.org/package=correspondenceTables)
<!-- badges: end -->

# CorrespondenceTables

Empowering the seamless creation and refinement of correspondence tables between two statistical classifications, such as NACE, CPA, CN, and more.
This toolkit not only streamlines the process of generating and updating these tables but also provides functionality for retrieving classification and correspondence tables.
It includes features for conducting quality control on classifications and offers a comprehensive analysis of correspondence tables with the ability to aggregate them effectively.

## Installation

'correspondenceTables' can be installed from [CRAN](https://CRAN.R-project.org/package=correspondenceTables) by 

```R
install.packages("correspondenceTables")
```

or use the development version from GitHub

```R
devtools::install_github("eurostat/correspondenceTables", build_vignettes = TRUE)
```

## Background

This package serves as a tool to assist you in creating and updating a correspondence table between two classifications, such as NACE, CPA, CN, and others. It enables the retrieval of classification tables and correspondence tables. Additionally, it facilitates quality control on a classification and allows for the analysis and aggregation of correspondence tables.

## Content

The package contains 11 functions:

1. `newCorrespondenceTable` to create a candidate correspondence table between two classifications when there are correspondence tables leading from the first classification to the second one via intermediate 'pivot' classifications.

2. `updateCorrespondenceTable` to update the correspondence table between two statistical classifications when one of the classifications gets updated to a new version.

3. `prefixList` Create a list of prefixes for both CELLAR and FAO repositories.

4. `correspondenceTableList` Overview of all the available correspondence classification from CELLAR and FAO repository.

5. `dataStructure`  Retrieve information about the structure of each classification tables from CELLAR and FAO repositories.

6. `classificationList` Retrieve a list of classification tables from CELLAR and FAO repositories or both.

7. `retrieveClassificationTable` Retrieve stastistical classification tables from CELLAR and FAO repositories.

8. `retrieveCorrespondenceTable` Retrieve correspondence tables between statistical classifications from CELLAR and FAO repositories.

9. `classificationQC` performs quality control checks on statistical classifications.

10. `analyseCorrespondanceTable` performs analysis on correspondence tables.  

11. `aggregateCorrespondenceTable` aggregates correspondence tables to higher hierarchical levels.

## Examples

For the examples see the vignettes.
```R
browseVignettes("correspondenceTables")
```

## ---- include = FALSE---------------------------------------------------------
library(correspondenceTables)
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
tmp_dir<-tempdir()
csv_files<-list.files(tmp_dir, pattern = ".csv")
if (length(csv_files)>0) unlink(csv_files)

## -----------------------------------------------------------------------------
A <- system.file("extdata", "CN2021.csv", package = "correspondenceTables")
AStar <- system.file("extdata", "CN2022.csv", package = "correspondenceTables")
B <- system.file("extdata", "CPA21.csv", package = "correspondenceTables")
AB <- system.file("extdata", "CN2021_CPA21.csv", package = "correspondenceTables")
AAStar <- system.file("extdata", "CN2021_CN2022.csv", package = "correspondenceTables")

## -----------------------------------------------------------------------------
UPC <- updateCorrespondenceTable(A, B, AStar, AB, AAStar, file.path(tmp_dir,"updateCorrespondenceTableCase1.csv"),
                                 "B", 0.4, 0.4)
print(UPC[[1]][1:10, 1:7])
print(UPC[[2]])

## -----------------------------------------------------------------------------
A <- system.file("extdata", "CN2021.csv", package = "correspondenceTables")
AStar <- system.file("extdata", "CN2022.csv", package = "correspondenceTables")
B <- system.file("extdata", "PRODCOM2021.csv", package = "correspondenceTables")
AB <- system.file("extdata", "CN2021_PRODCOM2021.csv", package = "correspondenceTables")
AAStar <- system.file("extdata", "CN2021_CN2022.csv", package = "correspondenceTables")

## ---- results = "hide"--------------------------------------------------------
UPC <- updateCorrespondenceTable(A, B, AStar, AB, AAStar, file.path(tmp_dir,"updateCorrespondenceTableCase2.csv"), "A", 0.4, 0.3)

## -----------------------------------------------------------------------------
A <- system.file("extdata", "NAICS2017.csv", package = "correspondenceTables")
AStar <- system.file("extdata", "NAICS2022.csv", package = "correspondenceTables")
B <- system.file("extdata", "NACE.csv", package = "correspondenceTables")
AB <- system.file("extdata", "NAICS2017_NACE.csv", package = "correspondenceTables")
AAStar <- system.file("extdata", "NAICS2017_NAICS2022.csv", package = "correspondenceTables")

## ---- results = "hide"--------------------------------------------------------
UPC <- updateCorrespondenceTable(A, B, AStar, AB, AAStar, file.path(tmp_dir,"updateCorrespondenceTableCase3.csv"), "none", 0.5, 0.3)

## -----------------------------------------------------------------------------
A <- system.file("extdata", "CN2021.csv", package = "correspondenceTables")
AStar <- system.file("extdata", "CN2022.csv", package = "correspondenceTables")
B <- system.file("extdata", "NST2007.csv", package = "correspondenceTables")
AB <- system.file("extdata", "CN2021_NST2007.csv", package = "correspondenceTables")
AAStar <- system.file("extdata", "CN2021_CN2022.csv", package = "correspondenceTables")

## ---- results = "hide"--------------------------------------------------------
UPC <- updateCorrespondenceTable(A, B, AStar, AB, AAStar, file.path(tmp_dir,"updateCorrespondenceTableCase4.csv"), "B", 0.4, 0.3)

## -----------------------------------------------------------------------------
A <- system.file("extdata", "CN2021.csv", package = "correspondenceTables")
AStar <- system.file("extdata", "CN2022.csv", package = "correspondenceTables")
B <- system.file("extdata", "SITC4.csv", package = "correspondenceTables")
AB <- system.file("extdata", "CN2021_SITC4.csv", package = "correspondenceTables")
AAStar <- system.file("extdata", "CN2021_CN2022.csv", package = "correspondenceTables")

## ---- results = "hide"--------------------------------------------------------
UPC <- updateCorrespondenceTable(A, B, AStar, AB, AAStar, file.path(tmp_dir,"updateCorrespondenceTableCase5.csv"), "B", 0.3, 0.7)

## -----------------------------------------------------------------------------
A <- system.file("extdata", "CN2021.csv", package = "correspondenceTables")
AStar <- system.file("extdata", "CN2022.csv", package = "correspondenceTables")
B <- system.file("extdata", "BEC4.csv", package = "correspondenceTables")
AB <- system.file("extdata", "CN2021_BEC4.csv", package = "correspondenceTables")
AAStar <- system.file("extdata", "CN2021_CN2022.csv", package = "correspondenceTables")

## ---- results = "hide"--------------------------------------------------------
UPC <- updateCorrespondenceTable(A, B, AStar, AB, AAStar, file.path(tmp_dir,"updateCorrespondenceTableCase6.csv"), "B", 0.3, 0.6)

## -----------------------------------------------------------------------------
fullPath <- function(CSVraw, CSVappended){
  NamesCsv <- system.file("extdata", CSVraw, package = "correspondenceTables")
  A <- read.csv(NamesCsv, header = FALSE, sep = ",")
   for (i in 1:nrow(A)) {
    for (j in 1:ncol(A)) {
      if (A[i,j]!="") {
        A[i, j] <- system.file("extdata", A[i, j], package = "correspondenceTables")
      }}}
  write.table(x = A, file = file.path(tmp_dir,CSVappended), row.names = FALSE, col.names = FALSE, sep = ",")
  return(A)
}

## ---- results = "hide"--------------------------------------------------------
fullPath("names1.csv", "names.csv")

## ---- results = "hide"--------------------------------------------------------
system.time(NCT <- newCorrespondenceTable(file.path(tmp_dir,"names.csv"), file.path(tmp_dir,"newCorrespondenceTableCase1.csv"), "A", 0.5))

## -----------------------------------------------------------------------------
print(NCT[[1]][1:10, 1:6])
print(NCT[[2]])

## ---- results = "hide"--------------------------------------------------------
fullPath("names2.csv", "names.csv")

## ---- results = "hide"--------------------------------------------------------
system.time(NCT <- newCorrespondenceTable(file.path(tmp_dir,"names.csv"), file.path(tmp_dir,"newCorrespondenceTableCase2.csv"), "B", 0.5))

## ---- results = "hide"--------------------------------------------------------
fullPath("names3.csv", "names.csv")

## ---- results = "hide"--------------------------------------------------------
system.time(NCT <- newCorrespondenceTable(file.path(tmp_dir,"names.csv"), file.path(tmp_dir,"newCorrespondenceTableCase3.csv"), "B", 0.5))

## ---- results = "hide"--------------------------------------------------------
fullPath("names4.csv", "names.csv")

## ---- results = "hide"--------------------------------------------------------
system.time(NCT <- newCorrespondenceTable(file.path(tmp_dir,"names.csv"), file.path(tmp_dir,"newCorrespondenceTableCase4.csv"), "none", 0.96))

## ----message=TRUE, warning=TRUE, include=FALSE--------------------------------
csv_files<-list.files(tmp_dir, pattern = ".csv")
if (length(csv_files)>0) unlink(csv_files)

## ----include=FALSE------------------------------------------------------------

knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  warning  = FALSE,
  message  = FALSE
)


## -----------------------------------------------------------------------------
library(correspondenceTables) 

## ----echo=FALSE, results="asis"-----------------------------------------------
cat("<style>table {  width: 100% !important;}table caption {  text-align: center;}
</style>")

## -----------------------------------------------------------------------------
lengths_example <- data.frame(
  charb = c(1, 3, 5),
  chare = c(2, 4, 7)
)

knitr::kable(
  lengths_example,
  caption = "Example of expected code lengths by hierarchical level",
  align = "c"
)


## -----------------------------------------------------------------------------
singleChildCode <- read.csv(
  system.file("extdata/test", "SingleChild.csv",
              package = "correspondenceTables")
)

knitr::kable(
  singleChildCode,
  caption = "Single-child code rules",
  align = "c"
)

## -----------------------------------------------------------------------------

sequencing <- read.csv(
  system.file("extdata/test", "Sequencing.csv",
              package = "correspondenceTables")
)

knitr::kable(
  sequencing,
  caption = "Example of sequencing rules by hierarchical level",
  align = "c"
)


## -----------------------------------------------------------------------------
classification <- read.csv(
  system.file("extdata/test", "Nace2_long.csv", package = "correspondenceTables")
)

lengths <- data.frame(
  charb = c(1, 2, 3, 5),
  chare = c(1, 2, 4, 5)
)


## -----------------------------------------------------------------------------
output <- classificationQC(
  classification   = classification,
  lengths          = lengths,
  fullHierarchy    = TRUE,
  labelUniqueness  = TRUE,
  labelHierarchy   = TRUE,
  singleChildCode  = NULL,
  sequencing       = NULL
)

qc_summary <- data.frame(
  Check            = c("No levels", "Orphan codes", "Childless codes"),
  Number_of_issues = c(
    nrow(output$QC_noLevels),
    nrow(output$QC_orphan),
    nrow(output$QC_childless)
  )
)

knitr::kable(
  qc_summary,
  caption = "Summary of quality control checks",
  align = "c"
)


## ----echo=FALSE, results="asis"-----------------------------------------------

tbl <- output$QC_noLevels

cat(sprintf(
  "**QC_noLevels**\n\n- Rows: %d  \n- Columns: %d\n\n",
  nrow(tbl), 
  ncol(tbl)
))


## ----echo=FALSE, results="asis"-----------------------------------------------

tbl2 <- output$QC_orphan

cat(sprintf(
  "**QC_orphan**\n\n- Rows: %d  \n- Columns: %d\n\n",
  nrow(tbl2), 
  ncol(tbl2) 
))

knitr::kable(
  head(tbl2[, 1:8], 5),
  caption = "Orphan codes (First 5 rows (first 7 columns))",
  align = "c"
)

## ----echo=FALSE, results="asis"-----------------------------------------------
tbl3 <- output$QC_childless

cat(sprintf(
  "**QC_childless**\n\n- Rows: %d  \n- Columns: %d\n\n",
 nrow(tbl3), 
 ncol(tbl3)
))

knitr::kable(
  head(tbl3[, 1:8], 5),
  caption = "First 5 rows (first 7 columns)",
  align = "c"
)


## -----------------------------------------------------------------------------

singleChildCode <- read.csv(
  system.file("extdata/test", "SingleChild.csv", package = "correspondenceTables")
)
knitr::kable(
  singleChildCode,
  caption = "singleChildCode argument",
  align = "c"
)

output2 <- classificationQC(
  classification   = classification,
  lengths          = lengths,
  fullHierarchy    = TRUE,
  labelUniqueness  = TRUE,
  labelHierarchy   = TRUE,
  singleChildCode  = singleChildCode,
  sequencing       = NULL
)

## ----echo=FALSE, results="asis"-----------------------------------------------
tbl4 <- output2$QC_orphan

cat(sprintf(
  "**QC_orphan**\n\n- Rows: %d  \n- Columns: %d\n\n",
  nrow(tbl4), 
  ncol(tbl4)
))


knitr::kable(
  head(tbl4[, 1:8], 5),
  caption = "First 5 rows (first 7 columns)",
  align = "c"
)


## ----echo=FALSE, results="asis"-----------------------------------------------
tbl6 <- output2$QC_childless

cat(sprintf(
  "**QC_childless**\n\n- Rows: %d  \n- Columns: %d\n\n",
 nrow(tbl6), 
  ncol(tbl6) 
))


knitr::kable(
  head(tbl6[, 1:8], 10),
  caption = "First 10 rows (first 7 columns)",
  align = "c"
)



## -----------------------------------------------------------------------------

singleChildCode <- read.csv(
  system.file("extdata/test", "SingleChild2.csv", package = "correspondenceTables")
)

sequencing <- read.csv(
  system.file("extdata/test", "Sequencing.csv",
              package = "correspondenceTables")
)

output3 <- classificationQC(
  classification   = classification, 
  lengths          = lengths,
  fullHierarchy    = TRUE,
  labelUniqueness  = TRUE,
  labelHierarchy   = TRUE,
  singleChildCode  = singleChildCode,
  sequencing       = sequencing
)



## ----echo=FALSE, results="asis"-----------------------------------------------
tbl7 <- output3$QC_gapBefore

cat(sprintf(
  "**QC_gapBefore**\n\n- Rows: %d  \n- Columns: %d\n\n",
  nrow(tbl7),
  ncol(tbl7)
))


knitr::kable(
  head(tbl7[, 1:8], 10),
  caption = "QC_gapBefore. First 10 rows (first 7 columns)",
  align = "c"
)



## ----echo=FALSE, results="asis"-----------------------------------------------
tbl8 <- output3$QC_lastSibling

cat(sprintf(
  "**QC_lastSibling**\n\n- Rows: %d  \n- Columns: %d\n\n",
  nrow(tbl8),
  ncol(tbl8)
))


knitr::kable(
  head(tbl8[, 1:8], 10),
  caption = "QC_lastSibling. First 10 rows (first 7 columns)",
  align = "c"
)



## ----echo=FALSE, results="asis"-----------------------------------------------
tbl9 <- output3$QC_output

cat(sprintf(
  "**QC_output**\n\n- Rows: %d  \n- Columns: %d\n\n",
  nrow(tbl9), 
  ncol(tbl9)
))


knitr::kable(
  head(tbl9[, 1:8], 10),
  caption = "First 10 rows (first 7 columns)",
  align = "c"
)

## ----setup, include=FALSE-----------------------------------------------------

knitr::opts_chunk$set(
  echo = TRUE,
  message = FALSE,
  warning = FALSE,
  cache = FALSE
)


## ----cars---------------------------------------------------------------------
library(correspondenceTables)

## ----echo=FALSE, results="asis"-----------------------------------------------
cat("
<style>
table {
  width: 100% !important;
}
table caption {
  text-align: center;
}
</style>
")

## -----------------------------------------------------------------------------
AB_path <- system.file("extdata/test", "ab_data.csv", package = "correspondenceTables")
A_path  <- system.file("extdata/test", "a_data.csv",  package = "correspondenceTables")
B_path  <- system.file("extdata/test", "b_data.csv",  package = "correspondenceTables")

stopifnot(nzchar(AB_path), nzchar(A_path), nzchar(B_path))

AB <- utils::read.csv(AB_path, stringsAsFactors = FALSE)
A  <- utils::read.csv(A_path,  stringsAsFactors = FALSE)
B  <- utils::read.csv(B_path,  stringsAsFactors = FALSE)

#For clarity and consistency, the correspondence table columns are renamed to the expected identifiers:

names(AB)[names(AB) == "NACE.Rev..2.Code"]   <- "from_code"
names(AB)[names(AB) == "NACE.Rev..2.1.Code"] <- "to_code"


res <- aggregateCorrespondenceTable(AB = AB, A = A, B = B)


knitr::kable(
  head(res$result),
  caption = "Aggregation using a correspondence table",
  align = "c"
)


## -----------------------------------------------------------------------------

# Correspondence table with weights  
AB <- data.frame(
  from_code = c("A1", "A1", "A2"),
  to_code   = c("B1", "B2", "B2"),
  weight    = c(0.7, 0.3, 1.0)
)

# Source classification with values  
A <- data.frame(
  code  = c("A1", "A2"), 
  value = c(100, 50)
)

# Target classification domain
B <- data.frame(
  code = c("B1", "B2")
)

res2 <- aggregateCorrespondenceTable(AB = AB, A = A, B = B)

knitr::kable(
  head(res2$result),
  caption = "Weighted correspondence (proportional allocation)",
  align = "c"
)

## ----include=FALSE------------------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")

## -----------------------------------------------------------------------------
library(correspondenceTables)  

## ----echo=FALSE, results="asis"-----------------------------------------------
cat("
<style>
table {
  width: 100% !important;
}
table caption {
  text-align: center;
}
</style>
")

## -----------------------------------------------------------------------------
# Load example correspondence table
AB <- read.csv(
  system.file("extdata/test", "ExempleAnnexe.csv", package = "correspondenceTables"),
  stringsAsFactors = FALSE
)


## -----------------------------------------------------------------------------
# Perform analysis
result <- analyseCorrespondenceTable(
  AB = AB,
  A  = NULL,
  B  = NULL,
  longestAcodeOnly = FALSE,
  longestBcodeOnly = FALSE
)

## -----------------------------------------------------------------------------
# Display results

knitr::kable(
  head(result$Inventory),
  caption = "Inventory results: Basic analysis of a correspondence table",
  align = "c"
)


knitr::kable(
  head(result$Analysis),
  caption = "Analysis results: Basic analysis of a correspondence table",
  align = "c"
)


## -----------------------------------------------------------------------------
# Load correspondence table (NACE Rev.2 -> NACE Rev.2.1)
AB <- read.csv(
  system.file("extdata/test", "ab_data.csv", package = "correspondenceTables"),
  stringsAsFactors = FALSE
)

# Load source classification (NACE Rev.2)
A <- read.csv(
  system.file("extdata/test", "a_data.csv", package = "correspondenceTables"),
  stringsAsFactors = FALSE
)

# Load target classification (NACE Rev.2.1)
B <- read.csv(
  system.file("extdata/test", "b_data.csv", package = "correspondenceTables"),
  stringsAsFactors = FALSE
)


## ----warning = FALSE----------------------------------------------------------
# Perform analysis using all function parameters
result2 <- analyseCorrespondenceTable(
  AB = AB,
  A  = A,
  B  = B,
  longestAcodeOnly = TRUE,
  longestBcodeOnly = TRUE
)


## -----------------------------------------------------------------------------
# Display results

knitr::kable(
  head(result2$Inventory),
  caption = "Inventory results: Full analysis with source and target classifications",
  align = "c"
) 


knitr::kable(
  head(result2$Analysis[, 1:5]),
  caption = "Analysis results: Full analysis with source and target classifications",
  align = "c"
)

#' @title List available correspondence tables from online services
#'
#' @description
#' Returns a data frame listing available correspondence tables (mappings between
#' statistical classifications) published by one or more supported online services.
#'
#' Use this function to discover valid `prefix` and `ID_table` values before
#' calling `retrieveCorrespondenceTable()`.
#'
#' @param endpoint Character. Which service(s) to query.
#'   One or more of the supported endpoints (e.g. "CELLAR", "FAO"),
#'   or "ALL" to query all available services. Case-insensitive.
#'   If multiple endpoints are supplied (including "ALL"), a **named list**
#'   is returned; otherwise a single `data.frame`.
#' @param showQuery Logical. If `TRUE`, the SPARQL queries used are emitted via
#'   `message()` (suppress with `suppressMessages()`).
#'
#' @details
#' Read-only queries; no side effects (no file writes or caching).
#'
#' @return
#' If a single endpoint is requested, a `data.frame` (one row per correspondence).
#' If multiple endpoints are requested (including "ALL"), a **named list**
#' of data frames, one per endpoint.
#'
#' @seealso retrieveCorrespondenceTable()
#' 
#' @examples
#' \dontrun{
#' # List correspondence tables available in CELLAR
#' ct <- correspondenceTableList(endpoint = "CELLAR")
#' head(ct)
#'
#' # List correspondence tables available in FAO
#' ct_fao <- correspondenceTableList(endpoint = "FAO")
#' head(ct_fao)
#'
#' # Get correspondence tables from both repositories
#' ct_all <- correspondenceTableList(endpoint = "ALL")
#' names(ct_all)
#' }
#' 
#' 
#' @export
correspondenceTableList <- function(endpoint = "ALL", showQuery = FALSE) {
  
  if (!requireNamespace("httr", quietly = TRUE)) {
    stop("Package 'httr' is required. Please install.packages('httr').", call. = FALSE)
  }
  
  # ---- Centralized endpoint handling ----
  eps <- .expand_endpoints(endpoint)  # validates + expands "ALL" + preserves vectors
  
  # ---- Worker for a single endpoint (kept local; you can externalize later) ----
  run_one <- function(ep) {
    
    endpoint_url <- .sparql_endpoint(ep)
    
    # 1) Discover classification schemes (best effort)
    schemes <- tryCatch(
      classificationList(endpoint = ep),
      error = function(e) {
        message("classificationList() failed for endpoint '", ep, "': ",
                conditionMessage(e), ". Returning empty result.")
        return(NULL)
      }
    )
    
    # Extract scheme URIs if available (used for non-FAO endpoints to filter noise)
    scheme_uri <- character(0)
    if (is.data.frame(schemes) && nrow(schemes)) {
      candidate_uri_cols <- c("SchemeURI", "ConceptSchemeURI", "URI", "SchemeUri", "ConceptSchemeUri")
      uri_cols_idx <- which(tolower(names(schemes)) %in% tolower(candidate_uri_cols))
      if (length(uri_cols_idx)) {
        scheme_uri <- unique(unlist(schemes[uri_cols_idx], use.names = FALSE))
        scheme_uri <- as.character(scheme_uri)
        scheme_uri <- scheme_uri[!is.na(scheme_uri) & grepl("^https?://", scheme_uri)]
      }
    }
    
    # 2) Build SPARQL query
    prefix_header <- paste(
      "PREFIX skos: <http://www.w3.org/2004/02/skos/core#>",
      "PREFIX xkos: <http://rdf-vocabulary.ddialliance.org/xkos#>",
      sep = "\n"
    )
    
    values_block <- if (length(scheme_uri)) {
      paste0(
        "VALUES ?KnownScheme { ",
        paste(sprintf("<%s>", scheme_uri), collapse = " "),
        " }"
      )
    } else {
      ""
    }
    
    SPARQL.query <- if (toupper(ep) == "FAO") {
      # FAO: list correspondences directly (src/tgt schemes often not exposed here)
      paste0(
        prefix_header, "\n",
        "SELECT DISTINCT ?corr ?Label\n",
        "WHERE {\n",
        "  ?corr a xkos:Correspondence .\n",
        "  OPTIONAL { ?corr skos:prefLabel ?Label . }\n",
        "}\n"
      )
    } else {
      # CELLAR/others: optionally bind known schemes; try to infer src/tgt schemes
      paste0(
        prefix_header, "\n",
        "SELECT DISTINCT ?corr ?Label ?srcScheme ?tgtScheme\n",
        "WHERE {\n",
        if (nzchar(values_block)) paste0(values_block, "\n") else "",
        "  ?corr a xkos:Correspondence .\n",
        "  OPTIONAL { ?corr skos:prefLabel ?Label . }\n",
        "  OPTIONAL {\n",
        "    ?assoc xkos:inCorrespondence ?corr .\n",
        "    OPTIONAL { ?assoc xkos:sourceConcept ?src . ?src skos:inScheme ?srcScheme . }\n",
        "    OPTIONAL { ?assoc xkos:targetConcept ?tgt . ?tgt skos:inScheme ?tgtScheme . }\n",
        "  }\n",
        "}\n"
      )
    }
    
    if (isTRUE(showQuery)) {
      message("SPARQL query (", ep, "):\n", SPARQL.query)
    }
    
    # 3) Execute (read-only)
    csv_txt <- tryCatch(
      .do_post_csv(endpoint_url, SPARQL.query, timeout_sec = 60L, attempts = 2L),
      error = function(e) {
        message("SPARQL query failed for '", ep, "': ", conditionMessage(e),
                ". Returning empty result.")
        return("")
      }
    )
    if (!nzchar(csv_txt)) {
      return(data.frame(
        Prefix                = character(),
        ID                    = character(),
        Source.Classification = character(),
        Target.Classification = character(),
        Table.Name            = character(),
        URI                   = character(),
        stringsAsFactors      = FALSE
      ))
    }
    
    df <- tryCatch(
      utils::read.csv(text = csv_txt, stringsAsFactors = FALSE, check.names = FALSE),
      error = function(e) data.frame()
    )
    if (!is.data.frame(df) || !nrow(df)) {
      return(data.frame(
        Prefix                = character(),
        ID                    = character(),
        Source.Classification = character(),
        Target.Classification = character(),
        Table.Name            = character(),
        URI                   = character(),
        stringsAsFactors      = FALSE
      ))
    }
    
    # 4) Normalize output column names (robust to absent columns)
    df_nms_lc <- tolower(names(df))
    pick <- function(needle) {
      i <- match(needle, df_nms_lc)
      if (is.na(i)) NULL else names(df)[i]
    }
    col_corr  <- pick("corr")
    col_label <- pick("label")
    col_src   <- pick("srcscheme")
    col_tgt   <- pick("tgtscheme")
    
    n <- nrow(df)
    corr_vec  <- if (!is.null(col_corr))  as.character(df[[col_corr]])  else character(n)
    label_vec <- if (!is.null(col_label)) as.character(df[[col_label]]) else character(n)
    src_vec   <- if (!is.null(col_src))   as.character(df[[col_src]])   else character(n)
    tgt_vec   <- if (!is.null(col_tgt))   as.character(df[[col_tgt]])   else character(n)
    
    out <- data.frame(
      ID                    = if (length(corr_vec)) sub(".*/", "", corr_vec) else character(n),
      Source.Classification = src_vec,
      Target.Classification = tgt_vec,
      Table.Name            = label_vec,
      URI                   = corr_vec,
      stringsAsFactors      = FALSE
    )
    
    # 5) Compute Prefix (best effort) by joining with 'schemes' if available
    out$Prefix <- ""
    
    if (is.data.frame(schemes) && nrow(schemes)) {
      schemes_nms_lc   <- tolower(names(schemes))
      cand_prefix_cols <- c("prefix", "catalogprefix", "token", "namespaceprefix")
      prefix_col_idx   <- which(schemes_nms_lc %in% cand_prefix_cols)
      prefix_col       <- if (length(prefix_col_idx)) names(schemes)[prefix_col_idx[1]] else NULL
      
      candidate_uri_cols <- c("SchemeURI", "ConceptSchemeURI", "URI", "SchemeUri", "ConceptSchemeUri")
      scheme_uri_col_idx <- which(tolower(names(schemes)) %in% tolower(candidate_uri_cols))
      scheme_uri_col     <- if (length(scheme_uri_col_idx)) names(schemes)[scheme_uri_col_idx[1]] else NULL
      
      if (!is.null(prefix_col) && !is.null(scheme_uri_col)) {
        map_df <- unique(schemes[, c(scheme_uri_col, prefix_col), drop = FALSE])
        names(map_df) <- c("SchemeURI", "SchemePrefix")
        map_df <- map_df[!duplicated(map_df$SchemeURI), , drop = FALSE]
        
        src_df <- merge(
          data.frame(SchemeURI = out$Source.Classification, stringsAsFactors = FALSE),
          map_df, by = "SchemeURI", all.x = TRUE
        )
        tgt_df <- merge(
          data.frame(SchemeURI = out$Target.Classification, stringsAsFactors = FALSE),
          map_df, by = "SchemeURI", all.x = TRUE
        )
        if (!nrow(src_df)) src_df <- data.frame(SchemePrefix = rep(NA_character_, nrow(out)))
        if (!nrow(tgt_df)) tgt_df <- data.frame(SchemePrefix = rep(NA_character_, nrow(out)))
        
        choose_pref <- function(a, b) {
          a <- ifelse(is.na(a), "", as.character(a))
          b <- ifelse(is.na(b), "", as.character(b))
          if (nzchar(a) && nzchar(b)) {
            if (identical(a, b)) a else ""
          } else if (nzchar(a)) {
            a
          } else if (nzchar(b)) {
            b
          } else {
            ""
          }
        }
        out$Prefix <- mapply(choose_pref, src_df$SchemePrefix, tgt_df$SchemePrefix, USE.NAMES = FALSE)
      }
    }
    
    # 6) Fallbacks: derive Prefix from URI, then from ID
    if (any(!nzchar(out$Prefix) & nzchar(out$URI))) {
      uri_to_token <- function(u) {
        u <- as.character(u); if (!nzchar(u)) return("")
        u <- sub("/+$", "", u)
        m <- regmatches(u, regexec("^https?://[^/]+/(.+)$", u))[[1]]
        if (length(m) >= 2) {
          parts <- strsplit(m[2], "/", fixed = TRUE)[[1]]
          if (length(parts) >= 2) {
            token <- parts[length(parts) - 1]
            token <- gsub("[^A-Za-z0-9_]", "", token)
            tolower(token)
          } else ""
        } else ""
      }
      idx <- which(!nzchar(out$Prefix) & nzchar(out$URI))
      if (length(idx)) out$Prefix[idx] <- vapply(out$URI[idx], uri_to_token, character(1))
    }
    
    if (any(!nzchar(out$Prefix) & nzchar(out$ID))) {
      id_to_token <- function(id) {
        id <- as.character(id); if (!nzchar(id)) return("")
        m <- regexpr("[-_]", id)
        token <- if (m > 0) substr(id, 1, m - 1) else id
        token <- gsub("[^A-Za-z0-9_]", "", token)
        tolower(token)
      }
      idx <- which(!nzchar(out$Prefix) & nzchar(out$ID))
      if (length(idx)) out$Prefix[idx] <- vapply(out$ID[idx], id_to_token, character(1))
    }
    
    # Reorder & deduplicate
    out <- out[, c("Prefix", "ID", "Source.Classification", "Target.Classification", "Table.Name", "URI")]
    out <- out[!duplicated(out), , drop = FALSE]
    out
  }
  
  # ---- Single vs multiple endpoints return type ----
  if (length(eps) > 1L) {
    res <- setNames(vector("list", length(eps)), eps)
    for (i in seq_along(eps)) res[[i]] <- run_one(eps[[i]])
    return(res)
  } else {
    return(run_one(eps[[1]]))
  }
}

# -------------------------------------------------------------------
# Internal SPARQL POST helper (CSV, Virtuoso-compatible)
# -------------------------------------------------------------------
.do_post_csv <- function(endpoint_url, query, timeout_sec = 120L, attempts = 3L) {
  
  resp <- httr::RETRY(
    verb  = "POST",
    url   = endpoint_url,
    times = attempts,
    pause_base = 1,
    pause_cap  = 10,
    terminate_on = c(200, 400, 403, 404, 500, 503),
    httr::accept("text/csv"),
    httr::content_type("application/x-www-form-urlencoded; charset=UTF-8"),
    body = list(
      query   = query,
      format  = "text/csv",
      timeout = 0        # Virtuoso server-side execution hint
    ),
    encode = "form",
    httr::user_agent("correspondenceTables (R)"),
    httr::timeout(timeout_sec)
  )
  
  httr::stop_for_status(resp)
  
  txt <- httr::content(resp, as = "text", encoding = "UTF-8")
  if (nzchar(txt)) txt <- sub("^\ufeff", "", txt)
  
  txt
}


# -------------------------------------------------------------------
# Internal helper: read CSV text returned by SPARQL endpoints
# -------------------------------------------------------------------
.read_sparql_csv <- function(csv_text) {
  
  if (!is.character(csv_text) || !nzchar(csv_text)) {
    return(data.frame())
  }
  
  utils::read.csv(
    text            = csv_text,
    sep             = ",",
    stringsAsFactors = FALSE,
    check.names     = FALSE
  )
}



# -------------------------------------------------------------------
# Internal helpers: detect non-CSV responses from SPARQL endpoints
# -------------------------------------------------------------------

# HTML pages (error pages, redirects, etc.)
is_html <- function(x) {
  if (!is.character(x) || !length(x)) return(FALSE)
  grepl("^\\s*<!DOCTYPE\\s+html|^\\s*<html", x, ignore.case = TRUE)
}

# XML responses (Virtuoso error XML, SPARQL XML output, etc.)
is_xml <- function(x) {
  if (!is.character(x) || !length(x)) return(FALSE)
  grepl("^\\s*<\\?xml|^\\s*<sparql", x, ignore.case = TRUE)
}

#' @title Perform quality control on a classification
#'
#' @description
#' This function performs a series of structural and logical quality-control
#' checks on a hierarchical classification. It verifies the consistency of the
#' hierarchical levels, the presence of parent–child relationships, the
#' uniqueness of labels, the compliance with admissible coding rules, and the
#' sequencing of children within each level.  
#' The output is a list of data frames flagging potential issues.
#'
#' @param classification A \code{data.frame} containing (at least) two columns: the codes
#'           and labels of the classification to be checked. This argument is mandatory.
#'
#' @param lengths A \code{data.frame} with one row per hierarchical level giving the
#'           initial and final positions of the segment of the code referring
#'           to that level. The number of rows implicitly defines the number of hierarchical levels
#'   (\eqn{k}). This argument is mandatory. The column names should be \code{charb} and \code{chare}, if this is not the case, they will be automatically changed and a warning will appear.
#'
#' @param fullHierarchy Logical.  
#'   If \code{FALSE}, the function checks that all positions at levels greater
#'   than 1 have a parent at the level immediately above (no orphans).  
#'   If \code{TRUE}, it additionally checks that positions at levels strictly
#'   lower than \eqn{k} have children in the next level (no childless nodes).
#'
#' @param labelUniqueness Logical.  
#'   When \code{TRUE}, the function checks whether labels are unique at each
#'   hierarchical level. Duplicates are listed in the
#'   \code{QC_duplicatesLabel} table.
#'
#' @param labelHierarchy Logical.  
#'   When \code{TRUE}, the function checks that:
#'   \itemize{
#'     \item single children share the same label as their parent;
#'     \item if a parent shares a label with one of its children, it must be a
#'           single-child parent.
#'   }
#'
#' @param singleChildCode A \code{data.frame} specifying admissible codes for single
#'           children at each hierarchical level.
#'   If not \code{NULL}, the function checks compliance with these codes.
#'
#' @param sequencing A \code{data.frame} defining admissible codes and their expected
#'           order for multiple-child situations at each level.
#'   If not \code{NULL}, the function checks sequencing consistency.
#'
#' @return
#' A list of data frames flagging potential issues in the classification:
#' \itemize{
#'   \item \code{QC_output}: full augmented table with computed levels,
#'         code segments, and all compliance flags;
#'   \item \code{QC_noLevels}: rows for which hierarchical level could not be
#'         determined;
#'   \item \code{QC_orphan}: positions that lack an expected parent;
#'   \item \code{QC_childless}: positions expected to have children but do not;
#'   \item \code{QC_duplicatesLabel}: positions with duplicate labels within
#'         the same level;
#'   \item \code{QC_duplicatesCode}: duplicated codes;
#'   \item \code{QC_singleChildMismatch}: label inconsistencies for
#'         single-children relationships;
#'   \item \code{QC_singleCodeError}: single children with invalid codes (if
#'         \code{singleChildCode} is provided);
#'   \item \code{QC_multipleCodeError}: non-single children with invalid codes
#'         (if \code{sequencing} is provided);
#'   \item \code{QC_gapBefore}: missing expected codes in sequence;
#'   \item \code{QC_lastSibling}: flag for last sibling in sequencing rules.
#' }
#'
#' @examples
#'
#'
#' classification <- read.csv(system.file("extdata/test", "Nace2_long.csv",
#'                                       package = "correspondenceTables"))
#' lengths <- data.frame(charb = c(1,2,3,5), chare = c(1,2,4,5))
#' 
#' Output <- classificationQC(
#'  classification = classification,
#'  lengths = lengths,
#'  fullHierarchy = TRUE,
#'  labelUniqueness = TRUE,
#'  labelHierarchy = TRUE,
#'  singleChildCode = NULL,
#'  sequencing = NULL
#')
#'
#' print(Output$QC_output)
#' print(Output$QC_orphan)
#' print(Output$QC_childless)
#' print(Output$QC_duplicatesLabel)
#' print(Output$QC_duplicatesCode)
#'
#' @export


classificationQC = function(classification, lengths, fullHierarchy = TRUE, labelUniqueness  = TRUE, labelHierarchy = TRUE, singleChildCode = NULL, sequencing = NULL) {
  
  if (!is.data.frame(classification)) {
    stop("The classification must be a data frame.")
  }
  
  classificationName <- colnames(classification)[1]
  
  
  colnames(classification)[1:2] = c("Code", "Label")
  
  if (!is.data.frame(lengths)) {
    stop("The 'lengths' argument must be a data frame.")
  }
  
  expected_headers <- c("charb", "chare")
  if (!all(expected_headers %in% colnames(lengths))) {
    warning("Variable names in 'lenghts' argument do not match expected headers. Renaming.")
    colnames(lengths)[1:2] <- expected_headers
  }
  
  # Check formatting requirements
  if (nrow(lengths) == 0) {
    stop("Lengths file must have at least one row.")
  }
  
  negative_lengths <- which(lengths[,1] < 1 | lengths[,2] < 1)
  if (length(negative_lengths) > 0) {
    stop(paste("Lengths must be strictly positive. Error at row:", negative_lengths))
  }
  
  na_lengths <- which(is.na(lengths[,1]) | is.na(lengths[,2]))
  if (length(na_lengths) > 0) {
    stop(paste("Lengths cannot be missing. Error at row:", na_lengths))
  }
  
  for (i in 1:(nrow(lengths)-1)) {
    if (lengths[i,2] >= lengths[i+1,1]) {
      stop(paste("Sequences should not overlap in the lengths definition. Error at row:", i+1))
    }
  }
  
  ### RULE 1 - Correctness of formatting requirements (lengths file)
  
  ## Create QC_output
  QC_output = classification
  
  # Add hierarchical level column
  QC_output$level <- sapply(classification$Code, function(x) {
    lvl <- which(nchar(x) == lengths$chare)
    if (length(lvl) == 0) return(NA_integer_)
    else return(lvl)
  })
  QC_output$level <- as.integer(QC_output$level)
  
  # Initialize the superior column
  QC_output$Parent = NA
  
  # Add Code and segment columns
  for (i in 1:nrow(lengths)) {
    charb = lengths$charb[i]
    chare = lengths$chare[i]
    
    segment_col_name = paste0("segment", i)
    code_col_name = paste0("Code", i)
    
    segments = character(nrow(QC_output))
    codes = character(nrow(QC_output))
    
    for (j in 1:nrow(QC_output)) {
      is_level_present <- !is.null(QC_output$level[j]) && !is.na(QC_output$level[j])
      is_code_present <- !is.null(QC_output$Code[j]) && !is.na(QC_output$Code[j])
      is_code_long_enough <- is_code_present && nchar(QC_output$Code[j]) >= chare
      is_level_eligible <- is_level_present && QC_output$level[j] >= i
      
      if (is_level_present && is_code_present && is_code_long_enough && is_level_eligible) {
        segments[j] <- substr(QC_output$Code[j], charb, chare)
        codes[j] <- substr(QC_output$Code[j], 1, chare)
      } else {
        segments[j] <- NA
        codes[j] <- NA
      }
    }
    
    
    QC_output[, segment_col_name] = segments
    QC_output[, code_col_name] = codes
  }
  
  # Add the 'superior' column based on the current level
  for (i in 1:nrow(QC_output)) {
    if (!is.na(QC_output$level[i]) && QC_output$level[i] > 1) {
      superior_col_name = paste0("Code", (QC_output$level[i] - 1))
      QC_output$Parent[i] = QC_output[i, superior_col_name]
    } else {
      QC_output$Parent[i] = NA
    }
  }
  
  ## RULE 2 - Compliance with formatting requirements (lengths file)
  na_level = which(is.na(QC_output$level))
  QC_noLevels = QC_output[na_level, ]
  
  if (length(na_level) > 0) {
    warning("Some codes have no specified level. There might be possible errors in the classification or in the length file (see 'QC_noLevels').")
  }
  
  
  ## RULE 3 - Uniqueness of codes
  dup =  duplicated(classification[,1])
  QC_output$duplicateCode = as.numeric(dup)
  if (any(dup)) {
    class_dup <- classification[dup,]
    errow_row <- which(dup)
    stop("Codes in classification file must be unique (see 'QC_duplicatesCode').")
  }
  QC_duplicatesCode = QC_output[which(QC_output$duplicateCode == 1), ]
  
  ## RULE 4 -	Fullness of hierarchy
  if (!is.null(fullHierarchy) && fullHierarchy != FALSE){
    QC_output$orphan = rep(NA, nrow(QC_output))
    
    for (k in nrow(lengths):2) {
      QC_output$exp_parents = rep(NA, nrow(QC_output))
      QC_output$exp_parents[which(QC_output$level == k)] = substr(QC_output[which(QC_output$level == k), paste0("Code", k)], 1, lengths[k-1,2])
      o_code = which(QC_output$exp_parents %in% QC_output[which(QC_output$level == k-1), paste0("Code", k-1)])
      QC_output$orphan[intersect(which(QC_output$level == k), o_code)] = 0
      noo_code = which(!QC_output$exp_parents %in% QC_output[which(QC_output$level == k-1), paste0("Code", k-1)])
      QC_output$orphan[intersect(which(QC_output$level == k), noo_code)] = 1
      #QC_output$orphan[which(is.na(QC_output$exp_parents))] = NA
    }
    QC_output = QC_output[, -which(colnames(QC_output) == "exp_parents")]
    
    #identify orphans
    orphan = which(QC_output$orphan == 1)
    if (length(orphan) > 0) {
      warning("Some codes at a lower level than 1 have no parents at higher levels ('see QC_orphan').")
    }
    
    QC_orphan = QC_output[orphan, ]
    
    #childless - if (fullHierarchy == TRUE)
    
    QC_output$childless = rep(NA, nrow(QC_output))
    
    for (k in 2:nrow(lengths)) {
      exp_parents = substr(QC_output[which(QC_output$level == k), paste0("Code", k)], 1, lengths[k-1,2])
      QC_output$childless[which(QC_output[, paste0("Code", k-1)] %in% exp_parents & QC_output$level == k-1)] = 0
      QC_output$childless[which(!QC_output[, paste0("Code", k-1)] %in% exp_parents & QC_output$level == k-1)] = 1
    }
    
    #identify childless
    childless = which(QC_output$childless == 1)
    if (length(childless) > 0) {
      warning(paste("Some codes at a higher level than ", nrow(lengths) ," have no children at lower levels ('see QC_childless')."))
    }
    QC_childless = QC_output[childless, ]
  }else{
    cat("FullHierarchy is NULL so no treatment")
  }
  ###  RULE 5 - Uniqueness of labels
  if (!is.null(labelUniqueness) && labelUniqueness != FALSE) {
    QC_output$duplicateLabel = 0
    
    # Get unique levels
    unique_levels <- unique(QC_output$level)
    
    # Iterate over each unique level
    for (le in unique_levels) {
      level_data = QC_output[QC_output$level == le,]
      
      if (nrow(level_data) != length(unique(level_data$Label))) {
        # There are duplicates, mark them in the QC output column
        # The outcome of the test should be reported in a new ‘QC output’ column (duplicateLabel) assuming the value 1 for positions involved in duplicates (0 otherwise).
        duplicate_labels = level_data$Label[duplicated(level_data$Label)]
        ## Here we add 1 if we have the same label.
        QC_output$duplicateLabel[QC_output$level == le & QC_output$Label %in% duplicate_labels] = 1
      }
    }
    
    # Identify duplicates
    duplicatesLabel = which(QC_output$duplicateLabel == 1)
    if (length(duplicatesLabel) > 0) {
      warning(paste("Some codes at the same hierarchical level have the same labels (see 'QC_duplicatesLabel')."))
    }
    QC_duplicatesLabel = QC_output[duplicatesLabel,]
  }
  
  ## Prepare label_nocode globally for reuse (labelHierarchy, singleChildCode, etc.)
  label_nocode <- gsub("(\\d)|\\.", "", QC_output$Label)
  label_nocode <- mapply(function(x, y) sub(x, "", y), QC_output$Code, label_nocode)
  label_nocode <- tolower(stringr::str_squish(label_nocode))
  ## RULE 6 - Hierarchical label dependencies
  if (!is.null(labelHierarchy) && labelHierarchy != FALSE){
    QC_output$singleChildMismatch = 0
    
    for (k in 1:(nrow(lengths)-1)) {
      
      #select only parents with children
      parents_k = QC_output$Code[which(QC_output$level == k)]
      #list the no. of child for each parent then select parents with single child / or multiple children
      child_ls = sapply(unique(parents_k), function(x) length(unique(na.omit(QC_output[which(QC_output[[paste0("Code", k)]] == x),  paste0("Code", k+1)]))))
      code_singlechild = QC_output[which(QC_output[[paste0("Code", k)]] %in% names(which(child_ls == 1)) & !is.na(QC_output[[paste0("Code", k+1)]]) & QC_output$level == k +1), c(paste0("Code", k), paste0("Code", k+1))]
      code_multichild = QC_output[which(QC_output[[paste0("Code", k)]] %in% names(which(child_ls > 1)) & !is.na(QC_output[[paste0("Code", k+1)]]) & QC_output$level == k +1), c(paste0("Code", k), paste0("Code", k+1))]
      
      #check if single child have different labels from their parents (=1)
      if (nrow(code_singlechild) != 0){
        for (c in 1:nrow(code_singlechild)){
          row_parent = which(QC_output$Code == as.character(code_singlechild[c,1]))
          label_parent = label_nocode[row_parent]
          row_child = which(QC_output$Code == as.character(code_singlechild[c,2]))
          label_child = label_nocode[row_child]
          if (label_parent != label_child) {
            QC_output$singleChildMismatch[row_child] = 1
          } else if (length(row_child) > 1) {
            QC_output$singleChildMismatch[row_child] = 9
          } else {
            QC_output$singleChildMismatch[row_child] = 0
          }
          
          
        }
      }
    }
    #identify mismatches -
    singleChildMismatch = which(QC_output$singleChildMismatch != 0)
    if (length(singleChildMismatch) > 0) {
      warning(paste("Some single child has a different label from their parent or some multiple child has the same label as their parent (see 'QC_singleChildMismatch')."))
    }
    QC_singleChildMismatch = QC_output[singleChildMismatch,]
    
  }else{
    cat("labelHierarchy is NULL so no treatment")
  }
  
  ## RULE 7 - Single child code compliance
  
  if (is.null(singleChildCode)) {
    
    QC_output$singleCodeError   <- NA_integer_
    QC_output$multipleCodeError <- NA_integer_
  } else {
    if (!is.data.frame(singleChildCode)) {
      stop("The singleChildCode argument must be a data frame.")
    }
    
    expected_headers <- c("level", "singleCode", "multipleCode")
    if (!all(expected_headers %in% colnames(singleChildCode))) {
      warning("Variable names do not match expected headers. Renaming to 'level', 'singleCode', 'multipleCode'.")
      colnames(singleChildCode)[1:3] <- expected_headers
    }
    
    QC_output$singleCodeError   <- 0
    QC_output$multipleCodeError <- 0
    
    for (k in 1:(nrow(lengths) - 1)) {
      segment_kplus1 <- na.omit(QC_output[[paste0("segment", k + 1)]])
      if (length(segment_kplus1) == 0) next
      
      if (length(unique(nchar(segment_kplus1))) > 1) {
        QC_output$singleCodeError[which(nchar(QC_output[[paste0("segment", k + 1)]]) > 1)]   <- NA
        QC_output$multipleCodeError[which(nchar(QC_output[[paste0("segment", k + 1)]]) > 1)] <- NA
        next
      }
      
      parents_k <- QC_output$Code[QC_output$level == k]
      child_ls <- sapply(unique(parents_k), function(x)
        length(unique(na.omit(QC_output[QC_output[[paste0("Code", k)]] == x, paste0("Code", k + 1)]))))
      
      code_singlechild <- QC_output[
        QC_output[[paste0("Code", k)]] %in% names(which(child_ls == 1)) &
          !is.na(QC_output[[paste0("Code", k + 1)]]) &
          QC_output$level == k + 1,
        c(paste0("Code", k), paste0("Code", k + 1))
      ]
      
      code_multichild <- QC_output[
        QC_output[[paste0("Code", k)]] %in% names(which(child_ls > 1)) &
          !is.na(QC_output[[paste0("Code", k + 1)]]) &
          QC_output$level == k + 1,
        c(paste0("Code", k), paste0("Code", k + 1))
      ]
      
      rule_row <- singleChildCode[singleChildCode$level == k, , drop = FALSE]
      single <- as.character(rule_row$singleCode)
      multi  <- as.character(rule_row$multipleCode)
      
      if (nrow(code_singlechild) != 0 && length(single) > 0) {
        for (c in 1:nrow(code_singlechild)) {
          row_child <- which(QC_output$Code == as.character(code_singlechild[c, 2]))
          end_code  <- stringr::str_sub(QC_output$Code[row_child],
                                        nchar(QC_output$Code[row_child]),
                                        nchar(QC_output$Code[row_child]))
          if (!end_code %in% single) {
            QC_output$singleCodeError[row_child] <- 1
          }
        }
      }
      
      if (nrow(code_multichild) != 0 && length(multi) > 0) {
        for (m in 1:nrow(code_multichild)) {
          row_child <- which(QC_output$Code == as.character(code_multichild[m, 2]))
          end_code  <- stringr::str_sub(QC_output$Code[row_child],
                                        nchar(QC_output$Code[row_child]),
                                        nchar(QC_output$Code[row_child]))
          if (!end_code %in% unlist(strsplit(multi, ""))) {
            QC_output$multipleCodeError[row_child] <- 1
          }
        }
      }
    }
    
    singleCodeError <- which(QC_output$singleCodeError != 0)
    if (length(singleCodeError) > 0) {
      warning("Some single children been wrongly coded (see 'QC_singleCodeError').")
    }
    QC_singleCodeError <- QC_output[singleCodeError, ]
    
    multipleCodeError <- which(QC_output$multipleCodeError != 0)
    if (length(multipleCodeError) > 0) {
      warning("Some multiple children been wrongly coded (see 'QC_multipleCodeError').")
    }
    QC_multipleCodeError <- QC_output[multipleCodeError, ]
  }
  
  ## RULE 8 - Sequencing of codes
  if (is.null(sequencing)) {
    QC_output$gapBefore  <- NA_integer_
    QC_output$lastSibling <- NA_integer_
  } else {
    if (!is.data.frame(sequencing)) {
      stop("The sequencing argument must be a data frame.")
    }
    
    expected_headers <- c("level", "multipleCode")
    if (!all(expected_headers %in% colnames(sequencing))) {
      warning("Variable names do not match expected headers. Renaming to 'level', 'multipleCode'.")
      colnames(sequencing)[1:2] <- expected_headers
    }
    
    QC_output$gapBefore  <- 0
    QC_output$lastSibling <- 0
    lengths$level <- seq_len(nrow(lengths))
    levels_to_filter <- unique(sequencing$level)
    lengths2 <- lengths[lengths$level %in% levels_to_filter, ]
    lengths2$level <- NULL
    
    for (k in 1:nrow(lengths2)) {
      segment_kplus1 <- na.omit(QC_output[[paste0("segment", k + 1)]])
      if (length(unique(nchar(segment_kplus1))) > 1) {
        warning(paste0("Sequencing of codes cannot be checked at level ", k + 1, " as segments of code have more than one character."))
        QC_output$gapBefore[which(nchar(QC_output[[paste0("segment", k + 1)]]) > 1)]  <- NA
        QC_output$lastSibling[which(nchar(QC_output[[paste0("segment", k + 1)]]) > 1)] <- NA
        next
      }
      
      parents_k <- QC_output$Code[QC_output$level == k]
      child_ls <- sapply(unique(parents_k), function(x)
        length(unique(na.omit(QC_output[QC_output[[paste0("Code", k)]] == x, paste0("Code", k + 1)]))))
      
      code_multichild <- QC_output[
        QC_output[[paste0("Code", k)]] %in% names(which(child_ls > 1)) &
          !is.na(QC_output[[paste0("Code", k + 1)]]) &
          QC_output$level == k + 1,
        c(paste0("Code", k), paste0("Code", k + 1))
      ]
      
      rule_row <- sequencing[sequencing$level == k, , drop = FALSE]
      multi <- if (nrow(rule_row) > 0) strsplit(as.character(rule_row$multipleCode), "")[[1]] else NULL
      
      if (length(multi) == 0 || nrow(code_multichild) == 0) next
      
      mcode_ls <- sapply(unique(code_multichild[, 1]), function(x)
        code_multichild[, 2][code_multichild[, 1] == x])
      ecode_ls <- lapply(mcode_ls, function(x) stringr::str_sub(x, nchar(x), nchar(x)))
      
      last_dig  <- unlist(lapply(ecode_ls, function(x) which(x == max(x, na.rm = TRUE))[1]))
      last_code <- as.vector(mapply(function(x, y) x[y], mcode_ls, last_dig))
      
      gap_find      <- lapply(ecode_ls, function(x) match(multi, x))
      code_gap      <- lapply(gap_find, function(x) which(is.na(x)) + 1)
      gapbefore_dig <- mapply(function(x, y) na.omit(x[y]), gap_find, code_gap)
      gapbefore_code <- as.vector(unlist(mapply(function(x, y) x[y], mcode_ls, gapbefore_dig)))
      
      QC_output$lastSibling[QC_output$Code %in% last_code]   <- 1
      QC_output$gapBefore[QC_output$Code %in% gapbefore_code] <- 1
      
      row_child <- which(QC_output$level == k & QC_output$multipleCodeError == 1)
      QC_output$gapBefore[row_child] <- 9
    }
    
    gap <- which(QC_output$gapBefore == 1)
    if (length(gap) > 0) {
      warning("There are gaps in the sequencing of multiple children coding (see 'QC_gapBefore').")
    }
    QC_gapBefore  <- QC_output[gap, ]
    lastSibling   <- which(QC_output$lastSibling == 1)
    QC_lastSibling <- QC_output[lastSibling, ]
  }
  
  ## RESULTS
  colnames(QC_output)[1:1] <- classificationName
  
  # Always start with QC_output
  return_ls <- list("QC_output" = QC_output)
  
  # Add optional data frames only if they exist
  if (exists("QC_noLevels")) {
    return_ls[["QC_noLevels"]] <- QC_noLevels
  }
  if (exists("QC_duplicatesCode")) {
    return_ls[["QC_duplicatesCode"]] <- QC_duplicatesCode
  }
  if (exists("QC_duplicatesLabel")) {
    return_ls[["QC_duplicatesLabel"]] <- QC_duplicatesLabel
  }
  if (exists("QC_orphan")) {
    return_ls[["QC_orphan"]] <- QC_orphan
  }
  if (exists("QC_childless")) {
    return_ls[["QC_childless"]] <- QC_childless
  }
  if (exists("QC_singleChildMismatch")) {
    return_ls[["QC_singleChildMismatch"]] <- QC_singleChildMismatch
  }
  if (exists("QC_singleCodeError")) {
    return_ls[["QC_singleCodeError"]] <- QC_singleCodeError
  }
  if (exists("QC_multipleCodeError")) {
    return_ls[["QC_multipleCodeError"]] <- QC_multipleCodeError
  }
  if (exists("QC_gapBefore")) {
    return_ls[["QC_gapBefore"]] <- QC_gapBefore
  }
  if (exists("QC_lastSibling")) {
    return_ls[["QC_lastSibling"]] <- QC_lastSibling
  }
  
  return(return_ls)
  
}

## Helper to read extdata CSVs as data.frames (strings not factors)
read_extdata_csv <- function(fname) {
  utils::read.csv(
    system.file("extdata/test", fname, package = "correspondenceTables"),
    stringsAsFactors = FALSE,
    check.names = FALSE
  )
}

############################################################
# [1] classificationQC() works with real NACE example
############################################################

{
  classification_df <- read_extdata_csv("NACE2.csv")
  lengths_df        <- data.frame(charb = c(1,2,3,5), chare = c(1,2,4,5))
  
  expect_warning(result <- classificationQC(
    classification  = classification_df,
    lengths         = lengths_df,
    fullHierarchy   = TRUE,
    labelUniqueness = TRUE,
    labelHierarchy  = TRUE,
    singleChildCode = NULL,
    sequencing      = NULL
  ))
  
  expect_inherits(result, "list")
  expect_true("QC_output" %in% names(result))
  expect_inherits(result$QC_output, "data.frame")
} 



############################################################
# [2] classification must be a data.frame (reject non-DF)
# (Replaces old "path must be csv" test)
############################################################
{
  lengths_df <- data.frame(charb = c(1,2,3,5), chare = c(1,2,4,5))
  
  expect_error(
    classificationQC(
      classification  = "not-a-data-frame",
      lengths         = lengths_df,
      fullHierarchy   = FALSE,
      labelUniqueness = FALSE,
      labelHierarchy  = FALSE
    ),
    "The classification must be a data frame\\."
  )
}


############################################################
# [3] lengths must be a data.frame (reject NA / non-DF)
# (Replaces old NA/wrong-extension tests)
############################################################
{
  classification_df <- read_extdata_csv("NACE2.csv")
  
  expect_error(
    classificationQC(
      classification  = classification_df,
      lengths         = NA,
      fullHierarchy   = FALSE,
      labelUniqueness = FALSE,
      labelHierarchy  = FALSE
    ),
    "The 'lengths' argument must be a data frame\\."
  )
  
  expect_error(
    classificationQC(
      classification  = classification_df,
      lengths         = "not-a-data-frame",
      fullHierarchy   = FALSE,
      labelUniqueness = FALSE,
      labelHierarchy  = FALSE
    ),
    "The 'lengths' argument must be a data frame\\."
  )
}

# Tests from 5 to 9 momentarily skipped as we are missing the datasets

############################################################
# [5] error on duplicate codes
############################################################
# {
#   classification_df <- read_extdata_csv("nace_duplicate_code.csv")
#   lengths_df        <- data.frame(charb = c(1,2,3,5), chare = c(1,2,4,5))
#   
#   expect_error(
#     classificationQC(
#       classification  = classification_df,
#       lengths         = lengths_df,
#       fullHierarchy   = FALSE,
#       labelUniqueness = FALSE,
#       labelHierarchy  = FALSE
#     ),
#     "Codes in classification file must be unique"
#   )
# }


############################################################
# [6] error on overlapping length segments
############################################################
# {
#   classification_df <- read_extdata_csv("NACE2.csv")
#   lengths_df        <- read_extdata_csv("nace_lengths_overlap.csv")
#   lengths_df        <- read.csv(system.file("extdata", "nace_lengths_overlap.csv", package = "correspondenceTables"))
#   
#   expect_error(
#     classificationQC(
#       classification  = classification_df,
#       lengths         = lengths_df,
#       fullHierarchy   = FALSE,
#       labelUniqueness = FALSE,
#       labelHierarchy  = FALSE
#     ),
#     "Sequences should not overlap in the lengths definition"
#   )
# }


############################################################
# [7] warn on duplicated labels at same level
############################################################
# {
#   classification_df <- read_extdata_csv("nace_duplicate_label.csv")
#   lengths_df        <- read_extdata_csv("lenghtsNace.csv")
#   lengths_df        <- data.frame(charb = c(1,2,3,5), chare = c(1,2,4,5))
#   
#   expect_warning(
#     classificationQC(
#       classification  = classification_df,
#       lengths         = lengths_df,
#       fullHierarchy   = FALSE,
#       labelUniqueness = TRUE,
#       labelHierarchy  = FALSE
#     ),
#     "same labels.*QC_duplicatesLabel"
#   )
# }


############################################################
# [8] warn on label hierarchy violations
############################################################
# {
#   classification_df <- read_extdata_csv("nace_label_hierarchy_test.csv")
#   lengths_df        <- read_extdata_csv("lenghtsNace.csv")
# lengths_df        <- data.frame(charb = c(1,2,3,5), chare = c(1,2,4,5))
#   
#   expect_warning(
#     classificationQC(
#       classification  = classification_df,
#       lengths         = lengths_df,
#       fullHierarchy   = FALSE,
#       labelUniqueness = FALSE,
#       labelHierarchy  = TRUE
#     ),
#     "same label as their parent.*QC_singleChildMismatch"
#   )
# }


############################################################
# [9] singleChildCode rule violations are reported
############################################################
# {
#   classification_df  <- read_extdata_csv("test_singleChildCode_classification.csv")
#   #lengths_df         <- read_extdata_csv("lenghtsNace.csv")
#   lengths_df        <- data.frame(charb = c(1,2,3,5), chare = c(1,2,4,5))
#   singleChildCode_df <- read_extdata_csv("test_singleChildCode_rules.csv")
#   
#   result <- suppressWarnings(
#     classificationQC(
#       classification  = classification_df,
#       lengths         = lengths_df,
#       fullHierarchy   = FALSE,
#       labelUniqueness = FALSE,
#       labelHierarchy  = FALSE,
#       singleChildCode = singleChildCode_df
#     )
#   )
#   
#   expect_true("QC_singleCodeError"   %in% names(result))
#   expect_true("QC_multipleCodeError" %in% names(result))
#   expect_true(nrow(result$QC_singleCodeError)   > 0)
#   expect_true(nrow(result$QC_multipleCodeError) > 0)
# }
# 

############################################################
# [10] Works with in-memory data.frames (minimal example)
############################################################
{
  classification_df <- data.frame(
    Code  = c("A", "A01", "A02", "B", "B01"),
    Label = c("Agriculture", "Crops", "Animals", "Mining", "Coal"),
    stringsAsFactors = FALSE
  )
  
  lengths_df <- data.frame(
    charb = c(1, 2),
    chare = c(1, 3)
  )
  
  expect_warning(
    result <- classificationQC(
      classification  = classification_df,
      lengths         = lengths_df,
      fullHierarchy   = TRUE,
      labelUniqueness = TRUE,
      labelHierarchy  = TRUE
    ) )
  
  expect_inherits(result, "list")
  expect_inherits(result$QC_output, "data.frame")
  expect_true(all(c("Code", "Label", "level") %in% names(result$QC_output)))
}


############################################################
# [11] Works with in-memory singleChildCode & sequencing
############################################################
{
  classification_df <- data.frame(
    Code  = c("A", "A01", "A02", "B", "B01"),
    Label = c("Agriculture", "Crops", "Animals", "Mining", "Coal"),
    stringsAsFactors = FALSE
  )
  
  lengths_df <- data.frame(
    charb = c(1, 2),
    chare = c(1, 3)
  )
  
  singleChildCode_df <- data.frame(
    level       = 2,
    singleCode  = "1",
    multipleCode = "2,3,4",
    stringsAsFactors = FALSE
  )
  
  sequencing_df <- data.frame(
    level        = 2,
    multipleCode = "1,2,3",
    stringsAsFactors = FALSE
  )
  
  result <- classificationQC(
    classification  = classification_df,
    lengths         = lengths_df,
    singleChildCode = singleChildCode_df,
    sequencing      = sequencing_df,
    fullHierarchy   = TRUE,
    labelUniqueness = FALSE,
    labelHierarchy  = FALSE
  )
  
  
  expect_true(all(c("singleCodeError", "gapBefore") %in% names(result$QC_output)))
}


############################################################
# [12] Structure sanity: always has QC_output; optional frames appear only if triggered
# (Safer than asserting an exact full list of names)
############################################################
{
  classification_df <- read_extdata_csv("NACE2.csv")
  lengths_df        <- data.frame(charb = c(1,2,3,5), chare = c(1,2,4,5))
  
  expect_warning(
    result <- classificationQC(
      classification  = classification_df,
      lengths         = lengths_df,
      fullHierarchy   = TRUE,
      labelUniqueness = TRUE,
      labelHierarchy  = TRUE
    )
  )
  
  expect_true("QC_output" %in% names(result))
  expect_inherits(result$QC_output, "data.frame")
  expect_true(all(c("Label", "level") %in% names(result$QC_output)))
  
  # Optional artifacts: if present, they must be data.frames with sensible columns
  optional <- intersect(
    names(result),
    c("QC_noLevels","QC_duplicatesCode","QC_duplicatesLabel",
      "QC_orphan","QC_childless","QC_singleChildMismatch",
      "QC_singleCodeError","QC_multipleCodeError",
      "QC_gapBefore","QC_lastSibling")
  )
  
  for (nm in optional) {
    expect_inherits(result[[nm]], "data.frame")
  }
}

#' @importFrom data.table fread fwrite
#' @importFrom stats aggregate
#' @importFrom stats na.omit setNames
#' @importFrom utils capture.output
#' @importFrom utils menu read.csv tail write.csv txtProgressBar setTxtProgressBar

.onLoad <- function(libname, pkgname) {
  if (getRversion() >= "2.15.1") {
    utils::globalVariables(c(
      "Acode", "Alevel", "Asuperior",
      "Bcode", "Blevel", "Bsuperior"
    ))
  }
  options(download.file.method = "libcurl")
  options(url.method = "libcurl")
}

#' Internal helper to derive code segment boundaries by level
#'
#' Computes, for a given classification, the character positions corresponding
#' to each hierarchical level within the compact code (i.e., without dots and
#' spaces). Used internally by higher-level routines; not part of the public API.
#'
#' @param endpoint Character; SPARQL endpoint (e.g., "CELLAR", "FAO").
#' @param prefix Character; catalogue prefix identifying the classification
#'   (e.g., "nace2", "cn2022").
#' @param conceptScheme Character; concept scheme identifier used by the SPARQL calls.
#' @param correction Logical; if TRUE (default), applies a small set of
#'   classification-specific corrections used in production.
#'
#' @return A data.frame with two columns:
#' \itemize{
#'   \item \code{charb}: starting character position (1-based) of the segment
#'         corresponding to each hierarchical level (in the compact code),
#'   \item \code{chare}: ending character position (1-based) of that segment.
#' }
#'
#' @keywords internal
#' @noRd
lengthsFile <- function(endpoint, prefix, conceptScheme, correction = TRUE) {
  
  # ---- Normalize and validate inputs ---------------------------------------
  endpoint <- .validate_endpoints(endpoint)  # accepts "CELLAR" / "FAO" (and would throw on invalid)
  if (identical(endpoint, "ALL") || length(endpoint) != 1L) {
    stop("'endpoint' must be a single endpoint (e.g., 'CELLAR' or 'FAO'), not 'ALL' or a vector.",
         call. = FALSE)
  }
  prefix <- tolower(trimws(prefix))
  
  if (!is.logical(correction) || length(correction) != 1L || is.na(correction)) {
    stop("'correction' must be a single logical value (TRUE or FALSE).", call. = FALSE)
  }
  
  # ---- 1) Retrieve level metadata ------------------------------------------
  level_dt <- dataStructure(endpoint, prefix, conceptScheme)
  
  # Apply small, classification-specific ordering/removal corrections used in production
  if (isTRUE(correction)) {
    # PRODCOM: reorder two highest levels in some vintages
    if (prefix %in% c("prodcom2019", "prodcom2021", "prodcom2022")) {
      # Expecting at least 3 rows; guard with drop=FALSE
      idx <- c(2, 1, 3)
      idx <- idx[idx <= nrow(level_dt)]
      if (length(idx)) level_dt <- level_dt[idx, , drop = FALSE]
    }
    
    # CN: drop the first level
    if (prefix %in% c("cn2017", "cn2018", "cn2019", "cn2020", "cn2021", "cn2022", "cn2023")) {
      if (nrow(level_dt) >= 2) level_dt <- level_dt[-1, , drop = FALSE]
    }
    
    # CPA 2.1: custom level ordering observed in some providers
    if (prefix %in% c("cpa21")) {
      # Make sure indices are within bounds
      idx <- c(1, 2, 3, 6, 4, 5)
      idx <- idx[idx <= nrow(level_dt)]
      if (length(idx)) level_dt <- level_dt[idx, , drop = FALSE]
    }
  }
  
  # For other classifications, ensure levels are in ascending order
  # based on the second column (assumed hierarchy depth indicator).
  special_prefix <- c(
    "prodcom2019", "prodcom2021", "prodcom2022",
    "cn2017", "cn2018", "cn2019", "cn2020", "cn2021", "cn2022", "cn2023",
    "cpa21"
  )
  if (!(prefix %in% special_prefix)) {
    ord <- order(suppressWarnings(as.numeric(level_dt[, 2])))
    if (length(ord)) level_dt <- level_dt[ord, , drop = FALSE]
  }
  
  nL <- nrow(level_dt)
  if (nL == 0L) {
    stop("dataStructure() returned no levels for this classification.", call. = FALSE)
  }
  
  # Assume the second column contains the level identifier (provider-dependent)
  level_ids <- level_dt[, 2]
  
  # Storage for total compact-length per level, and segment bounds
  total_len <- rep(NA_integer_, nL)
  charb     <- rep(NA_integer_, nL)
  chare     <- rep(NA_integer_, nL)
  
  # ---- 2) Iterate over levels ----------------------------------------------
  for (l in seq_len(nL)) {
    
    # Retrieve codes for a specific level
    cl <- retrieveClassificationTable(
      endpoint      = endpoint,
      prefix        = prefix,
      conceptScheme = conceptScheme,
      level         = level_ids[l]
    )
    dt <- cl$ClassificationTable
    
    # ---- Classification-specific corrections (kept for backward compatibility)
    if (isTRUE(correction)) {
      
      # ECOICOP: remove ".0" for 10, 11, 12 at level 1 only
      if (prefix %in% c("ecoicop") && l == 1L) {
        idx <- which(dt[, 1] %in% c("10.0", "11.0", "12.0"))
        if (length(idx) > 0L) dt[idx, 1] <- c("10", "11", "12")
      }
      
      # PRODCOM: remove rare odd codes at level 1
      if (prefix %in% c("prodcom2019", "prodcom2021", "prodcom2022", "prodcom2023", "prodcom2024", "prodcom2025", "prodcom2026") && l == 1L) {
        bad <- which(dt[, 1] %in% c("00.99.t", "00.99.z"))
        if (length(bad) > 0L) dt <- dt[-bad, , drop = FALSE]
      }
      
      # NACE / NACE 2.1 / CPA / ISIC: prepend "A" for levels > 1 (provider quirk)
      if (prefix %in% c("nace2", "nace21", "cpa21", "isicrev4") && l > 1L) {
        dt[, 1] <- paste0("A", dt[, 1])
      }
      
      # ICC_v11: add leading zero at level 2 only
      if (prefix %in% c("icc_v11") && l == 2L) {
        dt[, 1] <- sprintf("%.2f", dt[, 1])
      }
    }
    
    # ---- 3) Clean codes: remove spaces and dots to get compact form ----------
    codes       <- as.character(dt[, 1])
    codes_clean <- gsub("[ .]", "", codes)
    len_unique  <- unique(nchar(codes_clean))
    
    if (length(len_unique) != 1L || is.na(len_unique[1L])) {
      # Inconsistent code length for this level → cannot compute segment bounds reliably
      total_len[l] <- NA_integer_
      charb[l]     <- NA_integer_
      chare[l]     <- NA_integer_
      
      warning(
        "Inconsistent code lengths at level ", level_ids[l],
        " for prefix ", prefix, ". Segment boundaries may be unreliable for this level."
      )
      
    } else {
      total_len[l] <- len_unique[1L]
      
      if (l == 1L) {
        # First level: entire code is the first segment
        charb[l] <- 1L
        chare[l] <- total_len[l]
      } else {
        # Subsequent levels: require strictly increasing compact length
        if (is.na(total_len[l - 1L]) || total_len[l] <= total_len[l - 1L]) {
          charb[l] <- NA_integer_
          chare[l] <- NA_integer_
          
          warning(
            "Non-increasing or invalid compact length from level ", level_ids[l - 1L],
            " to level ", level_ids[l], " for prefix ", prefix,
            ". Segment boundaries may be unreliable for this level."
          )
          
        } else {
          seg_len  <- total_len[l] - total_len[l - 1L]
          charb[l] <- chare[l - 1L] + 1L
          chare[l] <- chare[l - 1L] + seg_len
        }
      }
    }
  }
  
  # ---- 3) Final table -------------------------------------------------------
  lengths <- data.frame(
    charb = charb,
    chare = chare,
    stringsAsFactors = FALSE
  )
  
  if (any(is.na(lengths$charb) | is.na(lengths$chare))) {
    warning(
      "Some levels have missing segment boundaries. ",
      "Please verify the classification data; the result may not be fully reliable."
    )
  }
  
  if (identical(correction, FALSE)) {
    warning(
      "Classification-specific corrections are disabled (correction = FALSE). ",
      "Results may differ from production behavior."
    )
  }
  
  lengths
}

#' @title Aggregate values from classification A to classification B
#'
#' @description
#' Aggregates numeric values from a source classification A into a target
#' classification B using a correspondence table AB.  
#' Each value in A is redistributed to one or more B codes according to weights
#' in AB (if available), or split equally when no valid weights exist.
#'
#' @details
#' This function is used when data expressed in one classification must be
#' converted into another classification using an A→B correspondence table.
#'
#' The workflow is:
#' \enumerate{
#'   \item Detect and standardize column names in \code{AB}, \code{A} and \code{B}
#'         (case-insensitive).
#'   \item For each code in A, determine all corresponding B codes.
#'   \item Apply proportional allocation:
#'         \itemize{
#'           \item If \code{AB} contains a numeric weight column, weights
#'                 are normalized per source code.
#'           \item If weights are missing, invalid, or sum to zero, values are
#'                 split equally across all mapped B codes.
#'         }
#'   \item Values are multiplied by the resulting weights and aggregated by B code.
#'   \item If \code{B} is provided, the result is aligned to its list of codes
#'         (unmatched codes receive 0).
#' }
#'
#' This function is appropriate for tasks such as converting statistical datasets
#' from NACE to CPA, CPA to CN, CPC to HS, or any situation where values have to
#' be re-expressed according to a different classification system.
#'
#' @param AB A \code{data.frame} containing A→B links, with optional numeric
#'           \code{weight}.
#' @param A  A \code{data.frame} with one (character) code column and one
#'           numeric value column.
#' @param B  Optional \code{data.frame} with a code column defining the output
#'           domain. Additional columns (e.g., labels) are preserved.
#'
#' @return A list with:
#' \itemize{
#'   \item \code{result}: data.frame with aggregated values for each B code
#'         (and any additional columns from \code{B}, if provided),
#'   \item \code{mapping}: standardized A→B mapping used internally,
#'   \item \code{diagnostics}: list with total input value, mapped value,
#'         coverage ratio, and codes in A that could not be mapped.
#' }
#'
#' @examples
#' ## Simple example (equal split)
#' AB <- data.frame(
#'   from_code = c("A1","A1","A2"),
#'   to_code   = c("B1","B2","B1")
#' )
#'
#'print(AB)
#'
#' A_tbl <- data.frame(
#'   code  = c("A1","A2"),
#'   value = c(100, 50)
#' )
#'
#'print(A_tbl)
#'
#' result <- aggregateCorrespondenceTable(AB, A_tbl)
#' 
#' print(result$result) 
#' print(result$mapping)
#'
#' @seealso \code{\link{retrieveCorrespondenceTable}} for downloading AB tables.
#'
#' @export

aggregateCorrespondenceTable <- function(AB, A, B = NULL) {
  # --- helpers ------------------------------------------
  stop_if <- function(cond, msg) if (isTRUE(cond)) stop(msg, call. = FALSE)
  
  trim_chr <- function(x) {
    x <- as.character(x)
    x <- trimws(x)
    x[x == ""] <- NA_character_
    x
  }
  
  # case-insensitive named-extractor for lists or named atomic vectors
  guess_col <- function(df, candidates, want_numeric = FALSE) {
    cols <- names(df)
    # direct match (case-sensitive)
    for (c in candidates) if (nzchar(c) && c %in% cols) return(c)
    # case-insensitive match
    low <- tolower(cols)
    for (c in candidates) {
      idx <- which(low == tolower(c))
      if (length(idx) == 1) return(cols[idx])
    }
    # fallback by type
    if (want_numeric) {
      nums <- cols[vapply(df, is.numeric, logical(1))]
      if (length(nums)) return(nums[1])
    } else {
      chrs <- cols[vapply(df, function(z) !(is.numeric(z) || is.logical(z)), logical(1))]
      if (length(chrs)) return(chrs[1])
    }
    NULL
  }
  
  # --- accept only data.frames ----------------------------------------------
  stop_if(!is.data.frame(AB), "Argument 'AB' must be a data.frame.")
  stop_if(!is.data.frame(A),  "Argument 'A' must be a data.frame.")
  if (!is.null(B)) stop_if(!is.data.frame(B), "Argument 'B' must be a data.frame or NULL.")
  
  # ----- Standardization helpers --------------------------------------------
  # Standardize AB to: code_A, code_B, weight (normalized per code_A)
  
  standardize_AB <- function(AB_df, A_codes = NULL) {
    
    # 1) First try explicit known names (your existing logic)
    A_cand <- c("code_A","from_code","A","from","source","a_code","Acode")
    B_cand <- c("code_B","to_code","B","to","target","dest","b_code","Bcode")
    W_cand <- c("weight","w","share","split","prop","ratio","coef")
    
    cA <- guess_col(AB_df, A_cand, want_numeric = FALSE)
    cB <- guess_col(AB_df, B_cand, want_numeric = FALSE)
    cW <- guess_col(AB_df, W_cand, want_numeric = TRUE)
    
    # 2) If explicit names NOT found, use a smarter heuristic
    if (is.null(cA) || is.null(cB) || identical(cA, cB)) {
      
      nms <- names(AB_df)
      nms_lc <- tolower(nms)
      
      # (a) ignore typical metadata columns from retrieveCorrespondenceTable()
      ignore_pat <- "^(correspondenceid|prefix|uri|url|table\\.name|comment)$|label|include|exclude|broader"
      keep <- !grepl(ignore_pat, nms_lc)
      
      # candidate columns are non-numeric and not ignored
      cand <- nms[keep & vapply(AB_df, function(z) !(is.numeric(z) || is.logical(z)), logical(1))]
      if (length(cand) < 2) {
        stop("Could not detect A/B code columns in 'AB'. Please provide columns like 'from_code'/'to_code' or two code columns.", call. = FALSE)
      }
      
      # (b) If CorrespondenceID exists like "NACE2_CPA21" and columns NACE2 + CPA21 exist, use them
      if ("CorrespondenceID" %in% nms) {
        id <- unique(as.character(AB_df[["CorrespondenceID"]]))
        id <- id[!is.na(id) & nzchar(id)]
        if (length(id)) {
          toks <- strsplit(id[1], "_", fixed = TRUE)[[1]]
          if (length(toks) >= 2 && all(toks[1:2] %in% nms)) {
            cA <- toks[1]
            cB <- toks[2]
          }
        }
      }
      
      # (c) If still not found, choose cA as the column that best matches A_codes
      if (is.null(cA) || is.null(cB) || identical(cA, cB)) {
        score_overlap <- function(col) {
          v <- trimws(as.character(AB_df[[col]]))
          v <- v[!is.na(v) & nzchar(v)]
          if (is.null(A_codes)) return(0L)
          sum(v %in% A_codes)
        }
        
        scores <- vapply(cand, score_overlap, integer(1))
        ord <- order(scores, decreasing = TRUE)
        
        cA <- cand[ord[1]]
        cB <- cand[ord[2]]
        
        # If overlap is useless (all 0), fall back to the first two remaining candidates
        if (all(scores == 0L)) {
          cA <- cand[1]
          cB <- cand[2]
        }
      }
    }
    
    # ---- Now proceed with your existing standardization logic ----
    code_A <- trim_chr(AB_df[[cA]])
    code_B <- trim_chr(AB_df[[cB]])
    keep   <- !is.na(code_A) & !is.na(code_B)
    
    out <- data.frame(code_A = code_A[keep], code_B = code_B[keep], stringsAsFactors = FALSE)
    stop_if(nrow(out) == 0, "Argument 'AB' has no valid records.")
    
    if (!is.null(cW) && is.numeric(AB_df[[cW]])) {
      w <- AB_df[[cW]][keep]
      w[!is.finite(w)] <- NA_real_
      out$weight <- as.numeric(w)
      out <- stats::aggregate(weight ~ code_A + code_B, data = out, FUN = function(z) sum(z, na.rm = TRUE))
      out <- out[, c("code_A","code_B","weight")]
    } else {
      n_per_A <- table(out$code_A)
      out$weight <- 1 / as.numeric(n_per_A[match(out$code_A, names(n_per_A))])
    }
    
    split_idx <- split(seq_len(nrow(out)), out$code_A)  
    for (idx in split_idx) {
      sw <- sum(out$weight[idx], na.rm = TRUE)
      if (isTRUE(sw > 0)) {
        out$weight[idx] <- out$weight[idx] / sw
      } else {
        out$weight[idx] <- 1 / length(idx)
      }
    }
    out
  }
  # Standardize A to: code_A, value
  standardize_A <- function(A_df) {
    A_code_cand <- c("code_A","A","from","source","code","item","prod","id","Acode")
    A_val_cand  <- c("value","val","amount","qty","quantity","count","obs_value","measure","num","n")
    
    cA <- guess_col(A_df, A_code_cand, want_numeric = FALSE)
    cV <- guess_col(A_df, A_val_cand,  want_numeric = TRUE)
    stop_if(is.null(cA), "Could not detect code column in 'A' (e.g., 'code').")
    stop_if(is.null(cV) || !is.numeric(A_df[[cV]]),
            "Could not detect numeric value column in 'A' (e.g., 'value').")
    
    code_A <- trim_chr(A_df[[cA]])
    value  <- as.numeric(A_df[[cV]])
    data.frame(code_A = code_A, value = value, stringsAsFactors = FALSE)
  }
  
  # Standardize B to: code_B (+ extra cols kept)
  standardize_B <- function(B_df) {
    if (is.null(B_df)) return(NULL)
    B_code_cand <- c("code_B","B","to","target","dest","code","id","Bcode")
    cB <- guess_col(B_df, B_code_cand, want_numeric = FALSE)
    stop_if(is.null(cB), "Could not detect code column in 'B' (e.g., 'code').")
    out <- B_df
    names(out)[names(out) == cB] <- "code_B"
    out$code_B <- trim_chr(out$code_B)
    out
  }
  
  # --- Standardize inputs ---------------------------------------------------
  A_std  <- standardize_A(A)
  AB_std <- standardize_AB(AB, A_codes = unique(A_std$code_A))
  B_std  <- standardize_B(B)
  
  # --- Join and proportional allocation ------------------------------------
  joined <- merge(A_std, AB_std, by = "code_A", all.x = TRUE)
  
  mapped_flag <- !is.na(joined$code_B)
  total_value <- sum(A_std$value, na.rm = TRUE)
  
  # multiply and aggregate (treat NA weights as 0)
  w <- joined$weight
  w[is.na(w)] <- 0
  joined$adj_value <- joined$value * w
  
  mapped_value      <- sum(joined$adj_value[mapped_flag], na.rm = TRUE)
  covered_a_codes   <- unique(joined$code_A[mapped_flag])
  is_not_empty_code <- function(x) !is.na(x) & nzchar(x)
  unmapped_a_codes  <- setdiff(unique(A_std$code_A[is_not_empty_code(A_std$code_A)]), covered_a_codes)
  
  # keep only mapped
  joined <- joined[mapped_flag, c("code_B","adj_value"), drop = FALSE]
  
  # aggregate by code_B
  if (nrow(joined)) {
    agg <- stats::aggregate(adj_value ~ code_B, data = joined, FUN = function(x) sum(x, na.rm = TRUE))
    names(agg) <- c("code_B","value")
  } else {
    agg <- data.frame(code_B = character(0), value = numeric(0), stringsAsFactors = FALSE)
  }
  
  # right-join with B domain if provided (preserve B; fill 0)
  if (!is.null(B_std)) {
    agg <- merge(B_std, agg, by = "code_B", all.x = TRUE, sort = FALSE)
    if (!("value" %in% names(agg))) agg$value <- 0
    agg$value[is.na(agg$value)] <- 0
    # put code_B first
    pos <- match("code_B", names(agg))
    ord <- c(pos, setdiff(seq_along(names(agg)), pos))
    agg <- agg[, ord, drop = FALSE]
  }
  
  diagnostics <- list(
    total_value_in_A       = total_value,
    mapped_value_to_B      = mapped_value,
    mapping_coverage_ratio = if (isTRUE(total_value > 0)) mapped_value / total_value else NA_real_,
    unmapped_A_codes       = unmapped_a_codes
  )
  
  list(result = agg, mapping = AB_std, diagnostics = diagnostics)
}

# --- 1) Wrong input: AB dataset with wrong column names ------------------
{
  AB_df <- read.csv(system.file("extdata/test", "ab_data.csv",
                                package = "correspondenceTables"),
                    stringsAsFactors = FALSE, check.names = FALSE)
  A_df  <- read.csv(system.file("extdata/test", "a_data.csv",
                                package = "correspondenceTables"),
                    stringsAsFactors = FALSE, check.names = FALSE)
  B_df  <- read.csv(system.file("extdata/test", "b_data.csv",
                                package = "correspondenceTables"),
                    stringsAsFactors = FALSE, check.names = FALSE)
  
  expect_error(aggregateCorrespondenceTable(AB_df, A_df, B_df))
}

# --- 2) Duplicate AB data: no error; weights normalized per code_A ----------
{
  AB_df <- read.csv(system.file("extdata/test", "ab_data_duplicate.csv",
                                package = "correspondenceTables"),
                    stringsAsFactors = FALSE, check.names = FALSE, col.names = c('from_code','to_code') )
  A_df  <- read.csv(system.file("extdata/test", "a_data.csv",
                                package = "correspondenceTables"),
                    stringsAsFactors = FALSE, check.names = FALSE)
  B_df  <- read.csv(system.file("extdata/test", "b_data.csv",
                                package = "correspondenceTables"),
                    stringsAsFactors = FALSE, check.names = FALSE)
  
  res_try <- try(aggregateCorrespondenceTable(AB_df, A_df, B_df), silent = TRUE)
  expect_false(inherits(res_try, "try-error"))
  
  ACT <- res_try
  expect_true(is.list(ACT))
  expect_true(is.data.frame(ACT$mapping))
  expect_true(all(c("code_A", "code_B", "weight") %in% names(ACT$mapping)))
  
  # Sum of weights per code_A must be (approximately) 1
  wsum <- aggregate(ACT$mapping$weight,
                    by = list(ACT$mapping$code_A),
                    FUN = function(z) sum(z, na.rm = TRUE))
  names(wsum) <- c("code_A", "sum_w")
  expect_true(all(abs(wsum$sum_w - 1) < 1e-12))
}

# --- 3) Empty AB data: expect error ----------------------------------------
{
  AB_df <- read.csv(system.file("extdata/test", "ab_data_empty.csv",
                                package = "correspondenceTables"),
                    stringsAsFactors = FALSE, check.names = FALSE, col.names = c('from_code','to_code') )
  A_df  <- read.csv(system.file("extdata/test", "a_data.csv",
                                package = "correspondenceTables"),
                    stringsAsFactors = FALSE, check.names = FALSE)
  B_df  <- read.csv(system.file("extdata/test", "b_data.csv",
                                package = "correspondenceTables"),
                    stringsAsFactors = FALSE, check.names = FALSE)
  
  expect_error(aggregateCorrespondenceTable(AB_df, A_df, B_df))
}

# --- 4) Missing code columns in AB: expect no error, the function removes it ---------------------------
{
  AB_df <- read.csv(system.file("extdata/test", "ab_data_missingcode.csv",
                                package = "correspondenceTables"),
                    stringsAsFactors = FALSE, check.names = FALSE, col.names = c('from_code','to_code') )
  is.na(AB_df)
  A_df  <- read.csv(system.file("extdata/test", "a_data.csv",
                                package = "correspondenceTables"),
                    stringsAsFactors = FALSE, check.names = FALSE)
  B_df  <- read.csv(system.file("extdata/test", "b_data.csv",
                                package = "correspondenceTables"),
                    stringsAsFactors = FALSE, check.names = FALSE)
  
  res_try <- try(aggregateCorrespondenceTable(AB_df, A_df, B_df), silent = TRUE)
  expect_false(inherits(res_try, "try-error"))
}

# --- 5) Text suppression / non-numeric A$value: expect no error, the function saves the unmatched values ---------------
{
  AB_df <- read.csv(system.file("extdata/test", "ab_data.csv",
                                package = "correspondenceTables"),
                    stringsAsFactors = FALSE, check.names = FALSE, col.names = c('from_code','to_code'))
  A_df  <- read.csv(system.file("extdata/test", "a_data_textsup.csv",
                                package = "correspondenceTables"),
                    stringsAsFactors = FALSE, check.names = FALSE)
  B_df  <- read.csv(system.file("extdata/test", "b_data.csv",
                                package = "correspondenceTables"),
                    stringsAsFactors = FALSE, check.names = FALSE)
  
  res_try <- try(aggregateCorrespondenceTable(AB_df, A_df, B_df), silent = TRUE)
  expect_false(inherits(res_try, "try-error"))
}

# --- 6) Return structure: result/mapping/diagnostics present ----------------
{
  AB_df <- read.csv(system.file("extdata/test", "ab_data.csv",
                                package = "correspondenceTables"),
                    stringsAsFactors = FALSE, check.names = FALSE, col.names = c('from_code','to_code'))
  A_df  <- read.csv(system.file("extdata/test", "a_data.csv",
                                package = "correspondenceTables"),
                    stringsAsFactors = FALSE, check.names = FALSE)
  B_df  <- read.csv(system.file("extdata/test", "b_data.csv",
                                package = "correspondenceTables"),
                    stringsAsFactors = FALSE, check.names = FALSE)
  
  ACT <- aggregateCorrespondenceTable(AB_df, A_df, B_df)
  
  expect_true(is.list(ACT))
  expect_true(is.data.frame(ACT$result))
  expect_true(is.data.frame(ACT$mapping))
  expect_true(is.list(ACT$diagnostics))
  
  expect_true(all(c("total_value_in_A",
                    "mapped_value_to_B",
                    "mapping_coverage_ratio",
                    "unmapped_A_codes") %in% names(ACT$diagnostics)))
}

# --- 7) Alignment to B domain: unmapped B codes get 0 ----------------------
{
  A_df <- data.frame(code = c("A1","A2"),
                     value = c(100, 50),
                     stringsAsFactors = FALSE)
  B_df <- data.frame(code = c("B1","B2"),
                     label = c("one","two"),
                     stringsAsFactors = FALSE)
  AB_df <- data.frame(from_code = "A2",
                      to_code   = "B2",
                      stringsAsFactors = FALSE)
  
  ACT <- aggregateCorrespondenceTable(AB_df, A_df, B_df)
  
  out <- ACT$result
  expect_true(all(c("code_B","label","value") %in% names(out)))
  expect_equal(out$value[out$code_B == "B1"], 0)
  expect_equal(out$value[out$code_B == "B2"], 50)
}

# This function is deprecated and kept only for backward compatibility.
# Users should now use \code{correspondenceTableList()} instead, which provides
# the same functionality with a clearer and more stable interface.

#   # Deprecated: please use correspondenceTableList() instead.
#   x <- correspondenceList("ALL", showQuery = FALSE)
#   str(x)

correspondenceList <- function(endpoint = "ALL", showQuery = FALSE) {
  .Deprecated("correspondenceTableList")
  correspondenceTableList(endpoint = endpoint, showQuery = showQuery)
}

# correspondenceList <- function(endpoint) {
#   
#   # Normaliser l'endpoint
#   endpoint <- toupper(endpoint)
#   
#   # Cas 1 : on veut tout → on rappelle la fonction pour chaque endpoint
#   if (endpoint == "ALL") {
#     res_cellar <- correspondenceList("CELLAR")
#     res_fao    <- correspondenceList("FAO")
#     
#     return(list(
#       CELLAR = res_cellar,
#       FAO    = res_fao
#     ))
#   }
#   
#   # Vérification basique
#   if (!endpoint %in% c("CELLAR", "FAO")) {
#     stop("endpoint must be one of 'CELLAR', 'FAO' or 'ALL'.")
#   }
#   
#   # cycle = 1 pour CELLAR, 2 pour FAO, comme dans ton code
#   if (endpoint == "CELLAR") {
#     cycle <- 1L
#   } else if (endpoint == "FAO") {
#     cycle <- 2L
#   }
#   
#   data <- list()
#   
#   for (j in cycle) {
#     e <- c("CELLAR", "FAO")[j]
#     
#     # Définition du SPARQL endpoint + séparateur
#     if (e == "CELLAR") {
#       source <- "http://publications.europa.eu/webapi/rdf/sparql"
#       sep    <- "_"
#       rm     <- 1:16
#     } else if (e == "FAO") {
#       source <- "https://stats.fao.org/caliper/sparql/AllVocs"
#       sep    <- "-"
#       rm     <- 1:16
#     }
#     
#     ## Préfixes SPARQL
#     prefixlist <- prefixList(e)
#     prefixlist <- as.character(paste(prefixlist, collapse = "\n"))
#     
#     prefixes_loop <- unlist(
#       lapply(
#         strsplit(as.character(prefixList(e)), " "),
#         function(x) x[2]
#       )
#     )
#     prefixes_loop <- prefixes_loop[-rm]
#     
#     data_t <- list()
#     
#     for (i in seq_along(prefixes_loop)) {
#       prefix <- prefixes_loop[i]
#       
#       SPARQL.query <- paste0(
#         prefixlist, "
#        SELECT ?ID_table ?A ?B ?Table ?URL 
# 
#        WHERE {
#              ?s a xkos:Correspondence ;
#                 skos:prefLabel ?Label .
#     
#             BIND (STR(?s) AS ?URL)
#             BIND (STR(?Label) AS ?Table)
#   
#             BIND (STRAFTER(STR(?s), STR(", prefix, ")) AS ?ID_table)
#             BIND (STRAFTER(STR(?ID_table), '", sep, "') AS ?B)
#             BIND (STRBEFORE(STR(?ID_table), '", sep, "') AS ?A)  
#   
#             FILTER (STRLEN(?ID_table) != 0) 
#         }
#       "
#       )
#       
#       # Appel HTTP : on demande explicitement du CSV, et on récupère le texte brut
#       response <- httr::POST(
#         url    = source,
#         body   = list(query = SPARQL.query),
#         encode = "form",
#         httr::accept("text/csv")
#       )
#       
#       csv_txt <- httr::content(response, as = "text", encoding = "UTF-8")
#       
#       if (!nzchar(csv_txt)) {
#         # Réponse vide : data.frame vide avec les colonnes attendues
#         df <- data.frame(
#           ID_table = character(),
#           A        = character(),
#           B        = character(),
#           Table    = character(),
#           URL      = character(),
#           stringsAsFactors = FALSE
#         )
#       } else {
#         # Lecture du CSV renvoyé
#         df <- utils::read.csv(
#           text = csv_txt,
#           stringsAsFactors = FALSE
#         )
#       }
#       
#       # Ajout de la colonne prefix (même logique que ton code)
#       if (nrow(df) == 0) {
#         df <- cbind(prefix = character(), df)
#       } else {
#         df <- cbind(
#           prefix = rep(gsub(":", "", prefix), nrow(df)),
#           df
#         )
#       }
#       
#       data_t[[i]] <- df
#       names(data_t)[i] <- prefix
#     }
#     
#     data[[j]] <- data_t
#     names(data)[j] <- c("CELLAR", "FAO")[j]
#   }
#   
#   return(data)
# }

#' @title List available classification schemes from CELLAR or FAO
#'
#' @description
#' List all classification schemes available on the selected repository
#' (CELLAR, FAO, or both). Returns, for each scheme, its prefix, identifier,
#' URI, English title, and the list of languages detected on its concepts.
#'
#' @param endpoint Character scalar. One of "CELLAR", "FAO" or "ALL".
#'                 "ALL" queries both repositories and returns a named list.
#' @param showQuery Logical; if TRUE, returns also the SPARQL query used.
#'
#' @return
#' If endpoint is "CELLAR" or "FAO" and showQuery = FALSE: a data.frame with
#' columns Prefix, ConceptScheme, URI, Title, Languages.
#'
#' If showQuery = TRUE: a list with "ClassificationList" and "SPARQL.query".
#'
#' If endpoint = "ALL": a named list with $CELLAR and $FAO. If showQuery = TRUE,
#' each element is itself a list (ClassificationList + SPARQL.query).
#'
#' @examples
#' \dontrun{
#' head(classificationList("CELLAR"))
#' res <- classificationList("FAO", showQuery = TRUE)
#' names(res)
#' }
#'
#' @import httr
#' @importFrom utils read.csv
#' @export
classificationList <- function(endpoint = "ALL", showQuery = FALSE) {
  # Validate and normalize endpoint using the internal helper.
  # .expand_endpoints() returns a vector of endpoint names.
  # Example: "ALL" -> c("CELLAR", "FAO")
  endpoints <- .expand_endpoints(endpoint)
  
  # If multiple endpoints were requested, return a named list.
  if (length(endpoints) > 1L) {
    out <- setNames(vector("list", length(endpoints)), endpoints)
    for (ep in endpoints) {
      out[[ep]] <- .ct_classification_list_single(ep, showQuery = showQuery)
    }
    return(out)
  }
  
  # Single endpoint case.
  .ct_classification_list_single(endpoints, showQuery = showQuery)
}

# -------------------------------------------------------------------
# Internal: run the SPARQL query for a single endpoint
# -------------------------------------------------------------------
#' @keywords internal
#' @noRd
.ct_classification_list_single <- function(endpoint, showQuery = FALSE) {
  
  # Resolve the endpoint URL via internal single source of truth
  endpoint_url <- .sparql_endpoint(endpoint)
  
  # Choose the SPARQL query by endpoint
  if (identical(endpoint, "CELLAR")) {
    SPARQL.query <- "
PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
SELECT
  (?scheme AS ?URI)
  (COALESCE(STR(?notation),
            REPLACE(REPLACE(STR(?scheme), '^.*/', ''), '[^A-Za-z]+.*', '')) AS ?Prefix)
  (COALESCE(SAMPLE(STR(?title_en)), '') AS ?Title)
  (GROUP_CONCAT(DISTINCT LANG(?clab); SEPARATOR=', ') AS ?Languages)
WHERE {
  ?scheme a skos:ConceptScheme ;
          skos:prefLabel ?title_en .
  FILTER (LANG(?title_en) = 'en')

  FILTER (
    STRSTARTS(STR(?scheme), 'http://data.europa.eu/') ||
    STRSTARTS(STR(?scheme), 'https://data.europa.eu/')
  )

  ?scheme ?p <http://publications.europa.eu/resource/authority/corporate-body/ESTAT> .

  OPTIONAL { ?scheme skos:notation ?notation }

  ?concept a skos:Concept ;
           skos:inScheme ?scheme ;
           skos:prefLabel ?clab .
  FILTER (LANG(?clab) != '')
}
GROUP BY ?scheme ?notation
ORDER BY ?Prefix
"
  } else if (identical(endpoint, "FAO")) {
    SPARQL.query <- "
PREFIX rdf:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
SELECT
  (?scheme AS ?URI)
  (COALESCE(STR(?notation),
            REPLACE(REPLACE(STR(?scheme), '^.*/', ''), '[^A-Za-z]+.*', '')) AS ?Prefix)
  (COALESCE(SAMPLE(STR(?title_en)), '') AS ?Title)
  (GROUP_CONCAT(DISTINCT LANG(?clab); SEPARATOR=', ') AS ?Languages)
WHERE {
  ?scheme rdf:type skos:ConceptScheme ;
          skos:prefLabel ?title_en .
  FILTER (LANG(?title_en) = 'en')

  OPTIONAL { ?scheme skos:notation ?notation }

  ?concept a skos:Concept ;
           skos:inScheme ?scheme ;
           skos:prefLabel ?clab .
  FILTER (LANG(?clab) != '')
}
GROUP BY ?scheme ?notation
ORDER BY ?Prefix
"
  } else {
    stop("Unsupported endpoint in .ct_classification_list_single().")
  }

# Optionally print the query for debugging
if (isTRUE(showQuery)) {
  message("Endpoint: ", endpoint_url)
  message(SPARQL.query)
}

# Execute the request and parse CSV response
out <- tryCatch({
  resp <- httr::POST(
    url    = endpoint_url,
    httr::accept("text/csv"),
    body   = list(query = SPARQL.query),
    encode = "form"
  )
  txt <- httr::content(resp, "text", encoding = "UTF-8")
  df  <- utils::read.csv(text = txt, sep = ",", stringsAsFactors = FALSE, check.names = FALSE)
  
  # Basic post-conditions: expected columns should exist
  needed <- c("URI", "Prefix", "Title", "Languages")
  if (!all(needed %in% names(df))) {
    stop("Unexpected columns returned by SPARQL endpoint. Got: ",
         paste(names(df), collapse = ", "))
  }
  
  # Prepare result structure; normalize prefix, extract basename of URI
  uri    <- df[["URI"]]
  prefix <- if (identical(endpoint, "CELLAR")) gsub("\\.", "", df[["Prefix"]]) else df[["Prefix"]]
  prefix <- .ct_clean_prefix(prefix)
  title  <- df[["Title"]]
  langs  <- df[["Languages"]]
  
  result <- data.frame(
    Prefix        = prefix,
    ConceptScheme = .ct_basename_uri(uri),
    URI           = uri,
    Title         = title,
    Languages     = langs,
    stringsAsFactors = FALSE
  )
  rownames(result) <- seq_len(nrow(result))
  result
}, error = function(e) {
  # Provide helpful diagnostics while keeping a clean error for callers
  message("The following SPARQL code was used in the call:\n", SPARQL.query)
  message("The above SPARQL call to ", endpoint_url,
          " generated the following error message:\n", conditionMessage(e))
  stop(simpleError(
    paste0("Error in classificationList(", endpoint, "): endpoint unavailable or returned unexpected data")
  ))
})

# Return either the data.frame alone or a list with the query
if (isTRUE(showQuery)) {
  return(list("SPARQL.query" = SPARQL.query, "ClassificationList" = out))
} else {
  return(out)
}
}

# -------------------------------------------------------------------
# Internal utilities
# -------------------------------------------------------------------

#' @keywords internal
#' @noRd
.ct_basename_uri <- function(x) {
  # Keep only the last path/token after '/' or '#', then sanitize
  out <- sub("^.*[/#]", "", x)
  gsub("[^A-Za-z0-9_\\-\\.]", "", out)
}

#' @keywords internal
#' @noRd
.ct_clean_prefix <- function(prefix) {
  # Normalize prefixes: lowercase, drop common version suffixes,
  # remove dots/spaces and keep a safe alphanumeric/underscore set
  prefix <- tolower(prefix)
  prefix <- sub("(_v[0-9.]+)$", "", prefix)
  prefix <- sub("(-v[0-9.]+)$", "", prefix)
  prefix <- gsub("\\.", "", prefix)
  prefix <- gsub(" ", "", prefix)
  prefix <- gsub("[^a-z0-9_]", "", prefix)
  trimws(prefix)
}

# -------- 1) Input validation: endpoint -------------------------------------

expect_error(
  dataStructure(endpoint = "INVALID", prefix = "CPC20", conceptScheme = "CPC20"),
  "(?i)endpoint.*(CELLAR|FAO)|single value|must be",
  fixed = FALSE
)

if (at_home()) {
  
  # -------- 2) Online path: CELLAR returns a data.frame -----------------------
  
  prefix        <- "cn2022"
  conceptScheme <- "cn2022"
  
  res <- dataStructure(
    endpoint      = "CELLAR",
    prefix        = prefix,
    conceptScheme = conceptScheme,
    language      = "en",
    showQuery     = FALSE
  )
  
  expect_inherits(res, "data.frame")
  expect_true(is.numeric(nrow(res)))
  
  
  # -------- 3) Online path: FAO returns a data.frame ---------------------------
  
  
  prefix        <- "icc11"
  conceptScheme <- "scheme"
  
  res <- dataStructure(
    endpoint      = "FAO",
    prefix        = prefix,
    conceptScheme = conceptScheme,
    language      = "en",
    showQuery     = FALSE
  )
  
  expect_inherits(res, "data.frame")
  expect_true(is.numeric(nrow(res)))
  
  # -------- 4) showQuery = TRUE: structure per 'return' -----------------------
  
  
  prefix        <- "icc11"
  conceptScheme <- "scheme"
  
  # return = "summary"
  res_sum <- suppressMessages( dataStructure(
    endpoint      = "FAO",
    prefix        = prefix,
    conceptScheme = conceptScheme,
    language      = "en",
    showQuery     = TRUE,
    return        = "summary"
  ) )
  expect_true(is.list(res_sum))
  expect_true(all(c("resolved", "query.summary", "dataStructure") %in% names(res_sum)))
  expect_true(is.character(res_sum[["query.summary"]]))
  expect_inherits(res_sum[["dataStructure"]], "data.frame")
  
  # return = "details"
  res_det <- suppressMessages( dataStructure(
    endpoint      = "FAO",
    prefix        = prefix,
    conceptScheme = conceptScheme,
    language      = "en",
    showQuery     = TRUE,
    return        = "details"
  ) )
  expect_true(is.list(res_det))
  expect_true(all(c("resolved", "query.details", "details") %in% names(res_det)))
  expect_true(is.character(res_det[["query.details"]]))
  expect_inherits(res_det[["details"]], "data.frame")
  
  # return = "both"
  res_both <- suppressMessages(dataStructure(
    endpoint      = "FAO",
    prefix        = prefix,
    conceptScheme = conceptScheme,
    language      = "en",
    showQuery     = TRUE,
    return        = "both"
  ) )
  expect_true(is.list(res_both))
  expect_true(all(c("resolved", "query.summary", "query.details", "summary", "details") %in% names(res_both)))
  expect_true(is.character(res_both[["query.summary"]]))
  expect_true(is.character(res_both[["query.details"]]))
  expect_inherits(res_both[["summary"]], "data.frame")
  expect_inherits(res_both[["details"]], "data.frame")
}
# -------- 5) Unresolvable prefix/conceptScheme for FAO -----------------------

# This test does not need internet (it fails at registry resolution),
# but classificationList('FAO') must be available in the package.
expect_error(
  suppressWarnings(
    dataStructure(
      endpoint      = "FAO",
      prefix        = "XXX", 
      conceptScheme = "YYY",  
      language      = "en"
    )
  ),
  "(?i)Could not resolve the requested classification",
  fixed = FALSE
)

#' @title Download a correspondence (mapping) table between two classifications
#'
#' @description
#' Downloads a correspondence table, a mapping of codes, from one statistical
#' classification (A) to another (B) from an online service (CELLAR or
#' FAO), and returns it as a plain data frame. The function has no side effects
#' (it does not read or write files).
#'
#' A correspondence table tells you which codes in classification A relate
#' to which codes in classification B (e.g., \code{NACE2} → \code{CPA21}).
#'
#' @param endpoint Character. The online service to query. Case-insensitive.
#'   Supported values are those returned by the internal endpoint registry
#'   (e.g., \code{"CELLAR"}, \code{"FAO"}).
#' @param prefix Character. Catalogue prefix where the correspondence is
#'   published (e.g., \code{"nace2"}, \code{"cpa21"}, \code{"cn2022"}).
#'   Use \code{\link{correspondenceTableList}()} to discover valid values.
#' @param ID_table Character. Identifier of the correspondence, typically of
#'   the form \code{"A_B"} such as \code{"NACE2_CPA21"} or \code{"CN2022_NACE2"}.
#'   Discover identifiers via \code{\link{correspondenceTableList}()}.
#' @param language Character. Preferred label language as a BCP‑47 code.
#'   Defaults to \code{"en"} (English). Examples: \code{"fr"}, \code{"de"}.
#' @param showQuery Logical. If \code{TRUE}, returns a list with the SPARQL
#'   query and the result data frame; otherwise (default) returns just the data frame.
#'
#' @details
#' \strong{When to use this.} Use \code{retrieveCorrespondenceTable()} when you
#' know (or have discovered) the \code{prefix} and \code{ID_table} of a
#' correspondence and want the mapping rows as a standard data frame.
#'
#' \strong{How it works (short).}
#' \enumerate{
#'   \item Validates the selected service (\code{endpoint}).
#'   \item Optionally resolves a stable anchor via \code{\link{correspondenceTableList}()}.
#'   \item Builds a SPARQL query returning source/target codes, labels, notes, and link URI.
#'   \item Issues an HTTP request and parses the response into a data frame.
#' }
#'
#'
#' @return
#' A \code{data.frame} where each row represents one mapping from the source
#' classification (\strong{A}) to the target classification (\strong{B}).
#'
#' Columns (when available):
#' \itemize{
#'   \item \code{CorrespondenceID}:  Requested identifier (e.g., \code{"NACE2_CPA21"}).
#'   \item \code{Prefix}: Catalogue prefix (constant per table).
#'   \item \code{A}: Source code (\code{A} parsed from \code{ID_table}).
#'   \item \code{B}: Target code (\code{B} parsed from \code{ID_table}).
#'   \item \code{Label_A}, \code{Label_B}: Human‑readable labels in \code{language}.
#'   \item \code{Include_A}, \code{Exclude_A}, \code{Include_B}, \code{Exclude_B}: Notes, if available.
#'   \item \code{Comment}: Free‑text comment on the mapping, if any.
#'   \item \code{URL}: URI of the mapping association.
#' }
#'
#' Attributes attached for traceability:
#' \itemize{
#'   \item \code{endpoint}, \code{endpoint_url}, \code{prefix}, \code{ID_table}
#'   \item \code{A}, \code{B}, \code{language}, \code{normalized_codes}, \code{anchor_mode}, \code{anchor}
#' }
#'
#' @section Notes:
#' \itemize{
#'   \item The function focuses on retrieval. Any code normalization is left to
#'         downstream steps.
#' }
#'
#' @examples
#' \dontrun{
#' # 1) Discover available correspondences at CELLAR
#' ct <- correspondenceTableList(endpoint = "CELLAR")
#' subset(ct, grepl("NACE2", ID) & grepl("CPA21", ID))
#'
#' # 2) Retrieve one correspondence (English labels)
#' x <- retrieveCorrespondenceTable(
#'   endpoint = "CELLAR",
#'   prefix   = "nace2",
#'   ID_table = "NACE2_CPA21",
#'   language = "en"
#' )
#' head(x)
#'
#' # 3) Save to CSV explicitly if needed (no automatic writing)
#' # utils::write.csv(x, "NACE2_CPA21.csv", row.names = FALSE)
#'
#' # 4) Inspect the generated SPARQL query
#' dbg <- retrieveCorrespondenceTable(
#'   endpoint = "FAO",
#'   prefix   = "cpa21",
#'   ID_table = "NACE2_CPA21",
#'   showQuery = TRUE
#' )
#' message(substr(dbg$SPARQL.query, 1, 400), "...")
#' }
#'
#' @importFrom httr GET content stop_for_status timeout accept user_agent
#' @export
retrieveCorrespondenceTable <- function(
    endpoint,
    prefix,
    ID_table,
    language  = "en",
    showQuery = FALSE
) {
  # -------- internal defaults (no user-visible side effects) --------
  precheck <- TRUE  # ENABLE correspondence discovery
  normalize_codes <- FALSE
  norm_case       <- "lower"
  norm_strip      <- "[[:space:].-]"
  timeout_sec <- 120L   
  max_retries     <- 1L
  
  # ------------------------------------------------------------------
  # Define empty result skeleton (used when no correspondence is found)
  # ------------------------------------------------------------------
  empty_result_df <- data.frame(
    CorrespondenceID = character(0),
    Prefix           = character(0),
    stringsAsFactors = FALSE,
    check.names      = FALSE
  )
  
  # ---- endpoint validation & URL (via internal helpers) ----
  endpoint <- .validate_endpoints(endpoint)
  if (identical(endpoint, "ALL")) {
    stop("'endpoint' must be a single endpoint (e.g., 'CELLAR' or 'FAO'), not 'ALL'.",
         call. = FALSE)
  }
  if (length(endpoint) != 1L) {
    stop("'endpoint' must be a single value.", call. = FALSE)
  }
  endpoint_url <- .sparql_endpoint(endpoint)
  
  # ---- sanitize inputs ----
  prefix   <- trimws(prefix)
  ID_table <- trimws(ID_table)
  language <- trimws(language)
  
  # ---- derive A and B tokens from ID_table (robust to spacing and hyphens) ----
  ID_table_std <- gsub("-", "_", ID_table, fixed = TRUE)
  ID_table_std <- gsub("__", "_", ID_table_std, fixed = TRUE)
  A_raw <- sub("_.*", "", ID_table_std)
  B_raw <- sub(".*_", "", ID_table_std)
  
  # ---- make SPARQL variable names safe ----
  .safe_var <- function(x) {
    x <- gsub("[^A-Za-z0-9_]", "_", x)
    if (!grepl("^[A-Za-z_]", x)) x <- paste0("X_", x)
    x
  }
  A <- .safe_var(A_raw)
  B <- .safe_var(B_raw)
  
  # ---- optional precheck to get a stable correspondence URI ----
  
  corr_uri <- NULL
  if (isTRUE(precheck)) {
    ct_list <- tryCatch(
      correspondenceTableList(endpoint),
      error = function(e) NULL
    )
    
    if (is.data.frame(ct_list) && nrow(ct_list)) {
      ct_df <- ct_list
      
      # Normalize column names from correspondenceTableList(): Prefix, ID, URI -> prefix, id, uri
      names(ct_df) <- tolower(gsub("\\s+", ".", names(ct_df)))  # [1](https://teams.microsoft.com/l/meeting/details?eventId=AAMkADU5YjkwZDczLWU5NTEtNGZkMC05MjUwLTg1ZmE0ZmFkZGM5YwBGAAAAAAAY52HbrUuXSYgNfQribIjRBwCJfZVPsYxqQJf_ENZMnpx2AAAAAAENAACJfZVPsYxqQJf_ENZMnpx2AAAA8uebAAA%3d)
      
      # Harmonize URI column name from correspondenceTableList()
      # (often returned as URL, not URI)
      if ("url" %in% names(ct_df) && !"uri" %in% names(ct_df)) {
        ct_df$uri <- ct_df$url
      }
      
      # correspondenceTableList() provides ID (-> 'id' after tolower); harmonize to id_table
      if ("id" %in% names(ct_df) && !"id_table" %in% names(ct_df)) {
        ct_df$id_table <- ct_df$id
      }
      
      # 1) Primary match: prefix + id_table (case-insensitive)
      hit <- integer(0)
      if (all(c("prefix", "id_table") %in% names(ct_df))) {
        hit <- which(
          tolower(trimws(ct_df$prefix))   == tolower(trimws(prefix)) &
            tolower(trimws(ct_df$id_table)) == tolower(trimws(ID_table))
        )
      }
      
      # 2) If not found, try by URI tail: .../<ID_table>
      if (!length(hit) && "uri" %in% names(ct_df)) {
        uri_tail <- function(u) {
          u <- as.character(u)
          u <- sub("/+$", "", u)
          sub(".*/", "", u)
        }
        tails <- vapply(ct_df$uri, uri_tail, character(1))
        cand  <- which(tolower(trimws(tails)) == tolower(trimws(ID_table)))
        
        if (length(cand)) {
          if ("prefix" %in% names(ct_df)) {
            rel <- which(tolower(trimws(ct_df$prefix[cand])) == tolower(trimws(prefix)))
            hit <- if (length(rel)) cand[rel[1]] else cand[1]
          } else {
            hit <- cand[1]
          }
        }
      }
      
      # 3) Take the URI if we found a match
      if (length(hit) && "uri" %in% names(ct_df)) {
        u <- ct_df$uri[hit[1]]
        if (is.character(u) && nzchar(trimws(u))) corr_uri <- trimws(u)
      }
    }
  }
  
  
  # ------------------------------------------------------------------
  # Derive scheme URI (namespace) from correspondence URI
  # ------------------------------------------------------------------
  if (is.null(corr_uri) || !nzchar(corr_uri)) {
    
    # If discovery is disabled, do NOT call correspondenceTableList()
    if (!isTRUE(precheck)) {
      warning(
        sprintf("No correspondence URI resolved for ID '%s' in %s (discovery disabled).",
                ID_table, endpoint),
        call. = FALSE
      )
      if (isTRUE(showQuery)) {
        return(list(SPARQL.query = NA_character_, CorrespondenceTable = empty_result_df))
      }
      return(empty_result_df)
    }
    
    # discovery-enabled: suggest close matches
    ct_list <- tryCatch(correspondenceTableList(endpoint), error = function(e) NULL)
    suggestion <- NULL
    if (is.data.frame(ct_list) && nrow(ct_list)) {
      names(ct_list) <- tolower(names(ct_list))
      if ("id" %in% names(ct_list)) {
        cand <- ct_list$id[
          grepl(A_raw, ct_list$id, ignore.case = TRUE) |
            grepl(B_raw, ct_list$id, ignore.case = TRUE) |
            grepl(ID_table, ct_list$id, ignore.case = TRUE)
        ]
        cand <- unique(cand)
        if (length(cand)) suggestion <- head(cand, 10)
      }
    }
    
    msg <- sprintf("No correspondence table with ID '%s' found in %s.", ID_table, endpoint)
    if (!is.null(suggestion)) msg <- paste0(msg, " Closest IDs: ", paste(suggestion, collapse = ", "))
    stop(msg, call. = FALSE)
    
    if (isTRUE(showQuery)) {
      return(list(SPARQL.query = NA_character_, CorrespondenceTable = empty_result_df))
    }
    return(empty_result_df)
  }
  
  # MPORTANT: corr_uri exists -> now define scheme_uri
  # Example corr_uri:  http://data.europa.eu/xsp/cn2018/CN2018_CPA21
  # scheme_uri becomes: http://data.europa.eu/xsp/cn2018/
  
  scheme_uri <- sub("/[^/]+$", "/", trimws(corr_uri))
  
  # ------------------------------------------------------------------
  # Build PREFIX block (REAL SPARQL, not HTML-escaped)
  # ------------------------------------------------------------------
  stopifnot(nzchar(prefix))
  stopifnot(nzchar(scheme_uri))
  
  prefix_block <- paste(
    "PREFIX dc:   <http://purl.org/dc/elements/1.1/>",
    "PREFIX dct:  <http://purl.org/dc/terms/>",
    "PREFIX rdf:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>",
    "PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>",
    "PREFIX skos: <http://www.w3.org/2004/02/skos/core#>",
    "PREFIX xkos: <http://rdf-vocabulary.ddialliance.org/xkos#>",
    "PREFIX xsd:  <http://www.w3.org/2001/XMLSchema#>",
    sprintf("PREFIX %s: <%s>", prefix, scheme_uri),
    sep = "\n"
  )
  
  # ---- anchor mode: use full URI if available; else identifier ----
  if (!is.null(corr_uri) && nzchar(trimws(corr_uri))) {
    anchor_mode  <- "uri"
    anchor_info  <- paste0("<", trimws(corr_uri), ">")
    ct_bind_line <- paste0("  BIND (", anchor_info, " AS ?CT)\n")
    
    
  } else {
    anchor_mode  <- "identifier"
    anchor_info  <- ID_table
    ct_bind_line <- paste0('  ?CT dct:identifier "', ID_table, "\" .\n")
  }
  
  
  # ---- build SPARQL query (PAGED, lightweight) ----
  build_query_paged <- function(lang, limit, offset) {
    paste0(
      prefix_block, "\n",
      "SELECT ?SourceCode ?TargetCode ?Comment ?URL\n",
      "WHERE {\n",
      ct_bind_line,
      "  ?CT xkos:madeOf ?Associations .\n",
      "  ?Associations xkos:sourceConcept ?Source .\n",
      "  OPTIONAL { ?Associations xkos:targetConcept ?Target . }\n",
      "  OPTIONAL { ?Associations rdfs:comment ?Comment . }\n",
      "  ?Source skos:notation ?SourceNotation .\n",
      "  OPTIONAL { ?Target skos:notation ?TargetNotation . }\n",
      "  BIND (STR(?SourceNotation) AS ?SourceCode)\n",
      "  BIND (IF(BOUND(?TargetNotation), STR(?TargetNotation), '') AS ?TargetCode)\n",
      "  BIND (STR(?Associations) AS ?URL)\n",
      "}\n",
      "ORDER BY ?SourceCode\n",
      "LIMIT ", as.integer(limit), "\n",
      "OFFSET ", as.integer(offset), "\n"
    )
  }
  
  # ------------------------------------------------------------------
  # Helpers to detect non-CSV responses
  # ------------------------------------------------------------------
  is_html <- function(txt) {
    is.character(txt) && length(txt) == 1L &&
      nzchar(txt) &&
      grepl("^\\s*<!DOCTYPE\\s+html|<html", txt, ignore.case = TRUE)
  }
  is_xml <- function(txt) {
    is.character(txt) && length(txt) == 1L &&
      nzchar(txt) &&
      grepl("^\\s*<\\?xml|<sparql", txt, ignore.case = TRUE)
  }
  
  
  # ---- initialise paging state ----
  parts      <- list()
  offset     <- 0L
  chunk_size <- 5000L
  
  # ---- run paged queries ----
  repeat {
    SPARQL.query <- build_query_paged(language, chunk_size, offset)
    
    csv_text <- tryCatch(
      .do_post_csv(endpoint_url, SPARQL.query,
                   timeout_sec = timeout_sec,
                   attempts    = max_retries),
      error = function(e) ""
    )
    
    if (!nzchar(csv_text)) break
    if (isTRUE(is_html(csv_text)) || isTRUE(is_xml(csv_text))) break
    
    df_chunk <- .read_sparql_csv(csv_text)
    if (!nrow(df_chunk)) break
    
    parts[[length(parts) + 1L]] <- df_chunk
    
    if (nrow(df_chunk) < chunk_size) break
    offset <- offset + chunk_size
  }
  
  # ---- assemble result ----
  data <- if (length(parts)) do.call(rbind, parts) else data.frame()
  
  # ---- rename columns for downstream logic ----
  if (nrow(data)) {
    data[[A]] <- data[["SourceCode"]]
    data[[B]] <- data[["TargetCode"]]
  }
  
  if (!is.data.frame(data) || !nrow(data)) {
    warning(sprintf("No correspondence rows returned for ID '%s' in %s.", ID_table, endpoint),
            call. = FALSE)
    if (isTRUE(showQuery)) {
      return(list(SPARQL.query = SPARQL.query, CorrespondenceTable = empty_result_df))
    }
    return(empty_result_df)
  }
  
  # ---- optional normalization (off by default) ----
  ct_norm_code <- function(x, strip = "[[:space:].-]", case = c("lower","upper","asis")) {
    case <- match.arg(case)
    x <- as.character(x)
    x <- trimws(x)
    if (!is.null(strip) && nzchar(strip)) x <- gsub(strip, "", x)
    if (case == "lower") x <- tolower(x) else if (case == "upper") x <- toupper(x)
    x
  }
  if (isTRUE(normalize_codes) && is.data.frame(data) && nrow(data)) {
    if (nzchar(A) && A %in% names(data)) data[[A]] <- ct_norm_code(data[[A]], strip = norm_strip, case = norm_case)
    if (nzchar(B) && B %in% names(data)) data[[B]] <- ct_norm_code(data[[B]], strip = norm_strip, case = norm_case)
  }
  
  # ---- prepend CorrespondenceID and Prefix; attach attributes ----
  if (is.data.frame(data) && nrow(data) >= 0) {
    data <- cbind(
      CorrespondenceID = rep(ID_table, nrow(data)),
      Prefix           = rep(prefix,   nrow(data)),
      data
    )
  } 
  
  attr(data, "endpoint")         <- endpoint
  attr(data, "prefix")           <- prefix
  attr(data, "ID_table")         <- ID_table
  attr(data, "A")                <- A
  attr(data, "B")                <- B
  attr(data, "language")         <- language
  attr(data, "normalized_codes") <- isTRUE(normalize_codes)
  attr(data, "endpoint_url")     <- endpoint_url
  attr(data, "anchor_mode")      <- anchor_mode
  attr(data, "anchor")           <- anchor_info
  attr(data, "corr_uri")         <- corr_uri
  attr(data, "scheme_uri")       <- scheme_uri
  
  if (isTRUE(showQuery)) {
    return(list(SPARQL.query = SPARQL.query, CorrespondenceTable = data))
  } else {
    return(data)
  }
}

# --- 1) retrieveCorrespondenceTable returns a data.frame (basic sanity) ----

if(at_home()){
  {
    res_try <- try(
      retrieveCorrespondenceTable("FAO", "v21", "CPC21-CPC20"
      ),
      silent = TRUE
    )
    
    # If endpoint unreachable, skip gracefully
    if (inherits(res_try, "try-error")) {
      msg <- conditionMessage(attr(res_try, "condition"))
      message("Skipping test: SPARQL endpoint unreachable: ", msg)
    } else {
      res <- res_try
      expect_true(is.data.frame(res))
      expect_true(nrow(res) > 0)
    }
  }
} 

# --- 2) Reject invalid endpoint --------------------------------------------
{
  expect_error(
    retrieveCorrespondenceTable(
      prefix   = "cn2022",
      endpoint = "INVALID",
      ID_table = "CN2022_NST2007"
    ),
    "Unsupported endpoint"
  )
}

# --- 3) Error when Prefix and ID_table not available ---
{
  expect_error(retrieveCorrespondenceTable("FAO", prefix   = "other", ID_table = "OTHER_TABLE"))
}

#' @title Retrieve a classification structure (levels and concepts) from a supported service
#'
#' @description
#' Queries the selected service (e.g., CELLAR or FAO) to retrieve the structure of a
#' classification scheme (levels, depth, and concept rows), with language-aware
#' label fallback. You can return a compact summary table, a detailed
#' concept-level table, or both.
#'
#' The function first resolves the input classification using the in-package registry
#' returned by \code{classificationList()}, then retrieves the structure and returns
#' it as data frames. Network errors are handled with retry logic.
#'
#' @param endpoint Character scalar. Accepted values are \code{"CELLAR"} or \code{"FAO"}.
#' @param prefix Character. A classification prefix used in the registry
#'   (e.g. \code{"NACE"}, \code{"CPA"}, \code{"CN2022"}). Used to resolve the scheme.
#' @param conceptScheme Character or \code{NULL}. Optional alternative identifier
#'   for the concept scheme. The resolver tries both \code{prefix} and \code{conceptScheme}
#'   using basic normalization (case-insensitive; dot removal).
#' @param language Character scalar. Two-letter language code used to select
#'   labels (default: \code{"en"}). Fallback chain: requested language -> English (\code{"en"})
#'   -> no-language literal -> the local resource name.
#' @param showQuery Logical. If \code{TRUE}, the generated queries are emitted via
#'   \code{message()} and also included in the returned list.
#' @param return One of \code{c("summary", "details", "both")}. Controls which table(s)
#'   are fetched and returned.
#' @param timeout_sec Numeric. HTTP timeout in seconds. Default: \code{60}.
#' @param retries Integer. Number of automatic retries on transient HTTP errors.
#'   Default: \code{3}.
#'
#' @details
#' \strong{Resolution of the scheme}
#'
#' The function calls \code{classificationList(endpoint)} and tries to match either
#' \code{prefix} or \code{conceptScheme} using normalized keys. On success it extracts
#' the classification namespace and scheme identifier needed to compose requests.
#'
#' \strong{Language fallback}
#'
#' Labels for both the classification levels and concepts follow a multi-step
#' fallback: requested \code{language} -> \code{"en"} -> no-language literal -> local name.
#'
#' @return
#' A \code{data.frame} (for \code{"summary"} or \code{"details"}), or a
#' \code{list(summary, details)} when \code{return = "both"}.
#'
#' \strong{Columns:}
#' \itemize{
#'   \item All returned tables include a leading \code{Prefix} column (lower-case token).
#'   \item \code{summary}: \code{Concept_Scheme}, \code{Depth}, \code{Level}, \code{Count}.
#'   \item \code{details}: \code{Concept}, \code{Code}, \code{Label}, \code{Depth}, \code{Level},
#'         \code{BroaderList}, \code{BroaderCodeList}.
#' }
#'
#' If \code{showQuery = TRUE}, the return value is a list containing:
#' \itemize{
#'   \item \code{resolved}: metadata (\code{endpoint}, inputs, \code{scheme_id}, \code{ns_uri}, and optionally \code{title})
#'   \item the query text (one or two elements depending on \code{return})
#'   \item the retrieved table(s)
#' }
#'
#' @section Errors and Warnings:
#' \itemize{
#'   \item Errors if \code{endpoint} is not \code{"CELLAR"} or \code{"FAO"}, or if the
#'         scheme cannot be resolved from the registry.
#'   \item Warns (and still returns rows) if no explicit levels are declared.
#'   \item Propagates HTTP errors from the service after retry policy.
#' }
#'
#' @seealso \code{\link{classificationList}}
#'
#' @examples
#' \dontrun{
#' ds_cn <- dataStructure(
#'   endpoint      = "CELLAR",
#'   prefix        = "cn2022",
#'   conceptScheme = "cn2022",
#'   language      = "en",
#'   return        = "summary"
#' )
#' head(ds_cn)
#' }
#'
#' @importFrom httr RETRY stop_for_status content accept add_headers user_agent timeout
#' @importFrom utils read.csv head
#' @export
dataStructure <- function(endpoint, prefix, conceptScheme = NULL,
                          language = "en",
                          showQuery = FALSE,
                          return = c("summary", "details", "both"),
                          timeout_sec = 60, retries = 3) {
  
  return <- match.arg(return)
  
  # ---- Input validation -----------------------------------------------------
  ep <- .validate_endpoints(endpoint)
  if (identical(ep, "ALL") || length(ep) != 1L) {
    stop("'endpoint' must be a single value ('CELLAR' or 'FAO'), not 'ALL' or a vector.", call. = FALSE)
  }
  endpoint <- ep
  
  if (!is.character(language) || length(language) != 1L || !grepl("^[a-z]{2}$", language)) {
    stop("'language' must be a two-letter ISO 639-1 code (e.g., 'en', 'fr', 'de').", call. = FALSE)
  }
  
  # ---- Helpers (local) ------------------------------------------------------
  normalize_key <- function(x) tolower(gsub("[^A-Za-z0-9._-]", "", x))
  
  get_registry_table <- function(endpoint) {
    ep  <- toupper(endpoint)
    lst <- classificationList(ep)
    if (is.data.frame(lst)) return(lst)
    if (is.list(lst) && !is.null(lst[[ep]]) && is.data.frame(lst[[ep]])) return(lst[[ep]])
    stop(sprintf("No classification registry available for endpoint '%s'.", endpoint), call. = FALSE)
  }
  
  resolve_scheme <- function(endpoint, key1, key2 = NULL) {
    tbl <- get_registry_table(endpoint)
    if (!is.data.frame(tbl) || !nrow(tbl)) {
      stop(sprintf("No classification registry available for endpoint '%s'.", endpoint), call. = FALSE)
    }
    # Expected columns: Prefix, ConceptScheme, URI, (optional) Title
    tbl$Prefix_clean              <- normalize_key(tbl$Prefix)
    tbl$ConceptScheme_clean       <- normalize_key(tbl$ConceptScheme)
    tbl$Prefix_clean_nodot        <- gsub("\\.", "", tbl$Prefix_clean)
    tbl$ConceptScheme_clean_nodot <- gsub("\\.", "", tbl$ConceptScheme_clean)
    
    keys_raw         <- unique(na.omit(c(key1, key2)))
    keys_clean       <- normalize_key(keys_raw)
    keys_clean_nodot <- gsub("\\.", "", keys_clean)
    
    pick <- function(ix) {
      row <- tbl[ix, , drop = FALSE]
      uri <- row$URI
      ns_uri    <- sub("/[^/]+$", "/", uri)
      scheme_id <- row$ConceptScheme
      list(scheme_id = scheme_id, ns_uri = ns_uri, row = row)
    }
    
    hit <- which(tbl$ConceptScheme %in% keys_raw);                      if (length(hit)) return(pick(hit[1]))
    hit <- which(tbl$ConceptScheme_clean %in% keys_clean);              if (length(hit)) return(pick(hit[1]))
    hit <- which(tbl$Prefix_clean %in% keys_clean);                     if (length(hit)) return(pick(hit[1]))
    hit <- which(tbl$ConceptScheme_clean_nodot %in% keys_clean_nodot);  if (length(hit)) return(pick(hit[1]))
    hit <- which(tbl$Prefix_clean_nodot %in% keys_clean_nodot);         if (length(hit)) return(pick(hit[1]))
    
    suggestions <- utils::head(paste0(tbl$Prefix, " (", tbl$ConceptScheme, ")"), 10)
    stop(
      paste0(
        "Could not resolve the requested classification '", paste(keys_raw, collapse = "', '"),
        "' for endpoint '", endpoint, "'.\nTry one of:\n  - ",
        paste(suggestions, collapse = "\n  - ")
      ),
      call. = FALSE
    )
  }
  
  warn_if_no_levels <- function(df) {
    if (is.null(df) || !is.data.frame(df) || nrow(df) == 0L) {
      warning(
        "The classification appears to have no declared levels; returning available rows only. ",
        "If you rely on explicit levels, consider using `level = 'ALL'` in retrieveClassificationTable().",
        call. = FALSE,
        immediate. = TRUE
      )
      return(TRUE)
    }
    FALSE
  }
  
  coerce_types_summary <- function(df) {
    if (!nrow(df)) return(df)
    if ("Depth" %in% names(df)) df$Depth <- suppressWarnings(as.integer(df$Depth))
    if ("Count" %in% names(df)) df$Count <- suppressWarnings(as.integer(df$Count))
    df
  }
  
  coerce_types_details <- function(df) {
    if (!nrow(df)) return(df)
    if ("Depth" %in% names(df)) df$Depth <- suppressWarnings(as.integer(df$Depth))
    df
  }
  
  # NEW: add Prefix as first column (length-safe; preserves empty frames)
  add_prefix_col <- function(df, token) {
    if (!is.data.frame(df)) return(df)
    if (!nrow(df)) {
      df$Prefix <- character(0)
      df <- df[, c("Prefix", setdiff(names(df), "Prefix")), drop = FALSE]
      return(df)
    }
    cbind(Prefix = rep(token, nrow(df)), df)
  }
  
  # ---- Resolve endpoint URL and scheme -------------------------------------
  source_url <- .sparql_endpoint(endpoint)
  
  resolved   <- resolve_scheme(endpoint, prefix, conceptScheme)
  scheme_id  <- resolved$scheme_id
  ns_uri     <- resolved$ns_uri
  title_val  <- if (!is.null(resolved$row$Title)) resolved$row$Title else NA_character_
  
  # NEW: printable prefix token (from registry if available; fallback to input)
  prefix_token <- if (!is.null(resolved$row$Prefix)) as.character(resolved$row$Prefix) else prefix
  prefix_token <- tolower(trimws(prefix_token))
  
  # Build PREFIX block with fixed alias 'cs:' for the classification namespace
  prefix_block <- paste(
    "PREFIX skos: <http://www.w3.org/2004/02/skos/core#>",
    "PREFIX xkos: <http://rdf-vocabulary.ddialliance.org/xkos#>",
    "PREFIX rdf:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>",
    sprintf("PREFIX cs:   <%s>", ns_uri),
    sep = "\n"
  )
  
  # -------------------------------
  # 1) SUMMARY query (uses cs:<scheme_id>)
  # -------------------------------
  q_summary <- paste0(
    prefix_block, "
SELECT DISTINCT ?Concept_Scheme ?Depth ?Level (COUNT (DISTINCT ?s) AS ?Count)
WHERE {
  ?s skos:prefLabel ?Label ;
     skos:inScheme cs:", scheme_id, " ;
     skos:inScheme ?Scheme ;
     ^skos:member ?Member ;
     skos:notation ?notation .

  ?Member a xkos:ClassificationLevel .
  OPTIONAL { ?Member xkos:levels ?levels_temp . }
  OPTIONAL { ?Member xkos:depth  ?Depth . }

  OPTIONAL { ?Member skos:prefLabel ?LevelLabel_req . FILTER (lang(?LevelLabel_req) = '", language, "') }
  OPTIONAL { ?Member skos:prefLabel ?LevelLabel_en  . FILTER (lang(?LevelLabel_en)  = 'en') }
  OPTIONAL { ?Member skos:prefLabel ?LevelLabel_nolang . FILTER (lang(?LevelLabel_nolang) = '') }

  FILTER (?Scheme = cs:", scheme_id, ")
  FILTER (lang(?Label) = '", language, "')

  BIND('", ns_uri, "' AS ?CLASS_URL)

  BIND (STRAFTER(STR(?Scheme), ?CLASS_URL) AS ?Concept_Scheme)
  BIND (STRAFTER(STR(?Member), ?CLASS_URL) AS ?LevelLocal)

  BIND (COALESCE(?LevelLabel_req, ?LevelLabel_en, ?LevelLabel_nolang, ?LevelLocal) AS ?LevelRaw)
  BIND (STR(?LevelRaw) AS ?Level)

  FILTER (STRLEN(?Concept_Scheme) != 0)
  FILTER (STRLEN(?Level) != 0)
}
GROUP BY ?Concept_Scheme ?Depth ?Level
ORDER BY ?Concept_Scheme ?Depth ?Level
")
  
  # -------------------------------
  # 2) DETAILS query (uses cs:<scheme_id>)
  # -------------------------------
  q_details <- paste0(
    prefix_block, "
SELECT ?Concept
       (MIN(STR(?Notation)) AS ?Code)
       (SAMPLE(?Label0)     AS ?Label)
       (SAMPLE(?Depth0)     AS ?Depth)
       (SAMPLE(?LevelTxt)   AS ?Level)
       (GROUP_CONCAT(DISTINCT STR(?Broader);         separator=' | ')  AS ?BroaderList)
       (GROUP_CONCAT(DISTINCT STR(?BroaderNotation); separator=' | ')  AS ?BroaderCodeList)
WHERE {
  ?Concept a skos:Concept ;
           skos:inScheme cs:", scheme_id, " ;
           skos:notation ?Notation .

  OPTIONAL { ?Concept xkos:depth ?DepthFromConcept }
  OPTIONAL {
    ?Concept ^skos:member ?Member .
    ?Member a xkos:ClassificationLevel .
    OPTIONAL { ?Member xkos:depth ?DepthFromLevel }

    OPTIONAL { ?Member skos:prefLabel ?Lvl_req FILTER (LANG(?Lvl_req) = '", language, "') }
    OPTIONAL { ?Member skos:prefLabel ?Lvl_en  FILTER (LANG(?Lvl_en)  = 'en') }
    OPTIONAL { ?Member skos:prefLabel ?Lvl_no  FILTER (LANG(?Lvl_no)  = '') }
    BIND (COALESCE(?Lvl_req, ?Lvl_en, ?Lvl_no, STRAFTER(STR(?Member), '", ns_uri, "')) AS ?LevelTxt)
  }
  BIND (COALESCE(?DepthFromLevel, ?DepthFromConcept) AS ?Depth0)

  OPTIONAL { ?Concept skos:prefLabel ?Pref FILTER (LANG(?Pref) = '", language, "') }
  OPTIONAL { ?Concept skos:altLabel  ?Alt  FILTER (LANG(?Alt)  = '", language, "') }
  BIND (COALESCE(?Pref, ?Alt) AS ?Label0)

  OPTIONAL { ?Concept skos:broader ?Broader . ?Broader skos:notation ?BroaderNotation }
}
GROUP BY ?Concept
ORDER BY ?Code
")
  
  # ---- HTTP runner (wrap ONLY the network call in tryCatch) -----------------
  run_q <- function(query) {
    if (isTRUE(showQuery)) message(query)
    resp <- tryCatch(
      httr::RETRY(
        verb   = "POST",
        url    = source_url,
        body   = list(query = query, format = "text/csv"),
        encode = "form",
        httr::accept("text/csv"),
        httr::add_headers(`Accept-Encoding` = "gzip"),
        httr::user_agent("correspondenceTables (R)"),
        httr::timeout(timeout_sec),
        times  = retries,
        pause_base = 1, pause_cap = 8,
        terminate_on = c(400, 401, 403, 404)
      ),
      error = function(e) {
        stop(sprintf("Network error while contacting the service: %s", conditionMessage(e)), call. = FALSE)
      }
    )
    httr::stop_for_status(resp)
    txt <- httr::content(resp, as = "text", encoding = "UTF-8")
    
    if (!nzchar(txt)) {
      # Service returned an empty payload instead of a CSV table
      return(data.frame())
    }
    
    df <- try(utils::read.csv(text = txt, check.names = FALSE, stringsAsFactors = FALSE), silent = TRUE)
    if (inherits(df, "try-error")) {
      stop("The service returned a non-parsable response instead of a CSV table.", call. = FALSE)
    }
    df
  }
  
  # ---- Fetch according to 'return' -----------------------------------------
  if (return == "summary") {
    tab <- coerce_types_summary(run_q(q_summary))
    warn_if_no_levels(tab)
    
    if (nrow(tab) && "Depth" %in% names(tab)) {
      ord <- do.call(order, c(list(tab$Depth), if ("Level" %in% names(tab)) list(tab$Level) else list(), list(na.last = TRUE)))
      tab <- tab[ord, , drop = FALSE]
    }
    
    # NEW: add Prefix column
    tab <- add_prefix_col(tab, prefix_token)
    
    if (isTRUE(showQuery)) {
      return(list(
        resolved      = list(endpoint = endpoint, input_prefix = prefix, input_conceptScheme = conceptScheme,
                             scheme_id = scheme_id, ns_uri = ns_uri, title = title_val),
        query.summary = q_summary,
        dataStructure = tab
      ))
    }
    return(tab)
  }
  
  if (return == "details") {
    det <- coerce_types_details(run_q(q_details))
    
    if (is.null(det) || !is.data.frame(det) || nrow(det) == 0L ||
        (!"Level" %in% names(det)) || all(is.na(det$Level))) {
      warning(
        "The classification appears to have no declared levels; returning available rows only. ",
        "If you rely on explicit levels, consider using `level = 'ALL'` in retrieveClassificationTable().",
        call. = FALSE, immediate. = TRUE
      )
    }
    
    # NEW: add Prefix column
    det <- add_prefix_col(det, prefix_token)
    
    if (isTRUE(showQuery)) {
      return(list(
        resolved     = list(endpoint = endpoint, input_prefix = prefix, input_conceptScheme = conceptScheme,
                            scheme_id = scheme_id, ns_uri = ns_uri, title = title_val),
        query.details = q_details,
        details       = det
      ))
    }
    return(det)
  }
  
  # return == "both"
  tab <- coerce_types_summary(run_q(q_summary))
  warn_if_no_levels(tab)
  if (nrow(tab) && "Depth" %in% names(tab)) {
    ord <- do.call(order, c(list(tab$Depth), if ("Level" %in% names(tab)) list(tab$Level) else list(), list(na.last = TRUE)))
    tab <- tab[ord, , drop = FALSE]
  }
  det <- coerce_types_details(run_q(q_details))
  
  # NEW: add Prefix column to both
  tab <- add_prefix_col(tab, prefix_token)
  det <- add_prefix_col(det, prefix_token)
  
  if (isTRUE(showQuery)) {
    return(list(
      resolved          = list(endpoint = endpoint, input_prefix = prefix, input_conceptScheme = conceptScheme,
                               scheme_id = scheme_id, ns_uri = ns_uri, title = title_val),
      query.summary     = q_summary,
      query.details     = q_details,
      summary           = tab,
      details           = det
    ))
  }
  list(summary = tab, details = det)
}

# R/endpoints.R
# Centralized endpoint handling (internal-only)

#' @keywords internal
.sparql_endpoints <- function() {
  c(
    "CELLAR" = "https://publications.europa.eu/webapi/rdf/sparql",
    "FAO"    = "https://caliper.integratedmodelling.org/caliper/sparql"
  )
}

#' @keywords internal
.valid_endpoints <- function() {
  names(.sparql_endpoints())
}

# Optional: small normalizer to keep code DRY across helpers
#' @keywords internal
.normalize_endpoints <- function(x) {
  x <- as.character(x)
  toupper(trimws(x))
}

#' @keywords internal
.validate_endpoints <- function(x) {
  x <- .normalize_endpoints(x)
  
  # Guard against NULL/empty/NA
  if (is.null(x) || !length(x) || any(is.na(x))) {
    stop("'endpoint' must be a non-empty character value.", call. = FALSE)
  }
  
  # Disallow mixing "ALL" with others to avoid ambiguity
  if (any(x == "ALL") && length(x) > 1L) {
    stop("'ALL' cannot be combined with specific endpoints. Use 'ALL' alone or list endpoints explicitly.",
         call. = FALSE)
  }
  
  # Allow "ALL" to pass through (expanded by .expand_endpoints())
  if (identical(x, "ALL")) return("ALL")
  
  # Validate against registry
  bad <- setdiff(x, .valid_endpoints())
  if (length(bad)) {
    stop(
      sprintf(
        "Unsupported endpoint(s): %s. Supported endpoints are: %s",
        paste(bad, collapse = ", "),
        paste(.valid_endpoints(), collapse = ", ")
      ),
      call. = FALSE
    )
  }
  x
}

#' @keywords internal
.expand_endpoints <- function(x) {
  x <- .validate_endpoints(x)
  if (identical(x, "ALL")) .valid_endpoints() else x
}

#' @keywords internal
.sparql_endpoint <- function(x) {
  key  <- .normalize_endpoints(x)
  urls <- .sparql_endpoints()
  if (!nzchar(key) || !(key %in% names(urls))) {
    stop(sprintf(
      "Unsupported endpoint '%s'. Supported endpoints are: %s",
      x, paste(names(urls), collapse = ", ")
    ), call. = FALSE)
  }
  unname(urls[[key]])
}

if (at_home()) {
  
  # Minimal SPARQL query: asks only if the endpoint can answer a basic ASK request
  ask_query <- "ASK { ?s ?p ?o }"
  
  ping_endpoint <- function(endpoint_name) {
    # Resolve endpoint URL via internal helper used by the package
    endpoint_url <- .sparql_endpoint(endpoint_name)
    
    # Perform a minimal POST request to check reachability
    resp <- httr::POST(
      url    = endpoint_url,
      body   = list(query = ask_query),
      encode = "form",
      httr::accept("text/plain")
    )
    
    status <- httr::status_code(resp)
    txt    <- httr::content(resp, as = "text", encoding = "UTF-8")
    
    # Basic reachability: HTTP 2xx
    expect_true(
      status >= 200 && status < 300,
      info = paste0("Endpoint ", endpoint_name, " reachable (HTTP ", status, ")")
    )
    
    # Basic sanity: non-empty response body
    expect_true(
      nchar(txt) > 0,
      info = paste0("Endpoint ", endpoint_name, " returned a non-empty response")
    )
  }
  
  # Check both supported endpoints
  ping_endpoint("CELLAR")
  ping_endpoint("FAO")
}

# This function is deprecated and kept only for backward compatibility.
# Users should now use \code{classificationList()} instead, which provides
# the same functionality with a clearer and more stable interface.

#   Deprecated: please use classificationList() instead.
#   cl <- classificationEndpoint("ALL", showQuery = FALSE)
#   print(cl)

 classificationEndpoint <- function(endpoint = "ALL", showQuery = FALSE) {
   .Deprecated("classificationList")
   classificationList(endpoint = endpoint, showQuery = showQuery)
 }
 
# classificationEndpoint <- function(endpoint = "ALL") {
#   
#   endpoint <- match.arg(endpoint, c("ALL", "CELLAR", "FAO"))
#   
#   #-----------------------------#
#   # Helper : CELLAR             #
#   #-----------------------------#
#   fetch_cellar <- function() {
#     endpoint_cellar <- "http://publications.europa.eu/webapi/rdf/sparql"
#     
#     SPARQL.query_cellar <- "
#       SELECT DISTINCT ?s ?Title
#       WHERE { ?s a skos:ConceptScheme ;
#               skos:prefLabel ?Title ;
#               ?p <http://publications.europa.eu/resource/authority/corporate-body/ESTAT> 
#                FILTER (LANG(?Title) = 'en')}
#       ORDER BY ?Title
#     "
#     
#     resp <- tryCatch(
#       httr::POST(
#         url   = endpoint_cellar,
#         httr::accept("text/csv"),
#         body  = list(query = SPARQL.query_cellar),
#         encode = "form"
#       ),
#       error = function(e) {
#         warning(
#           sprintf(
#             "CELLAR endpoint is currently unavailable (network error: %s). Classifications from CELLAR will not be returned.",
#             conditionMessage(e)
#           ),
#           call. = FALSE
#         )
#         return(NULL)
#       }
#     )
#     
#     if (is.null(resp)) return(NULL)
#     
#     if (httr::http_error(resp)) {
#       status <- httr::status_code(resp)
#       warning(
#         sprintf(
#           "CELLAR endpoint is currently unavailable (HTTP status %s). Classifications from CELLAR will not be returned.",
#           status
#         ),
#         call. = FALSE
#       )
#       return(NULL)
#     }
#     
#     data_cellar <- utils::read.csv(
#       text = httr::content(resp, "text", encoding = "UTF-8"),
#       sep  = ",",
#       stringsAsFactors = FALSE
#     )
#     
#     if (nrow(data_cellar) == 0L) {
#       warning("CELLAR SPARQL query returned no rows. Classifications from CELLAR will not be returned.",
#               call. = FALSE)
#       return(NULL)
#     }
#     
#     # --- Post-traitement ---
#     uri_vec <- data_cellar[, 1]
#     title   <- data_cellar[, 2]
#     
#     str_list <- strsplit(as.character(uri_vec), "/+")
#     mat_str  <- suppressWarnings(do.call(rbind, str_list))
#     
#     if (ncol(mat_str) < 5) {
#       warning(
#         "CELLAR URI structure not as expected; cannot derive Prefix/ConceptScheme. Classifications from CELLAR will not be returned.",
#         call. = FALSE
#       )
#       return(NULL)
#     }
#     
#     uri <- paste0(mat_str[, 1], "/", "/", mat_str[, 2], "/", mat_str[, 3], "/", mat_str[, 4])
#     prefix <- gsub("\\.", "", mat_str[, 4])
#     conceptscheme <- mat_str[, 5]
#     
#     out <- data.frame(
#       Prefix        = prefix,
#       ConceptScheme = conceptscheme,
#       URI           = uri,
#       Title         = title,
#       stringsAsFactors = FALSE
#     )
#     rownames(out) <- NULL
#     out
#   }
#   
#   #-----------------------------#
#   # Helper : FAO                #
#   #-----------------------------#
#   fetch_fao <- function(
#     endpoint_fao = getOption(
#       "correspondenceTables.fao_endpoint",
#       "https://caliper.integratedmodelling.org/caliper/sparql"
#     )
#   ) {
#     SPARQL.query_fao <- "
#       PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
#       PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
#       PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
# 
#       SELECT ?scheme ?notation ?label_en WHERE {
#         ?scheme rdf:type skos:ConceptScheme .
#         ?scheme skos:notation ?notation .
#         ?scheme skos:prefLabel ?label_en . FILTER(lang(?label_en)='en') .
#       }
#       ORDER BY ASC(?notation)
#     "
#     
#     resp <- tryCatch(
#       httr::POST(
#         url   = endpoint_fao,
#         httr::accept("text/csv"),
#         body  = list(query = SPARQL.query_fao),
#         encode = "form"
#       ),
#       error = function(e) {
#         warning(
#           sprintf(
#             "FAO endpoint is currently unavailable (network error: %s). Classifications from FAO will not be returned.",
#             conditionMessage(e)
#           ),
#           call. = FALSE
#         )
#         return(NULL)
#       }
#     )
#     
#     if (is.null(resp)) return(NULL)
#     
#     if (httr::http_error(resp)) {
#       status <- httr::status_code(resp)
#       warning(
#         sprintf(
#           "FAO endpoint is currently unavailable (HTTP status %s). Classifications from FAO will not be returned.",
#           status
#         ),
#         call. = FALSE
#       )
#       return(NULL)
#     }
#     
#     data_fao <- utils::read.csv(
#       text = httr::content(resp, "text", encoding = "UTF-8"),
#       sep  = ",",
#       stringsAsFactors = FALSE
#     )
#     
#     if (nrow(data_fao) == 0L) {
#       warning("FAO SPARQL query returned no rows. Classifications from FAO will not be returned.",
#               call. = FALSE)
#       return(NULL)
#     }
#     
#     uri_vec <- data_fao[, 1]
#     prefix  <- data_fao[, 2]
#     
#     str_list   <- strsplit(uri_vec, "/")
#     mat_str_dt <- suppressWarnings(do.call(rbind, str_list))
#     df_str_dt  <- as.data.frame(mat_str_dt, stringsAsFactors = FALSE)
#     
#     if (ncol(df_str_dt) >= 4) {
#       uri_root <- paste(df_str_dt[, 1], df_str_dt[, 2], df_str_dt[, 3], df_str_dt[, 4], sep = "/")
#       uri <- paste(uri_root, prefix, sep = "/")
#     } else {
#       uri <- uri_vec
#     }
#     
#     if (ncol(df_str_dt) >= 6) {
#       ConceptScheme <- paste0(df_str_dt[, 5], df_str_dt[, 6])
#     } else if (ncol(df_str_dt) >= 5) {
#       ConceptScheme <- df_str_dt[, 5]
#     } else {
#       ConceptScheme <- prefix
#     }
#     
#     out <- data.frame(
#       Prefix        = prefix,
#       ConceptScheme = ConceptScheme,
#       URI           = uri,
#       Title         = data_fao[, 3],
#       stringsAsFactors = FALSE
#     )
#     rownames(out) <- NULL
#     out
#   }
#   
#   #-----------------------------#
#   # Corps principal             #
#   #-----------------------------#
#   res <- list()
#   
#   if (endpoint %in% c("ALL", "CELLAR")) {
#     cellar <- fetch_cellar()
#     if (!is.null(cellar)) res$CELLAR <- cellar
#   }
#   
#   if (endpoint %in% c("ALL", "FAO")) {
#     fao <- fetch_fao()
#     if (!is.null(fao)) res$FAO <- fao
#   }
#   
#   if (endpoint == "ALL") {
#     if (!("CELLAR" %in% names(res))) {
#       message("CELLAR endpoint is unavailable; only FAO classifications are returned.")
#     }
#     if (!("FAO" %in% names(res))) {
#       message("FAO endpoint is unavailable; only CELLAR classifications are returned.")
#     }
#   }
#   
#   if (endpoint != "ALL" && length(res) == 0L) {
#     stop("Endpoint '", endpoint, "' is currently unavailable (see warnings above).", call. = FALSE)
#   }
#   
#   if (endpoint == "ALL" && length(res) == 0L) {
#     stop("No endpoint could be reached (CELLAR and FAO both failed).", call. = FALSE)
#   }
#   
#   # ---- Logique de sortie ----
#   if (endpoint == "ALL") {
#     return(res)      # LISTE
#   } else {
#     return(res[[1]]) # DATA.FRAME
#   }
# }

#' Internal helper to correct raw classification tables
#'
#' This function applies a set of prefix-specific correction rules to a
#' classification table (e.g. from CELLAR or FAO). It is for internal use only
#' and is not part of the public API.
#'
#' @keywords internal
#' @noRd
ClassificationCorrection <- function(classification, prefix) {
  # Basic checks on inputs
  if (!is.data.frame(classification)) {
    stop("`classification` must be a data.frame.")
  }
  
  if (missing(prefix) || !is.character(prefix) || length(prefix) != 1L) {
    stop("`prefix` must be a single character string (e.g. 'nace2', 'cn2022').")
  }
  
  # Normalise prefix (we expect lower-case, consistent with classificationList)
  prefix <- tolower(prefix)
  
  # Identify code and label columns:
  # - code column is assumed to be named as the prefix (e.g. "cn2022")
  # - label column is assumed to be "Name"
  code_col  <- prefix
  label_col <- "Name"
  
  if (!(code_col %in% colnames(classification))) {
    stop(paste0("Column '", code_col, "' not found in `classification`."))
  }
  
  if (!(label_col %in% colnames(classification))) {
    stop("Column 'Name' not found in `classification`.")
  }
  
  # Save original codes and number of rows to detect corrections later
  original_codes <- classification[[code_col]]
  original_nrow  <- nrow(classification)
  
  # Create normalized temporary columns for internal processing
  classification$Code  <- classification[[code_col]]
  classification$Label <- classification[[label_col]]
  
  # ------------------------------------------------------------------
  # 1) Add letters to codes for NACE, NACE 2.1, CPA21, ISIC rev.4
  # ------------------------------------------------------------------
  if (prefix %in% c("nace2", "nace21", "cpa21", "isicrev4")) {
    
    two_digits <- substr(classification$Code, 1, 2)
    
    add_letter <- function(rows, letter) {
      if (length(rows) > 0L) {
        classification$Code[rows] <<- paste0(letter, classification$Code[rows])
      }
    }
    
    add_letter(which(two_digits %in% c("01","02","03")), "A")
    add_letter(which(two_digits %in% c("05","06","07","08","09")), "B")
    add_letter(which(two_digits %in% sprintf("%02d", 10:33)), "C")
    add_letter(which(two_digits == "35"), "D")
    add_letter(which(two_digits %in% c("36","37","38","39")), "E")
    add_letter(which(two_digits %in% c("41","42","43")), "F")
    add_letter(which(two_digits %in% c("45","46","47")), "G")
    add_letter(which(two_digits %in% c("49","50","51","52","53")), "H")
    add_letter(which(two_digits %in% c("55","56")), "I")
    add_letter(which(two_digits %in% c("58","59","60","61","62","63")), "J")
    add_letter(which(two_digits %in% c("64","65","66")), "K")
    add_letter(which(two_digits == "68"), "L")
    add_letter(which(two_digits %in% c("69","70","71","72","73","74","75")), "M")
    add_letter(which(two_digits %in% c("77","78","79","80","81","82")), "N")
    add_letter(which(two_digits == "84"), "O")
    add_letter(which(two_digits == "85"), "P")
    add_letter(which(two_digits %in% c("86","87","88")), "Q")
    add_letter(which(two_digits %in% c("90","91","92","93")), "R")
    add_letter(which(two_digits %in% c("94","95","96")), "S")
    add_letter(which(two_digits %in% c("97","98")), "T")
    add_letter(which(two_digits == "99"), "U")
  }
  
  # ------------------------------------------------------------------
  # 2) ECOICOP: remove trailing ".0" for main groups 10.0, 11.0, 12.0
  # ------------------------------------------------------------------
  if (prefix == "ecoicop") {
    idx <- which(classification$Code %in% c("10.0", "11.0", "12.0"))
    classification$Code[idx] <- c("10", "11", "12")
  }
  
  # ------------------------------------------------------------------
  # 3) PRODCOM 2019: remove weird codes "00.99.t" and "00.99.z"
  # ------------------------------------------------------------------
  if (prefix == "prodcom2019") {
    weird <- which(classification$Code %in% c("00.99.t", "00.99.z"))
    if (length(weird) > 0L) {
      classification <- classification[-weird, , drop = FALSE]
    }
  }
  
  # ------------------------------------------------------------------
  # 4) CN classifications: remove section rows (codes containing letters)
  # ------------------------------------------------------------------
  if (prefix %in% c("cn2017","cn2018","cn2019","cn2020",
                    "cn2021","cn2022","cn2023")) {
    has_letters <- grepl("[A-Za-z]", classification$Code)
    classification <- classification[!has_letters, , drop = FALSE]
  }
  
  # ------------------------------------------------------------------
  # 5) CBF: remove trailing "." from the code (if present)
  # ------------------------------------------------------------------
  if (prefix == "cbf10") {
    classification$Code <- sub("\\.$", "", classification$Code)
  }
  
  # Write the corrected code back into the original code column
  classification[[code_col]] <- classification$Code
  
  # Drop internal helper columns
  classification$Code  <- NULL
  classification$Label <- NULL
  
  # Detect if any correction was applied
  corrections_applied <- !identical(classification[[code_col]], original_codes) ||
    nrow(classification) != original_nrow
  
  attr(classification, "corrections_applied") <- corrections_applied
  
  classification
}

#' @title Retrieve a full classification table from CELLAR or FAO
#'
#' @description
#' Retrieves the complete structure of a statistical classification published in
#' either CELLAR or FAO repositories based on a Prefix–ConceptScheme pair obtained
#' using \code{classificationList()}.
#'
#' The resulting table is suitable for classification browsing, integrity checks,
#' hierarchical analysis, documentation, and downstream correspondence mapping.
#'
#' @param endpoint Character. Repository to query. Must be either "CELLAR" or "FAO".
#' @param prefix Character. Classification prefix used for matching and URI resolution
#'   (e.g. "cn2022", "cpc21", "isic4").
#' @param conceptScheme Character. Local identifier of the scheme (often identical to
#'   \code{prefix}). The function automatically resolves this to the canonical
#'   ConceptScheme URI published in the endpoint.
#' @param language Character. Desired language for labels, scope notes and exclusion notes.
#'   Default: "en".
#' @param level Character. One of:
#'   \itemize{
#'     \item "ALL" (default): return all levels in the hierarchy;
#'     \item a specific depth value (e.g. "2") to filter concepts at that depth only.
#'   }
#' @param showQuery Logical.
#'   \itemize{
#'     \item FALSE (default): returns only the classification table;
#'     \item TRUE: returns a list containing the SPARQL query, the resolved scheme URI, and the table itself.
#'   }
#' @param knownSchemes Optional. A data.frame supplying authoritative mappings of the
#'   form \code{Prefix, ConceptScheme, URI[, Endpoint]}. When provided, this overrides
#'   automatic discovery. To be obtained using \code{classificationList(endpoint)}.
#' @param preferMappingOnly Logical. If TRUE, the function never attempts SPARQL
#'   discovery and uses only information in \code{knownSchemes} or \code{classificationList(endpoint)}.
#'   Default: FALSE.
#'
#' @return
#' If \code{showQuery = FALSE}, a \code{data.frame} with one row per concept and columns:
#' \itemize{
#'   \item \strong{Concept}: full concept URI;
#'   \item \strong{Code}: \code{skos:notation} coerced to string;
#'   \item \strong{Label}: preferred or alternative label in the requested language;
#'   \item \strong{Depth}: \code{xkos:depth} if available;
#'   \item \strong{BroaderList}: broader concept URIs (pipe-separated);
#'   \item \strong{BroaderCodeList}: broader concept notations (pipe-separated);
#'   \item \strong{IncludeNotes}: \code{skos:scopeNote} values (double-pipe-separated);
#'   \item \strong{ExcludeNotes}: \code{xkos:exclusionNote} values (double-pipe-separated).
#' }
#'
#' If \code{showQuery = TRUE}, returns a list with:
#' \itemize{
#'   \item \code{SPARQL.query} – the executed SPARQL string;
#'   \item \code{scheme_uri} – the resolved ConceptScheme URI;
#'   \item \code{ClassificationTable} – the data.frame described above.
#' }
#'
#' @examples
#' \dontrun{
#' # Retrieve full CN2022 classification (English)
#' cn <- retrieveClassificationTable(
#'   endpoint      = "CELLAR",
#'   prefix        = "cn2022",
#'   conceptScheme = "cn2022",
#'   language      = "en"
#' )
#'
#' # Retrieve CPC v2.1 from FAO and inspect query
#' out <- retrieveClassificationTable(
#'   endpoint      = "FAO",
#'   prefix        = "cpc21",
#'   conceptScheme = "cpc21",
#'   showQuery     = TRUE
#' )
#' cat(out$SPARQL.query)
#' }
#'
#' @import httr
#' @export

retrieveClassificationTable <- function(endpoint,
                                        prefix,
                                        conceptScheme,
                                        language  = "en",
                                        level     = "ALL",
                                        showQuery = FALSE,
                                        knownSchemes = NULL,
                                        preferMappingOnly = FALSE) {
  # --- Endpoint validation & URL resolution (using internal helpers) ----------
  endpoint <- .validate_endpoints(endpoint)   # normalizes and validates
  if (identical(endpoint, "ALL")) {
    stop("'ALL' is not allowed in retrieveClassificationTable(); provide a single endpoint.",
         call. = FALSE)
  }
  endpoint_url <- .sparql_endpoint(endpoint)
  
  # --- Sanitize basic inputs --------------------------------------------------
  prefix        <- trimws(prefix)
  conceptScheme <- trimws(conceptScheme)
  level         <- trimws(level)
  
  # --- Resolve ConceptScheme URI (mapping -> discovery -> CELLAR fallback) ----
  scheme_uri <- .resolve_scheme_uri(
    endpoint_url      = endpoint_url,
    endpoint          = endpoint,
    prefix            = prefix,
    conceptScheme     = conceptScheme,
    language          = language,
    knownSchemes      = knownSchemes,      
    preferMappingOnly = preferMappingOnly  # TRUE = mapping only, no discovery/fallback
  )
  
  if (length(scheme_uri) != 1L || is.na(scheme_uri) || !nzchar(scheme_uri)) {
    stop("Prefix or ConceptScheme not present in this endpoint.", call. = FALSE)
  }
  
  
  # --- Existence check (extra robustness) ------------------------------------
  
  ask_ok <- try(.ask_scheme_exists(endpoint_url, scheme_uri), silent = TRUE)
  
  if (inherits(ask_ok, "try-error")) {
    stop(sprintf("Scheme URI '%s' could not be checked against endpoint '%s'.",
                 scheme_uri, endpoint),
         call. = FALSE)
  }
  
  if (!isTRUE(ask_ok)) {
    
    # If mapping was provided and preferMappingOnly=TRUE -> the mapping entry is invalid
    if (!is.null(knownSchemes) && isTRUE(preferMappingOnly)) {
      stop(sprintf("The URI '%s' from knownSchemes is not valid for endpoint '%s'.",
                   scheme_uri, endpoint),
           call. = FALSE)
    }
    
    stop(paste0(
      "Prefix or ConceptScheme not present in this endpoint: ", endpoint,
      " (resolved URI was: ", scheme_uri, ")"
    ), call. = FALSE)
  }
  
  # --- SPARQL prefixes-------
  prefix_block <- paste(
    "PREFIX skos: <http://www.w3.org/2004/02/skos/core#>",
    "PREFIX xkos: <http://rdf-vocabulary.ddialliance.org/xkos#>",
    "PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>",
    sep = "\n"
  )
  
  # --- Optional depth filter -------
  filter_depth_inner <- ""
  if (!identical(toupper(level), "ALL")) {
    filter_depth_inner <- paste0("FILTER (STR(?Depth0) = '", level, "')")
  }
  
  # --- Query (flat select + normalized aggregations)-----
  SPARQL.query <- paste0(
    prefix_block, "\n",
    "SELECT ?Concept\n",
    "       (MIN(STR(?Notation)) AS ?Code)\n",
    "       (SAMPLE(?Label0)     AS ?Label)\n",
    "       (SAMPLE(?Depth0)     AS ?Depth)\n",
    "       (GROUP_CONCAT(DISTINCT STR(?Broader);         separator=\" | \")  AS ?BroaderList)\n",
    "       (GROUP_CONCAT(DISTINCT STR(?BroaderNotation); separator=\" | \")  AS ?BroaderCodeList)\n",
    "       (GROUP_CONCAT(DISTINCT ?Include;              separator=\" || \") AS ?IncludeNotes)\n",
    "       (GROUP_CONCAT(DISTINCT ?Exclude;              separator=\" || \") AS ?ExcludeNotes)\n",
    "WHERE {\n",
    "  ?Concept a skos:Concept ; skos:inScheme <", scheme_uri, "> ; skos:notation ?Notation .\n",
    "  # Depth from concept (fallback)\n",
    "  OPTIONAL { ?Concept xkos:depth ?DepthFromConcept }\n",
    "  # Depth via classification level (common case)\n",
    "  OPTIONAL {\n",
    "    ?Concept ^skos:member ?Level .\n",
    "    ?Level a xkos:ClassificationLevel .\n",
    "    ?Level xkos:depth ?DepthFromLevel .\n",
    "  }\n",
    "  BIND (COALESCE(?DepthFromLevel, ?DepthFromConcept) AS ?Depth0)\n",
    "  OPTIONAL { ?Concept skos:broader ?Broader . ?Broader skos:notation ?BroaderNotation }\n",
    "  OPTIONAL { ?Concept skos:prefLabel ?Pref FILTER (LANG(?Pref) = '", language, "') }\n",
    "  OPTIONAL { ?Concept skos:altLabel  ?Alt  FILTER (LANG(?Alt)  = '", language, "') }\n",
    "  BIND (COALESCE(?Pref, ?Alt) AS ?Label0)\n",
    "  OPTIONAL { ?Concept skos:scopeNote     ?Include FILTER (LANG(?Include) = '", language, "') }\n",
    "  OPTIONAL { ?Concept xkos:exclusionNote ?Exclude FILTER (LANG(?Exclude) = '", language, "') }\n",
    if (nzchar(filter_depth_inner)) paste0("  ", filter_depth_inner, "\n") else "",
    "}\n",
    "GROUP BY ?Concept\n",
    "ORDER BY ?Code\n"
  )
  
  # --- Execute: POST (CSV) with GET fallback ----
  resp <- httr::POST(
    url    = endpoint_url,
    body   = list(query = SPARQL.query, format = "text/csv"),
    encode = "form",
    httr::accept("text/csv"),
    httr::user_agent("correspondenceTables (R)")
  )
  
  if (httr::status_code(resp) >= 400L) {
    resp_get <- httr::GET(
      url   = endpoint_url,
      query = list(query = SPARQL.query, format = "text/csv"),
      httr::accept("text/csv"),
      httr::user_agent("correspondenceTables (R)")
    )
    httr::stop_for_status(resp_get)
    csv_text <- httr::content(resp_get, as = "text", encoding = "UTF-8")
  } else {
    csv_text <- httr::content(resp, as = "text", encoding = "UTF-8")
    # Rare: server responds XML on POST -> retry with GET
    if (grepl("^\\s*<\\?xml|<sparql", csv_text, ignore.case = TRUE)) {
      resp_get <- httr::GET(
        url   = endpoint_url,
        query = list(query = SPARQL.query, format = "text/csv"),
        httr::accept("text/csv"),
        httr::user_agent("correspondenceTables (R)")
      )
      httr::stop_for_status(resp_get)
      csv_text <- httr::content(resp_get, as = "text", encoding = "UTF-8")
    }
  }
  
  data <- .read_sparql_csv(csv_text)
  
  # --- Cleanup--
  data[] <- lapply(data, function(x) if (is.character(x)) gsub("\n", " ", x) else x)
  
  # --- Return --
  if (isTRUE(showQuery)) {
    return(list(SPARQL.query = SPARQL.query, scheme_uri = scheme_uri, ClassificationTable = data))
  } else {
    return(data)
  }
}

# Internal helpers -------------------

# Read CSV with row-name sanitization (prevents duplicate colname issues)
#' @keywords internal
#' @noRd
.read_sparql_csv <- function(csv_text) {
  if (!nzchar(csv_text)) return(data.frame())
  nl <- regexpr("\n", csv_text, fixed = TRUE)
  if (nl > 0) {
    header <- substr(csv_text, 1, nl - 1)
    body   <- substr(csv_text, nl + 1, nchar(csv_text))
    header <- sub("^row.names(,|$)", "row_names\\1", header)
    header <- gsub(",row.names,", ",row_names,", header, fixed = TRUE)
    header <- sub(",row.names$", ",row_names", header)
    csv_text <- paste0(header, "\n", body)
  }
  utils::read.csv(
    textConnection(csv_text),
    stringsAsFactors = FALSE,
    row.names = NULL,
    check.names = FALSE,
    comment.char = ""
  )
}

# ASK if scheme exists on the endpoint
#' @keywords internal
#' @noRd
.ask_scheme_exists <- function(endpoint_url, scheme_uri) {
  ask_q <- sprintf("
PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
ASK { ?c skos:inScheme <%s> }", scheme_uri)
  resp <- httr::GET(
    url   = endpoint_url,
    query = list(query = ask_q),
    httr::add_headers(Accept = "text/boolean, text/plain, application/sparql-results+json"),
    httr::user_agent("correspondenceTables (R) / ASK")
  )
  httr::stop_for_status(resp)
  txt <- httr::content(resp, as = "text", encoding = "UTF-8")
  grepl("\\btrue\\b", tolower(txt))
}

# Lightweight ConceptScheme discovery (SPARQL)
#' @keywords internal
#' @noRd
.discover_scheme_uri <- function(endpoint_url, prefix, conceptScheme, language = "en") {
  probe <- sprintf("
PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
SELECT ?scheme ?label WHERE {
  ?scheme a skos:ConceptScheme ;
          skos:prefLabel ?label .
  FILTER (LANG(?label) = '%s')
  FILTER (
    CONTAINS(LCASE(STR(?scheme)), LCASE('%s')) ||
    CONTAINS(LCASE(STR(?scheme)), LCASE('%s')) ||
    CONTAINS(LCASE(STR(?label)),  LCASE('%s')) ||
    CONTAINS(LCASE(STR(?label)),  LCASE('%s'))
  )
}
ORDER BY ?scheme", language, prefix, conceptScheme, prefix, conceptScheme)
  
  resp <- httr::GET(
    url   = endpoint_url,
    query = list(query = probe),
    httr::add_headers(Accept = "application/sparql-results+csv"),
    httr::user_agent("correspondenceTables (R) / SPARQL")
  )
  httr::stop_for_status(resp)
  csv <- httr::content(resp, as = "text", encoding = "UTF-8")
  df  <- .read_sparql_csv(csv)
  if (nrow(df)) {
    score <- function(u, lbl) {
      s <- 0L
      if (grepl("/scheme/?$", u)) s <- s + 3L
      if (grepl(prefix, u, ignore.case = TRUE)) s <- s + 2L
      if (grepl(conceptScheme, u, ignore.case = TRUE)) s <- s + 2L
      if (grepl(prefix, lbl, ignore.case = TRUE)) s <- s + 1L
      if (grepl(conceptScheme, lbl, ignore.case = TRUE)) s <- s + 1L
      s
    }
    df$._score <- mapply(score, df$scheme, df$label, USE.NAMES = FALSE)
    df <- df[order(-df$._score, df$scheme), ]
    return(df$scheme[1])
  }
  
  # broader probe on prefix only
  probe2 <- sprintf("
PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
SELECT ?scheme ?label WHERE {
  ?scheme a skos:ConceptScheme ;
          skos:prefLabel ?label .
  FILTER (LANG(?label) = '%s')
  FILTER (
    CONTAINS(LCASE(STR(?scheme)), LCASE('%s')) ||
    CONTAINS(LCASE(STR(?label)),  LCASE('%s'))
  )
}
ORDER BY ?scheme
LIMIT 50", language, prefix, prefix)
  
  resp2 <- httr::GET(
    url   = endpoint_url,
    query = list(query = probe2),
    httr::add_headers(Accept = "application/sparql-results+csv"),
    httr::user_agent("correspondenceTables (R) / SPARQL")
  )
  httr::stop_for_status(resp2)
  csv2 <- httr::content(resp2, as = "text", encoding = "UTF-8")
  df2  <- .read_sparql_csv(csv2)
  if (nrow(df2)) {
    df2$._ends_scheme <- grepl("/scheme/?$", df2$scheme)
    df2 <- df2[order(-as.integer(df2$._ends_scheme), df2$scheme), ]
    return(df2$scheme[1])
  }
  
  NA_character_
}

# Normalize an external mapping table (columns and trimming)
#' @keywords internal
#' @noRd
.normalize_known_schemes <- function(df) {
  if (is.null(df) || !is.data.frame(df) || !nrow(df)) return(NULL)
  nms <- names(df)
  nms <- sub("^concept\\s*scheme$", "ConceptScheme", nms, ignore.case = TRUE)
  nms <- sub("^conceptscheme$",     "ConceptScheme", nms, ignore.case = TRUE)
  nms <- sub("^concept_scheme$",    "ConceptScheme", nms, ignore.case = TRUE)
  nms <- sub("^prefix$",            "Prefix",        nms, ignore.case = TRUE)
  nms <- sub("^uri$",               "URI",           nms, ignore.case = TRUE)
  nms <- sub("^source$",            "Endpoint",      nms, ignore.case = TRUE)  # optional
  names(df) <- nms
  
  req_cols <- c("Prefix", "ConceptScheme", "URI")
  if (!all(req_cols %in% names(df))) return(NULL)
  
  keep <- intersect(c(req_cols, "Endpoint"), names(df))
  out  <- df[, keep, drop = FALSE]
  
  for (col in intersect(c("Prefix", "ConceptScheme", "URI", "Endpoint"), names(out))) {
    if (is.character(out[[col]])) out[[col]] <- trimws(out[[col]])
  }
  out
}

# Get mapping from override or classificationList(); uses internal endpoint helpers
#' @keywords internal
#' @noRd
.get_known_schemes_from_classificationList <- function(endpoint,
                                                       knownSchemes_override = NULL) {
  endpoint <- .validate_endpoints(endpoint)  
  
  # 1) external override (if provided)
  ks <- .normalize_known_schemes(knownSchemes_override) #if NULL, it stays NULL
  if (!is.null(ks)) {
    if ("Endpoint" %in% names(ks)) {
      ks_src <- ks[toupper(ks$Endpoint) == endpoint, setdiff(names(ks), "Endpoint"), drop = FALSE]
      if (nrow(ks_src)) return(ks_src)
      # otherwise: use the whole table
    }
    return(ks)
  }
  
  # 2) fallback: call classificationList(endpoint) (public function)
  if (!exists("classificationList", mode = "function")) return(NULL)
  df <- try(classificationList(endpoint), silent = TRUE)
  if (inherits(df, "try-error") || is.null(df)) return(NULL)
  ks2 <- .normalize_known_schemes(df)
  if (is.null(ks2)) return(NULL)
  ks2[, c("Prefix","ConceptScheme","URI"), drop = FALSE]
}

# Resolver: mapping -> discovery -> (CELLAR pattern fallback), all via internal helpers
#' @keywords internal
#' @noRd
.resolve_scheme_uri <- function(endpoint_url,
                                endpoint,
                                prefix,
                                conceptScheme,
                                language = "en",
                                knownSchemes = NULL,
                                preferMappingOnly = FALSE) {
  endpoint <- .validate_endpoints(endpoint)
  
  # 1) authoritative mapping (override -> classificationList())
  known_schemes <- .get_known_schemes_from_classificationList(
    endpoint = endpoint,
    knownSchemes_override = knownSchemes
  )
  
  if (!is.null(known_schemes) && nrow(known_schemes)) {
    ks_pref   <- tolower(trimws(known_schemes[["Prefix"]]))
    ks_cs     <- tolower(trimws(known_schemes[["ConceptScheme"]]))
    want_pref <- tolower(trimws(prefix))
    want_cs   <- tolower(trimws(conceptScheme))
    idx <- which(ks_pref == want_pref & ks_cs == want_cs)
    if (length(idx)) {
      hit <- known_schemes[idx, , drop = FALSE]
      # heuristic: prefer URIs that end with /scheme
      ends_scheme <- grepl("/scheme/?$", hit[["URI"]])
      ord <- order(!ends_scheme, hit[["URI"]], decreasing = TRUE)
      hit <- hit[ord, , drop = FALSE]
      uri <- hit[["URI"]][1]
      
      # NEW: if mapping-only, return the mapped URI AS-IS (ASK is done by caller)
      if (isTRUE(preferMappingOnly)) return(uri)
      # Otherwise (not mapping-only), do a sanity ASK here to avoid extra hop later
      if (isTRUE(.ask_scheme_exists(endpoint_url, uri))) return(uri)
      # If ASK fails and not mapping-only, continue to discovery/fallback below
    } else if (isTRUE(preferMappingOnly)) {
      # Mapping-only and no pair in mapping -> hard NA
      return(NA_character_)
    }
  } else if (isTRUE(preferMappingOnly)) {
    # Mapping-only and no mapping table available at all -> hard NA
    return(NA_character_)
  }
  
  # 2) SPARQL discovery (ONLY if not mapping-only)
  if (!isTRUE(preferMappingOnly)) {
    disc <- try(.discover_scheme_uri(endpoint_url, prefix, conceptScheme, language), silent = TRUE)
    if (!inherits(disc, "try-error") && is.character(disc) && nzchar(disc)) {
      if (.ask_scheme_exists(endpoint_url, disc)) return(disc)
    }
    
    # 3) CELLAR-only fallback pattern (validated by ASK)
    if (identical(endpoint, "CELLAR")) {
      pat <- paste0("http://data.europa.eu/xsp/", tolower(prefix), "/", conceptScheme)
      if (.ask_scheme_exists(endpoint_url, pat)) return(pat)
    }
  }
  
  NA_character_
}

# ---- 1) rejects invalid endpoint -------------------------------------------

expect_error(
  retrieveClassificationTable(
    endpoint      = "INVALID",
    prefix        = "cn2022",
    conceptScheme = "cn2022"
  ),
  pattern = "Unsupported endpoint"
)



if(at_home()){
  
  # ---- 2) returns a data.frame when showQuery = FALSE ------------------------
  
  res_df <- 
    retrieveClassificationTable(
      endpoint      = "FAO",
      prefix        = "coicop1999",
      conceptScheme = "scheme",
      language      = "en",
      level         = "ALL",
      showQuery     = FALSE
      
    )
  
  expect_inherits(res_df, "data.frame")
  # Columns depend on endpoint response; keep only stable checks.
  expect_true(all(c("Concept", "Code", "Label", "Depth") %in% names(res_df)))
  
  # ---- 3) returns query and table when showQuery = TRUE ----------------------
  
  
  
  
  res_q <- (
    retrieveClassificationTable(
      endpoint      = "FAO",
      prefix        = "coicop1999",
      conceptScheme = "scheme",
      language      = "en",
      level         = "ALL",
      showQuery     = TRUE
    )
  )
  
  expect_true(is.list(res_q))
  expect_true(all(c("SPARQL.query", "scheme_uri", "ClassificationTable") %in% names(res_q)))
  expect_true(is.character(res_q$SPARQL.query))
  expect_true(is.character(res_q$scheme_uri))
  expect_inherits(res_q$ClassificationTable, "data.frame")
  
  # ---- 4) level filter is applied when level != ALL ----------------
  # This adds a light semantic check on 'level' using the live endpoint.
  
  
  
  
  df_all <- (
    retrieveClassificationTable(
      endpoint      = "FAO",
      prefix        = "coicop1999",
      conceptScheme = "scheme",
      language      = "en",
      level         = "ALL",
      showQuery     = FALSE
    )
  )
  
  df_2 <- (
    retrieveClassificationTable(
      endpoint      = "FAO",
      prefix        = "coicop1999",
      conceptScheme = "scheme",
      language      = "en",
      level         = "2",
      showQuery     = FALSE
    )
  )
  
  expect_inherits(df_all, "data.frame")
  expect_inherits(df_2,  "data.frame")
  expect_true(nrow(df_2) <= nrow(df_all))
  # Depth may be character; compare as character to avoid coercion issues
  if ("Depth" %in% names(df_2) && nrow(df_2) > 0) {
    expect_true(all(as.character(df_2$Depth) == "2"))
  }
  
}

# Tests for knownSchemes & preferMappingOnly
# These tests do NOT require network calls.

# Fake knownSchemes mapping (valid)
ks <- data.frame(
  Prefix        = "cbf",
  ConceptScheme = "cbf",
  URI           = "http://data.europa.eu/38u/cbf1.0/cbf",
  stringsAsFactors = FALSE
)

res <- retrieveClassificationTable(
  endpoint      = "CELLAR",
  prefix        = "cbf",
  conceptScheme = "cbf",
  knownSchemes  = ks,
  showQuery     = TRUE,  # to extract resolved scheme_uri
  preferMappingOnly = TRUE # to stick to the given dataset
)

# The scheme_uri MUST equal the mapping (no SPARQL discovery)
expect_true(is.list(res))
expect_true(res$scheme_uri == "http://data.europa.eu/38u/cbf1.0/cbf")



ks_matto <- data.frame(
  Prefix        = c("c","aooo"),
  ConceptScheme = c("c","aooo"),
  URI           = c("http://example.com/fake/scheme","casa"),
  stringsAsFactors = FALSE
)


expect_error(
  retrieveClassificationTable(
    endpoint      = "CELLAR",
    prefix        = "c",
    conceptScheme = "c",
    knownSchemes  = ks_matto,
    showQuery     = TRUE,  
    preferMappingOnly = T
  ),
  pattern = "The URI 'http://example.com/fake/scheme' from knownSchemes is not valid for endpoint 'CELLAR'"
)


# Mapping intentionally misses the (prefix, scheme) pair
ks_missing <- data.frame(
  Prefix        = "OTHER",
  ConceptScheme = "OTHER",
  URI           = "http://example.com/other/scheme",
  stringsAsFactors = FALSE
)

expect_error(
  retrieveClassificationTable(
    endpoint          = "CELLAR",
    prefix            = "cbf",
    conceptScheme     = "cbf",
    knownSchemes      = ks_missing,
    preferMappingOnly = TRUE
  ),
  pattern = "Prefix or ConceptScheme not present in this endpoint."
)



# Fake knownSchemes mapping (valid)
ks <- data.frame(
  Prefix        = "cbf",
  ConceptScheme = "cbf",
  URI           = "http://data.europa.eu/38u/cbf1.0/cbf",
  stringsAsFactors = FALSE
)

res <- retrieveClassificationTable(
  endpoint      = "CELLAR",
  prefix        = "cbf",
  conceptScheme = "cbf",
  knownSchemes  = ks,
  showQuery     = TRUE,  # to extract resolved scheme_uri
  preferMappingOnly = TRUE # to stick to the given dataset
)

# The scheme_uri MUST equal the mapping (no SPARQL discovery)
expect_true(is.list(res))
expect_true(res$scheme_uri == "http://data.europa.eu/38u/cbf1.0/cbf")



ks_matto <- data.frame(
  Prefix        = c("c","aooo"),
  ConceptScheme = c("c","aooo"),
  URI           = c("http://example.com/fake/scheme","casa"),
  stringsAsFactors = FALSE
)


expect_error(
  retrieveClassificationTable(
    endpoint      = "CELLAR",
    prefix        = "c",
    conceptScheme = "c",
    knownSchemes  = ks_matto,
    showQuery     = TRUE,  
    preferMappingOnly = T
  ),
  pattern = "The URI 'http://example.com/fake/scheme' from knownSchemes is not valid for endpoint 'CELLAR'"
)



# Mapping intentionally misses the (prefix, scheme) pair
ks_missing <- data.frame(
  Prefix        = "OTHER",
  ConceptScheme = "OTHER",
  URI           = "http://example.com/other/scheme",
  stringsAsFactors = FALSE
)

expect_error(
  retrieveClassificationTable(
    endpoint          = "CELLAR",
    prefix            = "cbf",
    conceptScheme     = "cbf",
    knownSchemes      = ks_missing,
    preferMappingOnly = TRUE
  ),
  pattern = "Prefix or ConceptScheme not present in this endpoint"
)

#' @title Correspondence table creation
#'
#' @description
#' Update the correspondence table between two classifications when one of them has been updated.
#'
#' @param A A data frame containing the original classification A. The first column holds the codes;
#'          optional subsequent columns may hold labels and supplementary info.
#' @param B A data frame containing the target classification B. Same structure expectations as \code{A}.
#' @param AStar A data frame containing the updated version A*. Same structure expectations as \code{A}.
#' @param AB A data frame containing the previous correspondence table A:B (at least two columns: A-codes, B-codes).
#' @param AAStar A data frame containing the concordance table A:A* (at least two columns: A-codes, A*-codes).
#' @param Reference The reference classification among A and B. Valid values: \code{"none"}, \code{"A"}, \code{"B"}.
#'                  If \code{"A"} or \code{"B"}, a "Review" flag column is included (see "Explanation of the flags").
#' @param MismatchToleranceB Maximum acceptable proportion of rows in the updated table with no code of B
#'                           among those that have a code for A and/or A*. Default \code{0.2}, range [0,1].
#' @param MismatchToleranceAStar Maximum acceptable proportion of rows in the updated table with no code of A*
#'                               among those that have a code for A and/or B. Default \code{0.2}, range [0,1].
#' @param Redundancy_trim Logical. If \code{TRUE} (default) trims correct redundant records by collapsing A-side fields to "Multiple"
#'                        when appropriate; if \code{FALSE}, keeps redundancies and adds a \code{Redundancy_keep} indicator.
#'
#' @details
#' *Input data frame requirements*
#' \itemize{
#'   \item \code{A}, \code{B}, \code{AStar}: at least 1 column and at least 1 row. The first column contains the codes; the first row is the header.
#'   \item \code{AB}, \code{AAStar}: at least 2 columns and at least 1 row; the first two columns contain code pairs; the first row is the header.
#'   \item Codes in the first column of \code{A}, \code{B}, \code{AStar} must be unique (no duplicates).
#'   \item Pairs (first two columns) in \code{AB} and \code{AAStar} must be unique (no duplicate pairs).
#'   \item If additional columns are present in classification or correspondence tables, the second column is treated as a label.
#'   \item All columns are coerced to character.
#' }
#'
#' *Minimum interdependencies*
#' \itemize{
#'   \item At least one code of A must appear in both \code{AAStar} and \code{AB}.
#'   \item At least one code of A* must appear in both \code{AStar} and \code{AAStar}.
#'   \item At least one code of B must appear in both \code{B} and \code{AB}.
#' }
#'
#' *Mismatch tolerance*
#' \itemize{
#'   \item If the share of rows with \code{NoMatchToAStar == 1} exceeds \code{MismatchToleranceAStar}, the function stops with an error.
#'   \item If the share of rows with \code{NoMatchToB == 1} exceeds \code{MismatchToleranceB}, the function stops with an error.
#' }
#'
#' @return
#' \code{updateCorrespondenceTable()} returns a list with two data frames:
#' \itemize{
#'   \item \code{updateCorrespondenceTable}: the updated correspondence A*:B with flags "CodeChange", "Review" (if applicable),
#'         "Redundancy", "NoMatchToAStar", "NoMatchToB", "NoMatchFromAStar", "NoMatchFromB", "LabelChange",
#'         plus any additional columns coming from \code{A}, \code{B}, \code{AStar}, \code{AB}, \code{AAStar}.
#'   \item \code{classificationNames}: the names of the classifications (A, B, A*) read from the first column headers.
#' }
#'
#' @section Explanation of the flags:
#'
#' \itemize{
#'     \item For each row of the updated correspondence table, the value of "CodeChange" is equal to \code{1} if the code of A (or A*)
#'      contained in this row maps -in this or any other row of the table- to a different code of A* (or A), otherwise the 
#'      "CodeChange" is equal to \code{0}. The value of "CodeChange" is empty if either the code of A, or the code of A*, or both are missing.
#'     \item The "Review" flag is produced only if argument \code{Reference} has been set equal to "\code{A}" or "\code{B}".
#'     For each row of the updated correspondence table, if \code{Reference} = "\code{A}" the value of "Review" is equal to
#'     \code{1} if the code of B maps to more than one code of A*, and \code{0} otherwise. If \code{Reference} = "\code{B}" the
#'     value of "Review" is equal to \code{1} if the code of A* maps to more than one code of B, and \code{0} otherwise. The value
#'     of the flag is empty if either the code of A*, or the code of B, or both are missing.
#'     \item For each row of the updated correspondence table, the value of "Redundancy" is equal to \code{1} if the row contains
#'     a combination of codes of A* and B that also appears in at least one other row of the updated correspondence table. The
#'     value of the flag is empty if both the code of A* and the code of B are missing.
#'     \item When "Redundancy_Trim" is equal to \code{FALSE} the "Redundancy_keep" flag is created to identify with value \code{1}
#'     the records that will be kept if trimming is performed.
#'     \item For each row of the updated correspondence table, the value of "NoMatchToAStar" is equal to \code{1} if there is a
#'     code for A,  for B, or for both, but no code for A*. The value of the flag is \code{0} if there are codes for both A and
#'     A* (regardless of whether there is a code for B or not). Finally, the value of "NoMatchToAStar" is empty if neither A nor B
#'     have a code in this row.
#'     \item For each row of the updated correspondence table, the value of "NoMatchToB" is equal to \code{1} if there is a code
#'     for A,  for A*, or for both, but no code for B. The value of the flag is \code{0} if there are codes for both A and B
#'     (regardless of whether there is a code for A* or not). Finally, the value of "NoMatchToB" is empty if neither A nor
#'     A* have a code in this row.
#'     \item For each row of the updated correspondence table, the value of "NoMatchFromAStar" is equal to \code{1} if the row
#'     contains a code of A* that appears in the table of classification A* but not in the concordance table A:A*. The value of
#'     the flag is \code{0} if the row contains a code of A* that appears in both the table of classification
#'     A* and the concordance table A:A*. Finally, the value of the flag is empty if the row contains no code of A* or if it
#'     contains a code of A* that appears in the concordance table A:A* but not in the table of classification A*.
#'     \item For each row of the updated correspondence table, the value of "NoMatchFromB" is equal to \code{1} if the row
#'     contains a code of B that appears in the table of classification B but not in the correspondence table A:B. The value of
#'     the flag is \code{0} if the row contains a code of B that appears in both the table of classification B and the
#'     correspondence table A:B. Finally, the value of the flag is empty if the row contains no code of B or if it contains a code
#'     of B that appears in the correspondence table A:B but not in the table of classification B.
#'     \item For each row of the updated correspondence table, the value of "LabelChange" is equal to \code{1} if the labels of
#'     the codes of A and A* are different, and \code{0} if they are the same. Finally, the value of "LabelChange" is empty if
#'     either of the labels, or both labels, are missing. Lower and upper case are considered the same, and punctuation characters
#'     are ignored when comparing code labels.
#'     \item The argument "Redundancy_trim" is used to delete all the redundancies which are mapping correctly. If the analysis 
#'    concludes that the A*code / Bcode mapping is correct for all cases involving redundancies, then an action is needed to remove 
#'    the redundancies. If the selected value is \code{TRUE}, all redundant records are removed and kept only one record for each unique 
#'    combination. For this record retained, the Acodes, the Alabel and the Asupp information is replaced with ‘multiple’. If the multiple 
#'    A records are the same, their value will not be replaced. If the selected value is \code{FALSE}, no trimming is executed so redundant 
#'    records are shown, together with the redundancy flag.
#' }
#'
#' @examples
#' \dontrun{
#' 
#'   # Read CSVs outside, pass data frames in:
#'   
#'   A_df      <- utils::read.csv(system.file("extdata/test", "NAICS2017.csv",
#'   package = "correspondenceTables"),
#'   sep = ",", header = TRUE, check.names = FALSE, 
#'   colClasses = "character", encoding = "UTF-8")
#'                                
#'   AStar_df  <- utils::read.csv(system.file("extdata/test", "NAICS2022.csv",
#'   package = "correspondenceTables"),
#'   sep = ",", header = TRUE, check.names = FALSE,
#'   colClasses = "character", encoding = "UTF-8")
#'                                
#'   B_df      <- utils::read.csv(system.file("extdata/test", "NACE.csv",
#'   package = "correspondenceTables"),
#'   sep = ",", header = TRUE, check.names = FALSE,
#'   colClasses = "character", encoding = "UTF-8")
#'   
#'   AB_df     <- utils::read.csv(system.file("extdata/test", "NAICS2017_NACE.csv",
#'   package = "correspondenceTables"),
#'   sep = ",", header = TRUE, check.names = FALSE,
#'   colClasses = "character", encoding = "UTF-8")
#'                                
#'   AAStar_df <- utils::read.csv(system.file("extdata/test", "NAICS2017_NAICS2022.csv",
#'   package = "correspondenceTables"),
#'   sep = ",", header = TRUE, check.names = FALSE,
#'   colClasses = "character", encoding = "UTF-8")
#'
#'   UPC <- updateCorrespondenceTable(
#'     A = A_df, B = B_df, AStar = AStar_df, AB = AB_df, AAStar = AAStar_df,
#'     Reference = "none", MismatchToleranceB = 0.5, MismatchToleranceAStar = 0.3,
#'     Redundancy_trim = FALSE
#'   )
#'
#'   summary(UPC)
#'   head(UPC$updateCorrespondenceTable)
#'   UPC$classificationNames
#' }
#'
#' @export
updateCorrespondenceTable <- function(
    A, B, AStar, AB, AAStar,
    Reference = "none",
    MismatchToleranceB = 0.2,
    MismatchToleranceAStar = 0.2,
    Redundancy_trim = TRUE
) {
  # --- Helpers ---------------------------------------------------------------
  arg_names <- vapply(as.list(substitute(list(A, B, AStar, AB, AAStar)))[-1], deparse, "")
  names(arg_names) <- c("A", "B", "AStar", "AB", "AAStar")
  
  stop_if_not_df <- function(x, nm) {
    if (!is.data.frame(x)) {
      stop(
        simpleError(
          paste0(
            "As of this version, `updateCorrespondenceTable()` accepts only data frames. ",
            "Argument `", nm, "` is not a data frame.\n",
            "-> Read the CSV first (e.g., `read.csv()` or `readr::read_csv()`), then pass the data frame."
          )
        )
      )
    }
  }
  
  ensure_df_chars <- function(x, what) {
    stop_if_not_df(x, what)
    x[] <- lapply(x, function(col) if (is.factor(col)) as.character(col) else as.character(col))
    colnames(x) <- gsub("\\xef\\xbb\\xbf", "", colnames(x), useBytes = TRUE)  # strip BOM if present
    x
  }
  
  # --- Validate scalar params ------------------------------------------------
  if (!(Reference %in% c("A", "B", "none"))) {
    stop(simpleError('You entered a non-allowed value for Reference. Allowed: "A", "B", "none".'))
  }
  if (is.character(MismatchToleranceB) || MismatchToleranceB < 0 || MismatchToleranceB > 1) {
    stop(simpleError("MismatchToleranceB must be numeric in [0, 1]."))
  }
  if (is.character(MismatchToleranceAStar) || MismatchToleranceAStar < 0 || MismatchToleranceAStar > 1) {
    stop(simpleError("MismatchToleranceAStar must be numeric in [0, 1]."))
  }
  
  # --- Coerce/validate inputs as character data frames -----------------------
  classA     <- ensure_df_chars(A,      "A")
  classB     <- ensure_df_chars(B,      "B")
  classAStar <- ensure_df_chars(AStar,  "AStar")
  corrAB     <- ensure_df_chars(AB,     "AB")
  corrAAStar <- ensure_df_chars(AAStar, "AAStar")
  
  # --- Remove rows with empty first column & non-empty second in corr tables --
  NoCorrAAStar <- if (length(which(corrAAStar[, 1] == "" & corrAAStar[, 2] != "")) >= 1) {
    corrAAStar[-which(corrAAStar[, 1] == "" & corrAAStar[, 2] != ""), ]
  } else corrAAStar
  
  NoCorrAB <- if (length(which(corrAB[, 1] == "" & corrAB[, 2] != "")) >= 1) {
    corrAB[-which(corrAB[, 1] == "" & corrAB[, 2] != ""), ]
  } else corrAB
  
  # --- Dimension checks ------------------------------------------------------
  test.dimClass <- list(classA, classB, classAStar)
  for (i in 1:3) {
    if (ncol(test.dimClass[[i]]) < 1 || nrow(test.dimClass[[i]]) < 1) {
      stop(simpleError(
        paste0("Argument `", names(arg_names)[i], "` should have at least one column and one row (headers included).")
      ))
    }
  }
  
  test.dimCorr <- list(corrAB, corrAAStar)
  for (i in 1:2) {
    if (ncol(test.dimCorr[[i]]) <= 1 || nrow(test.dimCorr[[i]]) < 1) {
      stop(simpleError(
        paste0("Argument `", names(arg_names)[i + 3], "` should have at least two columns and one row (headers included).")
      ))
    }
  }
  
  # --- Uniqueness checks -----------------------------------------------------
  for (i in 1:3) {
    if (sum(duplicated(test.dimClass[[i]][, 1])) >= 1) {
      stop(simpleError(
        paste0("At least one code of ", colnames(test.dimClass[[i]])[1],
               " appears more than once in `", names(arg_names)[i], "`. Each code must appear only once.")
      ))
    }
  }
  
  for (i in 1:2) {
    if (nrow(test.dimCorr[[i]][, 1:2]) != nrow(unique(test.dimCorr[[i]][, 1:2]))) {
      stop(simpleError(
        paste0("At least one pair of codes (", colnames(test.dimCorr[[i]])[1], ", ",
               colnames(test.dimCorr[[i]])[2], ") appears more than once in `",
               names(arg_names)[i + 3], "`. Each pair must be unique.")
      ))
    }
  }
  
  # --- Interdependency checks ------------------------------------------------
  if (sum(!is.na(match(classAStar[, 1], corrAAStar[, 2]))) == 0) {
    stop(simpleError(
      paste0("No code of ", colnames(classAStar)[1], " appears in both `AStar` and `AAStar`. At least one is required.")
    ))
  }
  if (sum(!is.na(match(corrAAStar[, 1], corrAB[, 1]))) == 0) {
    stop(simpleError(
      paste0("No code of ", colnames(corrAAStar)[1], " appears in both `AAStar` and `AB`. At least one is required.")
    ))
  }
  if (sum(!is.na(match(classB[, 1], corrAB[, 2]))) == 0) {
    stop(simpleError(
      paste0("No code of ", colnames(classB)[1], " appears in both `B` and `AB`. At least one is required.")
    ))
  }
  if (sum(!is.na(match(classA[, 1], corrAAStar[, 1]))) == 0) {
    message(paste0("WARNING: No code of ", colnames(classA)[1], " appears in both `A` and `AAStar`. Please double-check."))
  }
  if (sum(!is.na(match(classA[, 1], corrAB[, 1]))) == 0) {
    message(paste0("WARNING: No code of ", colnames(classA)[1], " appears in both `A` and `AB`. Please double-check."))
  }
  
  # --- Set indices based on Reference ---------------------------------------
  tryCatch({
    if (Reference == "A") {
      idx <- 2; idx.thres <- 7
    } else if (Reference == "B") {
      idx <- 1; idx.thres <- 7
    } else { # "none"
      idx <- 3; idx.thres <- 6
    }
    
    # --- codeChange ----------------------------------------------------------
    codeChange <- function(X) {
      diffX <- as.matrix(X[, 1:2])
      Diff <- rep(0, nrow(diffX))
      Diff[which(diffX[, 1] != diffX[, 2])] <- 1
      difference <- data.frame(diffX, Diff)
      
      dif1 <- difference[order(difference[, 1]), ]
      dif1.values <- lapply(split(seq_along(dif1[, 1]), dif1[, 1]), function(x) dif1[x, 3])
      dif1 <- cbind(dif1, unlist(mapply(rep, lapply(dif1.values, max), lapply(dif1.values, length))))
      
      dif2 <- difference[order(difference[, 2]), ]
      dif2.values <- lapply(split(seq_along(dif2[, 2]), dif2[, 2]), function(x) dif2[x, 3])
      dif2 <- cbind(dif2, unlist(mapply(rep, lapply(dif2.values, max), lapply(dif2.values, length))))
      
      dif2.new <- dif2[order(dif2[, 1]), ]
      final.diff <- cbind(dif1, dif2.new[, 4])
      final.diff <- cbind(dif1, dif2.new[, 4], apply(final.diff[, 4:5], 1, max))
      final.diff
    }
    
    # --- review --------------------------------------------------------------
    review <- function(X, Y) {
      X1 <- X[!is.na(match(X[, 1], Y[, 1])), ]
      Y1 <- Y[!is.na(match(Y[, 1], X[, 1])), ]
      X1 <- X1[order(X1[, 1]), ]
      Y1 <- Y1[order(Y1[, 1]), ]
      
      x1 <- unlist(lapply(split(seq_along(X1[, 1]), X1[, 1]), length))
      y1 <- unlist(lapply(split(seq_along(Y1[, 1]), Y1[, 1]), length))
      Review <- data.frame(
        X1[rep(1:nrow(X1), rep(y1, x1)), 1:2],
        Y1[unlist(rep(split(seq_along(Y1[, 1]), Y1[, 1]), x1)), 2],
        0
      )
      colnames(Review) <- c("R1", "R2", "R3", "R4")
      
      q1 <- 0
      if (length(which(is.na(match(Y[, 1], X[, 1])) == TRUE)) >= 1) {
        Y2 <- matrix(unlist(Y[is.na(match(Y[, 1], X[, 1])), 1:2]), ncol = 2)
        Y2 <- cbind(Y2[, 1], rep("", nrow(Y2)), Y2[, 2], 0); colnames(Y2) <- c("R1", "R2", "R3", "R4")
        q1 <- 1
      }
      
      q2 <- 0
      if (length(which(is.na(match(X[, 1], Y[, 1])) == TRUE)) >= 1) {
        YY2 <- matrix(unlist(X[is.na(match(X[, 1], Y[, 1])), 1:2]), ncol = 2)
        YY2 <- cbind(YY2, rep("", nrow(YY2)), 0); colnames(YY2) <- c("R1", "R2", "R3", "R4")
        q2 <- 1
      }
      
      if (q1 == 1) Review <- rbind(Review, Y2)
      if (q2 == 1) Review <- rbind(Review, YY2)
      Review <- Review[!duplicated(Review[, c(1:3)]), ]
      
      F1 <- Review[apply(Review, 1, function(x) length(which(x == ""))) == 0, ]
      F2 <- Review[apply(Review, 1, function(x) length(which(x == ""))) >= 1, ]
      f  <- stats::aggregate(unique(F1[, 2:3])[, idx], list(num = unique(F1[, 2:3])[, idx]), length)[
        which(stats::aggregate(unique(F1[, 2:3])[, idx], list(num = unique(F1[, 2:3])[, idx]), length)[, 2] > 1), 1
      ]
      F1[which(F1[, idx + 1] %in% f), 4] <- 1
      Review <- rbind(F1, F2)
      ChangeReview <- data.frame(Review[, 1:3], 0, Review[, 4])
      ChangeReview[which(ChangeReview[, 1] %in% unique(codeChange(X)[which(codeChange(X)[, 6] == 1), 1])), 4] <- 1
      ChangeReview[which(ChangeReview[, 2] == ""), 4] <- ""
      ChangeReview
    }
    
    # --- redundancy (unchanged core) ----------------------------------------
    X1 <- NoCorrAAStar[!is.na(match(NoCorrAAStar[, 1], NoCorrAB[, 1])), ]
    Y1 <- NoCorrAB[!is.na(match(NoCorrAB[, 1], NoCorrAAStar[, 1])), ]
    X1 <- X1[order(X1[, 1]), ]; Y1 <- Y1[order(Y1[, 1]), ]
    x1 <- unlist(lapply(split(seq_along(X1[, 1]), X1[, 1]), length))
    y1 <- unlist(lapply(split(seq_along(Y1[, 1]), Y1[, 1]), length))
    Redun <- data.frame(X1[rep(1:nrow(X1), rep(y1, x1)), 1:2],
                        Y1[unlist(rep(split(seq_along(Y1[, 1]), Y1[, 1]), x1)), 2], 0)
    colnames(Redun) <- c("R1","R2","R3","R4")
    
    q1 <- 0
    if (length(which(is.na(match(NoCorrAB[, 1], NoCorrAAStar[, 1])) == TRUE)) >= 1) {
      Y2 <- matrix(unlist(NoCorrAB[is.na(match(NoCorrAB[, 1], NoCorrAAStar[, 1])), 1:2]), ncol = 2)
      Y2 <- cbind(Y2[, 1], rep("", nrow(Y2)), Y2[, 2], 0); colnames(Y2) <- c("R1","R2","R3","R4")
      q1 <- 1
    }
    q2 <- 0
    if (length(which(is.na(match(NoCorrAAStar[, 1], NoCorrAB[, 1])) == TRUE)) >= 1) {
      YY2 <- matrix(unlist(NoCorrAAStar[is.na(match(NoCorrAAStar[, 1], NoCorrAB[, 1])), 1:2]), ncol = 2)
      YY2 <- cbind(YY2, rep("", nrow(YY2)), 0); colnames(YY2) <- c("R1","R2","R3","R4")
      q2 <- 1
    }
    if (q1 == 1) Redun <- rbind(Redun, Y2)
    if (q2 == 1) Redun <- rbind(Redun, YY2)
    Redun <- Redun[!duplicated(Redun[, c(1:3)]), ]
    
    F1 <- Redun[apply(Redun, 1, function(x) length(which(x == ""))) == 0, ]
    F2 <- Redun[apply(Redun, 1, function(x) length(which(x == ""))) >= 1, ]
    
    if (idx != 3) {
      f1 <- stats::aggregate(F1[, 2:3], by = F1[, 2:3], length)[1:(ncol(F1[, 2:3]) + 1)][
        which(stats::aggregate(F1[, 2:3], by = F1[, 2:3], length)[1:(ncol(F1[, 2:3]) + 1)][, 3] >= 2), 1:2
      ]
      F1[which(apply(F1[, 2:3], 1, paste, collapse = " ") %in% apply(f1, 1, paste, collapse = " ")), 4] <- 1
      Redundancy <- rbind(F1, F2)
      ChangeReviewRedundancy <- data.frame(review(NoCorrAAStar, NoCorrAB), Redundancy[, 4])
      ChangeReviewRedundancy <- ChangeReviewRedundancy[order(ChangeReviewRedundancy[, 2], ChangeReviewRedundancy[, 3], decreasing = FALSE), ]
      List <- cbind(ChangeReviewRedundancy, 0, 0)
      List[which(ChangeReviewRedundancy[, 2] == ""), 7] <- 1
      List[which(ChangeReviewRedundancy[, 3] == ""), 8] <- 1
      colnames(List) <- c(colnames(classA[1]), colnames(classAStar[1]), colnames(classB[1]),
                          "CodeChange", "Review", "Redundancy", "NoMatchToAStar", "NoMatchToB")
    } else {
      if (nrow(F2) >= 1) { F2 <- cbind(F2, 0); colnames(F2) <- c("R1","R2","R3","R4","R5") }
      f1 <- stats::aggregate(F1[, 2:3], by = F1[, 2:3], length)[1:(ncol(F1[, 2:3]) + 1)][
        which(stats::aggregate(F1[, 2:3], by = F1[, 2:3], length)[1:(ncol(F1[, 2:3]) + 1)][, 3] >= 2), 1:2
      ]
      F1 <- data.frame(F1[, 1:3], 0, 0)
      F1[which(apply(F1[, 2:3], 1, paste, collapse = " ") %in% apply(f1, 1, paste, collapse = " ")), 5] <- 1
      colnames(F1) <- c("R1","R2","R3","R4","R5")
      ChangeRedundancy <- rbind(F1, F2)
      ChangeRedundancy[which(ChangeRedundancy[, 1] %in% unique(codeChange(NoCorrAAStar)[which(codeChange(NoCorrAAStar)[, 6] == 1), 1])), 4] <- 1
      ChangeRedundancy[which(ChangeRedundancy[, 2] == ""), 4] <- ""
      ChangeRedundancy <- ChangeRedundancy[order(ChangeRedundancy[, 2], ChangeRedundancy[, 3], decreasing = FALSE), ]
      List <- cbind(ChangeRedundancy, 0, 0)
      List[which(ChangeRedundancy[, 2] == ""), 6] <- 1
      List[which(ChangeRedundancy[, 3] == ""), 7] <- 1
      colnames(List) <- c(colnames(classA[1]), colnames(classAStar[1]), colnames(classB[1]),
                          "CodeChange", "Redundancy", "NoMatchToAStar", "NoMatchToB")
    }
    
    # --- Append missing-code rows & compute NoMatchFrom* ---------------------
    NoMatchFromAStar <- rep("", nrow(List))
    NoMatchFromB     <- rep("", nrow(List))
    List <- cbind(List, NoMatchFromAStar, NoMatchFromB)
    
    inA1 <- which(is.na(match(classA[, 1], corrAAStar[, 1])) == TRUE)
    inA2 <- which(is.na(match(classA[, 1], corrAB[, 1])) == TRUE)
    inA  <- intersect(inA1, inA2)
    if (length(inA) >= 1) {
      InA <- cbind(matrix(classA[inA, 1], length(inA), 1), matrix("", length(inA), 2), matrix("", length(inA), idx.thres - 2))
      InA <- cbind(InA, matrix("", length(inA), 2))
      InA <- data.frame(InA); colnames(InA) <- colnames(List); List <- rbind(List, InA)
    }
    
    inAStar <- which(is.na(match(classAStar[, 1], corrAAStar[, 2])) == TRUE)
    if (length(inAStar) >= 1) {
      InAStar <- cbind(matrix("", length(inAStar), 1), matrix(classAStar[inAStar, 1], length(inAStar), 1), matrix("", length(inAStar), idx.thres - 1))
      InAStar <- cbind(InAStar, matrix("", length(inAStar), 2))
      InAStar <- data.frame(InAStar); colnames(InAStar) <- colnames(List); List <- rbind(List, InAStar)
    }
    
    noInA <- which(corrAAStar[, 1] == "" & corrAAStar[, 2] != "")
    if (length(noInA) >= 1) {
      NoInA <- cbind(matrix("", length(noInA), 1), matrix(corrAAStar[noInA, 2], length(noInA), 1), matrix("", length(noInA), idx.thres - 1))
      NoInA <- cbind(NoInA, matrix("", length(noInA), 2))
      NoInA <- data.frame(NoInA); colnames(NoInA) <- colnames(List); List <- rbind(List, NoInA)
    }
    
    inB <- which(is.na(match(classB[, 1], corrAB[, 2])) == TRUE)
    if (length(inB) >= 1) {
      InB <- cbind(matrix("", length(inB), 2), matrix(classB[inB, 1], length(inB), 1), matrix("", length(inB), idx.thres - 2))
      InB <- cbind(InB, matrix("", length(inB), 2))
      InB <- data.frame(InB); colnames(InB) <- colnames(List); List <- rbind(List, InB)
    }
    
    noInB <- which(corrAB[, 1] == "" & corrAB[, 2] != "")
    if (length(noInB) >= 1) {
      NoInB <- cbind(matrix("", length(noInB), 2), matrix(corrAB[noInB, 2], length(noInB), 1), matrix("", length(noInB), idx.thres - 2))
      NoInB <- cbind(NoInB, matrix("", length(noInB), 2))
      NoInB <- data.frame(NoInB); colnames(NoInB) <- colnames(List); List <- rbind(List, NoInB)
    }
    
    yesAstarClass <- which(!is.na(match(List[, 2], classAStar[, 1])) == TRUE)
    yesAstarCorr  <- which(!is.na(match(List[, 2], corrAAStar[, 2])) == TRUE)
    noAstarCorr   <- which(is.na(match(List[, 2], corrAAStar[, 2])) == TRUE)
    List$NoMatchFromAStar[intersect(yesAstarClass, yesAstarCorr)] <- 0
    List$NoMatchFromAStar[intersect(yesAstarClass, noAstarCorr)]  <- 1
    
    yesBClass <- which(!is.na(match(List[, 3], classB[, 1])) == TRUE)
    yesBCorr  <- which(!is.na(match(List[, 3], corrAB[, 2])) == TRUE)
    noBCorr   <- which(is.na(match(List[, 3], corrAB[, 2])) == TRUE)
    List$NoMatchFromB[intersect(yesBClass, yesBCorr)] <- 0
    List$NoMatchFromB[intersect(yesBClass, noBCorr)]  <- 1
    
    # --- NoMatchTo flags -----------------------------------------------------
    noA <- which(List[, 1] == ""); yesA <- which(List[, 1] != "")
    noAstar <- which(List[, 2] == ""); yesAstar <- which(List[, 2] != "")
    noB <- which(List[, 3] == ""); yesB <- which(List[, 3] != "")
    
    List$NoMatchToAStar <- 1
    List$NoMatchToAStar[intersect(intersect(noA, noB), yesAstar)] <- ""
    List$NoMatchToAStar[intersect(intersect(yesA, yesAstar), noB)] <- 0
    List$NoMatchToAStar[intersect(intersect(yesA, yesAstar), yesB)] <- 0
    
    List$NoMatchToB <- 1
    List$NoMatchToB[intersect(intersect(noA, noAstar), yesB)] <- ""
    List$NoMatchToB[intersect(intersect(yesA, yesB), noAstar)] <- 0
    List$NoMatchToB[intersect(intersect(yesA, yesAstar), yesB)] <- 0
    
    # --- Final Review/Redundancy flags consistency --------------------------
    if (Reference %in% c("A", "B")) {
      List$Review[which(List[, 2] == "")] <- ""
      List$Review[which(List[, 3] == "")] <- ""
    }
    
    List$Redundancy <- 0
    f1 <- stats::aggregate(List[, 2:3], by = List[, 2:3], length)[1:(ncol(List[, 2:3]) + 1)][
      which(stats::aggregate(List[, 2:3], by = List[, 2:3], length)[1:(ncol(List[, 2:3]) + 1)][, 3] >= 2), 1:2
    ]
    List$Redundancy[which(apply(List[, 2:3], 1, paste, collapse = " ") %in% apply(f1, 1, paste, collapse = " "))] <- 1
    List$Redundancy[intersect(which(List[, 2] == ""), which(List[, 3] == ""))] <- ""
    
  }, error = function(e) {
    stop(simpleError("An error has occurred and execution needs to stop. Please check the input data."))
  })
  
  # --- Mismatch tolerance checks --------------------------------------------
  if (length(which(List[, idx.thres] == 1)) / length(which(List[, idx.thres] != "")) > MismatchToleranceAStar) {
    StopAAStar <- 1
  } else StopAAStar <- 0
  
  if (length(which(List[, idx.thres + 1] == 1)) / length(which(List[, idx.thres + 1] != "")) > MismatchToleranceB) {
    StopAB <- 1
  } else StopAB <- 0
  
  if (StopAAStar == 1 && StopAB == 0) {
    stop("Too many missing values for A* in the updated correspondence table (NoMatchToAStar).")
  } else if (StopAAStar == 0 && StopAB == 1) {
    stop("Too many missing values for B in the updated correspondence table (NoMatchToB).")
  } else if (StopAAStar == 1 && StopAB == 1) {
    stop("Too many missing values for both A* and B in the updated correspondence table.")
  } else {
    # --- Enrich with label/supp columns --------------------------------------
    tryCatch({
      if (ncol(classA) >= 2) {
        A1 <- as.matrix(classA[match(List[, 1], unlist(classA[, 1])), 2:ncol(classA)])
        A1[is.na(A1)] <- ""
        colnames(A1) <- paste(colnames(classA)[1], colnames(classA)[2:ncol(classA)], sep = "_")
        List <- cbind(List, A1)
      }
      if (ncol(classAStar) >= 2) {
        A2 <- as.matrix(classAStar[match(List[, 2], unlist(classAStar[, 1])), 2:ncol(classAStar)])
        A2[is.na(A2)] <- ""
        colnames(A2) <- paste(colnames(classAStar)[1], colnames(classAStar)[2:ncol(classAStar)], sep = "_")
        List <- cbind(List, A2)
      }
      if (ncol(classB) >= 2) {
        B1 <- as.matrix(classB[match(List[, 3], unlist(classB[, 1])), 2:ncol(classB)])
        B1[is.na(B1)] <- ""
        colnames(B1) <- paste(colnames(classB)[1], colnames(classB)[2:ncol(classB)], sep = "_")
        List <- cbind(List, B1)
      }
      if (ncol(corrAAStar) >= 3) {
        AA1 <- as.matrix(corrAAStar[match(data.frame(t(List[, c(1, 2)])), data.frame(t(corrAAStar[, 1:2]))), 3:ncol(corrAAStar)])
        AA1[is.na(AA1)] <- ""
        colnames(AA1) <- paste(paste(colnames(corrAAStar)[1], colnames(corrAAStar)[2], sep = " - "), colnames(corrAAStar)[3:ncol(corrAAStar)], sep = "_")
        List <- cbind(List, AA1)
      }
      if (ncol(corrAB) >= 3) {
        AB1 <- as.matrix(corrAB[match(data.frame(t(List[, c(1, 3)])), data.frame(t(corrAB[, 1:2]))), 3:ncol(corrAB)])
        AB1[is.na(AB1)] <- ""
        colnames(AB1) <- paste(paste(colnames(corrAB)[1], colnames(corrAB)[2], sep = " - "), colnames(corrAB)[3:ncol(corrAB)], sep = "_")
        List <- cbind(List, AB1)
      }
    }, error = function(e) {
      stop(simpleError("An error has occurred and execution needs to stop. Please check the input data."))
    })
  }
  
  # --- LabelChange -----------------------------------------------------------
  tryCatch({
    if (ncol(classA) >= 2 && ncol(classAStar) >= 2) {
      LabelChange <- rep(1, nrow(List))
      LabelChange[which(
        tolower(gsub("[[:punct:] ]+", " ", List[, idx.thres + 4])) ==
          tolower(gsub("[[:punct:] ]+", " ", List[, idx.thres + 4 + ncol(classA) - 1]))
      )] <- 0
      LabelChange[which(List[, idx.thres + 4] == "")] <- ""
      LabelChange[which(List[, idx.thres + 4 + ncol(classA) - 1] == "")] <- ""
      List <- cbind(List[, 1:(idx.thres + 3)], LabelChange, List[, (idx.thres + 4):ncol(List)])
    }
  }, error = function(e) {
    stop(simpleError("An error has occurred when computing LabelChange."))
  })
  
  # --- Redundancy trimming / keeping ----------------------------------------
  if (isTRUE(Redundancy_trim)) {
    col_multiple <- c(grep(colnames(List)[1], colnames(List), value = TRUE),
                      grep("LabelChange", colnames(List), value = TRUE))
    
    uniqueAstarB <- unique(List[which(List$Redundancy == 1), c(2, 3)])
    uniqueAstarB$id_to_use <- seq_len(nrow(uniqueAstarB))
    
    List <- merge(List, uniqueAstarB, by = colnames(List)[c(2, 3)], sort = FALSE, all.x = TRUE)[, union(names(List), names(uniqueAstarB))]
    col_link <- grep("id_to_use", colnames(List), value = TRUE)
    
    x_temp <- split(List[which(List$Redundancy == 1), col_multiple], List[which(List$Redundancy == 1), col_link])
    
    for (i in seq_len(nrow(uniqueAstarB))) {
      multiple_values <- apply(x_temp[[i]], 2, function(x) length(unique(x)))
      x_change <- which(multiple_values != 1)
      col_multiple_temp <- c(col_multiple)[x_change]
      List[which(List$Redundancy == 1 & List[, col_link] == i), col_multiple_temp] <- "Multiple"
    }
    
    List <- List[, !names(List) %in% "id_to_use"]
    
    dup <- as.numeric(duplicated(List[, 1:3]))
    List <- List[which(dup == 0), ]
    
    List[which(List$LabelChange == "Multiple"), grep("LabelChange", colnames(List), value = TRUE)] <- 1
    
  } else {
    dup <- as.numeric(duplicated(List[, c(2, 3)]))
    red_col <- which(colnames(List) == "Redundancy")
    List$Redundancy_keep <- rep(0, nrow(List))
    List$Redundancy_keep[which(dup == "0" & List$Redundancy == "1")] <- 1
    List <- List[, c(1:red_col, ncol(List), (red_col + 1):(ncol(List) - 1))]
  }
  
  # --- Classification names data frame --------------------------------------
  CsvNames <- data.frame(matrix(0, 3, 1))
  CsvNames[1, 1] <- paste("A:",     colnames(List)[1])
  CsvNames[2, 1] <- paste("B:",     colnames(List)[3])
  CsvNames[3, 1] <- paste("AStar:", colnames(List)[2])
  CsvNames <- data.frame(CsvNames)
  
  # --- Normalize spaces in Final (kept for parity) ---------------------------
  Final <- apply(List, 2, function(x) { gsub(" ", " ", x) })
  if (is.null(dim(Final))) {
    Final <- t(data.frame(Final)); rownames(Final) <- 1
  }
  
  # --- Return ---------------------------------------------------------------
  FinalResults <- list()
  FinalResults[[1]] <- data.frame(Final, check.names = FALSE, row.names = NULL)
  
  cols_to_convert <- c(
    "CodeChange", "Redundancy", "Redundancy_keep",
    "NoMatchToAStar", "NoMatchToB",
    "NoMatchFromAStar", "NoMatchFromB",
    "LabelChange", "Review"
  )
  
  for (col in cols_to_convert) {
    if (col %in% names(FinalResults[[1]])) {
      FinalResults[[1]][[col]] <- suppressWarnings(as.numeric(FinalResults[[1]][[col]]))
    }
  }
  FinalResults[[2]] <- CsvNames
  names(FinalResults) <- c("updateCorrespondenceTable", "classificationNames")
  FinalResults
}

############################################################
# TEST 1 — Redundancy trimming scenarios
############################################################

{
  A_df      <- utils::read.csv(system.file("extdata/test", "NACE2_Red.csv", package = "correspondenceTables"))
  AStar_df  <- utils::read.csv(system.file("extdata/test","NACE21_Red.csv", package = "correspondenceTables"))
  B_df      <- utils::read.csv(system.file("extdata/test","CPA_Red.csv", package = "correspondenceTables"))
  AB_df     <- utils::read.csv(system.file("extdata/test","NACECPA_Red.csv", package = "correspondenceTables"))
  AAStar_df <- utils::read.csv(system.file("extdata/test","NACE221_Red.csv", package = "correspondenceTables"))
  
  CT_nullT <- updateCorrespondenceTable(
    A = A_df, B = B_df, AStar = AStar_df, AB = AB_df, AAStar = AAStar_df,
    Reference = "none",
    MismatchToleranceB = 0.4, MismatchToleranceAStar = 0.4,
    Redundancy_trim = TRUE
  )
  CT_nullS <- updateCorrespondenceTable(
    A = A_df, B = B_df, AStar = AStar_df, AB = AB_df, AAStar = AAStar_df,
    Reference = "none",
    MismatchToleranceB = 0.4, MismatchToleranceAStar = 0.4,
    Redundancy_trim = FALSE
  )
  
  CT_AT <- updateCorrespondenceTable(
    A = A_df, B = B_df, AStar = AStar_df, AB = AB_df, AAStar = AAStar_df,
    Reference = "A",
    MismatchToleranceB = 0.4, MismatchToleranceAStar = 0.4,
    Redundancy_trim = TRUE
  )
  CT_AS <- updateCorrespondenceTable(
    A = A_df, B = B_df, AStar = AStar_df, AB = AB_df, AAStar = AAStar_df,
    Reference = "A",
    MismatchToleranceB = 0.4, MismatchToleranceAStar = 0.4,
    Redundancy_trim = FALSE
  )
  
  CT_BT <- updateCorrespondenceTable(
    A = A_df, B = B_df, AStar = AStar_df, AB = AB_df, AAStar = AAStar_df,
    Reference = "B",
    MismatchToleranceB = 0.4, MismatchToleranceAStar = 0.4,
    Redundancy_trim = TRUE
  )
  CT_BS <- updateCorrespondenceTable(
    A = A_df, B = B_df, AStar = AStar_df, AB = AB_df, AAStar = AAStar_df,
    Reference = "B",
    MismatchToleranceB = 0.4, MismatchToleranceAStar = 0.4,
    Redundancy_trim = FALSE
  )
  
  Ctext_nullT <- utils::read.csv(system.file("extdata/test","LblYY_RednullTrim.csv", package = "correspondenceTables"))
  Ctext_nullS <- utils::read.csv(system.file("extdata/test","LblYY_RednullShow.csv", package = "correspondenceTables"))
  
  Ctext_AT <- utils::read.csv(system.file("extdata/test","LblYY_RedATrim.csv", package = "correspondenceTables"))
  Ctext_AS <- utils::read.csv(system.file("extdata/test","LblYY_RedAShow.csv", package = "correspondenceTables"))
  
  Ctext_BT <- utils::read.csv(system.file("extdata/test","LblYY_RedBTrim.csv", package = "correspondenceTables"))
  Ctext_BS <- utils::read.csv(system.file("extdata/test","LblYY_RedBShow.csv", package = "correspondenceTables"),header = T)
  
  
  #transformation needed just for the test
  CT_nullT$updateCorrespondenceTable$NACE21code = as.numeric(CT_nullT$updateCorrespondenceTable$NACE21code)
  CT_nullT$updateCorrespondenceTable$CPAcode = as.numeric(CT_nullT$updateCorrespondenceTable$CPAcode)
  expect_equal(CT_nullT[[1]], Ctext_nullT)
  
  CT_AT$updateCorrespondenceTable$NACE21code = as.numeric(CT_AT$updateCorrespondenceTable$NACE21code)
  CT_AT$updateCorrespondenceTable$CPAcode = as.numeric(CT_AT$updateCorrespondenceTable$CPAcode)
  expect_equal(CT_AT[[1]],   Ctext_AT)
  
  CT_BT$updateCorrespondenceTable$NACE21code = as.numeric(CT_BT$updateCorrespondenceTable$NACE21code)
  CT_BT$updateCorrespondenceTable$CPAcode = as.numeric(CT_BT$updateCorrespondenceTable$CPAcode)
  expect_equal(CT_BT[[1]],   Ctext_BT)
  
  CT_nullS$updateCorrespondenceTable$NACE21code = as.numeric(CT_nullS$updateCorrespondenceTable$NACE21code)
  CT_nullS$updateCorrespondenceTable$NACE2code = as.numeric(CT_nullS$updateCorrespondenceTable$NACE2code)
  CT_nullS$updateCorrespondenceTable$CPAcode = as.numeric(CT_nullS$updateCorrespondenceTable$CPAcode)  
  expect_equal(CT_nullS[[1]], Ctext_nullS)
  
  CT_AS$updateCorrespondenceTable$NACE21code = as.numeric(CT_AS$updateCorrespondenceTable$NACE21code)
  CT_AS$updateCorrespondenceTable$NACE2code = as.numeric(CT_AS$updateCorrespondenceTable$NACE2code)
  CT_AS$updateCorrespondenceTable$CPAcode = as.numeric(CT_AS$updateCorrespondenceTable$CPAcode)  
  expect_equal(CT_AS[[1]],    Ctext_AS)
  
  CT_BS$updateCorrespondenceTable$NACE21code = as.numeric(CT_BS$updateCorrespondenceTable$NACE21code)
  CT_BS$updateCorrespondenceTable$NACE2code = as.numeric(CT_BS$updateCorrespondenceTable$NACE2code)
  CT_BS$updateCorrespondenceTable$CPAcode = as.numeric(CT_BS$updateCorrespondenceTable$CPAcode)    
  expect_equal(CT_BS[[1]],    Ctext_BS)
}


############################################################
# TEST 2 — Baseline (Reference = "A"), no trimming
############################################################
{
  A_df = read.csv(system.file("extdata/test", "NACE2.csv", package = "correspondenceTables"))
  AStar_df = read.csv(system.file("extdata/test", "NACE21.csv", package = "correspondenceTables"))
  B_df = read.csv(system.file("extdata/test", "CPA.csv", package = "correspondenceTables"))
  AB_df = read.csv(system.file("extdata/test", "NACECPA.csv", package = "correspondenceTables"))
  AAStar_df = read.csv(system.file("extdata/test", "NACE221.csv", package = "correspondenceTables"))
  
  CT_nullT <- updateCorrespondenceTable(
    A = A_df, B = B_df, AStar = AStar_df, AB = AB_df, AAStar = AAStar_df,
    Reference = "A",
    MismatchToleranceB = 0.4,
    MismatchToleranceAStar = 0.4,
    Redundancy_trim = FALSE
  )
  
  Ctext_nullT <- utils::read.csv(system.file("extdata/test","LblYY.csv", package = "correspondenceTables"))
  
  CT_nullT$updateCorrespondenceTable$NACE21code = as.numeric(CT_nullT$updateCorrespondenceTable$NACE21code)
  CT_nullT$updateCorrespondenceTable$NACE2code = as.numeric(CT_nullT$updateCorrespondenceTable$NACE2code)
  CT_nullT$updateCorrespondenceTable$CPAcode = as.numeric(CT_nullT$updateCorrespondenceTable$CPAcode)    
  expect_equal(CT_nullT[[1]], Ctext_nullT)
}


############################################################
# TEST 3 — “no” variants; compare first 10 columns
############################################################
{
  Ano_df = read.csv(system.file("extdata/test", "NACE2no.csv", package = "correspondenceTables"))
  AStarno_df = read.csv(system.file("extdata/test", "NACE21no.csv", package = "correspondenceTables"))
  Bno_df = read.csv(system.file("extdata/test", "CPA.csv", package = "correspondenceTables"))
  ABno_df = read.csv(system.file("extdata/test", "NACECPA.csv", package = "correspondenceTables"))
  AAStarno_df = read.csv(system.file("extdata/test", "NACE221.csv", package = "correspondenceTables"))
  
  CT_nullTno <- updateCorrespondenceTable(
    A = Ano_df, B = Bno_df, AStar = AStarno_df, AB = ABno_df, AAStar = AAStarno_df,
    Reference = "A",
    MismatchToleranceB = 0.4,
    MismatchToleranceAStar = 0.4,
    Redundancy_trim = FALSE
  )
  
  Ctext_nullTno <- utils::read.csv(system.file("extdata/test","LblNY.csv", package = "correspondenceTables"))
  
  CT_nullTno$updateCorrespondenceTable$NACE21code = as.numeric(CT_nullTno$updateCorrespondenceTable$NACE21code)
  CT_nullTno$updateCorrespondenceTable$NACE2code = as.numeric(CT_nullTno$updateCorrespondenceTable$NACE2code)
  CT_nullTno$updateCorrespondenceTable$CPAcode = as.numeric(CT_nullTno$updateCorrespondenceTable$CPAcode)    
  expect_equal(CT_nullTno[[1]][, 1:10], Ctext_nullTno[, 1:10])
}

############################################################
# TEST 4 — Upper-case fixtures; legend merge check
############################################################
{
  A_df      <- utils::read.csv(system.file("extdata/test","UcaseA.csv", package = "correspondenceTables"))
  AStar_df  <- utils::read.csv(system.file("extdata/test","UcaseAstar.csv", package = "correspondenceTables"))
  B_df      <- utils::read.csv(system.file("extdata/test","UcaseB.csv", package = "correspondenceTables"))
  AB_df     <- utils::read.csv(system.file("extdata/test","UcaseAB.csv", package = "correspondenceTables"))
  AAStar_df <- utils::read.csv(system.file("extdata/test","UcaseAAstar.csv", package = "correspondenceTables"))
  
  UCT <- updateCorrespondenceTable(
    A = A_df, B = B_df, AStar = AStar_df, AB = AB_df, AAStar = AAStar_df,
    Reference = "B",
    MismatchToleranceB = 0.8,
    MismatchToleranceAStar = 0.8,
    Redundancy_trim = FALSE
  )
  UCT[[1]]$sort <- seq_len(nrow(UCT[[1]]))
  
  UTC_T <- utils::read.csv(system.file("extdata/test","nomatch_test.csv", package = "correspondenceTables"))
  UTC_T$sort <- seq_len(nrow(UTC_T))
  
  legend <- utils::read.csv(system.file("extdata/test","legend_nomatch.csv", package = "correspondenceTables"))
  colnames(legend)[2:5] <- c("NoMatchToAStar","NoMatchToB","NoMatchFromAStar","NoMatchFromB")
  
  x <- as.data.frame(UCT[[1]])
  y <- as.data.frame(legend)
  
  merged <- merge(
    x, y,
    by = c("NoMatchToAStar","NoMatchToB","NoMatchFromAStar","NoMatchFromB"),
    all.x = TRUE
  )
  
  merged <- merged[order(as.numeric(merged$sort)), ]
  UTC_T  <- UTC_T[order(as.numeric(UTC_T$sort)), ]
  
  
  expect_equal(merged[,19], UTC_T[,8])
  
}

#' @title Perform analysis on correspondence tables
#' @description Perform analysis on a correspondence table using its associated
#' classifications to generate various statistics. It checks the validity of the
#' input data, identifies components, calculates correspondence types, and
#' creates summary tables.
#'
#' @param AB A mandatory argument containing a data frame provided by the user
#'   with the correspondence table data. The table must contain at least two
#'   columns containing the two classifications \code{A code} and \code{B code}. The name of the two columns is irrelevant.
#' @param Aname Optional. A character string specifying the name of the column
#'   in `AB` that should be used as the source classification (`Acode`). If
#'   `NULL`, the function assumes that the first column of `AB` represents the
#'   source classification.
#'
#' @param Bname Optional. A character string specifying the name of the column
#'   in `AB` that should be used as the target classification (`Bcode`). If
#'   `NULL`, the function assumes that the second column of `AB` represents the
#'   target classification.
#'    
#' @param A A data frame containing the source classification data with
#'   an \code{"Acode"} column. Can be \code{NULL} if no source classification
#'   table is provided.
#' @param B A data frame containing the target classification data with
#'   a \code{"Bcode"} column. Can be \code{NULL} if no target classification
#'   table is provided.
#' @param longestAcodeOnly Logical. If \code{TRUE}, source classification data
#'   are filtered based on \code{"Acode"} retaining only the maximum length,
#'   i.e. the lowest-level Acodes in the correspondence table.
#' @param longestBcodeOnly Logical. If \code{TRUE}, target classification data
#'   are filtered based on \code{"Bcode"} retaining only the maximum length,
#'   i.e. the lowest-level Bcodes in the correspondence table.
#'
#' @importFrom igraph graph.data.frame decompose.graph
#' @import igraph
#'
#' @return A list containing two data frames: \code{Inventory} and \code{Analysis}.
#'   \code{Inventory} contains statistics related to components, correspondence
#'   types, and source/target positions. \code{Analysis} contains statistics for
#'   each class in the correspondence table.
#'
#' @export
#'
#' @examples
#' # Use data from the folder extdata/test
#' 
#' AB = read.csv(system.file("extdata/test", "ExempleAnnexe.csv", package = "correspondenceTables"))
#' head(AB)
#' 
#' result <- analyseCorrespondenceTable(
#'   AB = AB,
#'   Aname = NULL,
#'   Bname = NULL,
#'   A  = NULL,
#'   B  = NULL,
#'   longestAcodeOnly          = FALSE,
#'   longestBcodeOnly          = FALSE
#' )
#' 
#' print(result$Inventory)
#' print(result$Analysis)
#' 
#' # the name of the columns does not matter:
#' AB_mod = AB
#' colnames(AB_mod) = c("a", "b")
#' 
#' result_mod <- analyseCorrespondenceTable(
#'   AB = AB_mod,
#'   Aname = NULL,
#'   Bname = NULL,
#'   A  = NULL,
#'   B  = NULL,
#'   longestAcodeOnly          = FALSE,
#'   longestBcodeOnly          = FALSE
#' )
#' 
#' print(result_mod$Inventory)
#' print(result_mod$Analysis)
#' 
#' 
#' # examples with using Aname and Bname
#' 
#' # fist we add two more column to the same dataset
#' set.seed(123)
#' AB_mod2 = AB_mod
#' 
#' AB_mod2 <- cbind(new_first_col = sample(1:100, nrow(AB_mod2), replace = TRUE), AB_mod2)
#' AB_mod2 <- cbind(
#'   AB_mod2[1:2], 
#'   new_middle_col = sample(1:100, nrow(AB_mod2), replace = TRUE),
#'   AB_mod2[3:ncol(AB_mod2)]
#' )
#' colnames(AB_mod2)
#' 
#' result_mod2 <- analyseCorrespondenceTable(
#'   AB = AB_mod2,
#'   Aname = "a",
#'   Bname = "b",
#'   A  = NULL,
#'   B  = NULL,
#'   longestAcodeOnly          = FALSE,
#'   longestBcodeOnly          = FALSE
#' )
#' 
#' print(result_mod2$Inventory)
#' print(result_mod2$Analysis)
#' 
#' # check if we obtained the same results in all the 3 examples
#' 
#' first = identical(result$Inventory, result_mod$Inventory) 
#' second = identical(result_mod$Inventory, result_mod2$Inventory)
#' first && second
#' # TRUE
#' 
#' # the following won't work as we changed column names in AB_mod, 
#' # and added new columns in AB_mod2 but the results are the same.
#' first = identical(result$Analysis, result_mod$Analysis)
#' second = identical(result_mod$Analysis, result_mod2$Analysis)
#' first && second



analyseCorrespondenceTable <- function(AB,
                                       Aname = NULL,
                                       Bname = NULL,
                                       A = NULL,
                                       B = NULL,
                                       longestAcodeOnly = FALSE,
                                       longestBcodeOnly = FALSE) {
  # checks  
  
  if (!inherits(AB, "data.frame")) {
    stop("AB must be a data.frame.")
  }
  
  if (!is.null(Aname) && !(is.character(Aname) && length(Aname) == 1)) {
    stop("Aname must be a character or NULL.")
  }
  
  if (!is.null(Bname) && !(is.character(Bname) && length(Bname) == 1)) {
    stop("Bname must be a character or NULL.")
  }
  
  if (!(is.null(Aname) == is.null(Bname))) {
    stop("Aname and Bname must be both NULL or both non-NULL.")
  }
  
  
  if (!is.null(Aname) && !is.null(Bname)) {
    
    # Check they exist in AB
    if (!all(c(Aname, Bname) %in% colnames(AB))) {
      stop("Both Aname and Bname must correspond to existing columns in AB.")
    }
    
    if (Aname == Bname) {
      stop("Aname and Bname must refer to two different columns.")
    }
  }
  
  
  if (!is.null(A) && !inherits(A, "data.frame")) {
    stop("A must be a data.frame or NULL.")
  }
  
  if (!is.null(B) && !inherits(B, "data.frame")) {
    stop("B must be a data.frame or NULL.")
  }
  
  if (!is.logical(longestAcodeOnly)) {
    stop("Argument 'longestAcodeOnly' must be TRUE or FALSE (logical).")
  }
  if (!is.logical(longestBcodeOnly)) {
    stop("Argument 'longestBcodeOnly' must be TRUE or FALSE (logical).")
  }
  
  # function
  
  ab_data = AB
  
  if (!is.null(Aname) && !is.null(Bname)){
    ColumnNames_ab = c(Aname, Bname)
    colnames(ab_data)[which(colnames(ab_data)==Aname)] = "Acode"
    colnames(ab_data)[which(colnames(ab_data)==Bname)] = "Bcode"
    unused_data_ab <- ab_data
    ab_data <- ab_data[,c("Acode","Bcode")]
    
  } else {
    ColumnNames_ab <- colnames(ab_data)[1:2]
    colnames(ab_data)[1:2] <- c("Acode", "Bcode")
    unused_data_ab <- ab_data
  }
  
  
  # Nrows > 0
  if (nrow(ab_data) == 0) {
    stop("No valid records found in the input correspondence table AB")
  }  
  # Missing value
  missing_code_rows <- ab_data[is.na(ab_data$Acode) | ab_data$Acode == "" | is.na(ab_data$Bcode) | ab_data$Bcode == "", ]
  if (nrow(missing_code_rows) > 0) stop(paste("Rows with missing values in the", ColumnNames_ab[1], "or", ColumnNames_ab[2], "column of the AB data:"))
  
  # Duplicates  Acode-Bcode check
  duplicate_pairs <- ab_data[duplicated(ab_data[c("Acode", "Bcode")]), c("Acode", "Bcode")]
  if (nrow(duplicate_pairs) > 0) {
    stop(paste0(
      "Duplicate row(s) found in AB file: the following Acode-Bcode pairs appear more than once:\n",
      paste(capture.output(print(duplicate_pairs)), collapse = "\n"),
      "\nPlease remove duplicate rows from your correspondence table."
    ))
  }
  
  
  # Filter rows in AB table based on longest Acode and/or Bcode
  if (longestAcodeOnly) {
    maxLengthA <- max(nchar(ab_data$Acode, type = "width"))
    ab_data <- ab_data[nchar(ab_data$Acode, type = "width") == maxLengthA, ]
  }
  
  if (longestBcodeOnly) {
    maxLengthB <- max(nchar(ab_data$Bcode, type = "width"))
    ab_data <- ab_data[nchar(ab_data$Bcode, type = "width") == maxLengthB, ]
  }
  
  # Final safety check
  if (nrow(ab_data) == 0) stop("No valid records after filtering longest codes.")
  
  
  if (!is.null(A)) {
    a_data <- A
    
    
    colnames(a_data)[1] <- "Acode"
    unused_data_a <- a_data
    
    if (nrow(a_data) == 0) stop("No valid records found in source classification table (A).")
    dups_A <- a_data$Acode[duplicated(a_data$Acode)]
    if (length(dups_A) > 0) {
      stop(paste0(
        "Duplicate Acode(s) found in A file:\n",
        paste(unique(dups_A), collapse = ", "),
        "\nEach Acode must be unique in the source classification."
      ))
    }
    
    unmatched_codes_A <- setdiff(a_data$Acode, ab_data$Acode)
    extra_codes_in_AB_A <- setdiff(ab_data$Acode, a_data$Acode)
    
    if (length(unmatched_codes_A) > 0) {
      warning(paste0(
        "Unmatched source classification codes in A: ",
        paste(unmatched_codes_A, collapse = ", ")
      ))
    }
    
    if (length(extra_codes_in_AB_A) > 0) {
      warning(paste0(
        "Source classification codes in AB not found in A: ",
        paste(extra_codes_in_AB_A, collapse = ", ")
      ))
    }
  }
  
  
  if (!is.null(B)) {
    b_data <- B
    
    colnames(b_data)[1] <- "Bcode"
    unused_data_b <- b_data
    
    if (nrow(b_data) == 0) stop("No valid records found in target classification table (B).")
    dups_B <- b_data$Bcode[duplicated(b_data$Bcode)]
    if (length(dups_B) > 0) {
      stop(paste0(
        "Duplicate Bcode(s) found in B file:\n",
        paste(unique(dups_B), collapse = ", "),
        "\nEach Bcode must be unique in the target classification."
      ))
    }
    
    unmatched_codes_B <- setdiff(b_data$Bcode, ab_data$Bcode)
    extra_codes_in_AB_B <- setdiff(ab_data$Bcode, b_data$Bcode)
    
    # Warnings explicites
    if (length(unmatched_codes_B) > 0) {
      warning(paste0(
        "Unmatched target classification codes in B: ",
        paste(unmatched_codes_B, collapse = ", ")
      ))
    }
    
    if (length(extra_codes_in_AB_B) > 0) {
      warning(paste0(
        "Target classification codes in AB not found in B: ",
        paste(extra_codes_in_AB_B, collapse = ", ")
      ))
    }
  }
  
  #graph biparti
  g <- igraph::graph_from_data_frame(ab_data, directed = FALSE)
  components <- igraph::decompose(g)
  component_codes <- lapply(components, function(comp) igraph::V(comp)$name)
  
  ab_data$component <- NA
  for (i in seq_along(component_codes)) {
    component <- component_codes[[i]]
    ab_data$component[ab_data$Acode %in% component | ab_data$Bcode %in% component] <- paste("Component", i)
  }
  
  component_stats <- lapply(components, function(comp) {
    component <- igraph::V(comp)$name
    n_unique_targets <- length(unique(ab_data[ab_data$Acode %in% component, "Bcode"]))
    
    correspondence_type <- if (n_unique_targets == 1) {
      if (length(unique(ab_data[ab_data$Bcode %in% component, "Acode"])) == 1) {
        "1:1"
      } else {
        "M:1"
      }
    } else {
      if (length(unique(ab_data[ab_data$Bcode %in% component, "Acode"])) == 1) {
        "1:M"
      } else {
        "M:M"
      }
    }
    
    source_positions <- unique(ab_data[ab_data$Acode %in% component, "Acode"])
    target_positions <- unique(ab_data[ab_data$Bcode %in% component, "Bcode"])
    n_source_positions <- length(source_positions)
    n_target_positions <- length(target_positions)
    
    component_name <- unique(ab_data$component[ab_data$Acode %in% component | ab_data$Bcode %in% component])
    
    list(
      Component = component_name,
      CorrespondenceType = correspondence_type,
      SourcePositions = source_positions,
      TargetPositions = target_positions,
      nSourcePositions = n_source_positions,
      nTargetPositions = n_target_positions
    )
  })
  
  Inventory <- do.call(rbind, lapply(component_stats, function(x) {
    data.frame(
      Component = x$Component,
      CorrespondenceType = x$CorrespondenceType,
      SourcePositions = I(list(x$SourcePositions)),
      TargetPositions = I(list(x$TargetPositions)),
      nSourcePositions = as.integer(x$nSourcePositions),
      nTargetPositions = as.integer(x$nTargetPositions),
      stringsAsFactors = FALSE
    )
  }))
  # Table Analysis
  Analysis <- data.frame(
    ClassC = ab_data$Acode,
    ClassD = ab_data$Bcode,
    nTargetClasses = NA,
    SourceToTargetMapping = NA,
    nSourceClasses = NA,
    TargetToSourceMapping = NA
  )
  
  Analysis$nTargetClasses <- sapply(Analysis$ClassC, function(c_code) length(unique(Analysis[Analysis$ClassC == c_code, "ClassD"])))
  Analysis$SourceToTargetMapping <- sapply(Analysis$ClassC, function(c_code) paste(unique(Analysis[Analysis$ClassC == c_code, "ClassD"]), collapse = ", "))
  Analysis$nSourceClasses <- sapply(Analysis$ClassD, function(d_code) length(unique(Analysis[Analysis$ClassD == d_code, "ClassC"])))
  Analysis$TargetToSourceMapping <- sapply(Analysis$ClassD, function(d_code) paste(unique(Analysis[Analysis$ClassD == d_code, "ClassC"]), collapse = ", "))
  
  colnames(Analysis)[1:2] <- c("Acode", "Bcode")
  
  Inventory_df <- Inventory
  Inventory_df$Component <- as.character(Inventory_df$Component)
  Inventory_df$CorrespondenceType <- as.character(Inventory_df$CorrespondenceType)
  Inventory_df$SourcePositions <- sapply(Inventory_df$SourcePositions, function(x) paste(x, collapse = ", "))
  Inventory_df$TargetPositions <- sapply(Inventory_df$TargetPositions, function(x) paste(x, collapse = ", "))
  
  
  Analysis_df <- merge(Analysis, unused_data_ab, by = c("Acode", "Bcode"), all = FALSE)
  
  if (!is.null(B)) {
    Analysis_df <- merge(Analysis_df, unused_data_b, by = "Bcode", all = FALSE)
  }
  
  if (!is.null(A)) {
    Analysis_df <- merge(Analysis_df, unused_data_a, by = "Acode", all = FALSE)
  }
  force_integer_if_whole <- function(df) {
    for (col in names(df)) {
      if (is.numeric(df[[col]]) && all(df[[col]] %% 1 == 0, na.rm = TRUE)) {
        df[[col]] <- as.integer(df[[col]])
      }
    }
    df
  }
  Analysis_df <- Analysis_df[, c("Acode", "Bcode", setdiff(names(Analysis_df), c("Acode", "Bcode")))]
  colnames(Analysis_df)[1:2] <- ColumnNames_ab[1:2]
  
  list(Inventory = Inventory_df, Analysis = Analysis_df)
}

############################################################
# [1] analyseCorrespondenceTable works with valid data and explicit A/B classifications
############################################################
{
  AB_path <- system.file("extdata/test", "TestAnalyse_ExempleAnnexe.csv", package = "correspondenceTables")
  A_path  <- system.file("extdata/test", "TestAnalyse_SourceClassification.csv", package = "correspondenceTables")
  B_path  <- system.file("extdata/test", "TestAnalyse_TargetClassification.csv", package = "correspondenceTables")
  
  AB_df <- read.csv(AB_path, encoding = "UTF-8", stringsAsFactors = FALSE)
  A_df  <- read.csv(A_path,  encoding = "UTF-8", stringsAsFactors = FALSE)
  B_df  <- read.csv(B_path,  encoding = "UTF-8", stringsAsFactors = FALSE)
  
  result <- analyseCorrespondenceTable(
    AB = AB_df, A = A_df, B = B_df,
    longestAcodeOnly = FALSE, longestBcodeOnly = FALSE
  )
  
  expect_true(is.list(result))
  expect_equal(names(result), c("Inventory", "Analysis"))
  
  expect_inherits(result$Inventory, "data.frame")
  expect_inherits(result$Analysis,  "data.frame")
  
  expect_true(nrow(result$Inventory) > 0)
  
  expect_equal(nrow(result$Analysis), nrow(AB_df))
  
  expected_cols_analysis <- c("Acode","Bcode","nTargetClasses","SourceToTargetMapping",
                              "nSourceClasses","TargetToSourceMapping")
  expect_true(all(expected_cols_analysis %in% colnames(result$Analysis)))
  
  expect_true(any(result$Analysis$Acode == "A1" & result$Analysis$Bcode == "B1"))
}

############################################################
# [2] emits warnings for unmatched codes in A and B
############################################################
{
  AB_path        <- system.file("extdata/test", "TestAnalyse_ExempleAnnexe.csv", package = "correspondenceTables")
  A_mismatch_path <- system.file("extdata/test", "TestAnalyse_SourceClassification_missing.csv", package = "correspondenceTables")
  B_mismatch_path <- system.file("extdata/test", "TestAnalyse_TargetClassification_missing.csv", package = "correspondenceTables")
  
  AB_df        <- read.csv(AB_path,         stringsAsFactors = FALSE)
  A_mismatchDF <- read.csv(A_mismatch_path, stringsAsFactors = FALSE)
  B_mismatchDF <- read.csv(B_mismatch_path, stringsAsFactors = FALSE)
  
  expect_warning(
    analyseCorrespondenceTable(
      AB = AB_df, A = A_mismatchDF, B = B_mismatchDF,
      longestAcodeOnly = FALSE, longestBcodeOnly = FALSE
    ),
    "Unmatched source classification codes in A"
  )
  
  expect_warning(
    analyseCorrespondenceTable(
      AB = AB_df, A = A_mismatchDF, B = B_mismatchDF,
      longestAcodeOnly = FALSE, longestBcodeOnly = FALSE
    ),
    "Unmatched target classification codes in B"
  )
}

############################################################
# [3] correctly filters longestAcodeOnly and longestBcodeOnly
############################################################
{
  AB_path <- system.file("extdata/test", "TestAnalyse_longest_AB.csv", package = "correspondenceTables")
  A_path  <- system.file("extdata/test", "TestAnalyse_longest_A.csv",  package = "correspondenceTables")
  B_path  <- system.file("extdata/test", "TestAnalyse_longest_B.csv",  package = "correspondenceTables")
  
  AB_df <- read.csv(AB_path, stringsAsFactors = FALSE)
  A_df  <- read.csv(A_path,  stringsAsFactors = FALSE)
  B_df  <- read.csv(B_path,  stringsAsFactors = FALSE)
  
  max_len_A <- max(nchar(as.character(A_df$A)))
  max_len_B <- max(nchar(as.character(B_df$B)))
  
  res_A <- suppressWarnings(analyseCorrespondenceTable(AB = AB_df, A = A_df, longestAcodeOnly = TRUE))
  expect_true(all(nchar(as.character(res_A$Inventory$Acode)) == max_len_A),
              info = "Some A codes are not of maximum length when longestAcodeOnly = TRUE")
  
  res_B <- suppressWarnings(analyseCorrespondenceTable(AB = AB_df, B = B_df, longestBcodeOnly = TRUE))
  expect_true(all(nchar(as.character(res_B$Inventory$Bcode)) == max_len_B),
              info = "Some B codes are not of maximum length when longestBcodeOnly = TRUE")
  
  res_both <- suppressWarnings(analyseCorrespondenceTable(
    AB = AB_df, A = A_df, B = B_df,
    longestAcodeOnly = TRUE, longestBcodeOnly = TRUE
  ))
  expect_true(all(nchar(as.character(res_both$Inventory$Acode)) == max_len_A),
              info = "Some A codes in combined test are not of maximum length")
  expect_true(all(nchar(as.character(res_both$Inventory$Bcode)) == max_len_B),
              info = "Some B codes in combined test are not of maximum length")
}

############################################################
# [4] detects duplicates in AB, A, and B (datasets loaded then passed as data.frames)
############################################################
{
  AB_dup_path <- system.file("extdata/test", "ExempleAnnexe_duplicated.csv", package = "correspondenceTables")
  A_path      <- system.file("extdata/test", "TestAnalyse_SourceClassification.csv", package = "correspondenceTables")
  B_path      <- system.file("extdata/test", "TestAnalyse_TargetClassification.csv", package = "correspondenceTables")
  
  A_dup_path  <- system.file("extdata/test", "SourceClassification_duplicated.csv", package = "correspondenceTables")
  B_dup_path  <- system.file("extdata/test", "TargetClassification_duplicated.csv", package = "correspondenceTables")
  
  AB_dup_df <- read.csv(AB_dup_path, stringsAsFactors = FALSE)
  A_df      <- read.csv(A_path,     stringsAsFactors = FALSE)
  B_df      <- read.csv(B_path,     stringsAsFactors = FALSE)
  
  A_dup_df  <- read.csv(A_dup_path, stringsAsFactors = FALSE)
  B_dup_df  <- read.csv(B_dup_path, stringsAsFactors = FALSE)
  
  expect_error(
    analyseCorrespondenceTable(AB = AB_dup_df, A = A_df, B = B_df),
    "Duplicate"
  )
  
  expect_error(
    analyseCorrespondenceTable(AB = AB_dup_df, A = A_dup_df, B = B_df),
    "Duplicate"
  )
  
  expect_error(
    analyseCorrespondenceTable(AB = AB_dup_df, A = A_df, B = B_dup_df),
    "Duplicate"
  )
}

############################################################
# [5] handles invalid input types (flags must be logical)
############################################################
{
  AB_path <- system.file("extdata/test", "TestAnalyse_ExempleAnnexe.csv", package = "correspondenceTables")
  AB_df   <- read.csv(AB_path, stringsAsFactors = FALSE)
  
  expect_error(
    analyseCorrespondenceTable(AB = AB_df, longestAcodeOnly = "YES"),
    "Argument 'longestAcodeOnly' must be TRUE or FALSE"
  )
  
  expect_error(
    analyseCorrespondenceTable(AB = AB_df, longestBcodeOnly = 1),
    "Argument 'longestBcodeOnly' must be TRUE or FALSE"
  )
}

############################################################
# [6] output structure of Inventory and Analysis is correct (loaded AB)
############################################################
{
  AB_path <- system.file("extdata/test", "ExempleAnnexe.csv", package = "correspondenceTables")
  AB_df   <- read.csv(AB_path, stringsAsFactors = FALSE)
  
  result <- analyseCorrespondenceTable(AB = AB_df)
  
  expect_true(is.list(result))
  expect_true( all(names(result)== c("Inventory", "Analysis")))
  
  expected_cols_inventory <- c("Component","CorrespondenceType","SourcePositions","TargetPositions",
                               "nSourcePositions","nTargetPositions")
  expected_cols_analysis  <- c("ClassA","ClassB","nTargetClasses",
                               "SourceToTargetMapping","nSourceClasses","TargetToSourceMapping")
  
  expect_inherits(result$Inventory, "data.frame")
  expect_true(all(expected_cols_inventory %in% names(result$Inventory)))
  
  expect_inherits(result$Analysis, "data.frame")
  expect_true(all(expected_cols_analysis %in% names(result$Analysis)))
  
  expect_true( class(result$Inventory$Component)==            "character")
  expect_true( class(result$Inventory$nSourcePositions) ==     "integer")
  expect_true( class(result$Inventory$nTargetPositions) ==     "integer")
  expect_true( class(result$Inventory$SourcePositions) ==      "character")
  expect_true( class(result$Inventory$TargetPositions) ==      "character")
  
  expect_true( class(result$Analysis$ClassA) ==                "character")
  expect_true( class(result$Analysis$ClassB) ==                "character")
  expect_true( class(result$Analysis$nTargetClasses) ==        "integer")
  expect_true( class(result$Analysis$nSourceClasses) ==        "integer")
  expect_true( class(result$Analysis$SourceToTargetMapping) == "character")
  expect_true( class(result$Analysis$TargetToSourceMapping) == "character")
}

############################################################
# [7] has AB, A, B as first arguments and runs on a simple in-memory example
############################################################
{
  
  
  args <- names(formals(analyseCorrespondenceTable))
  expect_identical(args, c("AB", "Aname", "Bname", "A", "B", "longestAcodeOnly", "longestBcodeOnly"))
  
  AB_df <- data.frame(
    Acode = c("A1","A2"),
    Bcode = c("B1","B2"),
    stringsAsFactors = FALSE
  )
  
  res <- analyseCorrespondenceTable(
    AB = AB_df, A = NULL, B = NULL,
    longestAcodeOnly = FALSE, longestBcodeOnly = FALSE
  )
  
  expect_true(is.list(res))
  expect_true(all(c("Inventory","Analysis") %in% names(res)))
}


# new parameters' tests!

############################################################
# [A1] Defaults: Aname/Bname NULL -> first two columns are used
############################################################
{
  AB_default <- data.frame(
    X = c("A1","A2","A2","A3"),
    Y = c("B1","B1","B2","B3"),
    Z = 1:4,
    stringsAsFactors = FALSE
  )
  
  # Case 1: Aname/Bname left NULL (first two columns used)
  res_null <- analyseCorrespondenceTable(
    AB = AB_default,
    Aname = NULL, Bname = NULL,
    A = NULL, B = NULL,
    longestAcodeOnly = FALSE, longestBcodeOnly = FALSE
  )
  
  # Case 2: Explicitly map the same columns via names
  res_named <- analyseCorrespondenceTable(
    AB = AB_default,
    Aname = "X", Bname = "Y",
    A = NULL, B = NULL,
    longestAcodeOnly = FALSE, longestBcodeOnly = FALSE
  )
  
  expect_true(is.list(res_null) && is.list(res_named))
  expect_true(all(c("Inventory","Analysis") %in% names(res_null)))
  expect_true(all(c("Inventory","Analysis") %in% names(res_named)))
  
  # Results should be equivalent modulo column order propagation
  expect_equal(res_null$Analysis[, c("X","Y")],
               res_named$Analysis[, c("X","Y")])
}

############################################################
# [A2] Works when A/B columns are not the first two (mapping by names)
############################################################
{
  AB_shuffled <- data.frame(
    extra1 = 101:104,
    Acode = c("A1","A2","A2","A21"),
    extra2 = letters[1:4],
    Bcode = c("B1","B1","B2","B2"),
    stringsAsFactors = FALSE
  )
  
  res <- analyseCorrespondenceTable(
    AB = AB_shuffled,
    Aname = "Acode", Bname = "Bcode",
    A = NULL, B = NULL,
    longestAcodeOnly = FALSE, longestBcodeOnly = FALSE
  )
  
  expect_true(is.list(res))
  expect_true(all(c("Inventory","Analysis") %in% names(res)))
  expect_inherits(res$Inventory, "data.frame")
  expect_inherits(res$Analysis,  "data.frame")
  expect_true(all(c("Acode","Bcode") %in% names(res$Analysis)))
  expect_true(nrow(res$Analysis) == nrow(AB_shuffled))
}

############################################################
# [A3] Error if only one of Aname/Bname is provided
############################################################
{
  AB_two <- data.frame(
    colA = c("A1","A2"),
    colB = c("B1","B2"),
    stringsAsFactors = FALSE
  )
  
  expect_error(
    analyseCorrespondenceTable(AB = AB_two, Aname = "colA", Bname = NULL),
    "Aname and Bname must be both NULL or both non-NULL"
  )
  
  expect_error(
    analyseCorrespondenceTable(AB = AB_two, Aname = NULL, Bname = "colB"),
    "Aname and Bname must be both NULL or both non-NULL"
  )
}

############################################################
# [A4] Error if Aname/Bname do not exist in AB
############################################################
{
  AB_cols <- data.frame(
    left  = c("A1","A2"),
    right = c("B1","B2"),
    stringsAsFactors = FALSE
  )
  
  expect_error(
    analyseCorrespondenceTable(AB = AB_cols, Aname = "foo", Bname = "right"),
    "Both Aname and Bname must correspond to existing columns in AB"
  )
  
  expect_error(
    analyseCorrespondenceTable(AB = AB_cols, Aname = "left", Bname = "bar"),
    "Both Aname and Bname must correspond to existing columns in AB"
  )
}

############################################################
# [A5] Type checks: Aname/Bname must be character scalars or NULL
############################################################
{
  AB_cols <- data.frame(
    left  = c("A1","A2"),
    right = c("B1","B2"),
    stringsAsFactors = FALSE
  )
  
  expect_error(
    analyseCorrespondenceTable(AB = AB_cols, Aname = 123, Bname = "right"),
    "Aname must be a character or NULL"
  )
  
  expect_error(
    analyseCorrespondenceTable(AB = AB_cols, Aname = "left", Bname = TRUE),
    "Bname must be a character or NULL"
  )
  
  expect_error(
    analyseCorrespondenceTable(AB = AB_cols, Aname = c("left","alt"), Bname = "right"),
    "Aname must be a character or NULL"
  )
  
  expect_error(
    analyseCorrespondenceTable(AB = AB_cols, Aname = "left", Bname = c("right","alt")),
    "Bname must be a character or NULL"
  )
}

############################################################
# [A6] Missing values in mapped columns should trigger the AB missing-code error
############################################################
{
  AB_with_na <- data.frame(
    src = c("A1", NA, "A2", ""),
    dst = c("B1", "B2", "",  "B3"),
    stringsAsFactors = FALSE
  )
  
  expect_error(
    analyseCorrespondenceTable(AB = AB_with_na, Aname = "src", Bname = "dst"),
    "Rows with missing values in the"
  )
}

############################################################
# [A7] Duplicate Acode-Bcode pairs detected after mapping by names
############################################################
{
  AB_dup <- data.frame(
    src = c("A1","A1"),
    dst = c("B1","B1"),
    stringsAsFactors = FALSE
  )
  
  expect_error(
    analyseCorrespondenceTable(AB = AB_dup, Aname = "src", Bname = "dst"),
    "Duplicate row\\(s\\) found in AB file"
  )
}

############################################################
# [A8] Result stability: explicit names vs default-first-two
############################################################
{
  AB_base <- data.frame(
    Acode = c("A1","A2","A21"),
    Bcode = c("B1","B2","B2"),
    extra = 1:3,
    stringsAsFactors = FALSE
  )
  
  # Default (first two columns)
  r_def <- analyseCorrespondenceTable(AB = AB_base)
  
  # Explicit naming
  r_exp <- analyseCorrespondenceTable(AB = AB_base, Aname = "Acode", Bname = "Bcode")
  
  expect_equal(r_def$Analysis[, c("Acode","Bcode")],
               r_exp$Analysis[, c("Acode","Bcode")])
}



############################################################
# [A9] Non acceptance of the same name
############################################################

expect_error(analyseCorrespondenceTable(AB = AB_base, Aname = "Acode", Bname = "Acode"),
             "Aname and Bname must refer to two different columns.")

check_n_columns <- function(df, source, num_columns) {
  caller <- sys.call(-1) #define the caller function
  tryCatch({
    if (ncol(df) < num_columns) {
      stop(paste("In",as.character(caller[1]), ", the data", source, " has less than the minimum required ", num_columns, "columns."))
    } else if (ncol(df) > num_columns) { 
      warning(paste("In", as.character(caller[1]),", the data", source, " has more than", num_columns, "columns."))
    } else {
      #print(paste("The data", source, " has exactly", num_columns, "columns."))
    }
  }, error = function(e) {
    print(paste("Error:", e$message))
  }, warning = function(e) {
    print(paste("Warning:", e$message))
  })
}

#' Normalize a SPARQL prefix label (pure)
#'
#' @description
#' Cleans a SPARQL prefix token so it is safe and consistent:
#' - lowercases,
#' - trims whitespace,
#' - removes trailing ":" and whitespace,
#' - keeps only [a-z0-9_] characters,
#' - drops empty results.
#'
#' @param x character vector of candidate prefix labels
#' @return cleaned, safe labels (lowercase, [a-z0-9_])
#' @keywords internal
#' 
normalize_prefix_token <- function(x) {
  x <- tolower(trimws(as.character(x)))
  x <- gsub("[:\\s]+$", "", x)
  x <- gsub("[^a-z0-9_]", "", x)
  x[nzchar(x)]
}

#' Normalize a namespace IRI (pure)
#'
#' @description
#' Ensures each IRI:
#' - has an HTTP(S) scheme (defaults to https if missing),
#' - has no whitespace,
#' - ends with exactly one trailing slash.
#'
#' @param uri character vector
#' @return normalized IRIs
#' @keywords internal
#' 
normalize_namespace_iri <- function(uri) {
  u <- trimws(as.character(uri))
  has_scheme <- grepl("^(?i)https?://", u, perl = TRUE)
  u[!has_scheme] <- paste0("https://", u[!has_scheme])
  u <- gsub("\\s+", "", u)
  u <- sub("/+$", "", u)
  paste0(u, "/")
}

#' Build PREFIX lines from tokens and IRIs (pure)
#'
#' @param tokens character vector of prefix tokens (will be normalized)
#' @param iris character vector of namespace IRIs (will be normalized)
#' @return character vector like "PREFIX token: <iri/>"
#' @keywords internal
build_prefix_lines <- function(tokens, iris) {
  stopifnot(length(tokens) == length(iris))
  tokens <- normalize_prefix_token(tokens)
  iris   <- normalize_namespace_iri(iris)
  unique(sprintf("PREFIX %s: <%s>", tokens, iris))
}

#' Validate that a prefix token is declared in a prefix block (pure)
#'
#' @param token a candidate prefix token (will be normalized)
#' @param prefix_block a single string containing the PREFIX lines
#' @return TRUE if declared, FALSE otherwise
#' @keywords internal
prefix_declared_in_block <- function(token, prefix_block) {
  token <- normalize_prefix_token(token)
  if (!length(token)) return(FALSE)
  grepl(paste0("\\bPREFIX\\s+", token, ":"), prefix_block, ignore.case = TRUE)
}






#' @title Build SPARQL PREFIX declarations for an endpoint
#'
#' @description
#' Returns a character vector of `PREFIX ...` lines used by package queries.
#' The result always includes core vocabularies (SKOS/XKOS/RDFS/DC/RDF/XSD),
#' and—when available—adds classification scheme namespaces discovered via
#' `classificationList(endpoint)`. You can filter to specific schemes with
#' the `prefix` argument.
#'
#' @param endpoint Character. Endpoint name (case-insensitive). Typically
#'   `"CELLAR"` or `"FAO"`. The set of supported names is defined by the
#'   package's internal endpoint registry (see `.valid_endpoints()`).
#' @param prefix Character vector of scheme tokens to include (e.g.,
#'   `c("cn2022","nace2")`). If `NULL` (default), returns all discovered schemes.
#'
#' @return
#' A character vector of SPARQL PREFIX declarations.
#' Use `paste(result, collapse = "\\n")` to inject into a query string.
#'
#' @details
#' - This function has no side effects and does not perform network I/O
#'   by itself. Discovery is delegated to `classificationList()`.
#'   
#' - If `classificationList()` is unavailable or fails, the function returns
#'   only the core vocabularies and emits a warning.
#'
#' @examples
#' \dontrun{
#' # All known scheme prefixes for the endpoint
#' pl <- prefixList("CELLAR")
#' cat(paste(pl, collapse = "\n"))
#'
#' # Only include CN2022 and NACE2 if they exist
#' pl2 <- prefixList("CELLAR", prefix = c("cn2022", "nace2"))
#' }
#'
#' @export

prefixList <- function(endpoint, prefix = NULL) {
  # --- helpers (non-exported) ---
  .norm_name <- function(x) toupper(trimws(as.character(x)))
  .core_prefixes <- function() {
    c(
      "PREFIX dc: <http://purl.org/dc/elements/1.1/>",
      "PREFIX dct: <http://purl.org/dc/terms/>",
      "PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>",
      "PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>",
      "PREFIX skos: <http://www.w3.org/2004/02/skos/core#>",
      "PREFIX xkos: <http://rdf-vocabulary.ddialliance.org/xkos#>",
      "PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>"
    )
  }
  
  # --- validate endpoint via internal registry if present ---
  ep <- .norm_name(endpoint)
  if (exists(".valid_endpoints", mode = "function")) {
    valid <- tryCatch(.valid_endpoints(), error = function(e) character())
    if (length(valid) && !(ep %in% valid)) {
      stop(sprintf(
        "Unsupported endpoint '%s'. Supported: %s",
        endpoint, paste(valid, collapse = ", ")
      ), call. = FALSE)
    }
  } else {
    if (!nzchar(ep)) stop("`endpoint` must be a non-empty string.", call. = FALSE)
  }
  
  # --- start with core vocabularies ---
  prefix_lines <- .core_prefixes()
  
  # --- discover scheme namespaces via classificationList() ---
  schemes <- tryCatch(
    classificationList(ep),
    error = function(e) {
      warning(
        "prefixList(): classificationList('", ep, "') failed: ", conditionMessage(e),
        ". Returning only core prefixes."
      )
      NULL
    }
  )
  
  # If discovery failed or returned nothing usable, return core only
  if (!is.data.frame(schemes) || nrow(schemes) == 0L) {
    return(prefix_lines)
  }
  
  # Normalize expected columns: need a scheme token ("Prefix") and scheme URI ("URI")
  if (!("Prefix" %in% names(schemes))) {
    if ("ConceptScheme" %in% names(schemes)) {
      schemes$Prefix <- tolower(schemes$ConceptScheme)
    } else if ("URI" %in% names(schemes)) {
      schemes$Prefix <- tolower(sub(".*/", "", schemes$URI))
    } else {
      warning("prefixList(): 'Prefix' not found and cannot be derived. Returning core prefixes only.")
      return(prefix_lines)
    }
  }
  
  if (!("URI" %in% names(schemes))) {
    if ("SchemeURI" %in% names(schemes)) {
      schemes$URI <- schemes$SchemeURI
    } else if ("ConceptSchemeURI" %in% names(schemes)) {
      schemes$URI <- schemes$ConceptSchemeURI
    } else {
      warning("prefixList(): no URI column found to build scheme prefixes. Returning core prefixes only.")
      return(prefix_lines)
    }
  }
  
  # Optionally filter requested tokens
  scheme_tokens <- schemes$Prefix
  scheme_uris   <- schemes$URI
  
  if (!is.null(prefix)) {
    want <- normalize_prefix_token(prefix)
    have <- normalize_prefix_token(scheme_tokens)
    keep <- have %in% want
    if (!any(keep)) {
      warning(
        "prefixList(): requested prefix(es) not found for endpoint ", ep, ": ",
        paste(prefix, collapse = ", "), ". Returning core prefixes only."
      )
      return(prefix_lines)
    }
    scheme_tokens <- scheme_tokens[keep]
    scheme_uris   <- scheme_uris[keep]
  }
  
  add <- build_prefix_lines(scheme_tokens, scheme_uris)
  out <- c(prefix_lines, add)
  # Drop duplicates while preserving order
  out[!duplicated(out)]
}

# Utility to extract token names from "PREFIX token: <...>" lines
.extract_tokens <- function(x) {
  sub("^PREFIX\\s+([^:]+):.*$", "\\1", x)
}

# Core tokens as hard-coded by prefixList()
.core_tokens <- c("dc", "dct", "rdfs", "rdf", "skos", "xkos", "xsd")

# 1) Validation of 'endpoint' argument ---------------------------------------

# Error for unsupported endpoints (message can vary if .valid_endpoints() is present)
expect_error(
  prefixList("ALL"),
  "(?i)(unsupported endpoint|'CELLAR'|FAO|non-empty|endpoint)",
  fixed  = FALSE
)

expect_error(
  prefixList("INVALID"),
  "(?i)(unsupported endpoint|'CELLAR'|FAO|non-empty|endpoint)",
  fixed  = FALSE
)

if (at_home()) {
  # Valid endpoints should not error (even if discovery fails => core-only)
  res_cellar <- try(prefixList("CELLAR"), silent = TRUE)
  expect_true(!inherits(res_cellar, "try-error"))
  
  res_fao <- try(prefixList("FAO"), silent = TRUE)
  expect_true(!inherits(res_fao, "try-error"))
  
  # 2) 'CELLAR' structure checks (character vector) ----------------------------
  
  res <- prefixList("CELLAR")
  expect_true(is.character(res))
  expect_true(length(res) >= length(.core_tokens))
  expect_true(all(grepl("^PREFIX\\s+", res)))
  
  
  # 3) 'FAO' structure checks (character vector) -------------------------------
  
  res <- prefixList("FAO")
  expect_true(is.character(res))
  expect_true(length(res) >= length(.core_tokens))
  expect_true(all(grepl("^PREFIX\\s+", res)))
  
  
  # 4) Filtering by subset of prefixes -----------------------------------------
  
  
  all_prefixes <- prefixList("FAO")
  prefix_names <- .extract_tokens(all_prefixes)
  
  # Consider "dynamic" as anything not in known core tokens
  dynamic_prefixes <- setdiff(prefix_names, .core_tokens)
  
  if (length(dynamic_prefixes) == 0L) {
    message("No dynamic prefixes found for FAO; skipping filtering test.")
  } else {
    target_prefix <- dynamic_prefixes[1L]
    
    filtered <- prefixList("FAO", prefix = target_prefix)
    expect_true(is.character(filtered))
    
    filtered_tokens <- .extract_tokens(filtered)
    # requested prefix should be present
    expect_true(target_prefix %in% filtered_tokens)
    # at least one core prefix should still be present
    expect_true(any(.core_tokens %in% filtered_tokens))
  }
  
  
  # 5) Requested prefixes do not exist -> warning + core-only result -----------
  
  # Current behavior: warn + return only core prefixes
  expect_warning(
    not_found <- prefixList("FAO", prefix = "THISDOESNOTEXIST"),
    "(?i)(requested prefix|not found|Returning.*core)",
    fixed  = FALSE
  )
  expect_true(is.character(not_found))
  expect_true(all(grepl("^PREFIX\\s+", not_found)))
  
  tokens_nf <- .extract_tokens(not_found)
  # Should NOT contain the requested token (note: tokens are normalized to lowercase)
  expect_false("thisdoesnotexist" %in% tokens_nf)
  # Should contain core tokens (at least these)
  expect_true(all(.core_tokens %in% tokens_nf))
  
}

#' @title Correspondence table creation
#' @description Create a candidate correspondence table between two classifications based on their correspondences with intermediate classifications
#' @param Tables A string of type character containing the name of a csv file which contains the names of the files that
#'   contain the classifications and the intermediate correspondence tables OR a list of vectors with the names of the dataframes
#'   classifications and the intermediate correspondence tables (see "Details" below).
#' @param Reference The reference classification among A and B. If a classification is the reference to the other, and hence
#'  \emph{hierarchically superior} to it, each code of the other classification is expected to be mapped to at most one code
#'   of the reference classification. The valid values are \code{"none"}, \code{"A"}, and \code{"B"}. If the selected value
#'   is \code{"A"} or \code{"B"}, a "Review" flag column (indicating the records violating this expectation) is included
#'   in the output (see "Explanation of the flags" below).
#' @param MismatchTolerance The maximum acceptable proportion of rows in the candidate correspondence table which contain
#'   no code for classification A or no code for classification B. The default value is \code{0.2}. The valid values are
#'   real numbers in the interval [0, 1].
#' @param Redundancy_trim An argument in the function containing the logical values \code{TRUE} or \code{FALSE}
#'   used to facilitate the trimming of the redundant records.
#'   The default value is \code{TRUE}, which removes all redundant records.
#'   The other values is \code{FALSE}, which shows redundant records together with the redundancy flag.
#' @param Progress  An argument in the function containing the logical values \code{TRUE} (default) or \code{FALSE}. 
#'   Used to switch ON (when \code{TRUE}) or OFF (when \code{FALSE}) the progress bar that interactively display the creation process of the new table in the console output.
#'   
#' @export
#' @details
#' File and file name requirements:
#'     \itemize{
#'         \item The file that corresponds to argument \code{Tables} and the files to which the contents of \code{Tables}
#'         lead, must be in \emph{csv format with comma as delimiter}. If full paths are not provided, then these files must
#'          be available in the working directory. No two filenames provided must be identical.
#'         \item The file that corresponds to argument \code{Tables} must contain filenames, \emph{and nothing else}, in
#'         a \eqn{(k+2)} × \eqn{(k+2)} table, where \eqn{k}, a positive integer, is the number of "pivot" classifications.
#'         The cells in the main diagonal of the table provide the filenames of the files which contain, with this order,
#'         the classifications  \eqn{A, C_1}, \eqn{\ldots}, \eqn{C_k} and \eqn{B}. The off-diagonal directly above the main
#'         diagonal contains the filenames of the files that contain, with this order, the correspondence tables
#'         \eqn{A:C_1}, \{\eqn{C_i}:\eqn{C_{i+1}}, \eqn{1 \le i \le k-1}\} and \eqn{B:C_k}. All other cells of the table
#'         must be empty.
#'     }
#' Classification table requirements:
#'     \itemize{
#'         \item Each of the files that contain classifications must contain at least one column and at least two rows.
#'         The first column contains the codes of the respective classification. The first row contains column headers.
#'         The header of the first column is the name of the respective classification (e.g., "CN 2021").
#'         \item The classification codes contained in a classification file (expected in its first column as mentioned
#'         above) must be unique. No two identical codes are allowed in the column.
#'         \item If any of the files that contain classifications has additional columns the first one of them is assumed
#'         to contain the labels of the respective classification codes.
#'     }
#' Correspondence table requirements:
#'     \itemize{
#'         \item The files that contain correspondence tables must contain at least two columns and at least two rows.
#'         The first column of the file that contains A:\eqn{C_1} contains the codes of classification A. The second column
#'         contains the codes of classification \eqn{C_1}. Similar requirements apply to the files that contain
#'         \eqn{C_i}:\eqn{C_{i+1}}, \eqn{1 \le i \le k-1} and B:\eqn{C_k}. The first row of each of the files that contain
#'         correspondence tables contains column headers. The names of the first two columns are the names of the respective
#'         classifications.
#'         \item The pairs of classification codes contained in a correspondence table file (expected in its first two columns
#'         as mentioned above) must be unique. No two identical pairs of codes are allowed in the first two columns.
#'     }
#' Interdependency requirements:
#'     \itemize{
#'         \item At least one code of classification A must appear in both the file of classification A and the file of
#'         correspondence table A:\eqn{C_1}.
#'         \item At least one code of classification B must appear in both the file of classification B and the file of
#'         correspondence table B:\eqn{C_k}, where \eqn{k}, \eqn{k\ge 1}, is the number of pivot classifications.
#'         \item If there is only one pivot classification, \eqn{C_1}, at least one code of it must appear in both the file of
#'         correspondence table A:\eqn{C_1} and the file of correspondence table B:\eqn{C_1}.
#'         \item If the pivot classifications are \eqn{k} with \eqn{k\ge 2} then at least one code of \eqn{C_1} must appear in
#'         both the file of correspondence table A:\eqn{C_1} and the file of correspondence table \eqn{C_1}:\eqn{C_2}, at least one
#'         code of each of the \eqn{C_i}, \eqn{i = 2, \ldots, k-1} (if \eqn{k\ge 3}) must appear in both the file of correspondence
#'         table \eqn{C_{i-1}}:\eqn{C_i} and the file of correspondence table \eqn{C_i}:\eqn{C_{i+1}}, and at least one code of
#'         \eqn{C_k} must appear in both the file of correspondence table \eqn{C_{k-1}}:\eqn{C_k} and the file of correspondence table
#'         B:\eqn{C_k}.
#'     }
#' Mismatch tolerance:
#'     \itemize{
#'         \item The ratio that is compared with \code{MismatchTolerance} has as numerator the number of rows in the candidate
#'         correspondence table which contain no code for classification A or no code for classification B and as denominator
#'         the total number of rows of this table. If the ratio exceeds \code{MismatchTolerance} the execution of the function
#'         is halted.
#'     }
#' If any of the conditions required from the arguments is violated an error message is produced and execution is stopped.
#'
#' @section Explanation of the flags:
#'
#' \itemize{
#'     \item  The "Review" flag is produced only if argument Reference has been set equal to "\code{A}" or "\code{B}". For
#'     each row of the candidate correspondence table, if \code{Reference} = "\code{A}" the value of "Review" is equal to
#'     \code{1} if the code of B maps to more than one code of A, and \code{0} otherwise. If \code{Reference} = "\code{B}"
#'     the value of "Review" is equal to \code{1} if the code of A maps to more than one code of B, and \code{0} otherwise.
#'     The value of the flag is empty if the row does not contain a code of A or a code of B.
#'     \item For each row of the candidate correspondence table, the value of "Redundancy" is equal to \code{1} if the row
#'     contains a combination of codes of A and B that also appears in at least one other row of the candidate
#'     correspondence table.
#'     \item When "Redundancy_Trim" is equal to \code{FALSE} the "Redundancy_keep" flag is created to identify with value \code{1}
#'     the records that will be kept if trimming is performed.
#'     \item For each row of the candidate correspondence table, the value of "Unmatched" is equal to \code{1} if the row
#'     contains a code of A but no code of B or if it contains a code of B but no code of A. The value of the flag is
#'     \code{0} if the row contains codes for both A and B.
#'     \item For each row of the candidate correspondence table, the value of "NoMatchFromA" is equal to \code{1} if the row
#'     contains a code of A that appears in the table of classification A but not in correspondence table A:\eqn{C_1}. The
#'     value of the flag is \code{0} if the row contains a code of A that appears in both the table of classification A and
#'     correspondencetable A:\eqn{C_1}. Finally, the value of the flag is empty if the row contains no code of A or if it
#'     contains a code of A that appears in correspondence table A:\eqn{C_1} but not in the table of classification A.
#'     \item For each row of the candidate correspondence table, the value of "NoMatchFromB" is equal to \code{1} if the row
#'     contains a code of B that appears in the table of classification B but not in correspondence table B:\eqn{C_k}. The
#'     value of the flag is \code{0} if the row contains a code of B that appears in both the table of classification B and
#'     correspondence table B:\eqn{C_k}. Finally, the value of the flag is empty if the row contains no code of B or if it
#'     contains a code of B that appears in correspondence table B:\eqn{C_k} but not in the table of classification B.
#'     \item The argument "Redundancy_trim" is used to delete all the redundancies which are mapping correctly.
#'     The valid logical values for this argument in the candidate correspondence table are \code{TRUE} or \code{FALSE}.
#'     If the selected value is \code{TRUE}, all redundant records are removed and kept exactly one record for each unique combination.
#'     For this retained record, the codes, the label and the supplementary information of the pivot classifications are replaced with
#'     'multiple'. If the multiple infomration of the pivot classifications are the same, their value will not be replaced.
#'     If the selected value is \code{FALSE}, no trimming is executed so redundant records are shown, together with the redundancy flag.
#'     If the logical values are missing the implementation of the function will stop.
#'
#' }
#'
#' @section Sample datasets included in the package:
#'
#' If the user wish to access the csv files with the sample data, they have two options:
#' \itemize{
#' \item Option 1: Unpack into any folder of their choice the tar.gz file into which the package has arrived. All sample
#' datasets may be found in the "inst/extdata" subfolder of this folder.
#' \item Option 2: Go to the "extdata" subfolder of the folder in which the package has been installed in their PC's \code{R}
#' library. All sample datasets may be found there.
#' }
#'
#' @return
#'  \code{newCorrespondenceTable()} returns a list with two elements, both of which are data frames.
#' \itemize{
#'     \item The first element is the candidate correspondence table A:B, including the codes of all "pivot" classifications,
#'      augmented with flags "Review" (if applicable), "Redundancy", "Unmatched", "NoMatchFromA", "NoMatchFromB" and with all
#'      the additional columns of the classification and intermediate correspondence table files.
#'     \item The second element contains the names of classification A, the "pivot" classifications and classification B as
#'     read from the top left-hand side cell of the respective input files.
#' }
#'
#' @examples
#' {
#' \dontrun{
#' 
#' code_in   <- data.frame(code_in  = c("A","B"))
#' code_mid  <- data.frame(code_mid = c("B","C","D"))
#' in_to_mid <- data.frame(code_in  = c("A","B")
#'                         , code_out = c("C","B"))
#' code_out  <- data.frame(code_out = c("B","C","D","E"))
#' mid_to_out <- data.frame(code_mid = c("B","C","D")
#'                          , code_out = c("B","C","D"))
#' 
#' write.csv(code_in,    file = "code_in.csv",   row.names  = FALSE)
#' write.csv(code_mid,   file = "code_mid.csv",  row.names  = FALSE)
#' write.csv(in_to_mid,  file = "in_to_mid.csv", row.names  = FALSE)
#' write.csv(mid_to_out, file = "mid_to_out.csv", row.names = FALSE)
#' write.csv(code_out,   file = "code_out.csv",   row.names = FALSE)
#' 
#' 
#' m <- matrix("",3,3)
#' diag(m) <- c("code_in.csv","code_mid.csv","code_out.csv")
#' m[1,2] <- "in_to_mid.csv"
#' m[2,3] <- "mid_to_out.csv"
#' 
#' 
#' 
#' write.table(m, file="merge_sequence.csv",row.names=FALSE,col.names=FALSE,sep=",")
#' 
#' 
#' out <- newCorrespondenceTable("merge_sequence.csv",MismatchTolerance = 0.9)
#' print(out$newCorrespondenceTable)
#' print(out$classificationNames)
#' 
#' 
#'     } }

newCorrespondenceTable <- function(
    Tables,
    Reference = "none",
    MismatchTolerance = 0.2,
    Redundancy_trim = TRUE,
    Progress = TRUE
){
  # Check if the Tables is a File or a list of vectors
  # Check if the file that contains the names of both classifications and correspondence tables exists in working directory
  
  if (is.list(Tables)) {
    x <- as.matrix(do.call(rbind, Tables))
    y <- Tables
    
    # The following if statement checks if the names of both classifications
    # and correspondence tables lists construct a sparse square matrix.
    is_square <- TRUE
    n <- length(y)
    
    for (i in seq_along(y)) {
      if (length(y[[i]]) != n) {
        is_square <- FALSE
      }
    }
    
    if (is_square) {
      # Define inputs as an empty list
      inputs <- list()
      inputs_name <- list()
      n <- length(y)
      # First diagonal element
      inputs[1] <- y[[1]][1]
      inputs_name[1] <- "Table[1,1]"
      # Next n diagonal elements (from 2nd diagonal to the n th)
      for (i in 2:n) {
        inputs[i] <- y[[i]][i]
        inputs_name[i]<- paste0("Table[",i, ",", i, "]")
        
      }
      counter <- length(inputs) + 1
      for (i in 1:(n - 1)) {
        for (j in (i + 1):n) {
          if (!is.null(y[[i]][[j]])) {  # Check that the entry is not empty
            inputs[counter] <- y[[i]][j]
            inputs_name[counter] <- paste0("Table[",i, ",", j, "]")
            counter <- counter + 1
          }
        }
      }
    } else {
      # Error message in case lists do not construct a
      # sparse square matrix.
      stop(paste("The List do not construct a sparse square matrix. \n Please verify that the appropriate number of elements are inserted in the appropriate cells."))
    }
    
    
  } else if (file.exists(Tables)) {
    # x <- as.matrix(utils::read.csv(Tables, sep = ",", header = FALSE, colClasses = c("character"),
    #                                 encoding = "UTF-8"))
    x <- as.matrix(data.table::fread(Tables, sep = ",", header = FALSE, colClasses = c("character"),
                                     encoding = "UTF-8"))
    mat.list <- apply(x, 2, function(x) {
      as.character(which(x != ""))
    })
    
    # Check if files exist in working directory
    test.names <- as.vector(x)[which(as.vector(x) != "")]
    if (!all(file.exists(test.names))) {
      for (i in which(file.exists(test.names) == FALSE)) {
        stop(simpleError(paste("The is no file with name", test.names[i], "in your working directory.")))
      }
    }
    if (length(which(duplicated(test.names) == TRUE)) >= 1) {
      stop(simpleError(paste("At least two of the filenames in", Tables, "are the same.")))
    }
    
    test.list <- list()
    test.list[[1]] <- "1"
    for (mat.index in 2:ncol(x)) {
      test.list[[mat.index]] <- as.character(c((mat.index - 1):mat.index))
    }
    
    # The following if statement checks if the names of both classifications
    # and correspondence tables in the 'names.csv' file construct a sparse
    # square matrix.
    if (all(unlist(Map(identical, mat.list, test.list)) == TRUE) && nrow(x) >= 3) {
      k <- nrow(x) - 2
      
    } else {
      # Error message in case the names of both classifications and
      # correspondence tables in the 'names.csv' file do not construct a
      # sparse square matrix.
      stop(paste("The filenames  in", Tables, "do not construct a sparse square matrix. \n Please verify that the appropriate number of filenames are inserted in the appropriate cells."))
    }
    
    # The list inputs includes the names of both classifications and
    # correspondence tables.
    inputs <- list()
    inputs[[1]] <- diag(x)[1]
    inputs[seq(k) + 1] = as.list(diag(x)[seq(k) + 1])
    inputs[[k + 2]] <- diag(x)[length(diag(x))]
    inputs[(k + 3):(k + 2 + length(as.list(x[upper.tri(x)][x[upper.tri(x)] != ""])))] <- as.list(x[upper.tri(x)][x[upper.tri(x)] !=
                                                                                                                   ""])
    
  } else {
    stop(simpleError(paste("There is no file with name", Tables, "in your working directory. Or the argument is not a list of vectors")))
  }
  
  
  
  if (is.list(Tables)) {
    RRR <- inputs
    k <- nrow(x) - 2
    
    inputs <- inputs_name
    
  } else {
    # Create a list of the classifications and the known correspondence tables
    # as data frames.
    RRR <- lapply(inputs[1:length(inputs)], function(x) {
      utils::read.csv(x, sep = ",", check.names = FALSE, colClasses = c("character"),
                      encoding = "UTF-8")
      # data.table::fread(x, sep = ",", check.names = FALSE, colClasses = c("character"),
      # encoding = "UTF-8")
    })
  }
  # Check Reference
  if (!(Reference %in% c("A", "B", "none"))) {
    stop(simpleError("You entered a non-allowed value for Reference. The allowed values are \"A\", \"B\" and \"none\"."))
  }
  
  # Check MismatchTolerance
  if (is.character(MismatchTolerance) || MismatchTolerance < 0 || MismatchTolerance >
      1) {
    stop(simpleError("You entered a non-allowed value for MismatchTolerance. The allowed values are numbers in the interval [0, 1]."))
  }
  
  
  
  removeBOM <- function(headers) {
    gsub("\\xef\\xbb\\xbf", "", headers, useBytes = T)
  }
  
  for (i in 1:length(RRR)) {
    colnames(RRR[[i]]) <- removeBOM(colnames(RRR[[i]]))
  }
  
  # Convert data frames into matrices.
  RR <- lapply(RRR, function(x) {
    matrix(unlist(x), ncol = ncol(x))
  })
  
  # Select the correspondence tables.
  R <- RR[utils::tail(c(1:length(RR)), (length(RR) - 1)/2)]
  
  # Check the dimensions of the files
  for (i in 1:nrow(x)) {
    if (ncol(RRR[[i]]) < 1 || nrow(RRR[[i]]) < 1) {
      stop(simpleError(paste("File", inputs[i], "should have at least one column and two rows (including the row of headers).")))
    }
  }
  
  for (i in 1:length(R)) {
    if (ncol(R[[i]]) <= 1 || nrow(R[[i]]) < 1) {
      stop(simpleError(paste("File", inputs[i + nrow(x)], "should have at least two columns and two rows (including the row of headers).")))
    }
  }
  
  # Check for entries dimensions of the files
  for (i in 1:nrow(x)) {
    if (sum(duplicated(RRR[[i]][, 1])) >= 1) {
      stop(simpleError(paste("At least one code of ", colnames(RRR[[i]])[1],
                             " appears more than once in file ", inputs[i], ". This is an error. Each code must appear only once in the file.",
                             sep = "")))
    }
  }
  
  for (i in 1:length(R)) {
    if (nrow(unique(R[[i]][, 1:2])) != nrow(R[[i]][, 1:2])) {
      stop(simpleError(paste("At least one pair of codes of ", colnames(RRR[[i +
                                                                               nrow(x)]])[1], " and ", colnames(RRR[[i + nrow(x)]])[2], " appears more than once in file ",
                             inputs[i + nrow(x)], ". This is an error. Each pair of codes must appear only once in the file.",
                             sep = "")))
    }
  }
  
  # Check for at least one match in classifications and correspondence
  # tables. In inputs there are the names of both classifications and
  # correspondence tables. Stop with error
  if (k == 1) {
    # A in A appears in A:C1
    if (sum(!is.na(match(unlist(RRR[[1]][, 1]), R[[1]][, 1]))) == 0) {
      stop(simpleError(paste("There is no code of ", colnames(RRR[[1]])[1],
                             " that appears in both ", inputs[1], " and ", inputs[1 + nrow(x)],
                             ". This is an error. The files should have at least one code of ",
                             colnames(RRR[[1]])[1], " in common to allow the generation of the candidate correspondence table.",
                             sep = "")))
    }
    
    # C1 in A:C1 appears in B:C1
    if (sum(!is.na(match(R[[1]][, 2], R[[2]][, 2]))) == 0) {
      stop(simpleError(paste("There is no code of ", colnames(RRR[[1 + nrow(x)]])[2],
                             " that appears in both ", inputs[1 + nrow(x)], " and ", inputs[2 +
                                                                                              nrow(x)], ". This is an error. The files should have at least one code of ",
                             colnames(RRR[[1 + nrow(x)]])[2], " in common to allow the generation of the candidate correspondence table.",
                             sep = "")))
    }
    
    # B in B:C1 appears in B
    if (sum(!is.na(match(R[[length(R)]][, 1], unlist(RRR[[nrow(x)]][, 1])))) == 0) {
      stop(simpleError(paste("There is no code of ", colnames(RRR[[length(R) +
                                                                     nrow(x)]])[1], " that appears in both ", inputs[nrow(x)], " and ",
                             inputs[length(R) + nrow(x)], ". This is an error. The files should have at least one code of ",
                             colnames(RRR[[length(R) + nrow(x)]])[1], " in common to allow the generation of the candidate correspondence table.",
                             sep = "")))
    }
    
  }
  
  if (k >= 2) {
    
    # A in A appears in A:C1
    if (sum(!is.na(match(unlist(RRR[[1]][, 1]), R[[1]][, 1]))) == 0) {
      stop(simpleError(paste("There is no code of ", colnames(RRR[[1]])[1],
                             " that appears in both ", inputs[1], " and ", inputs[1 + nrow(x)],
                             ". This is an error. The files should have at least one code of ",
                             colnames(RRR[[1]])[1], " in common to allow the generation of the candidate correspondence table.",
                             sep = "")))
    }
    
    # C1 in A:C1 appears in C1:C2 C2 in C1:C2 appears in C2:C3 ...
    for (i in 1:(k - 1)) {
      
      if (sum(!is.na(match(R[[i]][, 2], R[[i + 1]][, 1]))) == 0) {
        stop(simpleError(paste("There is no code of ", colnames(RRR[[i +
                                                                       nrow(x)]])[2], " that appears in both ", inputs[i + nrow(x)], " and ",
                               inputs[i + 1 + nrow(x)], ". This is an error. The files should have at least one code of ",
                               colnames(RRR[[i + nrow(x)]])[2], " in common to allow the generation of the candidate correspondence table.",
                               sep = "")))
      }
      
    }
    
    # Ck in C(k-1):Ck appears in B:Ck
    if (sum(!is.na(match(R[[k]][, 2], R[[k + 1]][, 2]))) == 0) {
      stop(simpleError(paste("There is no code of ", colnames(RRR[[k + nrow(x)]])[2],
                             " that appears in both ", inputs[k + nrow(x)], " and ", inputs[k +
                                                                                              1 + nrow(x)], ". This is an error. The files should have at least one code of ",
                             colnames(RRR[[k + nrow(x)]])[2], " in common to allow the generation of the candidate correspondence table.",
                             sep = "")))
    }
    
    # B in B:Ck appears in B
    if (sum(!is.na(match(R[[length(R)]][, 1], unlist(RRR[[nrow(x)]][, 1])))) == 0) {
      stop(simpleError(paste("There is no code of ", colnames(RRR[[length(R) +
                                                                     nrow(x)]])[1], " that appears in both ", inputs[nrow(x)], " and ",
                             inputs[length(R) + nrow(x)], ". This is an error. The files should have at least one code of ",
                             colnames(RRR[[length(R) + nrow(x)]])[1], " in common to allow the generation of the candidate correspondence table.",
                             sep = "")))
    }
    
  }
  
  # Warning
  if (k == 1) {
    
    # C1 in C1 appears in A:C1
    if (sum(!is.na(match(unlist(RRR[[2]][, 1]), R[[1]][, 2]))) == 0) {
      message(paste("WARNING: there is no code of ", colnames(RRR[[2]])[1], " that appears in both ",
                    inputs[2], " and ", inputs[1 + nrow(x)], ". When the execution of the function is over, please check the files to ensure that this is not the result of a mistake in their preparation or declaration.\n",
                    sep = ""))
    }
    
    # C1 in C1 appears in B:C1
    if (sum(!is.na(match(unlist(RRR[[2]][, 1]), R[[2]][, 2]))) == 0) {
      message(paste("WARNING: there is no code of ", colnames(RRR[[2]])[1], " that appears in both ",
                    inputs[2], " and ", inputs[2 + nrow(x)], ". When the execution of the function is over, please check the files to ensure that this is not the result of a mistake in their preparation or declaration.\n",
                    sep = ""))
    }
    
  }
  
  if (k == 2) {
    
    for (i in 2:k) {
      
      # C1 in C1 appears in A:C1 C2 in C2 appears in C1:C2 C3 in C3
      # appears in C2:C3
      if (sum(!is.na(match(unlist(RRR[[i]][, 1]), R[[i - 1]][, 2]))) == 0) {
        message(paste("WARNING: there is no code of ", colnames(RRR[[i]])[1],
                      " that appears in both ", inputs[i], " and ", inputs[i - 1 + nrow(x)],
                      ". When the execution of the function is over, please check the files to ensure that this is not the result of a mistake in their preparation or declaration.\n",
                      sep = ""))
      }
      
      # C1 in C1 appears in C1:C2 C2 in C2 appears in C2:C3 C3 in C3
      # appears in C3:C4
      if (sum(!is.na(match(unlist(RRR[[i]][, 1]), R[[i]][, 1]))) == 0) {
        message(paste("WARNING: there is no code of ", colnames(RRR[[i]])[1],
                      " that appears in both ", inputs[i], " and ", inputs[i + nrow(x)],
                      ". When the execution of the function is over, please check the files to ensure that this is not the result of a mistake in their preparation or declaration.\n",
                      sep = ""))
      }
    }
    
    # Ck in Ck appears in C(k-1):Ck
    if (sum(!is.na(match(unlist(RRR[[k + 1]][, 1]), R[[k]][, 2]))) == 0) {
      message(paste("WARNING: there is no code of ", colnames(RRR[[k + 1]])[1],
                    " that appears in both ", inputs[k + 1], " and ", inputs[k + nrow(x)],
                    ". When the execution of the function is over, please check the files to ensure that this is not the result of a mistake in their preparation or declaration.\n",
                    sep = ""))
    }
    
    # Ck in Ck appears in B:Ck
    if (sum(!is.na(match(unlist(RRR[[k + 1]][, 1]), R[[k + 1]][, 2]))) == 0) {
      message(paste("WARNING: there is no code of ", colnames(RRR[[k + 1]])[1],
                    " that appears in both ", inputs[k + 1], " and ", inputs[k + 1 +
                                                                               nrow(x)], ". When the execution of the function is over, please check the files to ensure that this is not the result of a mistake in their preparation or declaration.\n",
                    sep = ""))
    }
    
  }
  
  # Create the final correspondence table moving from the classification A to
  # the classification B.
  tryCatch({
    
    F_AtoB <- list()
    
    # The following if statement is used when we have only the
    # correspondence tables A:C1 and B:C1.
    counter <- 0
    if (length(R) == 2) {
      #creating a progress bar
      if (Progress) {
        message("Percentage of codes of ", colnames(RRR[[1]][1]), " processed:")
        pb <- txtProgressBar(min = 0, max = 100, style = 3, width = 50, char = "=")
      }
      
      # The following for loop creates the desirable correspondence
      # table.  The operations are conducted for each unique element of
      # classification A of the correspondence table A:C1.
      for (i in unique(R[[1]][, 1])) {
        
        # Print the percentage of codes that have been processed.
        counter <- counter + 1
        if (Progress) {
          setTxtProgressBar(pb, round(counter/length(unique(R[[1]][, 1])) * 100, digits = 0))
        }
        # Matrix TT contains the rows of correspondence table A:C1 for
        # a specific element of classification A.  Matrix T contains
        # the rows of correspondence table B:C1 that match with the
        # specific element of classification A based on classification
        # C1.
        x1 <- R[[1]][which(R[[1]][, 1] == i), 2]
        TT <- matrix(R[[1]][which(R[[1]][, 1] == i), 1:2], ncol = 2)
        T <- matrix(R[[2]][!is.na(match(R[[2]][, 2], x1)), 1:2], ncol = 2)
        
        # Create a list whose each element is a matrix that contains
        # all unique rows of matrix T based on the elements of
        # classification C1.
        t <- match(T[, 2], T[, 2])
        v1 <- sequence(rle(sort(t))$lengths)
        v1 <- split(seq_along(v1), cumsum(v1 == 1))
        Z <- lapply(v1, function(x) {
          T[order(t)[x], , drop = FALSE]
        })
        
        # Create a list whose each element is a matrix that contains
        # all unique rows of matrix TT that match with the unique
        # elements of the second column of matrix T.
        t1 <- match(TT[, 2], T[, 2])
        v1 <- sequence(rle(sort(t1))$lengths)
        v1 <- split(seq_along(v1), cumsum(v1 == 1))
        Z1 <- lapply(v1, function(x) {
          TT[order(t1)[x], , drop = FALSE]
        })
        
        # Keep matrices in Z that exist in Z1 based on their second
        # columns (elements of classification C1).
        Z <- Z[!is.na(match(lapply(Z, function(x) {
          unique(x[, 2])
        }), lapply(Z1, function(x) {
          unique(x[, 2])
        })))]
        
        # ZZ is a matrix that consists of matrices in Z1 expanded by
        # their corresponding matrices (based on the elements of
        # classification C1).
        a <- lapply(Z, function(x) {
          1:nrow(x)
        })
        a1 <- lapply(Z1, function(x) {
          1:nrow(x)
        })
        aa <- lapply(Map(function(x, y) {
          x[y[, 1], ]
        }, Z, Map(expand.grid, a, a1)), function(x) {
          matrix(x, ncol = 2)
        })
        aa1 <- lapply(Map(function(x, y) {
          x[y[, 1], ]
        }, Z1, Map(expand.grid, a1, a)), function(x) {
          matrix(x, ncol = 2)
        })
        ZZ <- do.call(rbind, Map(cbind, aa1, aa))
        
        # The records of A:C1 that do not exist in C1:C2 (in terms of
        # the values of classification C1) are adjusted to ZZ which
        # consists of records of A:C1 that exist in C1:C2 (in terms of
        # the values of classification C1).
        t1 <- matrix(TT[is.na(match(TT[, 2], ZZ[, 2])), ], ncol = 2)
        ZZ <- rbind(ZZ, cbind(t1, matrix("", nrow = nrow(t1), ncol = 2)))
        
        F_AtoB[[counter]] <- ZZ
        
      }
      if (Progress) close(pb)
    }
    
    # The following if statement is used when we have only the
    # correspondence tables A:C1, C1:C2 and B:C2.
    if (length(R) == 3) {
      #creating a progress bar
      if (Progress) {
        message("Percentage of codes of ", colnames(RRR[[1]][1]), " processed:")
        pb <- txtProgressBar(min = 0, max = 100, style = 3, width = 50, char = "=")
      }
      
      # The following for loop creates the desirable correspondence
      # table.  The operations are conducted for each unique element of
      # classification A of the correspondence table A:C1.
      for (i in unique(R[[1]][, 1])) {
        counter <- counter + 1
        if (Progress){
          setTxtProgressBar(pb, round(counter/length(unique(R[[1]][, 1])) * 100, digits = 0))
        }
        
        # Matrix T contains the rows of correspondence table C1:C2 that
        # match with the specific element of classification A based on
        # classification C1.
        x1 <- R[[1]][which(R[[1]][, 1] == i), 2]
        T <- matrix(R[[2]][!is.na(match(R[[2]][, 1], x1)), 1:2], ncol = 2)
        
        # The records of A:C1 that do not exist in C1:C2 (in terms of
        # the values of classification C1).
        if (length(which(is.na(match(x1, T[, 1])) == TRUE)) > 0) {
          M1 <- matrix(matrix(R[[1]][which(R[[1]][, 1] == i), 1:2], ncol = 2)[is.na(match(x1,
                                                                                          T[, 1])), ], ncol = 2)
        } else {
          M1 = matrix(0, 1, 2 * length(R))
          M1 = M1[FALSE, ]
        }
        
        if (nrow(M1) != 0) {
          for (times in 1:(2 * length(R) - ncol(M1))) {
            
            M1 <- cbind(M1, "")
            
          }
        }
        
        # Matrix TT contains the rows of correspondence table B:C2 that
        # match with the specific element of classification A based on
        # classification C1.
        x2 <- R[[2]][!is.na(match(R[[2]][, 1], x1)), 2]
        T1 <- matrix(R[[3]][!is.na(match(R[[3]][, 2], x2)), 1:2], ncol = 2)
        
        # The records of C1:C2 that do not exist in B:C2 (in terms of
        # the values of classification C2).
        if (length(which(is.na(match(x2, T1[, 2])) == TRUE)) > 0) {
          if (length(which(is.na(match(x2, T1[, 2])) == TRUE)) == 1) {
            M2 <- matrix(c(i, T[is.na(match(x2, T1[, 2])), 1], T[is.na(match(x2,
                                                                             T1[, 2])), ]), ncol = 4)
          } else {
            M2 <- cbind(i, T[is.na(match(x2, T1[, 2])), 1], T[is.na(match(x2,
                                                                          T1[, 2])), ])
          }
        } else {
          M2 = matrix(0, 1, 2 * length(R))
          M2 = M2[FALSE, ]
        }
        
        if (nrow(M2) != 0) {
          for (times in 1:(2 * length(R) - ncol(M2))) {
            
            M2 <- cbind(M2, "")
            
          }
        }
        
        # Create a list whose each element is a matrix that contains
        # all unique rows of matrix T based on the elements of
        # classification C1.
        t <- match(T[, 2], T[, 2])
        v1 <- sequence(rle(sort(t))$lengths)
        v1 <- split(seq_along(v1), cumsum(v1 == 1))
        Z <- lapply(v1, function(x) {
          T[order(t)[x], , drop = FALSE]
        })
        
        # Create a list whose each element is a matrix that contains
        # all unique rows of matrix TT that match with the unique
        # elements of the second column of matrix T.
        t1 <- match(T1[, 2], T[, 2])
        v1 <- sequence(rle(sort(t1))$lengths)
        v1 <- split(seq_along(v1), cumsum(v1 == 1))
        Z1 <- lapply(v1, function(x) {
          T1[order(t1)[x], , drop = FALSE]
        })
        
        # Keep matrices in Z that exist in Z1 based on their second
        # columns (elements of classification C1).
        Z <- Z[!is.na(match(lapply(Z, function(x) {
          unique(x[, 2])
        }), lapply(Z1, function(x) {
          unique(x[, 2])
        })))]
        
        # ZZ is a matrix that consists of matrices in Z1 expanded by
        # their corresponding matrices (based on the elements of
        # classification C1).
        a <- lapply(Z, function(x) {
          1:nrow(x)
        })
        a1 <- lapply(Z1, function(x) {
          1:nrow(x)
        })
        aa <- lapply(Map(function(x, y) {
          x[y[, 1], ]
        }, Z, Map(expand.grid, a, a1)), function(x) {
          matrix(x, ncol = 2)
        })
        aa1 <- lapply(Map(function(x, y) {
          x[y[, 1], ]
        }, Z1, Map(expand.grid, a1, a)), function(x) {
          matrix(x, ncol = 2)
        })
        ZZ <- do.call(rbind, Map(cbind, aa, aa1))
        
        # The records of both M1 and M2 are adjusted to ZZ which
        # consists of records of A:C1 that exist in C1:C2 (in terms of
        # the values of classification C1).
        if (is.null(dim(ZZ))) {
          F_AtoB[[counter]] <- rbind(M1, M2)
        } else {
          F_AtoB[[counter]] <- rbind(cbind(i, ZZ[, 1], ZZ), M1, M2)
        }
        
      }
      
      if (Progress) close(pb)
    }
    
    # The following if statement is used in the general situation, in which
    # we have the correspondence tables A:C1, Ci:C(i+1) for i = 1, ...,
    # (k-1) Ci and B:Ck.
    M <- list()
    if (length(R) >= 4) {
      if (Progress) {
        message("Percentage of codes of ", colnames(RRR[[1]][1]), " processed:")
        pb <- txtProgressBar(min = 0, max = 100, style = 3, width = 50, char = "=")
      }
      
      # The following for loop creates the desirable correspondence
      # table.  The operations are conducted for each unique element of
      # classification A of the correspondence table A:C1.
      for (i in unique(R[[1]][, 1])) {
        
        counter <- counter + 1
        if (Progress) {
          setTxtProgressBar(pb, round(counter/length(unique(R[[1]][, 1])) * 100, digits = 0))
        }
        
        
        for (j in 1:(length(R) - 2)) {
          # The same operations as in the case that we have only the
          # correspondence tables A:C1 and B:C1, but here for the
          # correspondence tables C1:C2 and C2:C3.
          if (j == 1) {
            
            x1 <- R[[j]][which(R[[j]][, 1] == i), 2]
            T <- matrix(R[[j + 1]][!is.na(match(R[[j + 1]][, 1], x1)), 1:2],
                        ncol = 2)
            
            # The records of A:C1 that do not exist in C1:C2 (in terms
            # of the values of classification C1)
            
            if (length(which(is.na(match(x1, T[, 1])) == TRUE)) > 0) {
              M1 <- matrix(matrix(R[[j]][which(R[[j]][, 1] == i), 1:2], ncol = 2)[is.na(match(x1,
                                                                                              T[, 1])), ], ncol = 2)
            } else {
              M1 = matrix(0, 1, 2 * length(R))
              M1 = M1[FALSE, ]
            }
            
            if (nrow(M1) != 0) {
              for (times in 1:(2 * length(R) - ncol(M1))) {
                
                M1 <- cbind(M1, "")
                
              }
            }
            
            x2 <- R[[j + 1]][!is.na(match(R[[j + 1]][, 1], x1)), 2]
            T1 <- matrix(R[[j + 2]][!is.na(match(R[[j + 2]][, 1], x2)), 1:2],
                         ncol = 2)
            
            if (length(which(is.na(match(x2, T1[, 1])) == TRUE)) > 0) {
              
              if (length(which(is.na(match(x2, T1[, 1])) == TRUE)) == 1) {
                M2 <- matrix(c(i, T[is.na(match(x2, T1[, 1])), 1], T[is.na(match(x2,
                                                                                 T1[, 1])), ]), ncol = 4)
                
              } else {
                M2 <- cbind(i, T[is.na(match(x2, T1[, 1])), 1], T[is.na(match(x2,
                                                                              T1[, 1])), ])
              }
            } else {
              M2 = matrix(0, 1, 2 * length(R))
              M2 = M2[FALSE, ]
            }
            
            if (nrow(M2) != 0) {
              for (times in 1:(2 * length(R) - ncol(M2))) {
                
                M2 <- cbind(M2, "")
                
              }
            }
            
            t <- match(T[, 2], T[, 2])
            v1 <- sequence(rle(sort(t))$lengths)
            v1 <- split(seq_along(v1), cumsum(v1 == 1))
            Z <- lapply(v1, function(x) {
              T[order(t)[x], , drop = FALSE]
            })
            
            t1 <- match(T1[, 1], T[, 2])
            v1 <- sequence(rle(sort(t1))$lengths)
            v1 <- split(seq_along(v1), cumsum(v1 == 1))
            Z1 <- lapply(v1, function(x) {
              T1[order(t1)[x], , drop = FALSE]
            })
            
            Z <- Z[!is.na(match(lapply(Z, function(x) {
              unique(x[, 2])
            }), lapply(Z1, function(x) {
              unique(x[, 1])
            })))]
            
            a <- lapply(Z, function(x) {
              1:nrow(x)
            })
            a1 <- lapply(Z1, function(x) {
              1:nrow(x)
            })
            aa <- lapply(Map(function(x, y) {
              x[y[, 1], ]
            }, Z, Map(expand.grid, a, a1)), function(x) {
              matrix(x, ncol = 2)
            })
            aa1 <- lapply(Map(function(x, y) {
              x[y[, 1], ]
            }, Z1, Map(expand.grid, a1, a)), function(x) {
              matrix(x, ncol = 2)
            })
            ZZ <- do.call(rbind, Map(cbind, aa, aa1))
            
          }
          
          # The same operations as in the case that we have only the
          # correspondence tables A:C1 and B:C1, but here for the pairs
          # of correspondence tables (C2:C3 - C3:C4), (C3:C4 - C4:C5),
          # ..., (C(k-2):C(k-1) - C(k-1):Ck).  For each value of j that
          # satisfies the if statement, the previous matrix ZZ created
          # is used.  For j = 2, the matrix ZZ created in the previous
          # if statement is used.
          if (j >= 2 && j <= (length(R) - 3) && length(R) != 4) {
            
            t <- match(ZZ[, ncol(ZZ)], ZZ[, ncol(ZZ)])
            v1 <- sequence(rle(sort(t))$lengths)
            v1 <- split(seq_along(v1), cumsum(v1 == 1))
            Z <- lapply(v1, function(x) {
              ZZ[order(t)[x], , drop = FALSE]
            })
            
            t1 <- match(R[[j + 2]][, 1], ZZ[, ncol(ZZ)])
            v1 <- sequence(rle(sort(t1))$lengths)
            v1 <- split(seq_along(v1), cumsum(v1 == 1))
            Z1 <- lapply(v1, function(x) {
              R[[j + 2]][order(t1)[x], 1:2, drop = FALSE]
            })
            
            if (length(which(is.na(match(ZZ[, ncol(ZZ)], R[[j + 2]][, 1])) ==
                             TRUE)) > 0) {
              if (length(which(is.na(match(ZZ[, ncol(ZZ)], R[[j + 2]][, 1])) ==
                               TRUE)) == 1) {
                M3 <- matrix(c(i, ZZ[is.na(match(ZZ[, ncol(ZZ)], R[[j + 2]][,
                                                                            1])), 1], ZZ[is.na(match(ZZ[, ncol(ZZ)], R[[j + 2]][, 1])),
                                                                            ]), ncol = ncol(ZZ) + 2)
              } else {
                M3 <- cbind(i, ZZ[is.na(match(ZZ[, ncol(ZZ)], R[[j + 2]][,
                                                                         1])), 1], ZZ[is.na(match(ZZ[, ncol(ZZ)], R[[j + 2]][, 1])),
                                                                         ])
              }
            } else {
              M3 = matrix(0, 1, 2 * length(R))
              M3 = M3[FALSE, ]
            }
            
            if (nrow(M3) != 0) {
              for (times in 1:(2 * length(R) - ncol(M3))) {
                
                M3 <- cbind(M3, "")
                
              }
            }
            
            M[[j - 1]] <- M3
            
            Z <- Z[!is.na(match(lapply(Z, function(x) {
              unique(x[, ncol(ZZ)])
            }), lapply(Z1, function(x) {
              unique(x[, 1])
            })))]
            
            a <- lapply(Z, function(x) {
              1:nrow(x)
            })
            a1 <- lapply(Z1, function(x) {
              1:nrow(x)
            })
            
            aa <- lapply(Map(function(x, y) {
              x[y[, 1], ]
            }, Z, Map(expand.grid, a, a1)), function(x) {
              matrix(x, ncol = ncol(ZZ))
            })
            aa1 <- lapply(Map(function(x, y) {
              x[y[, 1], ]
            }, Z1, Map(expand.grid, a1, a)), function(x) {
              matrix(x, ncol = 2)
            })
            
            ZZ <- do.call(rbind, Map(cbind, aa, aa1))
            
          }
          
          # The same operations as in the case that we have only the
          # correspondence tables A:C1 and B:C1, but here for the
          # correspondence tables C(k-1):Ck and B:Ck.  For the value of
          # j that satisfies the if statement, the matrix ZZ created in
          # the previous if statement is used.
          if (j == (length(R) - 2)) {
            
            t <- match(ZZ[, ncol(ZZ)], ZZ[, ncol(ZZ)])
            v1 <- sequence(rle(sort(t))$lengths)
            v1 <- split(seq_along(v1), cumsum(v1 == 1))
            Z <- lapply(v1, function(x) {
              ZZ[order(t)[x], , drop = FALSE]
            })
            
            t1 <- match(R[[length(R)]][, 2], ZZ[, ncol(ZZ)])
            v1 <- sequence(rle(sort(t1))$lengths)
            v1 <- split(seq_along(v1), cumsum(v1 == 1))
            Z1 <- lapply(v1, function(x) {
              R[[length(R)]][order(t1)[x], 1:2, drop = FALSE]
            })
            
            if (length(which(is.na(match(ZZ[, ncol(ZZ)], R[[length(R)]][,
                                                                        2])) == TRUE)) > 0) {
              if (length(which(is.na(match(ZZ[, ncol(ZZ)], R[[length(R)]][,
                                                                          2])) == TRUE)) == 1) {
                M4 <- matrix(c(i, ZZ[is.na(match(ZZ[, ncol(ZZ)], R[[length(R)]][,
                                                                                2])), 1], ZZ[is.na(match(ZZ[, ncol(ZZ)], R[[length(R)]][,
                                                                                                                                        2])), ]), ncol = ncol(ZZ) + 2)
              } else {
                M4 <- cbind(i, ZZ[is.na(match(ZZ[, ncol(ZZ)], R[[length(R)]][,
                                                                             2])), 1], ZZ[is.na(match(ZZ[, ncol(ZZ)], R[[length(R)]][,
                                                                                                                                     2])), ])
              }
            } else {
              M4 = matrix(0, 1, 2 * length(R))
              M4 = M4[FALSE, ]
            }
            
            
            if (nrow(M4) != 0) {
              for (times in 1:(2 * length(R) - ncol(M4))) {
                
                M4 <- cbind(M4, "")
                
              }
            }
            
            Z <- Z[!is.na(match(lapply(Z, function(x) {
              unique(x[, ncol(ZZ)])
            }), lapply(Z1, function(x) {
              unique(x[, 2])
            })))]
            
            a <- lapply(Z, function(x) {
              1:nrow(x)
            })
            a1 <- lapply(Z1, function(x) {
              1:nrow(x)
            })
            
            aa <- lapply(Map(function(x, y) {
              x[y[, 1], ]
            }, Z, Map(expand.grid, a, a1)), function(x) {
              matrix(x, ncol = ncol(ZZ))
            })
            aa1 <- lapply(Map(function(x, y) {
              x[y[, 1], ]
            }, Z1, Map(expand.grid, a1, a)), function(x) {
              matrix(x, ncol = 2)
            })
            
            ZZ <- do.call(rbind, Map(cbind, aa, aa1))
            
          }
        }
        
        if (is.null(dim(ZZ))) {
          F_AtoB[[counter]] <- rbind(M1, M2, do.call(rbind, M), M4)
        } else {
          F_AtoB[[counter]] <- rbind(cbind(i, ZZ[, 1], ZZ), M1, M2, do.call(rbind,
                                                                            M), M4)
        }
      }
      if (Progress) close(pb)
    }
    
    # Create the desired correspondence table for the selected element of
    # classification A.
    F_AtoB <- do.call(rbind, F_AtoB)
    
    # Keep in F the classifications A, C1, C2, ..., Ck, B once, based on
    # the number of the correspondence tables.
    if (length(R) == 2) {
      F_AtoB <- F_AtoB[, c(1, 2, 3)]
    }
    if (length(R) == 3) {
      F_AtoB <- F_AtoB[, c(1, 2, 4, 5)]
    }
    if (length(R) >= 4) {
      F_AtoB <- F_AtoB[, sort(c(1, seq(2, 2 * length(R) - 2, 2), 2 * length(R) -
                                  1))]
    }
    
    # Convert classifications as well as correspondence tables so as to
    # move from classification B to classification A.  Until the next
    # comment, all the lines are the same as in the case that we move from
    # classification A to classification B.
    RRR_BtoA <- RRR[c(rev(1:(k + 2)), rev(utils::tail(c(1:length(RRR)), (length(RRR) -
                                                                           1)/2)))]
    if (length(rev(utils::tail(c(1:length(RR)), (length(RR) - 1)/2))) >= 3) {
      for (rev in (k + 4):(length(RRR_BtoA) - 1)) {
        column_2 <- RRR_BtoA[[rev]][, 2]
        RRR_BtoA[[rev]][, 2] <- RRR_BtoA[[rev]][, 1]
        RRR_BtoA[[rev]][, 1] <- column_2
      }
    }
    
    RR <- lapply(RRR_BtoA, function(x) {
      matrix(unlist(x), ncol = ncol(x))
    })
    
    R <- RR[utils::tail(c(1:length(RR)), (length(RR) - 1)/2)]
    
    F_BtoA <- list()
    
    counter <- 0
    if (Progress) message("\n")
    if (length(R) == 2) {
      if (Progress) {
        message("Percentage of codes of ", colnames(RRR_BtoA[[1]][1]), " processed:")
        pb <- txtProgressBar(min = 0, max = 100, style = 3, width = 50, char = "=")
      }
      
      for (i in unique(R[[1]][, 1])) {
        
        counter <- counter + 1
        if (Progress) {
          setTxtProgressBar(pb, round(counter/length(unique(R[[1]][, 1])) * 100, digits = 0))
        }
        
        x1 <- R[[1]][which(R[[1]][, 1] == i), 2]
        TT <- matrix(R[[1]][which(R[[1]][, 1] == i), 1:2], ncol = 2)
        T <- matrix(R[[2]][!is.na(match(R[[2]][, 2], x1)), 1:2], ncol = 2)
        
        t <- match(T[, 2], T[, 2])
        v1 <- sequence(rle(sort(t))$lengths)
        v1 <- split(seq_along(v1), cumsum(v1 == 1))
        Z <- lapply(v1, function(x) {
          T[order(t)[x], , drop = FALSE]
        })
        
        t1 <- match(TT[, 2], T[, 2])
        v1 <- sequence(rle(sort(t1))$lengths)
        v1 <- split(seq_along(v1), cumsum(v1 == 1))
        Z1 <- lapply(v1, function(x) {
          TT[order(t1)[x], , drop = FALSE]
        })
        
        Z <- Z[!is.na(match(lapply(Z, function(x) {
          unique(x[, 2])
        }), lapply(Z1, function(x) {
          unique(x[, 2])
        })))]
        
        a <- lapply(Z, function(x) {
          1:nrow(x)
        })
        a1 <- lapply(Z1, function(x) {
          1:nrow(x)
        })
        aa <- lapply(Map(function(x, y) {
          x[y[, 1], ]
        }, Z, Map(expand.grid, a, a1)), function(x) {
          matrix(x, ncol = 2)
        })
        aa1 <- lapply(Map(function(x, y) {
          x[y[, 1], ]
        }, Z1, Map(expand.grid, a1, a)), function(x) {
          matrix(x, ncol = 2)
        })
        ZZ <- do.call(rbind, Map(cbind, aa1, aa))
        
        t1 <- matrix(TT[is.na(match(TT[, 2], ZZ[, 2])), ], ncol = 2)
        ZZ <- rbind(ZZ, cbind(t1, matrix("", nrow = nrow(t1), ncol = 2)))
        
        F_BtoA[[counter]] <- ZZ
        
      }
      if (Progress) close(pb)
    }
    
    if (length(R) == 3) {
      if (Progress) {
        message("Percentage of codes of ", colnames(RRR_BtoA[[1]][1]), " processed:")
        pb <- txtProgressBar(min = 0, max = 100, style = 3, width = 50, char = "=")
      }
      
      for (i in unique(R[[1]][, 1])) {
        
        counter <- counter + 1
        if (Progress) {
          setTxtProgressBar(pb, round(counter/length(unique(R[[1]][, 1])) * 100, digits = 0))
        }
        
        x1 <- R[[1]][which(R[[1]][, 1] == i), 2]
        T <- matrix(R[[2]][!is.na(match(R[[2]][, 1], x1)), 1:2], ncol = 2)
        
        if (length(which(is.na(match(x1, T[, 1])) == TRUE)) > 0) {
          M1 <- matrix(matrix(R[[1]][which(R[[1]][, 1] == i), 1:2], ncol = 2)[is.na(match(x1,
                                                                                          T[, 1])), ], ncol = 2)
        } else {
          M1 = matrix(0, 1, 2 * length(R))
          M1 = M1[FALSE, ]
        }
        
        if (nrow(M1) != 0) {
          for (times in 1:(2 * length(R) - ncol(M1))) {
            
            M1 <- cbind(M1, "")
            
          }
        }
        
        x2 <- R[[2]][!is.na(match(R[[2]][, 1], x1)), 2]
        T1 <- matrix(R[[3]][!is.na(match(R[[3]][, 2], x2)), 1:2], ncol = 2)
        
        if (length(which(is.na(match(x2, T1[, 2])) == TRUE)) > 0) {
          if (length(which(is.na(match(x2, T1[, 2])) == TRUE)) == 1) {
            M2 <- matrix(c(i, T[is.na(match(x2, T1[, 2])), 1], T[is.na(match(x2,
                                                                             T1[, 2])), ]), ncol = 4)
          } else {
            M2 <- cbind(i, T[is.na(match(x2, T1[, 2])), 1], T[is.na(match(x2,
                                                                          T1[, 2])), ])
          }
        } else {
          M2 = matrix(0, 1, 2 * length(R))
          M2 = M2[FALSE, ]
        }
        
        if (nrow(M2) != 0) {
          for (times in 1:(2 * length(R) - ncol(M2))) {
            
            M2 <- cbind(M2, "")
            
          }
        }
        
        t <- match(T[, 2], T[, 2])
        v1 <- sequence(rle(sort(t))$lengths)
        v1 <- split(seq_along(v1), cumsum(v1 == 1))
        Z <- lapply(v1, function(x) {
          T[order(t)[x], , drop = FALSE]
        })
        
        t1 <- match(T1[, 2], T[, 2])
        v1 <- sequence(rle(sort(t1))$lengths)
        v1 <- split(seq_along(v1), cumsum(v1 == 1))
        Z1 <- lapply(v1, function(x) {
          T1[order(t1)[x], , drop = FALSE]
        })
        
        Z <- Z[!is.na(match(lapply(Z, function(x) {
          unique(x[, 2])
        }), lapply(Z1, function(x) {
          unique(x[, 2])
        })))]
        
        a <- lapply(Z, function(x) {
          1:nrow(x)
        })
        a1 <- lapply(Z1, function(x) {
          1:nrow(x)
        })
        aa <- lapply(Map(function(x, y) {
          x[y[, 1], ]
        }, Z, Map(expand.grid, a, a1)), function(x) {
          matrix(x, ncol = 2)
        })
        aa1 <- lapply(Map(function(x, y) {
          x[y[, 1], ]
        }, Z1, Map(expand.grid, a1, a)), function(x) {
          matrix(x, ncol = 2)
        })
        ZZ <- do.call(rbind, Map(cbind, aa, aa1))
        
        if (is.null(dim(ZZ))) {
          F_BtoA[[counter]] <- rbind(M1, M2)
        } else {
          F_BtoA[[counter]] <- rbind(cbind(i, ZZ[, 1], ZZ), M1, M2)
        }
        
      }
      if (Progress) close(pb)
    }
    M <- list()
    if (length(R) >= 4) {
      if (Progress) {
        message("Percentage of codes of ", colnames(RRR_BtoA[[1]][1]), " processed:")
        pb <- txtProgressBar(min = 0, max = 100, style = 3, width = 50, char = "=")
      }
      
      for (i in unique(R[[1]][, 1])) {
        
        counter <- counter + 1
        if (Progress) {
          setTxtProgressBar(pb, round(counter/length(unique(R[[1]][, 1])) * 100, digits = 0))
        }
        
        for (j in 1:(length(R) - 2)) {
          if (j == 1) {
            
            x1 <- R[[j]][which(R[[j]][, 1] == i), 2]
            T <- matrix(R[[j + 1]][!is.na(match(R[[j + 1]][, 1], x1)), 1:2],
                        ncol = 2)
            
            if (length(which(is.na(match(x1, T[, 1])) == TRUE)) > 0) {
              M1 <- matrix(matrix(R[[j]][which(R[[j]][, 1] == i), 1:2], ncol = 2)[is.na(match(x1,
                                                                                              T[, 1])), ], ncol = 2)
            } else {
              M1 = matrix(0, 1, 2 * length(R))
              M1 = M1[FALSE, ]
            }
            
            if (nrow(M1) != 0) {
              for (times in 1:(2 * length(R) - ncol(M1))) {
                
                M1 <- cbind(M1, "")
                
              }
            }
            
            x2 <- R[[j + 1]][!is.na(match(R[[j + 1]][, 1], x1)), 2]
            T1 <- matrix(R[[j + 2]][!is.na(match(R[[j + 2]][, 1], x2)), 1:2],
                         ncol = 2)
            
            if (length(which(is.na(match(x2, T1[, 1])) == TRUE)) > 0) {
              
              if (length(which(is.na(match(x2, T1[, 1])) == TRUE)) == 1) {
                M2 <- matrix(c(i, T[is.na(match(x2, T1[, 1])), 1], T[is.na(match(x2,
                                                                                 T1[, 1])), ]), ncol = 4)
                
              } else {
                M2 <- cbind(i, T[is.na(match(x2, T1[, 1])), 1], T[is.na(match(x2,
                                                                              T1[, 1])), ])
              }
            } else {
              M2 = matrix(0, 1, 2 * length(R))
              M2 = M2[FALSE, ]
            }
            
            if (nrow(M2) != 0) {
              for (times in 1:(2 * length(R) - ncol(M2))) {
                
                M2 <- cbind(M2, "")
                
              }
            }
            
            t <- match(T[, 2], T[, 2])
            v1 <- sequence(rle(sort(t))$lengths)
            v1 <- split(seq_along(v1), cumsum(v1 == 1))
            Z <- lapply(v1, function(x) {
              T[order(t)[x], , drop = FALSE]
            })
            
            t1 <- match(T1[, 1], T[, 2])
            v1 <- sequence(rle(sort(t1))$lengths)
            v1 <- split(seq_along(v1), cumsum(v1 == 1))
            Z1 <- lapply(v1, function(x) {
              T1[order(t1)[x], , drop = FALSE]
            })
            
            Z <- Z[!is.na(match(lapply(Z, function(x) {
              unique(x[, 2])
            }), lapply(Z1, function(x) {
              unique(x[, 1])
            })))]
            
            a <- lapply(Z, function(x) {
              1:nrow(x)
            })
            a1 <- lapply(Z1, function(x) {
              1:nrow(x)
            })
            aa <- lapply(Map(function(x, y) {
              x[y[, 1], ]
            }, Z, Map(expand.grid, a, a1)), function(x) {
              matrix(x, ncol = 2)
            })
            aa1 <- lapply(Map(function(x, y) {
              x[y[, 1], ]
            }, Z1, Map(expand.grid, a1, a)), function(x) {
              matrix(x, ncol = 2)
            })
            ZZ <- do.call(rbind, Map(cbind, aa, aa1))
            
          }
          
          if (j >= 2 && j <= (length(R) - 3) && length(R) != 4) {
            
            t <- match(ZZ[, ncol(ZZ)], ZZ[, ncol(ZZ)])
            v1 <- sequence(rle(sort(t))$lengths)
            v1 <- split(seq_along(v1), cumsum(v1 == 1))
            Z <- lapply(v1, function(x) {
              ZZ[order(t)[x], , drop = FALSE]
            })
            
            t1 <- match(R[[j + 2]][, 1], ZZ[, ncol(ZZ)])
            v1 <- sequence(rle(sort(t1))$lengths)
            v1 <- split(seq_along(v1), cumsum(v1 == 1))
            Z1 <- lapply(v1, function(x) {
              R[[j + 2]][order(t1)[x], 1:2, drop = FALSE]
            })
            
            if (length(which(is.na(match(ZZ[, ncol(ZZ)], R[[j + 2]][, 1])) ==
                             TRUE)) > 0) {
              if (length(which(is.na(match(ZZ[, ncol(ZZ)], R[[j + 2]][, 1])) ==
                               TRUE)) == 1) {
                M3 <- matrix(c(i, ZZ[is.na(match(ZZ[, ncol(ZZ)], R[[j + 2]][,
                                                                            1])), 1], ZZ[is.na(match(ZZ[, ncol(ZZ)], R[[j + 2]][, 1])),
                                                                            ]), ncol = ncol(ZZ) + 2)
              } else {
                M3 <- cbind(i, ZZ[is.na(match(ZZ[, ncol(ZZ)], R[[j + 2]][,
                                                                         1])), 1], ZZ[is.na(match(ZZ[, ncol(ZZ)], R[[j + 2]][, 1])),
                                                                         ])
              }
            } else {
              M3 = matrix(0, 1, 2 * length(R))
              M3 = M3[FALSE, ]
            }
            
            if (nrow(M3) != 0) {
              for (times in 1:(2 * length(R) - ncol(M3))) {
                
                M3 <- cbind(M3, "")
                
              }
            }
            M[[j - 1]] <- M3
            
            Z <- Z[!is.na(match(lapply(Z, function(x) {
              unique(x[, ncol(ZZ)])
            }), lapply(Z1, function(x) {
              unique(x[, 1])
            })))]
            
            a <- lapply(Z, function(x) {
              1:nrow(x)
            })
            a1 <- lapply(Z1, function(x) {
              1:nrow(x)
            })
            
            aa <- lapply(Map(function(x, y) {
              x[y[, 1], ]
            }, Z, Map(expand.grid, a, a1)), function(x) {
              matrix(x, ncol = ncol(ZZ))
            })
            aa1 <- lapply(Map(function(x, y) {
              x[y[, 1], ]
            }, Z1, Map(expand.grid, a1, a)), function(x) {
              matrix(x, ncol = 2)
            })
            
            ZZ <- do.call(rbind, Map(cbind, aa, aa1))
            
          }
          
          if (j == (length(R) - 2)) {
            
            t <- match(ZZ[, ncol(ZZ)], ZZ[, ncol(ZZ)])
            v1 <- sequence(rle(sort(t))$lengths)
            v1 <- split(seq_along(v1), cumsum(v1 == 1))
            Z <- lapply(v1, function(x) {
              ZZ[order(t)[x], , drop = FALSE]
            })
            
            t1 <- match(R[[length(R)]][, 2], ZZ[, ncol(ZZ)])
            v1 <- sequence(rle(sort(t1))$lengths)
            v1 <- split(seq_along(v1), cumsum(v1 == 1))
            Z1 <- lapply(v1, function(x) {
              R[[length(R)]][order(t1)[x], 1:2, drop = FALSE]
            })
            
            if (length(which(is.na(match(ZZ[, ncol(ZZ)], R[[length(R)]][,
                                                                        2])) == TRUE)) > 0) {
              if (length(which(is.na(match(ZZ[, ncol(ZZ)], R[[length(R)]][,
                                                                          2])) == TRUE)) == 1) {
                M4 <- matrix(c(i, ZZ[is.na(match(ZZ[, ncol(ZZ)], R[[length(R)]][,
                                                                                2])), 1], ZZ[is.na(match(ZZ[, ncol(ZZ)], R[[length(R)]][,
                                                                                                                                        2])), ]), ncol = ncol(ZZ) + 2)
              } else {
                M4 <- cbind(i, ZZ[is.na(match(ZZ[, ncol(ZZ)], R[[length(R)]][,
                                                                             2])), 1], ZZ[is.na(match(ZZ[, ncol(ZZ)], R[[length(R)]][,
                                                                                                                                     2])), ])
              }
            } else {
              M4 = matrix(0, 1, 2 * length(R))
              M4 = M4[FALSE, ]
            }
            
            
            if (nrow(M4) != 0) {
              for (times in 1:(2 * length(R) - ncol(M4))) {
                
                M4 <- cbind(M4, "")
                
              }
            }
            
            Z <- Z[!is.na(match(lapply(Z, function(x) {
              unique(x[, ncol(ZZ)])
            }), lapply(Z1, function(x) {
              unique(x[, 2])
            })))]
            
            a <- lapply(Z, function(x) {
              1:nrow(x)
            })
            a1 <- lapply(Z1, function(x) {
              1:nrow(x)
            })
            
            aa <- lapply(Map(function(x, y) {
              x[y[, 1], ]
            }, Z, Map(expand.grid, a, a1)), function(x) {
              matrix(x, ncol = ncol(ZZ))
            })
            aa1 <- lapply(Map(function(x, y) {
              x[y[, 1], ]
            }, Z1, Map(expand.grid, a1, a)), function(x) {
              matrix(x, ncol = 2)
            })
            
            ZZ <- do.call(rbind, Map(cbind, aa, aa1))
            
          }
        }
        
        if (is.null(dim(ZZ))) {
          F_BtoA[[counter]] <- rbind(M1, M2, do.call(rbind, M), M4)
        } else {
          F_BtoA[[counter]] <- rbind(cbind(i, ZZ[, 1], ZZ), M1, M2, do.call(rbind,
                                                                            M), M4)
        }
        
        
      }
      if (Progress) close(pb)
    }
    
    F_BtoA <- do.call(rbind, F_BtoA)
    
    if (length(R) == 2) {
      F_BtoA <- F_BtoA[, c(1, 2, 3)]
    }
    if (length(R) == 3) {
      F_BtoA <- F_BtoA[, c(1, 2, 4, 5)]
    }
    if (length(R) >= 4) {
      F_BtoA <- F_BtoA[, sort(c(1, seq(2, 2 * length(R) - 2, 2), 2 * length(R) -
                                  1))]
    }
    
    
    F_BtoA <- F_BtoA[, rev(1:ncol(F_BtoA))]
    # Combine the results from moving from classification A to B, and vice
    # versa.  F_AtoB
    keep <- 0
    keepF_AtoB <- c(0)
    for (iterr in 1:nrow(F_AtoB)) {
      
      if (F_AtoB[iterr, 1] != "") {
        blanks <- F_AtoB[iterr, ] == ""
        
        if (all(blanks == FALSE)) {
          keep <- keep + 1
          keepF_AtoB[keep] <- iterr
        } else {
          blanks = which(F_AtoB[iterr, ] == "")
          if (all(c(blanks[1]:ncol(F_AtoB)) == "")) {
            keep <- keep + 1
            keepF_AtoB[keep] <- iterr
          }
        }
        
      }
    }
    
    NoNullF_AtoB <- matrix(F_AtoB[keepF_AtoB, ], ncol = k + 2)
    if (nrow(NoNullF_AtoB) != nrow(F_AtoB)) {
      if (length(keepF_AtoB) == 1 && keepF_AtoB == c(0)) {
        FNullAtoB <- matrix(F_AtoB, ncol = k + 2)
        for (iter in 1:nrow(FNullAtoB)) {
          FNullAtoB[iter, (which(FNullAtoB[iter, ] == "")[1]):(k + 2)] <- ""
        }
      } else {
        FNullAtoB <- matrix(F_AtoB[-keepF_AtoB, ], ncol = k + 2)
        for (iter in 1:nrow(FNullAtoB)) {
          FNullAtoB[iter, (which(FNullAtoB[iter, ] == "")[1]):(k + 2)] <- ""
        }
      }
    } else {
      FNullAtoB <- matrix(0, 1, k + 2)
      FNullAtoB <- FNullAtoB[FALSE, ]
    }
    
    # F_BtoA
    keep <- 0
    keepF_BtoA <- c(0)
    for (iterr in 1:nrow(F_BtoA)) {
      
      if (F_BtoA[iterr, ncol(F_AtoB)] != "") {
        blanks <- F_BtoA[iterr, ] == ""
        
        if (all(blanks == FALSE)) {
          keep <- keep + 1
          keepF_BtoA[keep] <- iterr
        } else {
          blanks <- which(F_BtoA[iterr, ] == "")
          if (all(c(1:length(blanks)) == "")) {
            keep <- keep + 1
            keepF_BtoA[keep] <- iterr
          }
        }
        
      }
    }
    
    # Combine all together
    
    NoNullF_BtoA <- matrix(F_BtoA[keepF_BtoA, ], ncol = k + 2)
    if (nrow(NoNullF_BtoA) != nrow(F_BtoA)) {
      if (length(keepF_BtoA) == 1 && keepF_BtoA == c(0)) {
        FNullBtoA <- matrix(F_BtoA, ncol = k + 2)
        for (iter in 1:nrow(FNullBtoA)) {
          FNullBtoA[iter, (which(FNullBtoA[iter, ] == "")[length(which(FNullBtoA[iter,
          ] == ""))]):1] <- ""
        }
      } else {
        FNullBtoA <- matrix(F_BtoA[-keepF_BtoA, ], ncol = k + 2)
        for (iter in 1:nrow(FNullBtoA)) {
          FNullBtoA[iter, (which(FNullBtoA[iter, ] == "")[length(which(FNullBtoA[iter,
          ] == ""))]):1] <- ""
        }
      }
    } else {
      FNullBtoA <- matrix(0, 1, k + 2)
      FNullBtoA <- FNullBtoA[FALSE, ]
    }
    
    F <- unique(rbind(NoNullF_AtoB, NoNullF_BtoA))
    F <- unique(rbind(F, unique(FNullAtoB), unique(FNullBtoA)))
    if (length(which(apply(F, 1, function(x) {
      length(which(x == ""))
    } == k + 2) == TRUE)) >= 1) {
      F <- F[-which(apply(F, 1, function(x) {
        length(which(x == ""))
      } == k + 2) == TRUE), ]
    }
    
    # The if statement is based on which of classifications A or B is the
    # reference one (if any).
    if (length(which(apply(F, 1, function(x) {
      length(which(x == ""))
    }) == 0)) >= 1) {
      
      if (Reference == "A") {
        idx <- k + 5
        
        # Creation of the review flag for the correspondence table A:B.
        F1 <- matrix(F[apply(F, 1, function(x) {
          length(which(x == ""))
        }) == 0, ], ncol = k + 2)
        F2 <- F[apply(F, 1, function(x) {
          length(which(x == ""))
        }) >= 1, ]
        F2 <- matrix(unlist(F2), ncol = k + 2)
        f <- stats::aggregate(matrix(unique(F1[, c(1, ncol(F1))]), ncol = 2)[, 2],
                              list(num = matrix(unique(F1[, c(1, ncol(F1))]), ncol = 2)[, 2]),
                              length)[which(stats::aggregate(matrix(unique(F1[, c(1, ncol(F1))]), ncol = 2)[,
                                                                                                            2], list(num = matrix(unique(F1[, c(1, ncol(F1))]), ncol = 2)[,
                                                                                                                                                                          2]), length)[, 2] > 1), 1]
        reviewF1 <- rep(0, nrow(F1))
        reviewF1[which(F1[, ncol(F1)] %in% f)] <- 1
        Review <- data.frame(cbind(rbind(F1, F2), c(reviewF1, rep(0, nrow(F2)))))
        
        # Creation of the redundancy flag for the correspondence table
        # A:B.
        F1 <- Review[apply(Review, 1, function(x) {
          length(which(x == ""))
        }) == 0, ]
        F1 <- matrix(unlist(F1), ncol = k + 3)
        F1 <- data.frame(F1)
        colnames(F1) <- colnames(Review)
        F2 <- Review[apply(Review, 1, function(x) {
          length(which(x == ""))
        }) >= 1, ]
        F2 <- matrix(unlist(F2), ncol = k + 3)
        F2 <- data.frame(F2)
        colnames(F2) <- colnames(F1)
        f1 <- stats::aggregate(F1[, c(1, ncol(F1) - 1)], by = F1[, c(1, ncol(F1) -
                                                                       1)], length)[1:(ncol(F1[, c(1, ncol(F1) - 1)]) + 1)][which(stats::aggregate(F1[,
                                                                                                                                                      c(1, ncol(F1) - 1)], by = F1[, c(1, ncol(F1) - 1)], length)[1:(ncol(F1[,
                                                                                                                                                                                                                             c(1, ncol(F1) - 1)]) + 1)][, 3] >= 2), 1:2]
        redundancyF1 <- rep(0, nrow(F1))
        redundancyF1[which(apply(F1[, c(1, ncol(F1) - 1)], 1, paste, collapse = "") %in%
                             apply(f1, 1, paste, collapse = ""))] <- 1
        correspondenceAB <- data.frame(cbind(rbind(F1, F2), c(redundancyF1,
                                                              rep(0, nrow(F2)))))
        
        # Creation of the unmatched flag for the correspondence table
        # A:B.
        correspondenceAB <- data.frame(correspondenceAB, 1)
        colnames(correspondenceAB) <- c(paste(colnames(RRR[[1]][1])), paste(unlist(lapply(RRR,
                                                                                          function(x) {
                                                                                            colnames(x)[1]
                                                                                          }))[seq(k) + 1]), paste(colnames(RRR[[k + 2]][1])), "Review", "Redundancy",
                                        "Unmatched")
        
      } else if (Reference == "B") {
        idx <- k + 5
        
        # Creation of the review flag for the correspondence table A:B.
        F1 <- matrix(F[apply(F, 1, function(x) {
          length(which(x == ""))
        }) == 0, ], ncol = k + 2)
        F2 <- F[apply(F, 1, function(x) {
          length(which(x == ""))
        }) >= 1, ]
        F2 <- matrix(unlist(F2), ncol = k + 2)
        f <- stats::aggregate(matrix(unique(F1[, c(1, ncol(F1))]), ncol = 2)[, 1],
                              list(num = matrix(unique(F1[, c(1, ncol(F1))]), ncol = 2)[, 1]),
                              length)[which(stats::aggregate(matrix(unique(F1[, c(1, ncol(F1))]), ncol = 2)[,
                                                                                                            1], list(num = matrix(unique(F1[, c(1, ncol(F1))]), ncol = 2)[,
                                                                                                                                                                          1]), length)[, 2] > 1), 1]
        reviewF1 <- rep(0, nrow(F1))
        reviewF1[which(F1[, 1] %in% f)] <- 1
        Review <- data.frame(cbind(rbind(F1, F2), c(reviewF1, rep(0, nrow(F2)))))
        
        # Creation of the redundancy flag for the correspondence table
        # A:B.
        F1 <- Review[apply(Review, 1, function(x) {
          length(which(x == ""))
        }) == 0, ]
        F1 <- matrix(unlist(F1), ncol = k + 3)
        F1 <- data.frame(F1)
        colnames(F1) <- colnames(Review)
        F2 <- Review[apply(Review, 1, function(x) {
          length(which(x == ""))
        }) >= 1, ]
        F2 <- matrix(unlist(F2), ncol = k + 3)
        F2 <- data.frame(F2)
        colnames(F2) <- colnames(F1)
        f1 <- stats::aggregate(F1[, c(1, ncol(F1) - 1)], by = F1[, c(1, ncol(F1) -
                                                                       1)], length)[1:(ncol(F1[, c(1, ncol(F1) - 1)]) + 1)][which(stats::aggregate(F1[,
                                                                                                                                                      c(1, ncol(F1) - 1)], by = F1[, c(1, ncol(F1) - 1)], length)[1:(ncol(F1[,
                                                                                                                                                                                                                             c(1, ncol(F1) - 1)]) + 1)][, 3] >= 2), 1:2]
        redundancyF1 <- rep(0, nrow(F1))
        redundancyF1[which(apply(F1[, c(1, ncol(F1) - 1)], 1, paste, collapse = "") %in%
                             apply(f1, 1, paste, collapse = ""))] <- 1
        correspondenceAB <- data.frame(cbind(rbind(F1, F2), c(redundancyF1,
                                                              rep(0, nrow(F2)))))
        
        # Creation of the unmatched flag for the correspondence table
        # A:B.
        correspondenceAB <- data.frame(correspondenceAB, 1)
        colnames(correspondenceAB) <- c(paste(colnames(RRR[[1]][1])), paste(unlist(lapply(RRR,
                                                                                          function(x) {
                                                                                            colnames(x)[1]
                                                                                          }))[seq(k) + 1]), paste(colnames(RRR[[k + 2]][1])), "Review", "Redundancy",
                                        "Unmatched")
        
      } else if (Reference == "none") {
        idx <- k + 4
        
        # Creation of the redundancy flag for the correspondence table
        # A:B.
        F1 <- data.frame(F[apply(F, 1, function(x) {
          length(which(x == ""))
        }) == 0, ])
        F1 <- matrix(unlist(F1), ncol = k + 2)
        F1 <- data.frame(F1)
        F2 <- data.frame(F[apply(F, 1, function(x) {
          length(which(x == ""))
        }) >= 1, ])
        F2 <- matrix(unlist(F2), ncol = k + 2)
        F2 <- data.frame(F2)
        colnames(F2) <- colnames(F1)
        f1 <- stats::aggregate(F1[, c(1, ncol(F1))], by = F1[, c(1, ncol(F1))],
                               length)[1:(ncol(F1[, c(1, ncol(F1))]) + 1)][which(stats::aggregate(F1[,
                                                                                                     c(1, ncol(F1))], by = F1[, c(1, ncol(F1))], length)[1:(ncol(F1[,
                                                                                                                                                                    c(1, ncol(F1))]) + 1)][, 3] >= 2), 1:2]
        redundancyF1 <- rep(0, nrow(F1))
        redundancyF1[which(apply(F1[, c(1, ncol(F1))], 1, paste, collapse = "") %in%
                             apply(f1, 1, paste, collapse = ""))] <- 1
        correspondenceAB <- data.frame(cbind(rbind(F1, F2), c(redundancyF1,
                                                              rep(0, nrow(F2)))))
        
        # Creation of the unmatched flag for the correspondence table
        # A:B.
        correspondenceAB <- data.frame(correspondenceAB, 1)
        colnames(correspondenceAB) <- c(paste(colnames(RRR[[1]][1])), paste(unlist(lapply(RRR,
                                                                                          function(x) {
                                                                                            colnames(x)[1]
                                                                                          }))[seq(k) + 1]), paste(colnames(RRR[[k + 2]][1])), "Redundancy",
                                        "Unmatched")
        
      }
    } else {
      if (Reference %in% c("A", "B")) {
        Review <- rep(0, nrow(F))
        Redundancy <- rep(0, nrow(F))
        Unmatched <- rep(1, nrow(F))
        correspondenceAB <- data.frame(cbind(F, Review, Redundancy, Unmatched))
        colnames(correspondenceAB) <- c(paste(colnames(RRR[[1]][1])), paste(unlist(lapply(RRR,
                                                                                          function(x) {
                                                                                            colnames(x)[1]
                                                                                          }))[seq(k) + 1]), paste(colnames(RRR[[k + 2]][1])), "Review", "Redundancy",
                                        "Unmatched")
      }
      if (Reference == "none") {
        Redundancy <- rep(0, nrow(F))
        Unmatched <- rep(1, nrow(F))
        correspondenceAB <- data.frame(cbind(F, Redundancy, Unmatched))
        colnames(correspondenceAB) <- c(paste(colnames(RRR[[1]][1])), paste(unlist(lapply(RRR,
                                                                                          function(x) {
                                                                                            colnames(x)[1]
                                                                                          }))[seq(k) + 1]), paste(colnames(RRR[[k + 2]][1])), "Redundancy",
                                        "Unmatched")
      }
    }
    
    
    # The final Unmatched and the NoMatchFrom flags are created
    NoMatchFromA <- rep("", nrow(correspondenceAB))
    NoMatchFromB <- rep("", nrow(correspondenceAB))
    correspondenceAB <- cbind(correspondenceAB, NoMatchFromA, NoMatchFromB)
    
    inA <- which(is.na(match(unlist(RRR[[1]][, 1]), correspondenceAB[, 1])) == TRUE)
    if (length(inA) >= 1) {
      InA <- cbind(matrix(RRR[[1]][inA, 1], length(inA), 1), matrix("", length(inA),
                                                                    idx - 1))
      InA <- cbind(InA, matrix("", length(inA), 2))
      InA <- data.frame(InA)
      colnames(InA) <- colnames(correspondenceAB)
      correspondenceAB <- rbind(correspondenceAB, InA)
    }
    
    inB <- which(is.na(match(unlist(RRR[[nrow(x)]][, 1]), correspondenceAB[, k + 2])) ==
                   TRUE)
    if (length(inB) >= 1) {
      InB <- cbind(matrix("", length(inB), k + 1), matrix(RRR[[nrow(x)]][inB,
                                                                         1], length(inB), 1), matrix("", length(inB), idx - k - 2))
      InB <- cbind(InB, matrix("", length(inB), 2))
      InB <- data.frame(InB)
      colnames(InB) <- colnames(correspondenceAB)
      correspondenceAB <- rbind(correspondenceAB, InB)
    }
    
    yesA <- which(!is.na(match(correspondenceAB[, 1], unlist(RRR[[1]][, 1]))) == TRUE)
    yesAC1 <- which(!is.na(match(correspondenceAB[, 1], unlist(RRR[[nrow(x) + 1]][,
                                                                                  1]))) == TRUE)
    noAC1 <- which(is.na(match(correspondenceAB[, 1], unlist(RRR[[nrow(x) + 1]][, 1]))) ==
                     TRUE)
    
    correspondenceAB$NoMatchFromA[intersect(yesA, yesAC1)] <- 0
    correspondenceAB$NoMatchFromA[intersect(yesA, noAC1)] <- 1
    
    yesB <- which(!is.na(match(correspondenceAB[, k + 2], unlist(RRR[[nrow(x)]][, 1]))) ==
                    TRUE)
    yesBCk <- which(!is.na(match(correspondenceAB[, k + 2], unlist(RRR[[length(RRR)]][,
                                                                                      1]))) == TRUE)
    noBCk <- which(is.na(match(correspondenceAB[, k + 2], unlist(RRR[[length(RRR)]][,
                                                                                    1]))) == TRUE)
    
    correspondenceAB$NoMatchFromB[intersect(yesB, yesBCk)] <- 0
    correspondenceAB$NoMatchFromB[intersect(yesB, noBCk)] <- 1
    
    yesFinalA <- which(correspondenceAB[, 1] != "")
    yesFinalB <- which(correspondenceAB[, k + 2] != "")
    correspondenceAB$Unmatched <- 1
    correspondenceAB$Unmatched[intersect(yesFinalA, yesFinalB)] <- 0
    
    if ((Reference %in% c("A", "B"))) {
      correspondenceAB$Review[which(correspondenceAB[, 1] == "")] <- ""
      correspondenceAB$Review[which(correspondenceAB[, k + 2] == "")] <- ""
    }
    
    # Final redundancy flag
    correspondenceAB$Redundancy <- 0
    f1 <- stats::aggregate(correspondenceAB[, c(1, k + 2)], by = correspondenceAB[,
                                                                                  c(1, k + 2)], length)[1:(ncol(correspondenceAB[, c(1, k + 2)]) + 1)][which(stats::aggregate(correspondenceAB[,
                                                                                                                                                                                               c(1, k + 2)], by = correspondenceAB[, c(1, k + 2)], length)[1:(ncol(correspondenceAB[,
                                                                                                                                                                                                                                                                                    c(1, k + 2)]) + 1)][, 3] >= 2), 1:2]
    correspondenceAB$Redundancy[which(apply(correspondenceAB[, c(1, k + 2)],
                                            1, paste, collapse = " ") %in% apply(f1, 1, paste, collapse = " "))] <- 1
    
  }, error = function(e) {
    stop(simpleError(paste("An error has occurred and execution needs to stop. Please check the input data.",e)))
  })
  
  if (Progress) message("\n")
  
  # Check the number of the unmatched codes.
  if (length(which(as.vector(correspondenceAB$Unmatched) == 1))/nrow(correspondenceAB) <
      MismatchTolerance) {
    
    tryCatch({
      
      # The following if statement is applied if there are any
      # supplementary information for the classification A, in order to
      # be adjusted next to the correspondence table A:B.
      if (ncol(RRR[[1]]) >= 2) {
        A1 <- RRR[[1]][match(correspondenceAB[, 1], unlist(RRR[[1]][, 1])),
                       2:ncol(RRR[[1]])]
        A1[is.na(A1)] <- ""
        A1 <- matrix(unlist(A1), ncol = length(2:ncol(RRR[[1]])))
        colnames(A1) <- paste(paste(colnames(RRR[[1]])[1]), colnames(RRR[[1]])[2:ncol(RRR[[1]])],
                              sep = "_")
        correspondenceAB <- cbind(correspondenceAB, A1)
      }
      
      # The following for loop is applied for the classfications C1, C2,
      # ..., Ck.
      for (i1 in c(2:(((length(RRR) + 1)/2) - 1))) {
        
        # The if statement is applied if there are any supplementary
        # information for the classfications C1, C2, ..., Ck, in order
        # to be adjusted next to the correspondence table A:B.
        if (ncol(RRR[[i1]]) >= 2) {
          A1 <- RRR[[i1]][match(correspondenceAB[, i1], unlist(RRR[[i1]][,
                                                                         1])), 2:ncol(RRR[[i1]])]
          A1[is.na(A1)] <- ""
          A1 <- matrix(unlist(A1), ncol = length(2:ncol(RRR[[i1]])))
          colnames(A1) <- paste(paste(colnames(RRR[[i1]])[1]), colnames(RRR[[i1]])[2:ncol(RRR[[i1]])],
                                sep = "_")
          correspondenceAB <- cbind(correspondenceAB, A1)
        }
        
      }
      
      # The following if statement is applied if there are any
      # supplementary information for the classification B, in order to
      # be adjusted next to the correspondence table A:B.
      if (ncol(RRR[[(length(RRR) + 1)/2]]) >= 2) {
        A1 <- RRR[[(length(RRR) + 1)/2]][match(correspondenceAB[, (length(RRR) +
                                                                     1)/2], unlist(RRR[[(length(RRR) + 1)/2]][, 1])), 2:ncol(RRR[[(length(RRR) +
                                                                                                                                     1)/2]])]
        A1[is.na(A1)] <- ""
        A1 <- matrix(unlist(A1), ncol = length(2:ncol(RRR[[(length(RRR) +
                                                              1)/2]])))
        colnames(A1) <- paste(paste(colnames(RRR[[k + 2]])[1]), colnames(RRR[[(length(RRR) +
                                                                                 1)/2]])[2:ncol(RRR[[(length(RRR) + 1)/2]])], sep = "_")
        correspondenceAB <- cbind(correspondenceAB, A1)
      }
      
      # Find which .csv files are the correspondence tables.
      Tail <- utils::tail(c(1:length(RRR)), (length(RRR) - 1)/2)
      
      # The following if statement is applied if there are any
      # supplementary information for the correspondence table A:C1, in
      # order to be adjusted next to the correspondence table A:B.
      if (ncol(RRR[[Tail[1]]]) >= 3) {
        A1 <- RRR[[Tail[1]]][match(data.frame(t(correspondenceAB[, 1:2])),
                                   data.frame(t(RRR[[Tail[1]]][, 1:2]))), 3:ncol(RRR[[Tail[1]]])]
        A1[is.na(A1)] <- ""
        A1 <- matrix(unlist(A1), ncol = length(3:ncol(RRR[[Tail[1]]])))
        colnames(A1) <- paste(paste(colnames(RRR[[Tail[1]]])[1]), colnames(RRR[[Tail[1]]])[3:ncol(RRR[[Tail[1]]])],
                              sep = "_")
        correspondenceAB <- cbind(correspondenceAB, A1)
      }
      
      # The following if statement is applied if there are any
      # supplementary information for the correspondence tables (C1:C2 -
      # C2:C3), (C2:C3 - C3:C4), ..., (C(k-2):C(k-1) - C(k-1):Ck), in
      # order to be adjusted next to the correspondence table A:B.
      if (length(Tail) >= 3) {
        for (i2 in 2:(length(Tail) - 1)) {
          if (ncol(RRR[[Tail[i2]]]) >= 3) {
            A1 <- RRR[[Tail[i2]]][match(data.frame(t(correspondenceAB[, c(i2,
                                                                          i2 + 1)])), data.frame(t(RRR[[Tail[i2]]][, 1:2]))), 3:ncol(RRR[[Tail[i2]]])]
            A1[is.na(A1)] <- ""
            A1 <- matrix(unlist(A1), ncol = length(3:ncol(RRR[[Tail[i2]]])))
            colnames(A1) <- paste(paste(colnames(RRR[[Tail[i2]]])[1]), colnames(RRR[[Tail[i2]]])[3:ncol(RRR[[Tail[i2]]])],
                                  sep = "_")
            correspondenceAB <- cbind(correspondenceAB, A1)
          }
        }
      }
      
      # The following if statement is applied if there are any
      # supplementary information for the correspondence table B:Ck, in
      # order to be adjusted next to the correspondence table A:B.
      if (ncol(RRR[[Tail[length(Tail)]]]) >= 3) {
        A1 <- RRR[[Tail[length(Tail)]]][match(data.frame(t(correspondenceAB[,
                                                                            c(((length(RRR) + 1)/2) - 1, (length(RRR) + 1)/2)])), data.frame(t(RRR[[Tail[length(Tail)]]][,
                                                                                                                                                                         c(2, 1)]))), 3:ncol(RRR[[Tail[length(Tail)]]])]
        A1[is.na(A1)] <- ""
        A1 <- matrix(unlist(A1), ncol = length(3:ncol(RRR[[Tail[length(Tail)]]])))
        colnames(A1) <- paste(paste(colnames(RRR[[Tail[length(Tail)]]])[1]),
                              colnames(RRR[[Tail[length(Tail)]]])[3:ncol(RRR[[Tail[length(Tail)]]])],
                              sep = "_")
        correspondenceAB <- cbind(correspondenceAB, A1)
      }
    }, error = function(e) {
      stop(simpleError(paste("An error has occurred and execution needs to stop. Please check the input data.",e)))
    })
    
  } else {
    # Error message in case the percentage of unmatched codes between A and
    # B is larger than the desired threshold.
    stop("Too many codes in either of classifications A and B cannot be mapped to any code in the other one.\n",
         round(length(which(as.vector(correspondenceAB$Unmatched) == 1))/nrow(correspondenceAB)*100,2),"% is unmatched which exceeds the mismatch tolerance of ", MismatchTolerance)
  }
  
  
  tryCatch({
    
    # The final correspondence table A:B is sorted, firstly, based on
    # classification A, and then, based on classification B.
    correspondenceAB <- correspondenceAB[order(correspondenceAB[, 1], correspondenceAB[,
                                                                                       (length(RRR) + 1)/2], decreasing = FALSE), ]
    
    
    # Redundancy_trim parameter (MP)
    # Find the columns which are related to linking datasets which values need to be recorded as "Multiple"
    ## 2 + n_link_data*2 + 1 = n_data
    if (is.list(Tables)){
      num_link = (length(inputs_name) - 3)/2
      
    }else {
      num_link = (length(test.names) - 3)/2
      
    }
    
    col_multiple = numeric(0)
    for (nl in 1:num_link){
      matches <- grep(colnames(correspondenceAB)[1 + nl], colnames(correspondenceAB), value = TRUE)
      col_multiple <- unique(c(col_multiple, matches))
      
    }
    max_col = num_link + 2
    
    #Do redundancy trim only if redundancy is there (MP)
    if (length(which(correspondenceAB$Redundancy == 1)) != 0){
      
      # Find unique combination of A and B and identify them with a number
      uniqueAB = unique(correspondenceAB[which(correspondenceAB$Redundancy == 1),c(1, max_col)])
      uniqueAB$id_to_use = 1:nrow(uniqueAB)
      
      correspondenceAB = merge(correspondenceAB, uniqueAB, by = colnames(correspondenceAB)[c(1,max_col)], all.x = TRUE)[, union(names(correspondenceAB), names(uniqueAB))]
      matches <- grep("id_to_use", colnames(correspondenceAB), value = TRUE)
      if (length(matches) == 1) {
        col_link <- matches
      } else {
        stop("The column 'id_to_use' is not uniquely identified in correspondenceAB.")
      }
      
      ### new but probably slower
      if (Redundancy_trim == TRUE){
        x_temp = split(correspondenceAB[which(correspondenceAB$Redundancy == 1), col_multiple], correspondenceAB[which(correspondenceAB$Redundancy == 1), col_link])
        
        for (i in 1:nrow(uniqueAB)){
          multiple_values = apply(x_temp[[i]], 2, function(x) length(unique(x)))
          x_change = which(multiple_values != 1)
          #replace with multiple
          col_multiple_temp = c(col_multiple)[x_change]
          correspondenceAB[which(correspondenceAB$Redundancy == 1 & correspondenceAB[,col_link] == i), unique(col_multiple_temp)] = "Multiple"
        }
        
        correspondenceAB = correspondenceAB[,!names(correspondenceAB) %in% "id_to_use"]
        
        ### old but probably faster
        ##replace with multiple
        ##replace with multiple
        #correspondenceAB[which(correspondenceAB$Redundancy == 1), unique(col_multiple)] = "Multiple"
        
        #eliminate duplicates
        dup = as.numeric(duplicated(correspondenceAB[,1:max_col]))
        correspondenceAB = correspondenceAB[which(dup == 0), ]
        
      }
    } else {
      
      correspondenceAB = correspondenceAB
    }
    
    if (Redundancy_trim==FALSE){
      #add a redundancy keep flag to indicate which row will be kept
      dup = as.numeric(duplicated(correspondenceAB[,c(1,max_col)]))
      red_col = which(colnames(correspondenceAB) == "Redundancy")
      correspondenceAB$Redundancy_keep = rep(0, nrow(correspondenceAB))
      correspondenceAB$Redundancy_keep[which(dup == "0" & correspondenceAB$Redundancy == "1")] = 1
      correspondenceAB = correspondenceAB[,c(1:red_col, ncol(correspondenceAB), (red_col+1):(ncol(correspondenceAB)-1))]
      correspondenceAB = correspondenceAB[,!names(correspondenceAB) %in% "id_to_use"]
    }
    
    
    # Create a data frame that contains the names of the classifications.
    CsvNames <- data.frame(matrix(0, k + 2, 1))
    
    CsvNames[1, 1] <- paste("A:", colnames(correspondenceAB)[1], sep = " ")
    
    CsvNames[k + 2, 1] <- paste("B:", colnames(correspondenceAB)[k + 2], sep = " ")
    
    for (i3 in seq(k) + 1) {
      CsvNames[i3, 1] <- paste(paste("C", i3 - 1, ":", sep = ""), colnames(correspondenceAB)[i3],
                               sep = " ")
    }
    
    CsvNames <- data.frame(CsvNames)
    
    colnames(CsvNames) <- paste("Classification:", "Name", sep = " ")
    
    # Create a data frame that contains the final correspondence table
    # (final desired table).
    Final <- apply(correspondenceAB, 2, function(x) {
      gsub(" ", " ", x)
    })
    
    
    if (is.null(dim(Final))) {
      Final <- t(data.frame(Final))
      rownames(Final) <- 1
    }
    
    
  }, error = function(e) {
    stop(simpleError(paste("An error has occurred and execution needs to stop. Please check the input data. \n Deatils line 1895:\n",e)))
  })
  
  # The final list that contains the final correspondence table (final
  # desired table) as a data frame as well as the names of classifications as
  # a data frame.
  tryCatch({
    
    FinalResults <- list()
    FinalResults[[1]] <- data.frame(Final, check.names = FALSE, row.names = NULL)
    FinalResults[[2]] <- CsvNames
    names(FinalResults) <- c("newCorrespondenceTable", "classificationNames")
    
    # newCorrespondenceTable function returns the final correspondence
    # table A:B, that contains the pivot classifications C1, C2, ..., Ck,
    # as well as any supplementary information about the classification
    # tables A, C1, C2, ..., Ck, B, and the correspondence tables A:C1,
    # (C1:C2 - C2:C3), (C2:C3 - C3:C4), ..., (C(k-2):C(k-1) - C(k-1):Ck),
    # B:Ck.
    
    return(FinalResults)
  }, error = function(e) {
    stop(simpleError(paste("An error has occurred and execution needs to stop. Please check the input data.",e)))
  })
  
}

read_test_csv <- function(fname) {
  utils::read.csv(
    system.file("extdata/test", fname, package = "correspondenceTables"),
    sep = ",", header = TRUE, check.names = FALSE,
    colClasses = c("character"), encoding = "UTF-8",
    stringsAsFactors = FALSE, fileEncoding = "UTF-8-BOM"
  )
}

# Helper: materialize names.csv by expanding fixture file names to absolute paths
# Writes to tempdir() to avoid modifying the installed package directory.
materialize_names_csv <- function(src_csv_name, out_csv_name = "names.csv") {
  templ <- system.file("extdata/test", src_csv_name, package = "correspondenceTables")
  
  A <- read.csv(templ, header = FALSE, sep = ",", stringsAsFactors = FALSE)
  for (i in seq_len(nrow(A))) {
    for (j in seq_len(ncol(A))) {
      if (!is.na(A[i, j]) && A[i, j] != "") {
        A[i, j] <- system.file("extdata/test", A[i, j], package = "correspondenceTables")
      }
    }
  }
  out <- file.path(tempdir(), out_csv_name)
  write.table(A, file = out, row.names = FALSE, col.names = FALSE, sep = ",")
  invisible(out)
}

cleanup_names_csv <- function() {
  p <- file.path(tempdir(), "names.csv")
  if (file.exists(p)) file.remove(p)
}

############################################################
# TEST N1 — Basic pipeline, no trimming
############################################################
{
  on.exit(cleanup_names_csv(), add = TRUE)
  
  names_path <- materialize_names_csv("names12.csv", "names.csv")
  
  TEST <- newCorrespondenceTable(
    Tables = names_path,
    Reference = "none",
    MismatchTolerance = 0.96,
    Redundancy_trim = FALSE,
    Progress = FALSE
  ) 
  TEST_trim <- newCorrespondenceTable(
    Tables = names_path,
    Reference = "none",
    MismatchTolerance = 0.96,
    Redundancy_trim = TRUE,
    Progress = FALSE
  ) 
  
  
  
  ABC      <- read_test_csv("ABC.csv")
  ABC_Trim <- read_test_csv("ABC_trim.csv")
  
  expect_equal(TEST[[1]][1:5],      ABC)
  expect_equal(TEST_trim[[1]][1:4], ABC_Trim)
}

############################################################
# TEST N2 — Legend merge check on “new” fixtures
############################################################
{
  on.exit(cleanup_names_csv(), add = TRUE)
  
  names_path <- materialize_names_csv("names2.csv", "names.csv")
  
  TEST2 <- newCorrespondenceTable(
    Tables = names_path,
    Reference = "none",
    MismatchTolerance = 0.96,
    Redundancy_trim = FALSE,
    Progress = FALSE
  )
  
  
  TEST2[[1]]$sort <- seq_len(nrow(TEST2[[1]]))
  TEST2_T <- read_test_csv("nomatch_test_new.csv")
  TEST2_T$sort <- seq_len(nrow(TEST2_T))
  
  legend <- read_test_csv("legend_nomatch_new.csv")
  
  x <- as.data.frame(TEST2[[1]])
  y <- as.data.frame(legend)
  
  merged <- merge(x, y, by = c("Unmatched","NoMatchFromA","NoMatchFromB"), all.x = TRUE)
  merged  <- merged[order(as.numeric(merged$sort)), ]
  TEST2_T <- TEST2_T[order(as.numeric(TEST2_T$sort)), ]
  
  ## Preserve original index references
  expect_equal(merged[, 15], TEST2_T[, 13])
}

# Deprecated alias for classificationList

# `classEndpoint()` is deprecated.  
# Please use [classificationList()] instead.

# classEndpoint("ALL")   # deprecated
# classificationList("ALL")  # use this instead

classEndpoint <- function(endpoint = "ALL") {
  .Deprecated("classificationList")
  classificationList(endpoint = endpoint)
}


# classEndpoint = function(endpoint) {
# 
#   ### Datasets in CELLAR
#   endpoint_cellar = "http://publications.europa.eu/webapi/rdf/sparql"
#   
#   SPARQL.query_cellar = paste0("
#   SELECT DISTINCT ?s ?Title
#   WHERE { ?s a skos:ConceptScheme ;
#           skos:prefLabel ?Title ;
#           ?p <http://publications.europa.eu/resource/authority/corporate-body/ESTAT> 
#            FILTER (LANG(?Title) = 'en')}
#   ORDER BY ?Title
#   ")
# 
#   response = httr::POST(url = endpoint_cellar, config = httr::accept("text/csv"), body = list(query = SPARQL.query_cellar), encode = "form")
#   data_cellar = read.csv(text=httr::content(response, "text"), sep= ",")  
# 
#   ## add prefix name
#   str_dt = t(sapply(data_cellar[,1], function(x) unlist(strsplit(as.character(x), "/+"))))
#   uri = paste0(str_dt[,1],"/", "/", str_dt[,2],"/",str_dt[,3],"/",str_dt[,4] )
#   prefix = str_dt[,4]
#   prefix = gsub("\\.","",prefix)
#   #key = str_dt[,4]
#   conceptscheme = str_dt[,5]
#   title = data_cellar[,2]
#   data_cellar = cbind(prefix, conceptscheme, uri, title)
#   rownames(data_cellar) = 1:nrow(data_cellar)
#   colnames(data_cellar) = c("Prefix", "ConceptScheme", "URI", "Title")
#   
#   ### Datasets in FAO
#   endpoint_fao = "https://stats.fao.org/caliper/sparql/AllVocs"
#   SPARQL.query_fao = paste0("
#       PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
#       SELECT DISTINCT ?classification ?label
# 
#       WHERE {
#           ?classification a skos:ConceptScheme .
#           ?classification skos:prefLabel ?label .
#           FILTER(regex(?label, 'classification', 'i'))
#       }
#     ORDER BY ?label
#   ")
#   
#   response = httr::POST(url = endpoint_fao, config = httr::accept("text/csv"), body = list(query = SPARQL.query_fao), encode = "form")
#   data_fao = read.csv(text=httr::content(response, "text"), sep= ",")                                 
#   
#   ## add prefix name
#   str_dt = t(sapply(data_fao[,1], function(x) unlist(strsplit(as.character(x), "/+"))))
#   prefix = paste0(str_dt[,4], str_dt[,5])
#   prefix = gsub("\\.","",prefix)
#   uri = paste0(str_dt[,1], "/", "/", str_dt[,2], "/", str_dt[,3], "/", str_dt[,4], "/", str_dt[,5])
#   #class = prefix
#   ConceptScheme = str_dt[,6]
#   data_fao = cbind(prefix, ConceptScheme, uri, data_fao[,2])
#   rownames(data_fao) = 1:nrow(data_fao)
#   colnames(data_fao) = c("Prefix", "ConceptScheme", "URI", "Title")
# 
#   if (endpoint == "ALL") {
#     data = list("CELLAR" = data_cellar, "FAO" = data_fao)
#   }
#   
#   if (endpoint == "CELLAR") {
#     data = list("CELLAR" = data_cellar)
#   }
#   
#   if (endpoint == "FAO") {
#     data = list("FAO" = data_fao)
#   }  
#   
#   return(data)
# 
# }
# 

---
title: "Sample datasets with the correspondenceTables package"
output: rmarkdown::html_vignette

vignette: >
  %\VignetteIndexEntry{Sample datasets with the correspondenceTables package}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
library(correspondenceTables)
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
tmp_dir<-tempdir()
csv_files<-list.files(tmp_dir, pattern = ".csv")
if (length(csv_files)>0) unlink(csv_files)
```

This vignette provides information about applying the correspondenceTables package on the sample datasets included in it.

ATTENTION: please set as working directory a folder different than the folder in which the package has been installed.

LOCAL COPIES OF THE SAMPLE DATA

If users wish so, they can make copies of the sample datasets in a local folder of their choice. There are two ways of doing this:

1) Unpack into any folder of your choice the tar.gz file in which the package has arrived. All sample datasets may be found in the "inst/extdata" subfolder of this folder.

2) Copy sample datasets from the "extdata" subfolder of the folder in which the package has been installed in your PC's R library.

--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

ACCESSING SAMPLE DATASETS FROM WITHIN THE PACKAGE

Application of function updateCorrespondenceTable().

Case 1

Execute the following code in order to get the path of the required input files.

```{r}
A <- system.file("extdata", "CN2021.csv", package = "correspondenceTables")
AStar <- system.file("extdata", "CN2022.csv", package = "correspondenceTables")
B <- system.file("extdata", "CPA21.csv", package = "correspondenceTables")
AB <- system.file("extdata", "CN2021_CPA21.csv", package = "correspondenceTables")
AAStar <- system.file("extdata", "CN2021_CN2022.csv", package = "correspondenceTables")
```

Execute the following code line to apply function updateCorrespondenceTable() on these data.

```{r}
UPC <- updateCorrespondenceTable(A, B, AStar, AB, AAStar, file.path(tmp_dir,"updateCorrespondenceTableCase1.csv"),
                                 "B", 0.4, 0.4)
print(UPC[[1]][1:10, 1:7])
print(UPC[[2]])
```

Case 2

Execute the following code in order to get the path of the required input files.

```{r}
A <- system.file("extdata", "CN2021.csv", package = "correspondenceTables")
AStar <- system.file("extdata", "CN2022.csv", package = "correspondenceTables")
B <- system.file("extdata", "PRODCOM2021.csv", package = "correspondenceTables")
AB <- system.file("extdata", "CN2021_PRODCOM2021.csv", package = "correspondenceTables")
AAStar <- system.file("extdata", "CN2021_CN2022.csv", package = "correspondenceTables")
```

Execute the following code line to apply function updateCorrespondenceTable() on these data.

```{r, results = "hide"}
UPC <- updateCorrespondenceTable(A, B, AStar, AB, AAStar, file.path(tmp_dir,"updateCorrespondenceTableCase2.csv"), "A", 0.4, 0.3)
```

Case 3

Execute the following code in order to get the path of the required input files.

```{r}
A <- system.file("extdata", "NAICS2017.csv", package = "correspondenceTables")
AStar <- system.file("extdata", "NAICS2022.csv", package = "correspondenceTables")
B <- system.file("extdata", "NACE.csv", package = "correspondenceTables")
AB <- system.file("extdata", "NAICS2017_NACE.csv", package = "correspondenceTables")
AAStar <- system.file("extdata", "NAICS2017_NAICS2022.csv", package = "correspondenceTables")
```

Execute the following code line to apply function updateCorrespondenceTable() on these data.

```{r, results = "hide"}
UPC <- updateCorrespondenceTable(A, B, AStar, AB, AAStar, file.path(tmp_dir,"updateCorrespondenceTableCase3.csv"), "none", 0.5, 0.3)
```

Case 4

Execute the following code in order to get the path of the required input files.

```{r}
A <- system.file("extdata", "CN2021.csv", package = "correspondenceTables")
AStar <- system.file("extdata", "CN2022.csv", package = "correspondenceTables")
B <- system.file("extdata", "NST2007.csv", package = "correspondenceTables")
AB <- system.file("extdata", "CN2021_NST2007.csv", package = "correspondenceTables")
AAStar <- system.file("extdata", "CN2021_CN2022.csv", package = "correspondenceTables")
```

Execute the following code line to apply function updateCorrespondenceTable() on these data.

```{r, results = "hide"}
UPC <- updateCorrespondenceTable(A, B, AStar, AB, AAStar, file.path(tmp_dir,"updateCorrespondenceTableCase4.csv"), "B", 0.4, 0.3)
```

Case 5

Execute the following code in order to get the path of the required input files.

```{r}
A <- system.file("extdata", "CN2021.csv", package = "correspondenceTables")
AStar <- system.file("extdata", "CN2022.csv", package = "correspondenceTables")
B <- system.file("extdata", "SITC4.csv", package = "correspondenceTables")
AB <- system.file("extdata", "CN2021_SITC4.csv", package = "correspondenceTables")
AAStar <- system.file("extdata", "CN2021_CN2022.csv", package = "correspondenceTables")
```

Execute the following code line to apply function updateCorrespondenceTable() on these data.

```{r, results = "hide"}
UPC <- updateCorrespondenceTable(A, B, AStar, AB, AAStar, file.path(tmp_dir,"updateCorrespondenceTableCase5.csv"), "B", 0.3, 0.7)
```

Case 6

Execute the following code in order to get the path of the required input files.

```{r}
A <- system.file("extdata", "CN2021.csv", package = "correspondenceTables")
AStar <- system.file("extdata", "CN2022.csv", package = "correspondenceTables")
B <- system.file("extdata", "BEC4.csv", package = "correspondenceTables")
AB <- system.file("extdata", "CN2021_BEC4.csv", package = "correspondenceTables")
AAStar <- system.file("extdata", "CN2021_CN2022.csv", package = "correspondenceTables")
```

Execute the following code line to apply function updateCorrespondenceTable() on these data.

```{r, results = "hide"}
UPC <- updateCorrespondenceTable(A, B, AStar, AB, AAStar, file.path(tmp_dir,"updateCorrespondenceTableCase6.csv"), "B", 0.3, 0.6)
```


--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
Application of function newCorrespondenceTable().

The function fullPath is used in all cases in order to get the path of the required input files.

```{r}
fullPath <- function(CSVraw, CSVappended){
  NamesCsv <- system.file("extdata", CSVraw, package = "correspondenceTables")
  A <- read.csv(NamesCsv, header = FALSE, sep = ",")
   for (i in 1:nrow(A)) {
    for (j in 1:ncol(A)) {
      if (A[i,j]!="") {
        A[i, j] <- system.file("extdata", A[i, j], package = "correspondenceTables")
      }}}
  write.table(x = A, file = file.path(tmp_dir,CSVappended), row.names = FALSE, col.names = FALSE, sep = ",")
  return(A)
}
```

Case 1

```{r, results = "hide"}
fullPath("names1.csv", "names.csv")
```

Execute the following code to apply function newCorrespondenceTable() on these data.

```{r, results = "hide"}
system.time(NCT <- newCorrespondenceTable(file.path(tmp_dir,"names.csv"), file.path(tmp_dir,"newCorrespondenceTableCase1.csv"), "A", 0.5))
```

```{r}
print(NCT[[1]][1:10, 1:6])
print(NCT[[2]])
```

Case 2

```{r, results = "hide"}
fullPath("names2.csv", "names.csv")
```

Execute the following code to apply function newCorrespondenceTable() on these data.

```{r, results = "hide"}
system.time(NCT <- newCorrespondenceTable(file.path(tmp_dir,"names.csv"), file.path(tmp_dir,"newCorrespondenceTableCase2.csv"), "B", 0.5))
```

Case 3

```{r, results = "hide"}
fullPath("names3.csv", "names.csv")
```

Execute the following code to apply function newCorrespondenceTable() on these data.

```{r, results = "hide"}
system.time(NCT <- newCorrespondenceTable(file.path(tmp_dir,"names.csv"), file.path(tmp_dir,"newCorrespondenceTableCase3.csv"), "B", 0.5))
```

Case 4

```{r, results = "hide"}
fullPath("names4.csv", "names.csv")
```

Execute the following code to apply function newCorrespondenceTable() on these data.

```{r, results = "hide"}
system.time(NCT <- newCorrespondenceTable(file.path(tmp_dir,"names.csv"), file.path(tmp_dir,"newCorrespondenceTableCase4.csv"), "none", 0.96))
```

```{r message=TRUE, warning=TRUE, include=FALSE}
csv_files<-list.files(tmp_dir, pattern = ".csv")
if (length(csv_files)>0) unlink(csv_files)
```

---
title: "Combining correspondence tables"
output:
  rmarkdown::html_vignette:
    toc: TRUE

vignette: >
  %\VignetteIndexEntry{Combining correspondence tables}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r include=FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
options(width = 300)
```



```{r, echo=FALSE, results="asis"}
cat("
<style>
table {
  width: 100% !important;
}
table caption {
  text-align: center;
}
</style>
")
```

<div style="text-align: justify;">

## Overview

A correspondence table serves as a translation between two statistical classifications. When a correspondence table between two classifications does not yet exist, but both are linked to one or more intermediate classifications through existing correspondence tables, a new correspondence table can be generated automatically.

For the general case, where classifications $A$ and $B$ are indirectly linked via one or more intermediate classifications $C_1, \dots ,C_k$, the `newCorrespondenceTable()` function can automatically generate a new correspondence table.

A special case occurs when a classification $A$ is updated to a new version $A^*$ (with the correspondence table $A:A^*$ assumed to have been created as part of this update), and a correspondence table $A:B$ between the old version of $A$ and another classification of interest $B$ already exists.

Here, the `updateCorrespondenceTable()` function can be used to automatically generate the new correspondence table $A^*:B$. (The `newCorrespondenceTable()` function could also be applied to achieve this, but the `updateCorrespondenceTable()` function takes into consideration the fact that $A$ and $A^*$ are two versions of the same classification, and is therefore recommended for this updating scenario.

```{r, echo = F}
library(correspondenceTables)

```

### Input

In the case of `newCorrespondenceTable()`, the number of intermediate classifications is variable.  
For this reason, the function accepts a flexible, matrix-like input structure that represents the relationships between classifications and their correspondence tables.

The input must be provided **either**:

- as a **square CSV file** that specifies the input structure by listing the file paths of the classification tables (on the diagonal) and the correspondence tables (on the off-diagonal), or  
- as a **square two-level list of data frames**.

In both cases, the diagonal elements of the structure correspond to classification tables (e.g. $A$, $B$, $C$), while the off-diagonal elements represent the correspondence tables linking consecutive classifications (e.g. $A:B$, $B:C$).

To generate a correspondence table between classifications $A$ and $C$ from the correspondence tables $A:B$ and $B:C$, the function requires a matrix-like input structure with classifications on the diagonal and correspondence tables on the off-diagonal. Schematically, this structure can be represented as follows:

\[
\begin{bmatrix}
A & A\!:\!B & \\
  & B       & B\!:\!C \\
  &         & C
\end{bmatrix}
\]

This representation naturally extends to cases with multiple intermediate classifications.

The input for `updateCorrespondenceTable()` simply requires the classifications ($A, A^*$ and $B$) and correspondence tables ($A:B$ and $A:A^*$) as data frames.

### Output


As output, both `newCorrespondenceTable()` and `updateCorrespondenceTable()` return a list containing:

- the resulting correspondence table as a data frame, and  
- a data frame reporting the names of the classifications involved in the correspondence



### Helper for the examples

When `newCorrespondenceTable()` is used with a CSV-based input structure, the CSV file that specifies the input layout must contain full file paths to the referenced CSV files, rather than file names alone. Accordingly, in the sample input, the file names appearing in the CSV table cells must be prefixed with their full path.

To streamline this task, the utility function `fullPath`, defined below, is used in all the following examples.

```{r}

tmp_dir <- tempdir()

fullPath <- function(CSVraw, CSVappended){
  NamesCsv <- system.file("extdata/test", CSVraw, package = "correspondenceTables")
  A <- read.csv(NamesCsv, header = FALSE, sep = ",")
   for (i in 1:nrow(A)) {
    for (j in 1:ncol(A)) {
      if (A[i,j]!="") {
        A[i, j] <- system.file("extdata/test", A[i, j], package = "correspondenceTables")
      }}}
  write.table(x = A, file = file.path(tmp_dir,CSVappended), row.names = FALSE, col.names = FALSE, sep = ",")
  return(A)
}
```


```{r, echo=FALSE}

files_to_clean <- setdiff(
  list.files(tmp_dir, pattern = "\\.csv$", full.names = TRUE),
  file.path(tmp_dir, "names.csv")
)

if (length(files_to_clean) > 0) unlink(files_to_clean)
```

</div>

## Creating correspondence tables: general case using `newCorrespondenceTable()`

### Example 1: ISIC Rev. 4 : CPA Ver. 2.1 (via CPC Ver. 2.1)

<div style="text-align: justify;">

```{r, results = "hide"}
fullPath("names1.csv", "names.csv")
```

Execute the following code to apply function `newCorrespondenceTable()` and generate the correspondence table linking ISIC Rev. 4 (classification A) to CPA 2.1 (classification B) through the intermediate classification CPC 2.1. When no trimming is executed (`Redundancy_trim = FALSE`), redundant records are shown, together with the redundancy flag.


```{r}
NCT <- newCorrespondenceTable(
        Tables = file.path(tmp_dir, "names.csv"),
        Reference = "A",
        MismatchTolerance = 0.5,
        Redundancy_trim = FALSE,
        Progress = FALSE
)

knitr::kable(
  (NCT[[1]][3748:3753, 1:9]),
  caption = "ISIC Rev. 4 to CPA Ver. 2.1 (via CPC Ver. 2.1): Subsample of the new Correspondence Table",
  align = "c"
)
```

The table above represents a subset of the correspondence table generated in this example. Each row represents a candidate correspondence between an ISIC code and a CPA code, possibly mediated by one or more intermediate classifications.

Here, the ISIC code `1030` is linked to several CPA codes:

- The rows linking `1030` to `10.39.23` and `10.39.24` are **unique and unambiguous**.  
  These rows have `Redundancy = 0`, `Unmatched = 0`, and no review or mismatch flags set.

- The CPA code `10.39.25` appears **multiple times** in combination with the same ISIC code `1030`, via different CPC codes.  
  These rows are therefore flagged with `Redundancy = 1`.

When `Redundancy_trim = FALSE`, all redundant rows are retained and an additional column, `Redundancy_keep`, is included:

- `Redundancy_keep = 1` identifies the record that would be kept if redundancy trimming were applied.
- Rows with `Redundancy_keep = 0` represent redundant alternatives.

All rows in this example have `Unmatched = 0`, indicating that each ISIC code is matched to at least one CPA code and vice versa.  
Similarly, `NoMatchFromA = 0` and `NoMatchFromB = 0` show that no codes from the original classification tables are missing from the correspondence tables involved in the construction.

Finally, the `Review` flag is equal to `0` for all rows, indicating that given the selected reference classification, no hierarchical inconsistencies are detected.


```{r}
knitr::kable(
  head(NCT[[2]]),
  caption = "ISIC Rev. 4 to CPA Ver. 2.1 (via CPC Ver. 2.1): Names of the classifications involved",
  align = "c"
)

```

The table above is the second element generated with `newCorrespondenceTable`, which simply is a data frame containing the names of all classifications involved.

</div>

### Example 2: NACE Rev. 2 : SITC 4 (via CPA Ver. 2.1 and CN 2022), many-to-many case.

<div style="text-align: justify;">

```{r, results = "hide"}
fullPath("names4.csv", "names.csv")
```

Execute the following code to apply function `newCorrespondenceTable()` and generate the correspondence table linking NACE Rev. 2 (classification A) to SITC 4 (classification B) through the intermediate classifications CPA Ver. 2.1 and CN 2022. Given the option `Redundancy_trim = TRUE`, when there are redundant records, these are removed and kept exactly one record for each unique combination.

```{r}
NCT <- newCorrespondenceTable(
        Tables = file.path(tmp_dir, "names.csv"),
        Reference = "none",
        MismatchTolerance = 0.96,
        Redundancy_trim = TRUE,
        Progress = FALSE
      )



knitr::kable(
  head(NCT[[1]][5442:5450, 1:8]),
  caption = "NACE Rev. 2 : SITC 4 (via CPA Ver. 2.1 and CN 2022): Subsample of the new Correspondence Table",
  align = "c"
)

```

Also in this case, the table above represents a subset of the correspondence table generated in this example. Each row corresponds to a correspondence between a NACE code and a SITC code, possibly mediated by multiple intermediate classifications.

In this example, the NACE code `28.41` is mapped to several SITC codes:

- The first four rows represent unique and unambiguous correspondences, where specific CPA and CN codes are associated with specific SITC codes.  
  These rows have `Redundancy = 0` and `Unmatched = 0`, indicating clear one-to-one mappings across all classifications involved.

- The last two rows are flagged with `Redundancy = 1`.  
  In these cases, multiple intermediate codes (in CPA and/or CN) contribute to the same NACE–SITC mapping. As a result, the corresponding intermediate classification values are reported as `"Multiple"`.

All rows have `Unmatched = 0`, indicating that each correspondence links a valid NACE code to a valid SITC code.  
Additionally, `NoMatchFromA = 0` and `NoMatchFromB = 0` for all rows confirm that no classification codes are missing from the correspondence tables used to construct the result.


```{r}
knitr::kable(
  head(NCT[[2]]),
  caption = "NACE Rev. 2 : SITC 4 (via CPA Ver. 2.1 and CN 2022): Names of the classifications involved",
  align = "c"
)
```

```{r message=TRUE, warning=TRUE, include=FALSE}
csv_files<-list.files(tmp_dir, pattern = ".csv")
if (length(csv_files)>0) unlink(csv_files)
```
The table above corresponds to the second element returned by `newCorrespondenceTable` and is a data frame containing the names of all the classifications involved in the process.

</div>


## Updating correspondence tables using `updateCorrespondenceTable()`

### Example 3: Updating CN 2021 : CPA Ver. 2.1  (triggered by CN update)

<div style="text-align: justify;">

Execute the following code in order to get the path of the required input files.

```{r}
A <- read.csv(
  system.file("extdata/test", "CN2021.csv", package = "correspondenceTables"),
  colClasses = "character"
)

AStar <- read.csv(
  system.file("extdata/test", "CN2022.csv", package = "correspondenceTables"),
  colClasses = "character"
)

B <- read.csv(
  system.file("extdata/test", "CPA21.csv", package = "correspondenceTables"),
  colClasses = "character"
)

AB <- read.csv(
  system.file("extdata/test", "CN2021_CPA21.csv", package = "correspondenceTables"),
  colClasses = "character"
)

AAStar <- read.csv(
  system.file("extdata/test", "CN2021_CN2022.csv", package = "correspondenceTables"),
  colClasses = "character"
)

```


Execute the following code line to apply function `updateCorrespondenceTable()` and generate the updated correspondence table. In this case the classification CN 2021 (A) has been updated to CN 2022 (A\*), and the correspondence to CPA 2.1 (B) is revised accordingly. Given the option `Redundancy_trim = TRUE`, when there are redundant records, these are removed and kept exactly one record for each unique combination.

```{r}
UPC <- updateCorrespondenceTable(
  A = A,
  B = B,
  AStar = AStar,
  AB = AB,
  AAStar = AAStar,
  Reference = "B",
  MismatchToleranceB = 0.4,
  MismatchToleranceAStar = 0.4,
  Redundancy_trim = TRUE
)


knitr::kable(
  (UPC[[1]][7950:7955, 1:11]),
  caption = "Updating CN 2021 : CPA Ver. 2.1  (triggered by CN update): Subsample of the new CorrespondenceTable",
  align = "c"
)

```
The table above represents a subset of the correspondence table generated in this example. 
Each row links a CN 2022 code to a CPA 2.1 code and reflects changes from the previous version.

In this example:

- The first three rows are flagged with `CodeChange = 1`, indicating that the original CN 2021 codes are associated with updated CN 2022 codes in a way that differs from the previous mapping.  
  These rows also have `LabelChange = 1`, meaning that the labels of the corresponding CN codes have changed between versions.

- Rows where `Review = 1` indicate potential hierarchical inconsistencies with respect to the selected reference classification, and therefore require manual inspection.

- The remaining rows have `CodeChange = 0` and `LabelChange = 0`, showing that both the code and its label remain unchanged between CN 2021 and CN 2022 for the given correspondence to CPA 2.1.

All rows have `Redundancy = 0`, meaning that each CN 2022–CPA 2.1 combination appears only once in the updated correspondence table.  
Similarly, `NoMatchToAStar = 0` and `NoMatchToB = 0` indicate that each row contains valid codes for both CN 2022 and CPA 2.1.

Finally, the flags `NoMatchFromAStar = 0` and `NoMatchFromB = 0` for all rows confirm that every code appearing in the updated correspondence is consistently represented in both the updated classification table and the underlying concordance tables.
 

```{r}
knitr::kable(
  head(UPC[[2]]),
  caption = "Updating CN 2021 : CPA Ver. 2.1  (triggered by CN update): Names of the classifications involved",
  align = "c",
  col.names = "Classification: Name"
)

```
The table above is the second element generated with `updateCorrespondenceTable`, which simply is a data frame containing the names of all classifications involved.

</div>

### Example 4: Updating NAICS : NACE (triggered by NAICS update)

<div style="text-align: justify;">

Execute the following code in order to get the path of the required input files.

```{r}
A <- read.csv(
  system.file("extdata/test", "NAICS2017.csv", package = "correspondenceTables"),
  colClasses = "character"
)

AStar <- read.csv(
  system.file("extdata/test", "NAICS2022.csv", package = "correspondenceTables"),
  colClasses = "character"
)

B <- read.csv(
  system.file("extdata/test", "NACE.csv", package = "correspondenceTables"),
  colClasses = "character"
)

AB <- read.csv(
  system.file("extdata/test", "NAICS2017_NACE.csv", package = "correspondenceTables"),
  colClasses = "character"
)

AAStar <- read.csv(
  system.file("extdata/test", "NAICS2017_NAICS2022.csv", package = "correspondenceTables"),
  colClasses = "character"
)

```

Execute the following code line to apply function `updateCorrespondenceTable()` and generate the updated correspondence table. In this case the classification NAICS 2017 (A) has been updated to NAICS 2022 (A\*), and the correspondence to NACE Rev. 2 (B) is revised accordingly. Given the option `Redundancy_trim = TRUE`, when there are redundant records, these are removed and kept exactly one record for each unique combination.

```{r}

UPC3 <- updateCorrespondenceTable(
  A = A,
  B = B,
  AStar = AStar,
  AB = AB,
  AAStar = AAStar,
  Reference = "none",
  MismatchToleranceB = 0.5,
  MismatchToleranceAStar = 0.8,
  Redundancy_trim = TRUE
)


knitr::kable(
  head(UPC3[[1]][1208:1218, 1:10]),
  caption = "Updating NAICS : NACE (triggered by NAICS update): Subsample of the new Correspondence Table",
  align = "c"
)

```
The table above represents a subset of the correspondence table generated in this example. Each row represents a candidate correspondence between a NAICS 2022 code and a NACE Rev. 2 code, derived from the previous version of the classification (NAICS 2017).

In this example:

- The NAICS code `332313` is unchanged between NAICS 2017 and NAICS 2022, as indicated by `CodeChange = 0` for all rows.
  This shows that the classification update did not introduce any code-level changes for this activity.

- The same NAICS code `332313` is mapped to multiple NACE Rev. 2 codes (`25.11`, `25.29`, `25.30`, `28.22`, `28.91`, `30.11`), reflecting a one-to-many correspondence that already existed and remains valid after the update.

- All rows have `Redundancy = 0`, meaning that each NAICS 2022–NACE Rev. 2 combination appears only once in the updated correspondence table.

- The flags `NoMatchToAStar = 0` and `NoMatchToB = 0` indicate that every row contains valid and consistent codes for both the updated classification (NAICS 2022) and the target classification (NACE Rev. 2).

- Similarly, `NoMatchFromAStar = 0` and `NoMatchFromB = 0` confirm that all codes appearing in the updated correspondence are present in the respective classification tables and supported by the underlying concordance tables.

- Finally, `LabelChange = 0` for all rows shows that the labels associated with the NAICS codes are identical between the 2017 and 2022 versions.

```{r}

  knitr::kable(
  head(UPC3[[2]]),
  caption = "Updating NAICS : NACE (triggered by NAICS update): Names of the classifications involved",
  align = "c",
  col.names = "Classification: Name"
)
```

The table above corresponds to the second element returned by `updateCorrespondenceTable` and is a data frame containing the names of all relevant classifications.


</div>

---
title: "Analyse Correspondence Table"
output:
  rmarkdown::html_vignette:  
    toc: TRUE
description: "This vignette explains how to use the `analyseCorrespondenceTable()` function to explore and diagnose correspondence tables."
vignette: >
  %\VignetteIndexEntry{Analyse Correspondence Table}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---


```{r include=FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

<div style="text-align: justify;">
## Overview  
 
This vignette demonstrates how to extract key diagnostic information from correspondence tables using the `analyseCorrespondenceTable()` function, including:

- the inventory of codes involved on both sides of the correspondence;
- the presence of one‑to‑one, one‑to‑many, or many‑to‑many relationships;
- potential structural issues that may affect downstream use (aggregation, updates, or quality checks).
</div>

```{r}
library(correspondenceTables)  
```

```{r, echo=FALSE, results="asis"}
cat("
<style>
table {
  width: 100% !important;
}
table caption {
  text-align: center;
}
</style>
")
```

## Application of `analyseCorrespondenceTable`

### Example 1: Basic analysis of a correspondence table
In this first example, only a correspondence table is provided.
No explicit source or target classification tables are supplied.


```{r}
# Load example correspondence table
AB <- read.csv(
  system.file("extdata/test", "ExempleAnnexe.csv", package = "correspondenceTables"),
  stringsAsFactors = FALSE
)

```


The analysis is then performed using default settings:


```{r}
# Perform analysis
result <- analyseCorrespondenceTable(
  AB = AB,
  A  = NULL,
  B  = NULL,
  longestAcodeOnly = FALSE,
  longestBcodeOnly = FALSE
)
```

 
The function returns a list containing multiple outputs.
Two key components are displayed below.

The tables below illustrate the first rows of the outputs returned by the function:


```{r}
# Display results

knitr::kable(
  head(result$Inventory),
  caption = "Inventory results: Basic analysis of a correspondence table",
  align = "c"
)


knitr::kable(
  head(result$Analysis),
  caption = "Analysis results: Basic analysis of a correspondence table",
  align = "c"
)

```

 <div style="text-align: justify;">
**Interpretation of the results**

The `analyseCorrespondenceTable()` function returns a list of outputs that
summarise the structure of the correspondence table from complementary
perspectives. In this example, two components are displayed:
**Inventory** and **Analysis**.

**Inventory table**

The `Inventory` table provides descriptive information about the codes
appearing on each side of the correspondence. Typical columns include:

- the identifiers of codes on side *A* and side *B*;
- the number of times each code appears in the correspondence;
- summary indicators describing how codes are connected.

This table is useful for identifying which codes are involved in the
correspondence and for detecting duplicates or unexpected patterns.

**Analysis table**

The `Analysis` table summarises the relationship structure between the
two classifications. In particular, it classifies the relationships between codes as:

- **1:1 (one‑to‑one)**: each code on side *A* corresponds to exactly one
  code on side *B*;
- **1:M (one‑to‑many)**: one code on side *A* corresponds to multiple
  codes on side *B*;
- **M:1 (many‑to‑one)**: multiple codes on side *A* correspond to one
  code on side *B*;
- **M:M (many‑to‑many)**: multiple codes on side *A* correspond to
  multiple codes on side *B*.

Many‑to‑many (M:M) relationships are typically the most complex and may require
special attention, as they can introduce ambiguity in aggregation,
conversion, or validation processes.


Together, the inventory and analysis outputs provide an overview of both
the content and the structural properties of the correspondence
table.
</div>

### Example 2: Full analysis with source and target classifications
In this second example, all main parameters of the function are used.

- `AB`: correspondence table between NACE Rev.2 and NACE Rev.2.1
- `A`: source classification (NACE Rev.2)  
- `B`: target classification (NACE Rev.2.1)

The analysis is restricted to the lowest‑level codes on both sides of the
correspondence.


```{r}
# Load correspondence table (NACE Rev.2 -> NACE Rev.2.1)
AB <- read.csv(
  system.file("extdata/test", "ab_data.csv", package = "correspondenceTables"),
  stringsAsFactors = FALSE
)

# Load source classification (NACE Rev.2)
A <- read.csv(
  system.file("extdata/test", "a_data.csv", package = "correspondenceTables"),
  stringsAsFactors = FALSE
)

# Load target classification (NACE Rev.2.1)
B <- read.csv(
  system.file("extdata/test", "b_data.csv", package = "correspondenceTables"),
  stringsAsFactors = FALSE
)

```


```{r warning = FALSE}
# Perform analysis using all function parameters
result2 <- analyseCorrespondenceTable(
  AB = AB,
  A  = A,
  B  = B,
  longestAcodeOnly = TRUE,
  longestBcodeOnly = TRUE
)

```


The tables illustrate the first rows of the outputs returned by the function:
The resulting inventory and analysis summary are shown below: 


```{r}
# Display results

knitr::kable(
  head(result2$Inventory),
  caption = "Inventory results: Full analysis with source and target classifications",
  align = "c"
) 


knitr::kable(
  head(result2$Analysis[, 1:5]),
  caption = "Analysis results: Full analysis with source and target classifications",
  align = "c"
)


```

Restricting the analysis to the lowest‑level codes can be useful when the
correspondence table is intended for aggregation or conversion at the most
granular classification level.

 <div style="text-align: justify;">
## Notes

If `A` and `B` are provided, the function performs additional consistency checks against the source and target classification domains.

The `longestAcodeOnly` and `longestBcodeOnly` arguments allow users to restrict the analysis to the most detailed codes, which is often desirable in practical statistical workflows.

This functionality supports the validation and interpretation of correspondence tables, particularly when preparing data for aggregation, transformation, or integration tasks.

</div>

---
title: "Classification Quality Control"
output:
  rmarkdown::html_vignette:  
    toc: TRUE
vignette: >
  %\VignetteIndexEntry{Classification Quality Control}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8} 
---

```{r include=FALSE}

knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  warning  = FALSE,
  message  = FALSE
)

```

<div style="text-align: justify;">

This vignette provides information about the `classificationQC()` function included in the `correspondenceTables` package, which is used to perform quality control on classifications.

</div>

```{r}
library(correspondenceTables) 
```

```{r, echo=FALSE, results="asis"}
cat("<style>table {  width: 100% !important;}table caption {  text-align: center;}
</style>")
```
 
## Overview

<div style="text-align: justify;">

The main function `classificationQC()` performs structural and logical quality 
control on hierarchical classifications. It returns a list of data frames,
including an enriched version of the classification (QC_output) and additional
tables flagging potential issues such as orphan codes, duplicate labels, or
sequencing problems.

The quality‑control checks identify several types of potential structural
or logical issues commonly observed in official classifications:

- **Missing hierarchy levels**: codes for which no hierarchical level can be
  inferred from the provided structure (for example because their code length
  does not match any declared level). Such codes cannot be positioned reliably
  in the hierarchy.

- **Orphan codes**: codes that do not have a valid parent code at the immediately
  higher hierarchical level. This indicates a break in the hierarchical
  structure and usually reflects missing or inconsistent higher‑level codes.

- **Childless codes**: internal codes that do not have any descendants at the
  immediately lower hierarchical level. While expected at the lowest level,
  this may signal incomplete structures or unintended dead ends at higher
  levels.

- **Duplicate labels**: identical labels occurring more than once within the
  same hierarchical level. This does not invalidate the hierarchy, but may
  reduce interpretability and cause ambiguity when classifications are used
  in tabular or statistical contexts.

- **Label‑hierarchy inconsistencies**: situations in which the label of a
  child code does not reflect the label of its parent where such inheritance
  is expected (for example in single‑child situations), suggesting potential
  inconsistencies in naming conventions.

- **Sequencing anomalies**: gaps or breaks in expected code sequences among
  sibling codes under the same parent. In structured coding systems where
  code values are meaningful (for example numeric ranges), such gaps may
  indicate missing or omitted codes.

**Main arguments**

The main arguments of the function are:

- **classification**: A data frame containing the classification codes and labels.

- **lengths**:  A data frame with one row per hierarchical level giving the initial and final positions of the segment of the code referring to that level. The number of rows implicitly defines the number of hierarchical levels ($k$). The column names should be `charb` and `chare`, if this is not the case, they will be automatically changed and a warning will appear.

- **fullHierarchy**: Logical. If `FALSE`, the function checks that all positions at levels greater than 1 have a parent at the level immediately above (no orphans). If `TRUE`, it additionally checks that positions at levels strictly lower than $k$ have children in the next level (no childless nodes). More specifically, the function checks the completeness of the hierarchical structure by applying two rules:
  - **Orphan check**: A new field in the QC output, named `orphan`, takes the value 1 for positions at hierarchical level $j > 1$ that lack a parent at the immediately higher level ($j − 1$), and 0 otherwise.
  - **Childless check**: A new field in the QC output, named `childless`, takes the value 1 for positions at hierarchical level $j < k$ that lack a child at the immediately lower level ($j + 1$), and 0 otherwise.


- **labelUniqueness**: Logical. When `TRUE`, the function checks whether labels are unique, at each hierarchical level. Duplicates are listed in the `QC_duplicatesLabel` table.

- **labelHierarchy**:Logical. When `TRUE`, the function checks that single children share the same label as their parent and if a parent shares a label with one of its children, it must be a single-child parent. The possible values are:
  - **1**: indicates a single child whose label does not match that of its parent;
  - **9**: indicates a child whose label matches that of its parent without being a single child;
  - **0**: indicates compliance.

- **singleChildCode**: an optional data frame defining admissible coding rules
  for single‑child and multiple‑child situations, with columns **level**,
  **singleCode**, and **multipleCode**. If these headers are missing or
  incorrect, they are automatically corrected with a warning.

- **sequencing**: an optional data frame defining admissible code‑range rules for
  multiple‑child situations, used to identify potential gaps in structured code
  sequences. The expected columns are **level** and **multipleCode**.

It is important to note that not all detected issues necessarily indicate errors. The quality‑control checks are diagnostic signals intended to support expert review of classification quality and consistency, and they do not impose constraints on the hierarchical structure itself.
In particular:

- parent codes may legitimately have any number of children;
- sequencing checks are not about the order in which children appear;
- sequencing diagnostics are used to identify gaps in expected, structured code ranges where code values carry semantic meaning.


</div>

## Auxiliary Tables for Classification Validation

<div style="text-align: justify;">

The validation procedures rely on a small set of auxiliary tables that define structural constraints, such as expected code lengths, single‑child rules, and sequencing between levels.

We load three auxiliary tables used for classification validation.  

</div>

### Definition of expected code lengths using the mandatory `lengths` argument

<div style="text-align: justify;">

The `lengths` table specifies the character positions at which each hierarchical level of a classification code starts and ends.
Specifically, column `charb` indicates the starting position of the segment (character beginning), while column `chare` indicates the ending position (character end).

For example, the following definition indicates that:

- level 1 codes start at the first position and end at the second,
- level 2 codes start at the third position and end at the fourth,
- level 3 codes start at the fifth position and end at the seventh.

An example of such a structure is shown below:

```{r}
lengths_example <- data.frame(
  charb = c(1, 3, 5),
  chare = c(2, 4, 7)
)

knitr::kable(
  lengths_example,
  caption = "Example of expected code lengths by hierarchical level",
  align = "c"
)

```

</div>

### Single‑child code constraints

<div style="text-align: justify;">

In some classifications, specific coding conventions are used to distinguish between situations where a parent code has a single child and situations where it has multiple children. These conventions do **not** restrict the hierarchical structure itself and do **not** limit the number of children per node.

Instead, they verify whether observed codes comply with predefined coding
patterns **when a single‑child or multiple‑child situation occurs**.

The `singleChildCode` table defines these admissible patterns and contains
the following columns:

- **level**: the hierarchical level at which the rule applies.
- **singleCode**: the expected coding pattern when a parent has exactly one
  child (for example, retaining the same code).
- **multipleCode**: the expected coding pattern when a parent has multiple
  children (for example, using a sequence of numeric or alphanumeric suffixes).

These checks do not modify the classification and do not enforce a specific hierarchical shape. They merely flag cases where observed coding does not match the declared conventions, which may indicate inconsistencies in code design.

```{r}
singleChildCode <- read.csv(
  system.file("extdata/test", "SingleChild.csv",
              package = "correspondenceTables")
)

knitr::kable(
  singleChildCode,
  caption = "Single-child code rules",
  align = "c"
)
```

</div>

### Sequencing rules between hierarchical levels

<div style="text-align: justify;">

Sequencing checks are not intended to impose an ordering on hierarchical trees.
In a pure tree structure, only parent‑child relationships matter.

However, in many official classifications, code values themselves convey
implicit structure (for example numeric or alphanumeric sequences).
In such systems, sibling codes are often expected to follow predefined ranges
or patterns.

The purpose of sequencing checks is therefore **diagnostic**, not normative:
they aim to detect gaps or breaks in otherwise structured code spaces, which
may indicate missing, omitted, or inconsistently defined codes.

Sequencing rules are defined through a table with the following columns:

- **level**: the hierarchical level at which sequencing rules apply.
- **multipleCode**: the expected pattern or range of sibling codes used to
  detect potential gaps under the same parent.

Sequencing anomalies do not invalidate the hierarchy, but they may point to
classification maintenance issues or incomplete implementations of official
coding schemes.

```{r}

sequencing <- read.csv(
  system.file("extdata/test", "Sequencing.csv",
              package = "correspondenceTables")
)

knitr::kable(
  sequencing,
  caption = "Example of sequencing rules by hierarchical level",
  align = "c"
)

```

</div>

## Example 1: Basic quality control using hierarchy definitions

<div style="text-align: justify;">

The following example applies `classificationQC()` to the NACE Rev.2
classification using additional parameters.

In this example, the user provides:

- a data frame containing the classification to be checked, and
- a data frame defining the hierarchical structure of the classification  through the `lengths` argument. 

This example demonstrates how different parameters of `classificationQC()` are
used to perform structural and logical quality checks.

```{r}
classification <- read.csv(
  system.file("extdata/test", "Nace2_long.csv", package = "correspondenceTables")
)

lengths <- data.frame(
  charb = c(1, 2, 3, 5),
  chare = c(1, 2, 4, 5)
)

```

We now apply the `classificationQC()` function using the previously defined
classification and hierarchy structure. The function performs structural and
logical quality checks on the NACE Rev.2 classification. For illustration
purposes, the output is summarised by reporting the number of detected issues
for selected quality checks.


```{r}
output <- classificationQC(
  classification   = classification,
  lengths          = lengths,
  fullHierarchy    = TRUE,
  labelUniqueness  = TRUE,
  labelHierarchy   = TRUE,
  singleChildCode  = NULL,
  sequencing       = NULL
)

qc_summary <- data.frame(
  Check            = c("No levels", "Orphan codes", "Childless codes"),
  Number_of_issues = c(
    nrow(output$QC_noLevels),
    nrow(output$QC_orphan),
    nrow(output$QC_childless)
  )
)

knitr::kable(
  qc_summary,
  caption = "Summary of quality control checks",
  align = "c"
)

```

</div>

### Codes with no hierarchy level (`QC_noLevels`)

<div style="text-align: justify;">

In this example, all classification codes have a properly defined hierarchy level.
As a result, the quality check `QC_noLevels` does not produce any output.


```{r, echo=FALSE, results="asis"}  

tbl <- output$QC_noLevels

cat(sprintf(
  "**QC_noLevels**\n\n- Rows: %d  \n- Columns: %d\n\n",
  nrow(tbl), 
  ncol(tbl)
))

```

</div>

### Orphan codes (`QC_orphan`)

<div style="text-align: justify;">

Orphan codes are codes that have no parent code at a higher hierarchical level.
This usually indicates breaks in the hierarchical structure.

```{r, echo=FALSE, results="asis"}

tbl2 <- output$QC_orphan

cat(sprintf(
  "**QC_orphan**\n\n- Rows: %d  \n- Columns: %d\n\n",
  nrow(tbl2), 
  ncol(tbl2) 
))

knitr::kable(
  head(tbl2[, 1:8], 5),
  caption = "Orphan codes (First 5 rows (first 7 columns))",
  align = "c"
)
```

</div>

### Childless codes (`QC_childless`)  

<div style="text-align: justify;">

Childless codes are codes at high level that have no descendants at lower hierarchical levels.
This can be expected at the lowest level of a classification, but may indicate
structural issues at higher levels.


```{r, echo=FALSE, results="asis"}
tbl3 <- output$QC_childless

cat(sprintf(
  "**QC_childless**\n\n- Rows: %d  \n- Columns: %d\n\n",
 nrow(tbl3), 
 ncol(tbl3)
))

knitr::kable(
  head(tbl3[, 1:8], 5),
  caption = "First 5 rows (first 7 columns)",
  align = "c"
)

```

</div> 

## Example 2: Quality control with single‑child coding rules

<div style="text-align: justify;">

The following example illustrates the quality control of the NACE Rev.2 classification from CELLAR using additional parameters, including the `singleChildCode` argument.


```{r}

singleChildCode <- read.csv(
  system.file("extdata/test", "SingleChild.csv", package = "correspondenceTables")
)
knitr::kable(
  singleChildCode,
  caption = "singleChildCode argument",
  align = "c"
)

output2 <- classificationQC(
  classification   = classification,
  lengths          = lengths,
  fullHierarchy    = TRUE,
  labelUniqueness  = TRUE,
  labelHierarchy   = TRUE,
  singleChildCode  = singleChildCode,
  sequencing       = NULL
)
```

This table lists orphan codes, i.e. codes that do not have a valid parent at the immediately higher hierarchical level.

```{r, echo=FALSE, results="asis"}
tbl4 <- output2$QC_orphan

cat(sprintf(
  "**QC_orphan**\n\n- Rows: %d  \n- Columns: %d\n\n",
  nrow(tbl4), 
  ncol(tbl4)
))


knitr::kable(
  head(tbl4[, 1:8], 5),
  caption = "First 5 rows (first 7 columns)",
  align = "c"
)

```


This table lists childless codes, i.e. codes that have no descendants at the immediately lower hierarchical level

```{r, echo=FALSE, results="asis"}
tbl6 <- output2$QC_childless

cat(sprintf(
  "**QC_childless**\n\n- Rows: %d  \n- Columns: %d\n\n",
 nrow(tbl6), 
  ncol(tbl6) 
))


knitr::kable(
  head(tbl6[, 1:8], 10),
  caption = "First 10 rows (first 7 columns)",
  align = "c"
)


```

</div>


## Example 3: Quality control with sequencing constraints

<div style="text-align: justify;">

In this final example, the `sequencing` parameter is used to detect potential gaps
in structured sequences of sibling codes within the hierarchy.

Sequencing rules are applied at hierarchical levels 3 and 4, as specified in the
`sequencing` input table. At these levels, the function identifies missing or
inconsistent code values within predefined numeric or alphanumeric ranges,
which may indicate incomplete or faulty classification structures.


```{r}

singleChildCode <- read.csv(
  system.file("extdata/test", "SingleChild2.csv", package = "correspondenceTables")
)

sequencing <- read.csv(
  system.file("extdata/test", "Sequencing.csv",
              package = "correspondenceTables")
)

output3 <- classificationQC(
  classification   = classification, 
  lengths          = lengths,
  fullHierarchy    = TRUE,
  labelUniqueness  = TRUE,
  labelHierarchy   = TRUE,
  singleChildCode  = singleChildCode,
  sequencing       = sequencing
)


```

The `QC_gapBefore` argument identifies gaps in expected code sequences among sibling codes within the same parent.

```{r, echo=FALSE, results="asis"}
tbl7 <- output3$QC_gapBefore

cat(sprintf(
  "**QC_gapBefore**\n\n- Rows: %d  \n- Columns: %d\n\n",
  nrow(tbl7),
  ncol(tbl7)
))


knitr::kable(
  head(tbl7[, 1:8], 10),
  caption = "QC_gapBefore. First 10 rows (first 7 columns)",
  align = "c"
)


```

This table lists the last sibling codes within each group of children, used to assess sequence completeness.

```{r, echo=FALSE, results="asis"}  
tbl8 <- output3$QC_lastSibling

cat(sprintf(
  "**QC_lastSibling**\n\n- Rows: %d  \n- Columns: %d\n\n",
  nrow(tbl8),
  ncol(tbl8)
))


knitr::kable(
  head(tbl8[, 1:8], 10),
  caption = "QC_lastSibling. First 10 rows (first 7 columns)",
  align = "c"
)


```

This table contains the full classification enriched with all quality‑control flags produced by the checks


```{r, echo=FALSE, results="asis"}
tbl9 <- output3$QC_output

cat(sprintf(
  "**QC_output**\n\n- Rows: %d  \n- Columns: %d\n\n",
  nrow(tbl9), 
  ncol(tbl9)
))


knitr::kable(
  head(tbl9[, 1:8], 10),
  caption = "First 10 rows (first 7 columns)",
  align = "c"
)


```

</div>

---
title: "Aggregate Correspondence Table"
output:
  rmarkdown::html_vignette:
    toc: TRUE
vignette: >
  %\VignetteIndexEntry{Aggregate Correspondence Table}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}  
---
     


```{r setup, include=FALSE}

knitr::opts_chunk$set(
  echo = TRUE,
  message = FALSE,
  warning = FALSE,
  cache = FALSE
)

```

<div style="text-align: justify;">
## Overview
This vignette illustrates how to aggregate numeric values across classification systems using the `aggregateCorrespondenceTable()` function from the `correspondenceTables` package.
The function aggregates numeric values expressed in a source classification (A) into a target classification (B), using a correspondence table that links A to B (denoted A → B).
When correspondence weights are available, values are redistributed proportionally according to these weights.
If no weights are provided, values are distributed equally across all corresponding target codes.

This type of aggregation is commonly used to convert statistics between classification systems, for example:

NACE → CPA

CPA → CN

PRODCOM → CPA

CPC → HS
</div>

```{r cars}
library(correspondenceTables)
```


```{r, echo=FALSE, results="asis"}
cat("
<style>
table {
  width: 100% !important;
}
table caption {
  text-align: center;
}
</style>
")
```


**Inputs** 

<div style="text-align: justify;">

The `aggregateCorrespondenceTable()` function expects the following inputs:

- **AB**: A data frame representing the correspondence table between classification A and classification B.  
  It must contain:
  - a source code column (`from_code`)
  - a target code column (`to_code`)
  - optionally, a weight column

- **A**: A data frame containing values expressed in the source classification A.
  It typically includes:
  - a source classification code
  - one or more numeric variables to be aggregated  

  By default, the function expects the source code column in **A** to be named `code`. 
  This can be adapted if the function supports custom column names.

- **B (optional)**: A data frame defining the domain of the target classification B.  
  If provided, all B codes are preserved in the output, and target codes with no matching contributions receive a value of zero.
 
</div> 
 
## Application of  `aggregateCorrespondenceTable()`

### Example 1: Basic aggregation using a correspondence table

<div style="text-align: justify;">

In this example, all inputs are read from sample datasets included in the package.


```{r}
AB_path <- system.file("extdata/test", "ab_data.csv", package = "correspondenceTables")
A_path  <- system.file("extdata/test", "a_data.csv",  package = "correspondenceTables")
B_path  <- system.file("extdata/test", "b_data.csv",  package = "correspondenceTables")

stopifnot(nzchar(AB_path), nzchar(A_path), nzchar(B_path))

AB <- utils::read.csv(AB_path, stringsAsFactors = FALSE)
A  <- utils::read.csv(A_path,  stringsAsFactors = FALSE)
B  <- utils::read.csv(B_path,  stringsAsFactors = FALSE)

#For clarity and consistency, the correspondence table columns are renamed to the expected identifiers:

names(AB)[names(AB) == "NACE.Rev..2.Code"]   <- "from_code"
names(AB)[names(AB) == "NACE.Rev..2.1.Code"] <- "to_code"


res <- aggregateCorrespondenceTable(AB = AB, A = A, B = B)


knitr::kable(
  head(res$result),
  caption = "Aggregation using a correspondence table",
  align = "c"
)

```


The function returns a list. 
The aggregated values are stored in the `result` element, which is a data frame 
structured according to the target classification B.

**Interpretation of the output**

In this example:

- Dataset **A** contains numeric values expressed in the source classification.
- The correspondence table **AB** specifies how each source code is linked to one or more target codes.
- No weights are supplied in the correspondence table.

For each source code in **A**:

- If it maps to **a single target code**, its full value is assigned to that target code.
- If it maps to **multiple target codes**, its value is **split equally** among them.

All allocated contributions are then **summed for each target code**.
The column containing numeric values in the output therefore represents the total value aggregated to each target classification code in B.

**Notes**

- The aggregation performed by `aggregateCorrespondenceTable()` is additive: values are redistributed and summed, not averaged or otherwise summarized.
- Supplying the **B** argument ensures that the output covers the full target classification domain; target codes with no matching contributions receive a value of zero.


</div>

### Example 2: Weighted correspondence (proportional allocation)

<div style="text-align: justify;">


This example illustrates aggregation when the correspondence table includes **explicit weights**.

Here:

- Source code **A1** is linked to two target codes:
  - 70% of its value goes to **B1**
  - 30% goes to **B2**
- Source code **A2** is linked entirely to **B2**

The function multiplies each source value by the corresponding weight for each
correspondence link and then sums all weighted contributions per target code.


```{r}

# Correspondence table with weights  
AB <- data.frame(
  from_code = c("A1", "A1", "A2"),
  to_code   = c("B1", "B2", "B2"),
  weight    = c(0.7, 0.3, 1.0)
)

# Source classification with values  
A <- data.frame(
  code  = c("A1", "A2"), 
  value = c(100, 50)
)

# Target classification domain
B <- data.frame(
  code = c("B1", "B2")
)

res2 <- aggregateCorrespondenceTable(AB = AB, A = A, B = B)

knitr::kable(
  head(res2$result),
  caption = "Weighted correspondence (proportional allocation)",
  align = "c"
)

```

**Interpretation of the output**

The values shown in the output represent the total weighted sums per target code.

For example:

- Target code **B1** receives 70% of the value associated with **A1**
- Target code **B2** receives:
  - 30% of **A1**
  - 100% of **A2**

All contributions are summed to produce the final totals.

**Tiny numeric illustration**

If **A1** has a value of 100:

- 70 is allocated to **B1**
- 30 is allocated to **B2**

If **A2** has a value of 50 and maps fully to **B2**, the final value for **B2** is:

\(30 + 50 = 80\)


</div>

---
title: "Retrieve classifications and correspondence tables stored as Linked Open Data"
output:
  rmarkdown::html_vignette:
    toc: TRUE

vignette: >
  %\VignetteIndexEntry{Retrieve classifications and correspondence tables stored as Linked Open Data}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}  
---

```{r, echo=FALSE, results="asis"}
cat("
<style>
h1.title {
  text-align: center;
}
</style>
")
```

```{r setup, include=FALSE}

knitr::opts_chunk$set(message = FALSE, warning = FALSE, fig.align = "center")
old <- getOption("useLocalDataForVignettes")
options(useLocalDataForVignettes = TRUE)
on.exit(options(useLocalDataForVignettes = old), add = TRUE)

```


## Overview  

<div style="text-align: justify;">

Statistical classifications and correspondence tables are published as Linked Open Data (LOD) by several organisations, notably the Publications Office of the European Union (OP, via CELLAR) and the Food and Agriculture Organization (FAO).

While these resources can be accessed directly using SPARQL, this requires specific technical expertise. The **correspondenceTables** package provides high‑level R functions that allow users to retrieve these data as standard R data frames, without writing SPARQL queries themselves.

Two core data retrieval functions are provided:

- `retrieveClassificationTable()`: retrieves the structure of a statistical
  classification (codes, labels, hierarchy).
- `retrieveCorrespondenceTable()`: retrieves a correspondence (mapping) table
  between two classifications.

Optionally, both functions can return the SPARQL query used for the retrieval, making the process transparent, inspectable, and reproducible.  

In addition, the `dataStructure()` utility allows users to inspect the hierarchical
structure of a classification (e.g. available levels and code depth) before retrieving
the data. This step is optional but recommended when working with hierarchical 
classifications, particularly when the desired level is not known in advance and is covered in this vignette, with illustrative examples provided
for both the CELLAR and FAO endpoints.  

</div>

```{r}
library(correspondenceTables)
```


```{r, echo=FALSE, results="asis"}
cat("<style>table {width: 100% !important;}table caption {text-align: center;}</style>")
```


## Discovering available data

<div style="text-align: justify;">

Before using the core retrieval functions
`retrieveClassificationTable()` and `retrieveCorrespondenceTable()`,
it is necessary to know which data can be retrieved and how it is identified.

In practice, this means answering the following questions:

- Which classifications or correspondence tables are available?
- From which endpoint (`CELLAR` or `FAO`)?
- Which identifiers (`prefix`, `conceptScheme`, `ID_table`) should be used?
- For hierarchical classifications, which levels are available?

The package provides lightweight discovery utilities to support this step
before data retrieval.

</div>

### Gathering information necessary for classification retrieval

<div style="text-align: justify;">

To retrieve a statistical classification, users first need to know which
classifications are available at a given endpoint and how they are identified.
The `classificationList()` utility provides this information.

</div>

### Example 1: Available classifications (CELLAR)

<div style="text-align: justify;">

The example below illustrates the typical output structure using a static
snapshot of the `CELLAR` classification list bundled with the package.

To retrieve updated information about available classifications,
users only need to execute the `classificationList()` function.


```{r}
list_data <- read.csv(
  system.file("extdata/test", "classificationList_CELLAR.csv",
              package = "correspondenceTables"),
  stringsAsFactors = FALSE
)

knitr::kable(
  head(list_data, 3),
  caption = "Example output of classificationList() (retrieved from CELLAR)"
)

```
  

For each classification, three identifiers are required to retrieve the data:

- `endpoint`: `"CELLAR"` or `"FAO"`
- `prefix`: namespace prefix used in the SPARQL endpoint
- `conceptScheme`: unique identifier of the classification

For example, the NACE Rev. 2 classification:

- is available from the `CELLAR` repository,  
- uses prefix `"nace2"`,
- uses concept scheme `"nace2"`.

</div>

## Inspecting the structure of hierarchical classifications

<div style="text-align: justify;">

Many statistical classifications are hierarchical.
If only a specific level is required (e.g. divisions or classes),
it is recommended to inspect the classification structure first.
The `dataStructure()` function provides this information.

</div> 

### Example 2: Classification structure (CN 2022, CELLAR)

<div style="text-align: justify;">

This example illustrates how to inspect the structural characteristics of a classification
stored in the `CELLAR` repository.
The `dataStructure()` function can be used to retrieve either a **summary view**, a
**detailed view**, or **both**.

To keep the vignette reproducible and independent of live SPARQL endpoints,
the function calls below are shown for documentation purposes only.

</div> 

#### Summary view of the classification structure

<div style="text-align: justify;">

The `summary` output provides an overview of the hierarchical organisation of the
classification.
For each level, it reports:

- the classification scheme identifier,
- the hierarchical depth,
- the level label,
- the number of classification items defined at that level.

This view is useful for quickly understanding the overall structure of a classification
and identifying which hierarchical levels are available.


```{r, eval = FALSE}

ds_cn <- dataStructure(
  endpoint      = "CELLAR",
  prefix        = "cn2022",
  conceptScheme = "cn2022",
  language      = "en",
  return        = "summary"
)

knitr::kable(head(ds_cn, 20), caption = "CN 2022 — dataStructure(summary)")

```

The Combined Nomenclature (CN 2022) follows a hierarchical product classification
structure defined at several levels. The summary output shows that it consists of
five hierarchical levels:

- **Level 1: Sections**: broad groupings of goods;
- **Level 2: Chapters**: main product divisions;
- **Level 3: Headings**: four‑digit product categories;
- **Level 4: HS subheadings**: six‑digit Harmonized System categories;
- **Level 5: CN subheadings**: eight‑digit CN‑specific product codes.

The `Count` column indicates the number of classification items defined at each
hierarchical depth.

</div>

#### Detailed view of classification items

<div style="text-align: justify;">

The `details` output returns one row per classification item.
It provides item‑level metadata, including:

- the classification code,
- the preferred label,
- the hierarchical level and depth,
- links to broader (parent) concepts where available.

This view is intended for detailed inspection of classification content, for example
when analysing parent-child relationships or validating code hierarchies.


```{r, eval = FALSE}

ds_cn_det <- dataStructure(
  endpoint      = "CELLAR",
  prefix        = "cn2022",
  conceptScheme = "cn2022",
  language      = "en",
  return        = "details"
)

knitr::kable(head(ds_cn_det, 20), caption = "CN 2022 — dataStructure(details)")

```


</div>

#### Summary and detailed views combined

<div style="text-align: justify;">

When `return = "both"`, the function returns a list containing both summary and detailed outputs.
This option can be convenient when both a structural overview and item‑level
information are required within a single workflow.

```{r, eval = FALSE}



ds_cn_both <- dataStructure(
  endpoint      = "CELLAR",
  prefix        = "cn2022",
  conceptScheme = "cn2022",
  language      = "en",
  return        = "both"
)

knitr::kable(head(ds_cn_both$summary, 20), caption = "CN 2022 — summary (from both)")
knitr::kable(head(ds_cn_both$details, 20), caption = "CN 2022 — details (from both)")
```

As with classifications retrieved from `CELLAR`, this inspection step can be skipped
if the required classification level is already known in advance.


</div>

### Example 3: Classification structure (CPC 2.1, FAO)

<div style="text-align: justify;">

The same approach can be applied to classifications hosted in the FAO repository.
This example illustrates how to inspect the structure of the Central Product
Classification (CPC), version 2.1.

As with CELLAR, the `dataStructure()` function can return a **summary view**, a
**detailed view**, or **both** representations of the classification structure.
In practice, the choice depends on whether a high-level overview or item-level
information is required.

To keep the vignette reproducible and independent of live SPARQL endpoints,
the function call below is provided for documentation purposes only.

</div>

#### Summary view of the classification structure

<div style="text-align: justify;">

The *summary* output provides a compact overview of the hierarchical organisation
of CPC 2.1. For each level, it reports:

- the classification scheme identifier,
- the hierarchical depth,
- the level label,
- the number of classification items defined at that level.

This view is useful for understanding the overall structure of the classification
before retrieving detailed content.

```{r, eval = FALSE}
endpoint <- "FAO"
prefix <- "CPC21"
conceptScheme <- "CPC21"

ds_cpc <- dataStructure(
  endpoint      = endpoint,
  prefix        = prefix,
  conceptScheme = conceptScheme,
  language      = "en",
  showQuery     = FALSE,
  return        = "summary"
)

knitr::kable(
  head(ds_cpc, 20),
  caption = "CPC 2.1 — dataStructure(summary, FAO)"
)
```

As in the CELLAR example, `return = "details"` retrieves item-level information,
while `return = "both"` returns both summary and detailed outputs in a single call.

</div>

## Retrieving classification tables

<div style="text-align: justify;">

Once the classification identifiers and (optionally) the desired level are known,
the `retrieveClassificationTable()` function can be used to retrieve the data.

The function returns a flat data frame suitable for:

- browsing and documentation;
- validation of codes and hierarchy;
- downstream correspondence analysis.

**Main arguments**

- `endpoint`: `"CELLAR"` or `"FAO"`
- `prefix`: Character. Classification prefix used for matching and URI resolution (e.g. "cn2022", "cpc21", "isic4").
- `conceptScheme`: Character. Local identifier of the scheme (often identical to `prefix`). The function automatically resolves this to the canonical ConceptScheme URI published in the endpoint.
- `language`: Character. Preferred label language as a BCP47 code. Defaults to "en" (English). Examples: "fr", "de".
- `level`: Character. One of:
  + `"ALL"` (default): return all levels in the hierarchy;
  + a specific depth value (e.g. "2") to filter concepts at that depth only.
- `showQuery`: Logical.
  + `FALSE` (default): returns only the classification table;
  + `TRUE`: returns a list containing the SPARQL query, the resolved scheme URI, and the table itself.
- `knownSchemes`: Optional. A data.frame supplying authoritative mappings of the form Prefix, ConceptScheme, URI. When provided, this overrides automatic discovery. To be obtained using `classificationList(endpoint)`.
- `preferMappingOnly`: Logical. If `TRUE`, the function never attempts SPARQL discovery and uses only information in `knownSchemes` or `classificationList(endpoint)`. Default: `FALSE`.

</div>

### Example 4: Class‑level NACE Rev. 2 in multiple languages

<div style="text-align: justify;">

The following example demonstrates how to retrieve level‑4 (“class”)
data for the German, French, and Bulgarian versions of **NACE Rev. 2**.
The code is **not executed** during vignette rendering as data availability and response times may vary.


```{r retrieve-nace-multilang, eval=FALSE}
endpoint <- "CELLAR"
prefix <- "nace2"
conceptScheme <- "nace2"
level <- "4"

languages <- c("de", "fr", "bg")

results <- lapply(languages, function(lang) {
  retrieveClassificationTable(
    endpoint = endpoint,
    prefix = prefix,
    conceptScheme = conceptScheme,
    language = lang,
    level = level,
    showQuery = FALSE
  )
})
```


The resulting object is a list of data frames, one per language, each containing
the class‑level codes and labels for NACE Rev. 2 in the selected language.

</div>

### Example 5: FAO classification at group level

<div style="text-align: justify;">

The FAO endpoint provides access to a limited subset of international classifications. Availability depends on the endpoint configuration.

The following example illustrates how a FAO classification would be retrieved. The code is not executed during vignette rendering.

This call queries the FAO repository and returns metadata describing all published classification schemes (prefix, concept scheme, title, etc.).

```{r, eval=FALSE}

cl_fao <- classificationList("FAO")

knitr::kable(
  head(cl_fao),
  caption = "Retrieving a classification table from the FAO endpoint"
)
```

**Inspect available prefix identifiers**
 
 The `Prefix` field identifies the catalogue or namespace under which  each FAO classification is published.

```{r, eval=FALSE}
knitr::kable(
  head(unique(cl_fao$Prefix)))
```

**Inspect available concept schemes**
 
 The `ConceptScheme` field identifies the underlying classification schemes
 that can be queried using `retrieveClassificationTable()`.

```{r, eval=FALSE}
knitr::kable(
  head(unique(cl_fao$ConceptScheme)))
```


**Retrieving a classification table from the FAO endpoint**

The following example illustrates how to retrieve a classification from the `FAO` repository using `retrieveClassificationTable()`.
Because `FAO` data availability and response times may vary, this example is shown for documentation purposes and is not executed in the vignette.


```{r retrieve-fao-classification, eval=FALSE}
endpoint <- "FAO"
prefix <- "cpc21"
conceptScheme <- "core"

out <- retrieveClassificationTable(
  endpoint      = endpoint,
  prefix        = prefix,
  conceptScheme = conceptScheme,
  language      = "en",
  level         = "2",
  showQuery     = TRUE
)
```


The `FAO` endpoint provides access to selected international and domain‑specific classifications maintained by `FAO`.
Not all `CELLAR` classifications are available via `FAO`, and vice versa.

</div>

### Example 6: Retrieving a classification table from a known data frame of classification tables

<div style="text-align: justify;">

Every time it is executed, the `retrieveClassificationTable()` function attempts to retrieve the list of all the available classifications for a selected endpoint, in order to have always the most up-to-date URI for a given pair of prefix-concept scheme. Since this step can be time consuming, it can be skipped entirely by providing a previously retrieved (and stored) classification list (obtained with `classificationList()`) using the `knownSchemes` argument. The example that follows, shows how to use this argument:


```{r , eval=FALSE}
cl_fao <- classificationList("FAO")
endpoint <- "FAO"
prefix <- "cpc21"
conceptScheme <- "core"

out <- retrieveClassificationTable(
  endpoint      = endpoint,
  prefix        = prefix,
  conceptScheme = conceptScheme,
  knownSchemes  = cl_fao
)
```

</div>

## Retrieving correspondence tables

<div style="text-align: justify;">

The `retrieveCorrespondenceTable()` function retrieves a correspondence (mapping) table between two statistical classifications from a SPARQL endpoint. Its interface is similar to `retrieveClassificationTable()`, with the main difference that correspondence tables are identified using `ID_table` (instead of `conceptScheme`). Correspondence tables are usually provided at the most granular level of the classifications involved.

**Main arguments**

- `endpoint`: Character. The online service to query. Case-insensitive. Supported values are those returned by the internal endpoint registry (e.g., `"CELLAR"`, `"FAO"`).

- `prefix`: Character. Catalogue prefix where the correspondence is published (e.g., "nace2", "cpa21", "cn2022"). Use `correspondenceTableList()` to discover valid values.

- `ID_table`: Character. Identifier of the correspondence, typically of the form "A_B" such as "NACE2_CPA21" or "CN2022_NACE2". Discover identifiers via `correspondenceTableList()`.

- `language`: Character. Preferred label language as a BCP47 code. Defaults to "en" (English). Examples: "fr", "de".

- `showQuery`: Logical. If `TRUE`, returns a list with the SPARQL query and the result data frame; otherwise (default) returns just the data frame.

</div>

### Example 7: Available correspondence tables

<div style="text-align: justify;">

Before retrieving a correspondence table, users need to identify which correspondences are available and how they are referenced at a given SPARQL endpoint. The `correspondenceTableList()` utility serves this purpose. It is analogous to `classificationList()`, but lists correspondence tables instead of classifications.

The following example illustrates how to list correspondence tables available from the `CELLAR` and `FAO` repositories.
It is shown for documentation purposes and not executed during vignette rendering to avoid reliance on live external SPARQL endpoints.

```{r, eval = FALSE}

corr_list = correspondenceTableList("ALL")

names(corr_list)
#Correspondence tables available from CELLAR
knitr::kable(
  head(corr_list$CELLAR, 10),
  caption = "Available correspondence tables from the CELLAR endpoint (preview)"
)
#Correspondence tables available from FAO
knitr::kable(
  head(corr_list$FAO, 10),
  caption = "Available correspondence tables from the FAO endpoint (preview)"
)
```

When executed interactively, this call returns a list whose elements correspond
to the selected endpoints (e.g. `CELLAR`, `FAO`). Each element is a data frame
describing the available correspondence tables, including their identifiers,
associated prefixes, and human-readable labels.

Each correspondence table is identified by:

- **endpoint**: `"CELLAR"` or `"FAO"`
- **prefix**: namespace associated with the source classification
- **ID_table**: unique identifier of the correspondence table

</div>

### Inspect available correspondence tables (CELLAR)

<div style="text-align: justify;">

The following examples illustrate how to inspect the correspondence tables
available from the `CELLAR` endpoints.

```{r corr-table-list, eval=FALSE}
# Inspect correspondence tables available from CELLAR
tbl_cellar <- correspondenceTableList("CELLAR")

#Correspondence tables available from CELLAR
knitr::kable(
  head(tbl_cellar, 10),
  caption = "Available correspondence tables from the CELLAR endpoint "
)
```

</div>

### Example 8: Retrieve a correspondence table from CELLAR

<div style="text-align: justify;">

The following example illustrates the retrieval of a correspondence table published by the Publications Office of the European Union via the `CELLAR` endpoint.
Users should note that the availability of correspondence data depends on what is currently exposed by the underlying SPARQL endpoint. Although a correspondence table may be listed by `correspondenceTableList()`, it can legitimately return an empty result when queried. For some `CELLAR` correspondences (including several PRODCOM‑related mappings), `retrieveCorrespondenceTable()` may therefore return a valid but empty data frame, which does not indicate a failure of the retrieval
process.


```{r retrieve-prodcom, eval=FALSE}
res <- retrieveCorrespondenceTable(
  endpoint  = "CELLAR",
  prefix    = "prodcom2023",
  ID_table  = "PRODCOM2023_CPA21",
  language  = "en",
  showQuery = FALSE
)
knitr::kable(
  head(res, 10),
  caption = "PRODCOM2023_CPA21 CorrespondenceTable from the CELLAR endpoint "
)
```

To reduce potential user confusion, it is helpful to include at least one
correspondence example that is more likely to return data when queried.

```{r, eval = FALSE}
res2 <- retrieveCorrespondenceTable(
  endpoint = "CELLAR",
  prefix   = "nace2",
  ID_table = "NACE2_CPA21",
  language = "en"
)
knitr::kable(
  head(res2, 10),
  caption = "NACE2_CPA21 CorrespondenceTable from the CELLAR endpoint "
)
```

For transparency and reproducibility, the SPARQL query used for retrieval can
also be inspected by setting `showQuery = TRUE`.

</div>

### Inspect available correspondence tables (FAO)

The following examples illustrate how to inspect the correspondence tables available from the `FAO` endpoint.

```{r, eval = FALSE}
# Inspect correspondence tables available from FAO
tbl_fao <- correspondenceTableList("FAO")
head(tbl_fao)

knitr::kable(
  head(tbl_fao, 10),
  caption = "correspondence tables available from FAO"
)

```

</div>

### Example 9: Retrieve a correspondence table from FAO: CPC 2.1 : ISIC Rev. 4

<div style="text-align: justify;">

The following example illustrates the retrieval of a correspondence table published by the Food and Agriculture Organization of the United Nations (FAO) via the `FAO` endpoint.

Users should note that the availability of correspondence data depends on what is currently exposed by the underlying SPARQL endpoint. Although a correspondence table may be listed by `correspondenceTableList()`, it can legitimately return an empty result when queried. In practice, however, correspondence tables exposed by the `FAO` endpoint tend to be more consistently populated than some of those available from `CELLAR.`

The English-language version of the CPC 2.1 : ISIC Rev. 4 correspondence table can be retrieved as follows. This example is not executed during vignette rendering.


```{r retrieve-cpc21-isic4_FAO , eval = FALSE}
Res <- retrieveCorrespondenceTable(
  endpoint = "FAO",
  prefix   = "CPC21",
  ID_table = "CPC21-ISIC4",
  language = "en"
)

knitr::kable(
  head(Res[, 1:5], 10),
  caption = "CPC21–ISIC4 correspondence tables available from FAO"
)
```
 
</div> 
 
### (Optional) Inspect the underlying SPARQL query 

<div style="text-align: justify;">

For transparency and reproducibility, the SPARQL query used for retrieval can also be inspected by setting `showQuery = TRUE`.


```{r, eval = FALSE}
Res2 <- retrieveCorrespondenceTable(
  endpoint = "FAO",
  prefix   = "CPC21",
  ID_table = "CPC21-ISIC4",
  language = "en",
  showQuery = TRUE
)

# Extract the SPARQL query used
SPARQLquery <- Res2$SPARQL.query
SPARQLquery 
```

</div>
 
## Summary

<div style="text-align: justify;">

The `correspondenceTables` package simplifies access to statistical classifications
and correspondence tables published as Linked Open Data (LOD), including those
provided by major repositories such as the EU Publications Office (CELLAR) and FAO.

It offers a high-level R interface to:

- identify available classifications and correspondences;
- retrieve classification hierarchies and mapping tables without writing SPARQL queries;
- explore classification structures to select relevant levels;
- ensure reproducibility by exposing the underlying SPARQL queries when needed.

This approach lowers the technical barrier to working with official classification
systems, enabling analysts to integrate them seamlessly into their workflows while
preserving transparency and reproducibility.

</div>

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/retrieveClassificationTable.R
\name{retrieveClassificationTable}
\alias{retrieveClassificationTable}
\title{Retrieve a full classification table from CELLAR or FAO}
\usage{
retrieveClassificationTable(
  endpoint,
  prefix,
  conceptScheme,
  language = "en",
  level = "ALL",
  showQuery = FALSE,
  knownSchemes = NULL,
  preferMappingOnly = FALSE
)
}
\arguments{
\item{endpoint}{Character. Repository to query. Must be either "CELLAR" or "FAO".}

\item{prefix}{Character. Classification prefix used for matching and URI resolution
(e.g. "cn2022", "cpc21", "isic4").}

\item{conceptScheme}{Character. Local identifier of the scheme (often identical to
\code{prefix}). The function automatically resolves this to the canonical
ConceptScheme URI published in the endpoint.}

\item{language}{Character. Desired language for labels, scope notes and exclusion notes.
Default: "en".}

\item{level}{Character. One of:
\itemize{
  \item "ALL" (default): return all levels in the hierarchy;
  \item a specific depth value (e.g. "2") to filter concepts at that depth only.
}}

\item{showQuery}{Logical.
\itemize{
  \item FALSE (default): returns only the classification table;
  \item TRUE: returns a list containing the SPARQL query, the resolved scheme URI, and the table itself.
}}

\item{knownSchemes}{Optional. A data.frame supplying authoritative mappings of the
form \code{Prefix, ConceptScheme, URI[, Endpoint]}. When provided, this overrides
automatic discovery. To be obtained using \code{classificationList(endpoint)}.}

\item{preferMappingOnly}{Logical. If TRUE, the function never attempts SPARQL
discovery and uses only information in \code{knownSchemes} or \code{classificationList(endpoint)}.
Default: FALSE.}
}
\value{
If \code{showQuery = FALSE}, a \code{data.frame} with one row per concept and columns:
\itemize{
  \item \strong{Concept}: full concept URI;
  \item \strong{Code}: \code{skos:notation} coerced to string;
  \item \strong{Label}: preferred or alternative label in the requested language;
  \item \strong{Depth}: \code{xkos:depth} if available;
  \item \strong{BroaderList}: broader concept URIs (pipe-separated);
  \item \strong{BroaderCodeList}: broader concept notations (pipe-separated);
  \item \strong{IncludeNotes}: \code{skos:scopeNote} values (double-pipe-separated);
  \item \strong{ExcludeNotes}: \code{xkos:exclusionNote} values (double-pipe-separated).
}

If \code{showQuery = TRUE}, returns a list with:
\itemize{
  \item \code{SPARQL.query} – the executed SPARQL string;
  \item \code{scheme_uri} – the resolved ConceptScheme URI;
  \item \code{ClassificationTable} – the data.frame described above.
}
}
\description{
Retrieves the complete structure of a statistical classification published in
either CELLAR or FAO repositories based on a Prefix–ConceptScheme pair obtained
using \code{classificationList()}.

The resulting table is suitable for classification browsing, integrity checks,
hierarchical analysis, documentation, and downstream correspondence mapping.
}
\examples{
\dontrun{
# Retrieve full CN2022 classification (English)
cn <- retrieveClassificationTable(
  endpoint      = "CELLAR",
  prefix        = "cn2022",
  conceptScheme = "cn2022",
  language      = "en"
)

# Retrieve CPC v2.1 from FAO and inspect query
out <- retrieveClassificationTable(
  endpoint      = "FAO",
  prefix        = "cpc21",
  conceptScheme = "cpc21",
  showQuery     = TRUE
)
cat(out$SPARQL.query)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prefixList.R
\name{prefixList}
\alias{prefixList}
\title{Build SPARQL PREFIX declarations for an endpoint}
\usage{
prefixList(endpoint, prefix = NULL)
}
\arguments{
\item{endpoint}{Character. Endpoint name (case-insensitive). Typically
`"CELLAR"` or `"FAO"`. The set of supported names is defined by the
package's internal endpoint registry (see `.valid_endpoints()`).}

\item{prefix}{Character vector of scheme tokens to include (e.g.,
`c("cn2022","nace2")`). If `NULL` (default), returns all discovered schemes.}
}
\value{
A character vector of SPARQL PREFIX declarations.
Use `paste(result, collapse = "\\n")` to inject into a query string.
}
\description{
Returns a character vector of `PREFIX ...` lines used by package queries.
The result always includes core vocabularies (SKOS/XKOS/RDFS/DC/RDF/XSD),
and—when available—adds classification scheme namespaces discovered via
`classificationList(endpoint)`. You can filter to specific schemes with
the `prefix` argument.
}
\details{
- This function has no side effects and does not perform network I/O
  by itself. Discovery is delegated to `classificationList()`.
  
- If `classificationList()` is unavailable or fails, the function returns
  only the core vocabularies and emits a warning.
}
\examples{
\dontrun{
# All known scheme prefixes for the endpoint
pl <- prefixList("CELLAR")
cat(paste(pl, collapse = "\n"))

# Only include CN2022 and NACE2 if they exist
pl2 <- prefixList("CELLAR", prefix = c("cn2022", "nace2"))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prefixList.R
\name{prefix_declared_in_block}
\alias{prefix_declared_in_block}
\title{Validate that a prefix token is declared in a prefix block (pure)}
\usage{
prefix_declared_in_block(token, prefix_block)
}
\arguments{
\item{token}{a candidate prefix token (will be normalized)}

\item{prefix_block}{a single string containing the PREFIX lines}
}
\value{
TRUE if declared, FALSE otherwise
}
\description{
Validate that a prefix token is declared in a prefix block (pure)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/analyseCorrespondanceTable.R
\name{analyseCorrespondenceTable}
\alias{analyseCorrespondenceTable}
\title{Perform analysis on correspondence tables}
\usage{
analyseCorrespondenceTable(
  AB,
  Aname = NULL,
  Bname = NULL,
  A = NULL,
  B = NULL,
  longestAcodeOnly = FALSE,
  longestBcodeOnly = FALSE
)
}
\arguments{
\item{AB}{A mandatory argument containing a data frame provided by the user
with the correspondence table data. The table must contain at least two
columns containing the two classifications \code{A code} and \code{B code}. The name of the two columns is irrelevant.}

\item{Aname}{Optional. A character string specifying the name of the column
in `AB` that should be used as the source classification (`Acode`). If
`NULL`, the function assumes that the first column of `AB` represents the
source classification.}

\item{Bname}{Optional. A character string specifying the name of the column
in `AB` that should be used as the target classification (`Bcode`). If
`NULL`, the function assumes that the second column of `AB` represents the
target classification.}

\item{A}{A data frame containing the source classification data with
an \code{"Acode"} column. Can be \code{NULL} if no source classification
table is provided.}

\item{B}{A data frame containing the target classification data with
a \code{"Bcode"} column. Can be \code{NULL} if no target classification
table is provided.}

\item{longestAcodeOnly}{Logical. If \code{TRUE}, source classification data
are filtered based on \code{"Acode"} retaining only the maximum length,
i.e. the lowest-level Acodes in the correspondence table.}

\item{longestBcodeOnly}{Logical. If \code{TRUE}, target classification data
are filtered based on \code{"Bcode"} retaining only the maximum length,
i.e. the lowest-level Bcodes in the correspondence table.}
}
\value{
A list containing two data frames: \code{Inventory} and \code{Analysis}.
  \code{Inventory} contains statistics related to components, correspondence
  types, and source/target positions. \code{Analysis} contains statistics for
  each class in the correspondence table.
}
\description{
Perform analysis on a correspondence table using its associated
classifications to generate various statistics. It checks the validity of the
input data, identifies components, calculates correspondence types, and
creates summary tables.
}
\examples{
# Use data from the folder extdata/test

AB = read.csv(system.file("extdata/test", "ExempleAnnexe.csv", package = "correspondenceTables"))
head(AB)

result <- analyseCorrespondenceTable(
  AB = AB,
  Aname = NULL,
  Bname = NULL,
  A  = NULL,
  B  = NULL,
  longestAcodeOnly          = FALSE,
  longestBcodeOnly          = FALSE
)

print(result$Inventory)
print(result$Analysis)

# the name of the columns does not matter:
AB_mod = AB
colnames(AB_mod) = c("a", "b")

result_mod <- analyseCorrespondenceTable(
  AB = AB_mod,
  Aname = NULL,
  Bname = NULL,
  A  = NULL,
  B  = NULL,
  longestAcodeOnly          = FALSE,
  longestBcodeOnly          = FALSE
)

print(result_mod$Inventory)
print(result_mod$Analysis)


# examples with using Aname and Bname

# fist we add two more column to the same dataset
set.seed(123)
AB_mod2 = AB_mod

AB_mod2 <- cbind(new_first_col = sample(1:100, nrow(AB_mod2), replace = TRUE), AB_mod2)
AB_mod2 <- cbind(
  AB_mod2[1:2], 
  new_middle_col = sample(1:100, nrow(AB_mod2), replace = TRUE),
  AB_mod2[3:ncol(AB_mod2)]
)
colnames(AB_mod2)

result_mod2 <- analyseCorrespondenceTable(
  AB = AB_mod2,
  Aname = "a",
  Bname = "b",
  A  = NULL,
  B  = NULL,
  longestAcodeOnly          = FALSE,
  longestBcodeOnly          = FALSE
)

print(result_mod2$Inventory)
print(result_mod2$Analysis)

# check if we obtained the same results in all the 3 examples

first = identical(result$Inventory, result_mod$Inventory) 
second = identical(result_mod$Inventory, result_mod2$Inventory)
first && second
# TRUE

# the following won't work as we changed column names in AB_mod, 
# and added new columns in AB_mod2 but the results are the same.
first = identical(result$Analysis, result_mod$Analysis)
second = identical(result_mod$Analysis, result_mod2$Analysis)
first && second
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prefixList.R
\name{normalize_prefix_token}
\alias{normalize_prefix_token}
\title{Normalize a SPARQL prefix label (pure)}
\usage{
normalize_prefix_token(x)
}
\arguments{
\item{x}{character vector of candidate prefix labels}
}
\value{
cleaned, safe labels (lowercase, [a-z0-9_])
}
\description{
Cleans a SPARQL prefix token so it is safe and consistent:
- lowercases,
- trims whitespace,
- removes trailing ":" and whitespace,
- keeps only [a-z0-9_] characters,
- drops empty results.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aggregateCorrespondenceTable.R
\name{aggregateCorrespondenceTable}
\alias{aggregateCorrespondenceTable}
\title{Aggregate values from classification A to classification B}
\usage{
aggregateCorrespondenceTable(AB, A, B = NULL)
}
\arguments{
\item{AB}{A \code{data.frame} containing A→B links, with optional numeric
\code{weight}.}

\item{A}{A \code{data.frame} with one (character) code column and one
numeric value column.}

\item{B}{Optional \code{data.frame} with a code column defining the output
domain. Additional columns (e.g., labels) are preserved.}
}
\value{
A list with:
\itemize{
  \item \code{result}: data.frame with aggregated values for each B code
        (and any additional columns from \code{B}, if provided),
  \item \code{mapping}: standardized A→B mapping used internally,
  \item \code{diagnostics}: list with total input value, mapped value,
        coverage ratio, and codes in A that could not be mapped.
}
}
\description{
Aggregates numeric values from a source classification A into a target
classification B using a correspondence table AB.  
Each value in A is redistributed to one or more B codes according to weights
in AB (if available), or split equally when no valid weights exist.
}
\details{
This function is used when data expressed in one classification must be
converted into another classification using an A→B correspondence table.

The workflow is:
\enumerate{
  \item Detect and standardize column names in \code{AB}, \code{A} and \code{B}
        (case-insensitive).
  \item For each code in A, determine all corresponding B codes.
  \item Apply proportional allocation:
        \itemize{
          \item If \code{AB} contains a numeric weight column, weights
                are normalized per source code.
          \item If weights are missing, invalid, or sum to zero, values are
                split equally across all mapped B codes.
        }
  \item Values are multiplied by the resulting weights and aggregated by B code.
  \item If \code{B} is provided, the result is aligned to its list of codes
        (unmatched codes receive 0).
}

This function is appropriate for tasks such as converting statistical datasets
from NACE to CPA, CPA to CN, CPC to HS, or any situation where values have to
be re-expressed according to a different classification system.
}
\examples{
## Simple example (equal split)
AB <- data.frame(
  from_code = c("A1","A1","A2"),
  to_code   = c("B1","B2","B1")
)

print(AB)

A_tbl <- data.frame(
  code  = c("A1","A2"),
  value = c(100, 50)
)

print(A_tbl)

result <- aggregateCorrespondenceTable(AB, A_tbl)

print(result$result) 
print(result$mapping)

}
\seealso{
\code{\link{retrieveCorrespondenceTable}} for downloading AB tables.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/updateCorrespondenceTable.R
\name{updateCorrespondenceTable}
\alias{updateCorrespondenceTable}
\title{Correspondence table creation}
\usage{
updateCorrespondenceTable(
  A,
  B,
  AStar,
  AB,
  AAStar,
  Reference = "none",
  MismatchToleranceB = 0.2,
  MismatchToleranceAStar = 0.2,
  Redundancy_trim = TRUE
)
}
\arguments{
\item{A}{A data frame containing the original classification A. The first column holds the codes;
optional subsequent columns may hold labels and supplementary info.}

\item{B}{A data frame containing the target classification B. Same structure expectations as \code{A}.}

\item{AStar}{A data frame containing the updated version A*. Same structure expectations as \code{A}.}

\item{AB}{A data frame containing the previous correspondence table A:B (at least two columns: A-codes, B-codes).}

\item{AAStar}{A data frame containing the concordance table A:A* (at least two columns: A-codes, A*-codes).}

\item{Reference}{The reference classification among A and B. Valid values: \code{"none"}, \code{"A"}, \code{"B"}.
If \code{"A"} or \code{"B"}, a "Review" flag column is included (see "Explanation of the flags").}

\item{MismatchToleranceB}{Maximum acceptable proportion of rows in the updated table with no code of B
among those that have a code for A and/or A*. Default \code{0.2}, range [0,1].}

\item{MismatchToleranceAStar}{Maximum acceptable proportion of rows in the updated table with no code of A*
among those that have a code for A and/or B. Default \code{0.2}, range [0,1].}

\item{Redundancy_trim}{Logical. If \code{TRUE} (default) trims correct redundant records by collapsing A-side fields to "Multiple"
when appropriate; if \code{FALSE}, keeps redundancies and adds a \code{Redundancy_keep} indicator.}
}
\value{
\code{updateCorrespondenceTable()} returns a list with two data frames:
\itemize{
  \item \code{updateCorrespondenceTable}: the updated correspondence A*:B with flags "CodeChange", "Review" (if applicable),
        "Redundancy", "NoMatchToAStar", "NoMatchToB", "NoMatchFromAStar", "NoMatchFromB", "LabelChange",
        plus any additional columns coming from \code{A}, \code{B}, \code{AStar}, \code{AB}, \code{AAStar}.
  \item \code{classificationNames}: the names of the classifications (A, B, A*) read from the first column headers.
}
}
\description{
Update the correspondence table between two classifications when one of them has been updated.
}
\details{
*Input data frame requirements*
\itemize{
  \item \code{A}, \code{B}, \code{AStar}: at least 1 column and at least 1 row. The first column contains the codes; the first row is the header.
  \item \code{AB}, \code{AAStar}: at least 2 columns and at least 1 row; the first two columns contain code pairs; the first row is the header.
  \item Codes in the first column of \code{A}, \code{B}, \code{AStar} must be unique (no duplicates).
  \item Pairs (first two columns) in \code{AB} and \code{AAStar} must be unique (no duplicate pairs).
  \item If additional columns are present in classification or correspondence tables, the second column is treated as a label.
  \item All columns are coerced to character.
}

*Minimum interdependencies*
\itemize{
  \item At least one code of A must appear in both \code{AAStar} and \code{AB}.
  \item At least one code of A* must appear in both \code{AStar} and \code{AAStar}.
  \item At least one code of B must appear in both \code{B} and \code{AB}.
}

*Mismatch tolerance*
\itemize{
  \item If the share of rows with \code{NoMatchToAStar == 1} exceeds \code{MismatchToleranceAStar}, the function stops with an error.
  \item If the share of rows with \code{NoMatchToB == 1} exceeds \code{MismatchToleranceB}, the function stops with an error.
}
}
\section{Explanation of the flags}{


\itemize{
    \item For each row of the updated correspondence table, the value of "CodeChange" is equal to \code{1} if the code of A (or A*)
     contained in this row maps -in this or any other row of the table- to a different code of A* (or A), otherwise the 
     "CodeChange" is equal to \code{0}. The value of "CodeChange" is empty if either the code of A, or the code of A*, or both are missing.
    \item The "Review" flag is produced only if argument \code{Reference} has been set equal to "\code{A}" or "\code{B}".
    For each row of the updated correspondence table, if \code{Reference} = "\code{A}" the value of "Review" is equal to
    \code{1} if the code of B maps to more than one code of A*, and \code{0} otherwise. If \code{Reference} = "\code{B}" the
    value of "Review" is equal to \code{1} if the code of A* maps to more than one code of B, and \code{0} otherwise. The value
    of the flag is empty if either the code of A*, or the code of B, or both are missing.
    \item For each row of the updated correspondence table, the value of "Redundancy" is equal to \code{1} if the row contains
    a combination of codes of A* and B that also appears in at least one other row of the updated correspondence table. The
    value of the flag is empty if both the code of A* and the code of B are missing.
    \item When "Redundancy_Trim" is equal to \code{FALSE} the "Redundancy_keep" flag is created to identify with value \code{1}
    the records that will be kept if trimming is performed.
    \item For each row of the updated correspondence table, the value of "NoMatchToAStar" is equal to \code{1} if there is a
    code for A,  for B, or for both, but no code for A*. The value of the flag is \code{0} if there are codes for both A and
    A* (regardless of whether there is a code for B or not). Finally, the value of "NoMatchToAStar" is empty if neither A nor B
    have a code in this row.
    \item For each row of the updated correspondence table, the value of "NoMatchToB" is equal to \code{1} if there is a code
    for A,  for A*, or for both, but no code for B. The value of the flag is \code{0} if there are codes for both A and B
    (regardless of whether there is a code for A* or not). Finally, the value of "NoMatchToB" is empty if neither A nor
    A* have a code in this row.
    \item For each row of the updated correspondence table, the value of "NoMatchFromAStar" is equal to \code{1} if the row
    contains a code of A* that appears in the table of classification A* but not in the concordance table A:A*. The value of
    the flag is \code{0} if the row contains a code of A* that appears in both the table of classification
    A* and the concordance table A:A*. Finally, the value of the flag is empty if the row contains no code of A* or if it
    contains a code of A* that appears in the concordance table A:A* but not in the table of classification A*.
    \item For each row of the updated correspondence table, the value of "NoMatchFromB" is equal to \code{1} if the row
    contains a code of B that appears in the table of classification B but not in the correspondence table A:B. The value of
    the flag is \code{0} if the row contains a code of B that appears in both the table of classification B and the
    correspondence table A:B. Finally, the value of the flag is empty if the row contains no code of B or if it contains a code
    of B that appears in the correspondence table A:B but not in the table of classification B.
    \item For each row of the updated correspondence table, the value of "LabelChange" is equal to \code{1} if the labels of
    the codes of A and A* are different, and \code{0} if they are the same. Finally, the value of "LabelChange" is empty if
    either of the labels, or both labels, are missing. Lower and upper case are considered the same, and punctuation characters
    are ignored when comparing code labels.
    \item The argument "Redundancy_trim" is used to delete all the redundancies which are mapping correctly. If the analysis 
   concludes that the A*code / Bcode mapping is correct for all cases involving redundancies, then an action is needed to remove 
   the redundancies. If the selected value is \code{TRUE}, all redundant records are removed and kept only one record for each unique 
   combination. For this record retained, the Acodes, the Alabel and the Asupp information is replaced with ‘multiple’. If the multiple 
   A records are the same, their value will not be replaced. If the selected value is \code{FALSE}, no trimming is executed so redundant 
   records are shown, together with the redundancy flag.
}
}

\examples{
\dontrun{

  # Read CSVs outside, pass data frames in:
  
  A_df      <- utils::read.csv(system.file("extdata/test", "NAICS2017.csv",
  package = "correspondenceTables"),
  sep = ",", header = TRUE, check.names = FALSE, 
  colClasses = "character", encoding = "UTF-8")
                               
  AStar_df  <- utils::read.csv(system.file("extdata/test", "NAICS2022.csv",
  package = "correspondenceTables"),
  sep = ",", header = TRUE, check.names = FALSE,
  colClasses = "character", encoding = "UTF-8")
                               
  B_df      <- utils::read.csv(system.file("extdata/test", "NACE.csv",
  package = "correspondenceTables"),
  sep = ",", header = TRUE, check.names = FALSE,
  colClasses = "character", encoding = "UTF-8")
  
  AB_df     <- utils::read.csv(system.file("extdata/test", "NAICS2017_NACE.csv",
  package = "correspondenceTables"),
  sep = ",", header = TRUE, check.names = FALSE,
  colClasses = "character", encoding = "UTF-8")
                               
  AAStar_df <- utils::read.csv(system.file("extdata/test", "NAICS2017_NAICS2022.csv",
  package = "correspondenceTables"),
  sep = ",", header = TRUE, check.names = FALSE,
  colClasses = "character", encoding = "UTF-8")

  UPC <- updateCorrespondenceTable(
    A = A_df, B = B_df, AStar = AStar_df, AB = AB_df, AAStar = AAStar_df,
    Reference = "none", MismatchToleranceB = 0.5, MismatchToleranceAStar = 0.3,
    Redundancy_trim = FALSE
  )

  summary(UPC)
  head(UPC$updateCorrespondenceTable)
  UPC$classificationNames
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prefixList.R
\name{build_prefix_lines}
\alias{build_prefix_lines}
\title{Build PREFIX lines from tokens and IRIs (pure)}
\usage{
build_prefix_lines(tokens, iris)
}
\arguments{
\item{tokens}{character vector of prefix tokens (will be normalized)}

\item{iris}{character vector of namespace IRIs (will be normalized)}
}
\value{
character vector like "PREFIX token: <iri/>"
}
\description{
Build PREFIX lines from tokens and IRIs (pure)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classificationList.R
\name{classificationList}
\alias{classificationList}
\title{List available classification schemes from CELLAR or FAO}
\usage{
classificationList(endpoint = "ALL", showQuery = FALSE)
}
\arguments{
\item{endpoint}{Character scalar. One of "CELLAR", "FAO" or "ALL".
"ALL" queries both repositories and returns a named list.}

\item{showQuery}{Logical; if TRUE, returns also the SPARQL query used.}
}
\value{
If endpoint is "CELLAR" or "FAO" and showQuery = FALSE: a data.frame with
columns Prefix, ConceptScheme, URI, Title, Languages.

If showQuery = TRUE: a list with "ClassificationList" and "SPARQL.query".

If endpoint = "ALL": a named list with $CELLAR and $FAO. If showQuery = TRUE,
each element is itself a list (ClassificationList + SPARQL.query).
}
\description{
List all classification schemes available on the selected repository
(CELLAR, FAO, or both). Returns, for each scheme, its prefix, identifier,
URI, English title, and the list of languages detected on its concepts.
}
\examples{
\dontrun{
head(classificationList("CELLAR"))
res <- classificationList("FAO", showQuery = TRUE)
names(res)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prefixList.R
\name{normalize_namespace_iri}
\alias{normalize_namespace_iri}
\title{Normalize a namespace IRI (pure)}
\usage{
normalize_namespace_iri(uri)
}
\arguments{
\item{uri}{character vector}
}
\value{
normalized IRIs
}
\description{
Ensures each IRI:
- has an HTTP(S) scheme (defaults to https if missing),
- has no whitespace,
- ends with exactly one trailing slash.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classificationQC.R
\name{classificationQC}
\alias{classificationQC}
\title{Perform quality control on a classification}
\usage{
classificationQC(
  classification,
  lengths,
  fullHierarchy = TRUE,
  labelUniqueness = TRUE,
  labelHierarchy = TRUE,
  singleChildCode = NULL,
  sequencing = NULL
)
}
\arguments{
\item{classification}{A \code{data.frame} containing (at least) two columns: the codes
and labels of the classification to be checked. This argument is mandatory.}

\item{lengths}{A \code{data.frame} with one row per hierarchical level giving the
        initial and final positions of the segment of the code referring
        to that level. The number of rows implicitly defines the number of hierarchical levels
(\eqn{k}). This argument is mandatory. The column names should be \code{charb} and \code{chare}, if this is not the case, they will be automatically changed and a warning will appear.}

\item{fullHierarchy}{Logical.
If \code{FALSE}, the function checks that all positions at levels greater
than 1 have a parent at the level immediately above (no orphans).  
If \code{TRUE}, it additionally checks that positions at levels strictly
lower than \eqn{k} have children in the next level (no childless nodes).}

\item{labelUniqueness}{Logical.
When \code{TRUE}, the function checks whether labels are unique at each
hierarchical level. Duplicates are listed in the
\code{QC_duplicatesLabel} table.}

\item{labelHierarchy}{Logical.
When \code{TRUE}, the function checks that:
\itemize{
  \item single children share the same label as their parent;
  \item if a parent shares a label with one of its children, it must be a
        single-child parent.
}}

\item{singleChildCode}{A \code{data.frame} specifying admissible codes for single
        children at each hierarchical level.
If not \code{NULL}, the function checks compliance with these codes.}

\item{sequencing}{A \code{data.frame} defining admissible codes and their expected
        order for multiple-child situations at each level.
If not \code{NULL}, the function checks sequencing consistency.}
}
\value{
A list of data frames flagging potential issues in the classification:
\itemize{
  \item \code{QC_output}: full augmented table with computed levels,
        code segments, and all compliance flags;
  \item \code{QC_noLevels}: rows for which hierarchical level could not be
        determined;
  \item \code{QC_orphan}: positions that lack an expected parent;
  \item \code{QC_childless}: positions expected to have children but do not;
  \item \code{QC_duplicatesLabel}: positions with duplicate labels within
        the same level;
  \item \code{QC_duplicatesCode}: duplicated codes;
  \item \code{QC_singleChildMismatch}: label inconsistencies for
        single-children relationships;
  \item \code{QC_singleCodeError}: single children with invalid codes (if
        \code{singleChildCode} is provided);
  \item \code{QC_multipleCodeError}: non-single children with invalid codes
        (if \code{sequencing} is provided);
  \item \code{QC_gapBefore}: missing expected codes in sequence;
  \item \code{QC_lastSibling}: flag for last sibling in sequencing rules.
}
}
\description{
This function performs a series of structural and logical quality-control
checks on a hierarchical classification. It verifies the consistency of the
hierarchical levels, the presence of parent–child relationships, the
uniqueness of labels, the compliance with admissible coding rules, and the
sequencing of children within each level.  
The output is a list of data frames flagging potential issues.
}
\examples{


classification <- read.csv(system.file("extdata/test", "Nace2_long.csv",
                                      package = "correspondenceTables"))
lengths <- data.frame(charb = c(1,2,3,5), chare = c(1,2,4,5))

Output <- classificationQC(
 classification = classification,
 lengths = lengths,
 fullHierarchy = TRUE,
 labelUniqueness = TRUE,
 labelHierarchy = TRUE,
 singleChildCode = NULL,
 sequencing = NULL
)

print(Output$QC_output)
print(Output$QC_orphan)
print(Output$QC_childless)
print(Output$QC_duplicatesLabel)
print(Output$QC_duplicatesCode)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataStructure.R
\name{dataStructure}
\alias{dataStructure}
\title{Retrieve a classification structure (levels and concepts) from a supported service}
\usage{
dataStructure(
  endpoint,
  prefix,
  conceptScheme = NULL,
  language = "en",
  showQuery = FALSE,
  return = c("summary", "details", "both"),
  timeout_sec = 60,
  retries = 3
)
}
\arguments{
\item{endpoint}{Character scalar. Accepted values are \code{"CELLAR"} or \code{"FAO"}.}

\item{prefix}{Character. A classification prefix used in the registry
(e.g. \code{"NACE"}, \code{"CPA"}, \code{"CN2022"}). Used to resolve the scheme.}

\item{conceptScheme}{Character or \code{NULL}. Optional alternative identifier
for the concept scheme. The resolver tries both \code{prefix} and \code{conceptScheme}
using basic normalization (case-insensitive; dot removal).}

\item{language}{Character scalar. Two-letter language code used to select
labels (default: \code{"en"}). Fallback chain: requested language -> English (\code{"en"})
-> no-language literal -> the local resource name.}

\item{showQuery}{Logical. If \code{TRUE}, the generated queries are emitted via
\code{message()} and also included in the returned list.}

\item{return}{One of \code{c("summary", "details", "both")}. Controls which table(s)
are fetched and returned.}

\item{timeout_sec}{Numeric. HTTP timeout in seconds. Default: \code{60}.}

\item{retries}{Integer. Number of automatic retries on transient HTTP errors.
Default: \code{3}.}
}
\value{
A \code{data.frame} (for \code{"summary"} or \code{"details"}), or a
\code{list(summary, details)} when \code{return = "both"}.

\strong{Columns:}
\itemize{
  \item All returned tables include a leading \code{Prefix} column (lower-case token).
  \item \code{summary}: \code{Concept_Scheme}, \code{Depth}, \code{Level}, \code{Count}.
  \item \code{details}: \code{Concept}, \code{Code}, \code{Label}, \code{Depth}, \code{Level},
        \code{BroaderList}, \code{BroaderCodeList}.
}

If \code{showQuery = TRUE}, the return value is a list containing:
\itemize{
  \item \code{resolved}: metadata (\code{endpoint}, inputs, \code{scheme_id}, \code{ns_uri}, and optionally \code{title})
  \item the query text (one or two elements depending on \code{return})
  \item the retrieved table(s)
}
}
\description{
Queries the selected service (e.g., CELLAR or FAO) to retrieve the structure of a
classification scheme (levels, depth, and concept rows), with language-aware
label fallback. You can return a compact summary table, a detailed
concept-level table, or both.

The function first resolves the input classification using the in-package registry
returned by \code{classificationList()}, then retrieves the structure and returns
it as data frames. Network errors are handled with retry logic.
}
\details{
\strong{Resolution of the scheme}

The function calls \code{classificationList(endpoint)} and tries to match either
\code{prefix} or \code{conceptScheme} using normalized keys. On success it extracts
the classification namespace and scheme identifier needed to compose requests.

\strong{Language fallback}

Labels for both the classification levels and concepts follow a multi-step
fallback: requested \code{language} -> \code{"en"} -> no-language literal -> local name.
}
\section{Errors and Warnings}{

\itemize{
  \item Errors if \code{endpoint} is not \code{"CELLAR"} or \code{"FAO"}, or if the
        scheme cannot be resolved from the registry.
  \item Warns (and still returns rows) if no explicit levels are declared.
  \item Propagates HTTP errors from the service after retry policy.
}
}

\examples{
\dontrun{
ds_cn <- dataStructure(
  endpoint      = "CELLAR",
  prefix        = "cn2022",
  conceptScheme = "cn2022",
  language      = "en",
  return        = "summary"
)
head(ds_cn)
}

}
\seealso{
\code{\link{classificationList}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/newCorrespondenceTable.R
\name{newCorrespondenceTable}
\alias{newCorrespondenceTable}
\title{Correspondence table creation}
\usage{
newCorrespondenceTable(
  Tables,
  Reference = "none",
  MismatchTolerance = 0.2,
  Redundancy_trim = TRUE,
  Progress = TRUE
)
}
\arguments{
\item{Tables}{A string of type character containing the name of a csv file which contains the names of the files that
contain the classifications and the intermediate correspondence tables OR a list of vectors with the names of the dataframes
classifications and the intermediate correspondence tables (see "Details" below).}

\item{Reference}{The reference classification among A and B. If a classification is the reference to the other, and hence
\emph{hierarchically superior} to it, each code of the other classification is expected to be mapped to at most one code
 of the reference classification. The valid values are \code{"none"}, \code{"A"}, and \code{"B"}. If the selected value
 is \code{"A"} or \code{"B"}, a "Review" flag column (indicating the records violating this expectation) is included
 in the output (see "Explanation of the flags" below).}

\item{MismatchTolerance}{The maximum acceptable proportion of rows in the candidate correspondence table which contain
no code for classification A or no code for classification B. The default value is \code{0.2}. The valid values are
real numbers in the interval [0, 1].}

\item{Redundancy_trim}{An argument in the function containing the logical values \code{TRUE} or \code{FALSE}
used to facilitate the trimming of the redundant records.
The default value is \code{TRUE}, which removes all redundant records.
The other values is \code{FALSE}, which shows redundant records together with the redundancy flag.}

\item{Progress}{An argument in the function containing the logical values \code{TRUE} (default) or \code{FALSE}. 
Used to switch ON (when \code{TRUE}) or OFF (when \code{FALSE}) the progress bar that interactively display the creation process of the new table in the console output.}
}
\value{
\code{newCorrespondenceTable()} returns a list with two elements, both of which are data frames.
\itemize{
    \item The first element is the candidate correspondence table A:B, including the codes of all "pivot" classifications,
     augmented with flags "Review" (if applicable), "Redundancy", "Unmatched", "NoMatchFromA", "NoMatchFromB" and with all
     the additional columns of the classification and intermediate correspondence table files.
    \item The second element contains the names of classification A, the "pivot" classifications and classification B as
    read from the top left-hand side cell of the respective input files.
}
}
\description{
Create a candidate correspondence table between two classifications based on their correspondences with intermediate classifications
}
\details{
File and file name requirements:
    \itemize{
        \item The file that corresponds to argument \code{Tables} and the files to which the contents of \code{Tables}
        lead, must be in \emph{csv format with comma as delimiter}. If full paths are not provided, then these files must
         be available in the working directory. No two filenames provided must be identical.
        \item The file that corresponds to argument \code{Tables} must contain filenames, \emph{and nothing else}, in
        a \eqn{(k+2)} × \eqn{(k+2)} table, where \eqn{k}, a positive integer, is the number of "pivot" classifications.
        The cells in the main diagonal of the table provide the filenames of the files which contain, with this order,
        the classifications  \eqn{A, C_1}, \eqn{\ldots}, \eqn{C_k} and \eqn{B}. The off-diagonal directly above the main
        diagonal contains the filenames of the files that contain, with this order, the correspondence tables
        \eqn{A:C_1}, \{\eqn{C_i}:\eqn{C_{i+1}}, \eqn{1 \le i \le k-1}\} and \eqn{B:C_k}. All other cells of the table
        must be empty.
    }
Classification table requirements:
    \itemize{
        \item Each of the files that contain classifications must contain at least one column and at least two rows.
        The first column contains the codes of the respective classification. The first row contains column headers.
        The header of the first column is the name of the respective classification (e.g., "CN 2021").
        \item The classification codes contained in a classification file (expected in its first column as mentioned
        above) must be unique. No two identical codes are allowed in the column.
        \item If any of the files that contain classifications has additional columns the first one of them is assumed
        to contain the labels of the respective classification codes.
    }
Correspondence table requirements:
    \itemize{
        \item The files that contain correspondence tables must contain at least two columns and at least two rows.
        The first column of the file that contains A:\eqn{C_1} contains the codes of classification A. The second column
        contains the codes of classification \eqn{C_1}. Similar requirements apply to the files that contain
        \eqn{C_i}:\eqn{C_{i+1}}, \eqn{1 \le i \le k-1} and B:\eqn{C_k}. The first row of each of the files that contain
        correspondence tables contains column headers. The names of the first two columns are the names of the respective
        classifications.
        \item The pairs of classification codes contained in a correspondence table file (expected in its first two columns
        as mentioned above) must be unique. No two identical pairs of codes are allowed in the first two columns.
    }
Interdependency requirements:
    \itemize{
        \item At least one code of classification A must appear in both the file of classification A and the file of
        correspondence table A:\eqn{C_1}.
        \item At least one code of classification B must appear in both the file of classification B and the file of
        correspondence table B:\eqn{C_k}, where \eqn{k}, \eqn{k\ge 1}, is the number of pivot classifications.
        \item If there is only one pivot classification, \eqn{C_1}, at least one code of it must appear in both the file of
        correspondence table A:\eqn{C_1} and the file of correspondence table B:\eqn{C_1}.
        \item If the pivot classifications are \eqn{k} with \eqn{k\ge 2} then at least one code of \eqn{C_1} must appear in
        both the file of correspondence table A:\eqn{C_1} and the file of correspondence table \eqn{C_1}:\eqn{C_2}, at least one
        code of each of the \eqn{C_i}, \eqn{i = 2, \ldots, k-1} (if \eqn{k\ge 3}) must appear in both the file of correspondence
        table \eqn{C_{i-1}}:\eqn{C_i} and the file of correspondence table \eqn{C_i}:\eqn{C_{i+1}}, and at least one code of
        \eqn{C_k} must appear in both the file of correspondence table \eqn{C_{k-1}}:\eqn{C_k} and the file of correspondence table
        B:\eqn{C_k}.
    }
Mismatch tolerance:
    \itemize{
        \item The ratio that is compared with \code{MismatchTolerance} has as numerator the number of rows in the candidate
        correspondence table which contain no code for classification A or no code for classification B and as denominator
        the total number of rows of this table. If the ratio exceeds \code{MismatchTolerance} the execution of the function
        is halted.
    }
If any of the conditions required from the arguments is violated an error message is produced and execution is stopped.
}
\section{Explanation of the flags}{


\itemize{
    \item  The "Review" flag is produced only if argument Reference has been set equal to "\code{A}" or "\code{B}". For
    each row of the candidate correspondence table, if \code{Reference} = "\code{A}" the value of "Review" is equal to
    \code{1} if the code of B maps to more than one code of A, and \code{0} otherwise. If \code{Reference} = "\code{B}"
    the value of "Review" is equal to \code{1} if the code of A maps to more than one code of B, and \code{0} otherwise.
    The value of the flag is empty if the row does not contain a code of A or a code of B.
    \item For each row of the candidate correspondence table, the value of "Redundancy" is equal to \code{1} if the row
    contains a combination of codes of A and B that also appears in at least one other row of the candidate
    correspondence table.
    \item When "Redundancy_Trim" is equal to \code{FALSE} the "Redundancy_keep" flag is created to identify with value \code{1}
    the records that will be kept if trimming is performed.
    \item For each row of the candidate correspondence table, the value of "Unmatched" is equal to \code{1} if the row
    contains a code of A but no code of B or if it contains a code of B but no code of A. The value of the flag is
    \code{0} if the row contains codes for both A and B.
    \item For each row of the candidate correspondence table, the value of "NoMatchFromA" is equal to \code{1} if the row
    contains a code of A that appears in the table of classification A but not in correspondence table A:\eqn{C_1}. The
    value of the flag is \code{0} if the row contains a code of A that appears in both the table of classification A and
    correspondencetable A:\eqn{C_1}. Finally, the value of the flag is empty if the row contains no code of A or if it
    contains a code of A that appears in correspondence table A:\eqn{C_1} but not in the table of classification A.
    \item For each row of the candidate correspondence table, the value of "NoMatchFromB" is equal to \code{1} if the row
    contains a code of B that appears in the table of classification B but not in correspondence table B:\eqn{C_k}. The
    value of the flag is \code{0} if the row contains a code of B that appears in both the table of classification B and
    correspondence table B:\eqn{C_k}. Finally, the value of the flag is empty if the row contains no code of B or if it
    contains a code of B that appears in correspondence table B:\eqn{C_k} but not in the table of classification B.
    \item The argument "Redundancy_trim" is used to delete all the redundancies which are mapping correctly.
    The valid logical values for this argument in the candidate correspondence table are \code{TRUE} or \code{FALSE}.
    If the selected value is \code{TRUE}, all redundant records are removed and kept exactly one record for each unique combination.
    For this retained record, the codes, the label and the supplementary information of the pivot classifications are replaced with
    'multiple'. If the multiple infomration of the pivot classifications are the same, their value will not be replaced.
    If the selected value is \code{FALSE}, no trimming is executed so redundant records are shown, together with the redundancy flag.
    If the logical values are missing the implementation of the function will stop.

}
}

\section{Sample datasets included in the package}{


If the user wish to access the csv files with the sample data, they have two options:
\itemize{
\item Option 1: Unpack into any folder of their choice the tar.gz file into which the package has arrived. All sample
datasets may be found in the "inst/extdata" subfolder of this folder.
\item Option 2: Go to the "extdata" subfolder of the folder in which the package has been installed in their PC's \code{R}
library. All sample datasets may be found there.
}
}

\examples{
{
\dontrun{

code_in   <- data.frame(code_in  = c("A","B"))
code_mid  <- data.frame(code_mid = c("B","C","D"))
in_to_mid <- data.frame(code_in  = c("A","B")
                        , code_out = c("C","B"))
code_out  <- data.frame(code_out = c("B","C","D","E"))
mid_to_out <- data.frame(code_mid = c("B","C","D")
                         , code_out = c("B","C","D"))

write.csv(code_in,    file = "code_in.csv",   row.names  = FALSE)
write.csv(code_mid,   file = "code_mid.csv",  row.names  = FALSE)
write.csv(in_to_mid,  file = "in_to_mid.csv", row.names  = FALSE)
write.csv(mid_to_out, file = "mid_to_out.csv", row.names = FALSE)
write.csv(code_out,   file = "code_out.csv",   row.names = FALSE)


m <- matrix("",3,3)
diag(m) <- c("code_in.csv","code_mid.csv","code_out.csv")
m[1,2] <- "in_to_mid.csv"
m[2,3] <- "mid_to_out.csv"



write.table(m, file="merge_sequence.csv",row.names=FALSE,col.names=FALSE,sep=",")


out <- newCorrespondenceTable("merge_sequence.csv",MismatchTolerance = 0.9)
print(out$newCorrespondenceTable)
print(out$classificationNames)


    } }
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/correspondenceTableList.R
\name{correspondenceTableList}
\alias{correspondenceTableList}
\title{List available correspondence tables from online services}
\usage{
correspondenceTableList(endpoint = "ALL", showQuery = FALSE)
}
\arguments{
\item{endpoint}{Character. Which service(s) to query.
One or more of the supported endpoints (e.g. "CELLAR", "FAO"),
or "ALL" to query all available services. Case-insensitive.
If multiple endpoints are supplied (including "ALL"), a **named list**
is returned; otherwise a single `data.frame`.}

\item{showQuery}{Logical. If `TRUE`, the SPARQL queries used are emitted via
`message()` (suppress with `suppressMessages()`).}
}
\value{
If a single endpoint is requested, a `data.frame` (one row per correspondence).
If multiple endpoints are requested (including "ALL"), a **named list**
of data frames, one per endpoint.
}
\description{
Returns a data frame listing available correspondence tables (mappings between
statistical classifications) published by one or more supported online services.

Use this function to discover valid `prefix` and `ID_table` values before
calling `retrieveCorrespondenceTable()`.
}
\details{
Read-only queries; no side effects (no file writes or caching).
}
\examples{
\dontrun{
# List correspondence tables available in CELLAR
ct <- correspondenceTableList(endpoint = "CELLAR")
head(ct)

# List correspondence tables available in FAO
ct_fao <- correspondenceTableList(endpoint = "FAO")
head(ct_fao)

# Get correspondence tables from both repositories
ct_all <- correspondenceTableList(endpoint = "ALL")
names(ct_all)
}


}
\seealso{
retrieveCorrespondenceTable()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/retrieveCorrespondenceTable.R
\name{retrieveCorrespondenceTable}
\alias{retrieveCorrespondenceTable}
\title{Download a correspondence (mapping) table between two classifications}
\usage{
retrieveCorrespondenceTable(
  endpoint,
  prefix,
  ID_table,
  language = "en",
  showQuery = FALSE
)
}
\arguments{
\item{endpoint}{Character. The online service to query. Case-insensitive.
Supported values are those returned by the internal endpoint registry
(e.g., \code{"CELLAR"}, \code{"FAO"}).}

\item{prefix}{Character. Catalogue prefix where the correspondence is
published (e.g., \code{"nace2"}, \code{"cpa21"}, \code{"cn2022"}).
Use \code{\link{correspondenceTableList}()} to discover valid values.}

\item{ID_table}{Character. Identifier of the correspondence, typically of
the form \code{"A_B"} such as \code{"NACE2_CPA21"} or \code{"CN2022_NACE2"}.
Discover identifiers via \code{\link{correspondenceTableList}()}.}

\item{language}{Character. Preferred label language as a BCP‑47 code.
Defaults to \code{"en"} (English). Examples: \code{"fr"}, \code{"de"}.}

\item{showQuery}{Logical. If \code{TRUE}, returns a list with the SPARQL
query and the result data frame; otherwise (default) returns just the data frame.}
}
\value{
A \code{data.frame} where each row represents one mapping from the source
classification (\strong{A}) to the target classification (\strong{B}).

Columns (when available):
\itemize{
  \item \code{CorrespondenceID}:  Requested identifier (e.g., \code{"NACE2_CPA21"}).
  \item \code{Prefix}: Catalogue prefix (constant per table).
  \item \code{A}: Source code (\code{A} parsed from \code{ID_table}).
  \item \code{B}: Target code (\code{B} parsed from \code{ID_table}).
  \item \code{Label_A}, \code{Label_B}: Human‑readable labels in \code{language}.
  \item \code{Include_A}, \code{Exclude_A}, \code{Include_B}, \code{Exclude_B}: Notes, if available.
  \item \code{Comment}: Free‑text comment on the mapping, if any.
  \item \code{URL}: URI of the mapping association.
}

Attributes attached for traceability:
\itemize{
  \item \code{endpoint}, \code{endpoint_url}, \code{prefix}, \code{ID_table}
  \item \code{A}, \code{B}, \code{language}, \code{normalized_codes}, \code{anchor_mode}, \code{anchor}
}
}
\description{
Downloads a correspondence table, a mapping of codes, from one statistical
classification (A) to another (B) from an online service (CELLAR or
FAO), and returns it as a plain data frame. The function has no side effects
(it does not read or write files).

A correspondence table tells you which codes in classification A relate
to which codes in classification B (e.g., \code{NACE2} → \code{CPA21}).
}
\details{
\strong{When to use this.} Use \code{retrieveCorrespondenceTable()} when you
know (or have discovered) the \code{prefix} and \code{ID_table} of a
correspondence and want the mapping rows as a standard data frame.

\strong{How it works (short).}
\enumerate{
  \item Validates the selected service (\code{endpoint}).
  \item Optionally resolves a stable anchor via \code{\link{correspondenceTableList}()}.
  \item Builds a SPARQL query returning source/target codes, labels, notes, and link URI.
  \item Issues an HTTP request and parses the response into a data frame.
}
}
\section{Notes}{

\itemize{
  \item The function focuses on retrieval. Any code normalization is left to
        downstream steps.
}
}

\examples{
\dontrun{
# 1) Discover available correspondences at CELLAR
ct <- correspondenceTableList(endpoint = "CELLAR")
subset(ct, grepl("NACE2", ID) & grepl("CPA21", ID))

# 2) Retrieve one correspondence (English labels)
x <- retrieveCorrespondenceTable(
  endpoint = "CELLAR",
  prefix   = "nace2",
  ID_table = "NACE2_CPA21",
  language = "en"
)
head(x)

# 3) Save to CSV explicitly if needed (no automatic writing)
# utils::write.csv(x, "NACE2_CPA21.csv", row.names = FALSE)

# 4) Inspect the generated SPARQL query
dbg <- retrieveCorrespondenceTable(
  endpoint = "FAO",
  prefix   = "cpa21",
  ID_table = "NACE2_CPA21",
  showQuery = TRUE
)
message(substr(dbg$SPARQL.query, 1, 400), "...")
}

}
